#!/usr/bin/env python
# -*- coding: utf-8 -*-


import sys
import os
import subprocess


def print_help():
    print("Wrong configuration flag has been provided. You must run build_kinebec.py as follows:")
    print("    python build_kinebec.py <IO flags> [build flags] [optional flags]")
    print(" ")
    print("IO flags (mandatory): you are required to provide at least one of the following ones:")
    print("  --with-hdf5              Use Hdf5 or parallel Hdf5 library (>= 1.8.16)")
    print("  --with-adios             Use Adios library (>= 1.9.0)")
    print(" ")
    print("Build flags: please provide one or several flags among the following ones for building the corresponding version of Kinebec:")
    print("  --serial                 Serial version")
    print("  --mpi                    MPI distributed memory version (requires OpenMPI version >= 1.10.2)")
    print("  --omp                    OpenMP shared memory version (requires OpenMP version >= 3.1)")
    print("  --cuda                   CUDA GPGPU version (requires Cuda version >= 9.0)")
    print(" ")
    print("Optional flags:")
    print("  --force-vectorization    Compile specific loops for better benefits of GCC vectorization (warning: high memory cost during the execution)")
    print(" ")
    exit()


build_serial      = False
build_openmp      = False
build_mpi         = False
build_cuda        = False

use_hdf5          = False
use_adios         = False

use_vectorization = False


if (len(sys.argv) < 1):
    print_help()
else:
    for i in range(len(sys.argv)-1):
        config = sys.argv[i+1]
        build_serial = build_serial or (config == "--serial")
        build_openmp = build_openmp or (config == "--omp")
        build_mpi = build_mpi or (config == "--mpi")
        build_cuda = build_cuda or (config == "--cuda")
        use_hdf5 = use_hdf5 or (config == "--with-hdf5")
        use_adios = use_adios or (config == "--with-adios")
        use_vectorization = use_vectorization or (config == "--force-vectorization")
    if (not(use_hdf5 or use_adios)):
        print_help()
    if (not(build_serial or build_openmp or build_mpi or build_cuda)):
        print_help()



list_builds = [build_serial, build_openmp, build_mpi, build_cuda]
name_builds = ["SERIAL", "OPENMP", "MPI", "CUDA"]

hostname = subprocess.check_output(['uname', '-n'])
hostname = hostname.decode("utf-8").split('\n')[0]

cmd = "sh cleanall"
os.system(cmd)

defaultbuf = ""
defaultbuf += "SET(USE_HDF5_LIBRARY "
if (use_hdf5):
    defaultbuf += "TRUE"
else:
    defaultbuf += "FALSE"
defaultbuf += ")\n"

defaultbuf += "SET(USE_ADIOS_LIBRARY "
if (use_adios):
    defaultbuf += "TRUE"
else:
    defaultbuf += "FALSE"
defaultbuf += ")\n"

defaultbuf += "SET(FORCE_VECTORIZATION "
if (use_vectorization):
    defaultbuf += "TRUE"
else:
    defaultbuf += "FALSE"
defaultbuf += ")\n"

if ("zeus" in hostname):
    cmd = "module load anaconda3/2020.02"
    os.system(cmd)
    cmd = "module load cmake/3.16.1"
    os.system(cmd)

for i in range(len(list_builds)):
    if (list_builds[i]):
        buf = defaultbuf
        for j in range(len(list_builds)):
            buf += "SET(BUILD_"+name_builds[j]+"_VERSION "
            if ((list_builds[i]) and (i == j)):
                buf += "TRUE"
            else:
                buf += "FALSE"
            buf += ")\n"
        f = open("CMakeOptions.txt", "w")
        f.write(buf)
        f.close()

        cmd = "sh cleanall"
        if ((name_builds[i] == "MPI") and ("zeus" in hostname)):
            cmd += " && module load gcc/4.8.5/openmpi/1.8.4"
        if ((name_builds[i] == "CUDA") and ("zeus" in hostname)):
            cmd += " && module load nvidia/cuda/10.0/compilers gcc/6.1.0/compilers"
        cmd += " && cmake ."
        cmd += " && make"
        os.system(cmd)