#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tests_kinebec import *


def extend_in_time_test_relaxation(setupfile, new_Tfinal, new_Nt):
	dic = read_setupfile(setupfile)
	old_Tfinal = dic["Tfinal"]
	old_Nt = dic["Nt"]
	old_ndigits = int(np.log10(np.ceil(old_Nt)))+1
	prefixfile = dic["outputdir"]+"/"+dic["prefix"]
	list_old_h5files = []
	list_old_xdmffiles = []
	list_old_bpfiles = []
	list_new_h5files = []
	list_new_xdmffiles = []
	list_new_bpfiles = []

	if ((new_Tfinal <= old_Tfinal) or (new_Nt <= old_Nt)):
		print("Error while extending a simulation time. The new final time and/or number of time iterations is lower than the old one.")
		exit()

	new_ndigits = int(np.log10(np.ceil(new_Nt)))+1

	dic["Nt"] = new_Nt
	dic["Tfinal"] = new_Tfinal
	print(dic)
	write_setupfile(dic, setupfile)

	for i in range(old_Nt+1):
		old_h5file = prefixfile+"_"+str(i).rjust(old_ndigits, '0')+".h5"
		old_xdmffile = prefixfile+"_"+str(i).rjust(old_ndigits, '0')+".xdmf"
		old_bpfile = prefixfile+"_"+str(i).rjust(old_ndigits, '0')+".bp"

		new_h5file = prefixfile+"_"+str(i).rjust(new_ndigits, '0')+".h5"
		new_xdmffile = prefixfile+"_"+str(i).rjust(new_ndigits, '0')+".xdmf"
		new_bpfile = prefixfile+"_"+str(i).rjust(new_ndigits, '0')+".bp"

		if (os.path.isfile(old_h5file) and os.path.isfile(old_xdmffile)):
			list_old_h5files.append(old_h5file)
			list_new_h5files.append(new_h5file)
			list_old_xdmffiles.append(old_xdmffile)
			list_new_xdmffiles.append(new_xdmffile)
		if (os.path.isfile(old_bpfile)):
			list_old_bpfiles.append(old_bpfile)
			list_new_bpfiles.append(new_bpfile)

	for i in range(len(list_old_xdmffiles)):
		old_fileh5 = "./"+dic["prefix"]+"_"+str(i).rjust(old_ndigits, '0')+".h5"
		new_fileh5 = "./"+dic["prefix"]+"_"+str(i).rjust(new_ndigits, '0')+".h5"
		f = open(list_old_xdmffiles[i], "r")
		lines = f.readlines()
		buf = ""
		for line in lines:
			buf += line.replace(old_fileh5, new_fileh5)
		f.close()
		f = open(list_new_xdmffiles[i], "w")
		f.write(buf)
		f.close()
		if (list_old_h5files[i] != list_new_h5files[i]):
			cmd = "mv "+list_old_h5files[i]+" "+list_new_h5files[i]
			os.system(cmd)
			cmd = "rm "+list_old_xdmffiles[i]
			os.system(cmd)

	for i in range(len(list_old_bpfiles)):
		if (list_old_bpfiles[i] != list_new_bpfiles[i]):
			cmd = "mv "+list_old_bpfiles[i]+" "+list_new_bpfiles[i]
			os.system(cmd)


def test_relaxation(dic, subdir="", parallel="serial", nCPUs="max", walltime=[1,0,0], memprofile_freq=-1., force_archlimits=False, run=False, resume=False):
	newdic = {}
	newdic["ndims"] = dic["ndims"]

	newdic["type_qparticles"] = dic["type_particles"]
	if (dic["type_particles"] == "classical"):
		newdic["type_collision"] = "classical"
	else:
		newdic["type_collision"] = "quantum"

	newdic["type_initialization"] = dic["type_initialization"]

	newdic = kinebec_default_setup(newdic)

	if (newdic["type_initialization"] == "classical_maxwellian"):
		list_keys = ["rho0", "u0", "sigma0"]
		if (type(dic["rho0"]) != type(1.0)):
			print(type(dic["rho0"]))
			print("rho0 = "+str(dic["rho0"]))
			exit()
		if (type(dic["u0"]) != type([1.0])):
			print("u0 = "+str(dic["u0"]))
			exit()
		elif (len(dic["u0"]) != dic["ndims"]):
			print("u0 = "+str(dic["u0"]))
			exit()
		else:
			for d in range(dic["ndims"]):
				if (type(dic["u0"][d]) != type(1.0)):
					print("u0 = "+str(dic["u0"]))
					exit()
		if (type(dic["sigma0"]) != type(1.0)):
			print("sigma0 = "+str(dic["sigma0"]))
			exit()
	elif (newdic["type_initialization"] == "classical_bimaxwellian"):
		list_keys = ["rho0", "u1", "u2", "sigma0"]
		if (type(dic["rho0"]) != type(1.0)):
			print("rho0 = "+str(dic["rho0"]))
			exit()
		if (type(dic["u1"]) != type([1.0])):
			print("u1 = "+str(dic["u1"]))
			exit()
		elif (len(dic["u1"]) != dic["ndims"]):
			print("u1 = "+str(dic["u1"]))
			exit()
		else:
			for d in range(dic["ndims"]):
				if (type(dic["u1"][d]) != type(1.0)):
					print("u1 = "+str(dic["u1"]))
					exit()
		if (type(dic["u2"]) != type([1.0])):
			print("u2 = "+str(dic["u2"]))
			exit()
		elif (len(dic["u2"]) != dic["ndims"]):
			print("u2 = "+str(dic["u2"]))
			exit()
		else:
			for d in range(dic["ndims"]):
				if (type(dic["u2"][d]) != type(1.0)):
					print("u2 = "+str(dic["u2"]))
					exit()
		if (type(dic["sigma0"]) != type(1.0)):
			print("sigma0 = "+str(dic["sigma0"]))
			exit()
	elif (newdic["type_initialization"] == "classical_multimaxwellian"):
		list_keys = ["rho0", "nmaxwellians", "u", "sigma0"]
		if (type(dic["rho0"]) != type(1.0)):
			print("rho0 = "+str(dic["rho0"]))
			exit()
		if (type(dic["nmaxwellians"]) != type(1)):
			print("nmaxwellians = "+str(dic["nmaxwellians"]))
			exit()
		if (type(dic["u"]) != type([1.0])):
			print("u = "+str(dic["u"]))
			exit()
		elif (len(dic["u"]) != dic["nmaxwellians"]):
			print("u = "+str(dic["u"]))
			exit()
		else:
			for u in dic["u"]:
				if (type(u) != type([1.0])):
					print("u = "+str(dic["u"]))
					exit()
				elif (len(u) != dic["ndims"]):
					print("u = "+str(dic["u"]))
					exit()
				else:
					for d in range(dic["ndims"]):
						if (type(u[d]) != type(1.0)):
							print("u = "+str(dic["u"]))
							exit()
		if (type(dic["sigma0"]) != type(1.0)):
			print("sigma0 = "+str(dic["sigma0"]))
			exit()

	elif (newdic["type_initialization"] == "quantum_maxwellian"):
		list_keys = ["rho0", "u0", "e0", "type_qparticles", "hbar"]
		if (type(dic["rho0"]) != type(1.0)):
			print("rho0 = "+str(dic["rho0"]))
			exit()
		if (type(dic["u0"]) != type([1.0])):
			print("u0 = "+str(dic["u0"]))
			exit()
		elif (len(dic["u0"]) != dic["ndims"]):
			print("u0 = "+str(dic["u0"]))
			exit()
		else:
			for d in range(dic["ndims"]):
				if (type(dic["u0"][d]) != type(1.0)):
					print("u0 = "+str(dic["u0"]))
					exit()
		if (type(dic["e0"]) != type(1.0)):
			print("e0 = "+str(dic["e0"]))
			exit()
		if (not(dic["type_qparticles"] in ["fermions", "bosons"])):
			print("type_qparticles = "+str(dic["type_qparticles"]))
			exit()
		if (type(dic["hbar"]) != type(1.0)):
			if (type(dic["hbar"] != type([1.0]))):
				print("hbar = "+str(dic["hbar"]))
				exit()
			else:
				for h in dic["hbar"]:
					if (type(h) != type(1.0)):
						print("hbar = "+str(dic["hbar"]))
						exit()

	elif (newdic["type_initialization"] == "ball_indicator"):
		list_keys = ["rho0", "u0", "e0"]
		if (type(dic["rho0"]) != type(1.0)):
			print("rho0 = "+str(dic["rho0"]))
			exit()
		if (type(dic["u0"]) != type([1.0])):
			print("u0 = "+str(dic["u0"]))
			exit()
		elif (len(dic["u0"]) != dic["ndims"]):
			print("u0 = "+str(dic["u0"]))
			exit()
		else:
			for d in range(dic["ndims"]):
				if (type(dic["u0"][d]) != type(1.0)):
					print("u0 = "+str(dic["u0"]))
					exit()
		if (type(dic["e0"]) != type(1.0)):
			print("e0 = "+str(dic["e0"]))
			exit()
	elif (newdic["type_initialization"] == "BKW"):
		list_keys = ["sigma0"]
		if (type(dic["sigma0"]) != type(1.0)):
			print("sigma0 = "+str(dic["sigma0"]))
			exit()
	
	list_keys += ["L", "Tfinal", "time_method", "Nt"]
	for key in 	list_keys:
		newdic[key] = dic[key]
	
	newdic["type_cutoff"] = "kS"
	newdic["cutoff_radius"] = 1.
	newdic["save_results"] = "all"

	if (newdic["type_collision"] == "quantum"):
		if (("hbarmax" in dic.keys()) and ("percent_hbarmax" in dic.keys())):
			# An analysis of relaxation regarding to the value of hbar is asked
			# It cannot be performed with the quantum maxwellian
			if (dic["type_initialization"] == "quantum_maxwellian"):
				print("You have asked a relaxation analysis regarding the value of hbar. It is a non-sense to run it with an initial data set to the quantum maxwellian.")
				exit()
			look_for_critical = True
			list_hbar = [dic["hbarmax"]*dic["percent_hbarmax"][i] for i in range(len(dic["percent_hbarmax"]))]
		else:
			look_for_critical = False
			list_hbar = dic["hbar"]
	else:
		if (dic["type_initialization"] != "quantum_maxwellian"):
			# Indicate that there is no hbar in the next lines
			list_hbar = [-1.]
	
	list_N = dic["N"]
	list_L = dic["L"]
	list_time_method = dic["time_method"]
	list_residual = dic["use_residual"]
	list_rescaling = dic["rescaling"]

	list_convolution = dic["force_convolution"]
	list_int_degrees = dic["int_degrees"]

	for mylist in [list_N, list_L, list_time_method, list_residual, list_rescaling, list_hbar]:
		if (type(mylist) != type([1.0])):
			print(mylist)
			exit()

	hostname = subprocess.check_output(['uname', '-n'])
	hostname = hostname.decode("utf-8").split('\n')[0]

	oldmessage = ""
	batch_mode = activate_batch_mode()
	if (batch_mode):
		mailuser = ""
		memory = ""
	
	# Environment variables
	hostname = subprocess.check_output(['uname', '-n'])
	hostname = hostname.decode("utf-8").split('\n')[0]
	if (hostname == 'mouton-Latitude-5480'):
		rootdir = "/home/mouton/KINEBEC/code"
	elif (hostname == "mouton-OptiPlex-7050"):
		rootdir = "/home/mouton/KINEBEC/code"
	elif ("mathcalc" in hostname):
		rootdir = "/home/mouton/KINEBEC"
	elif ("zeus" in hostname):
		rootdir = "/home/amouton/KINEBEC"
	
	nameexec_kinebec = rootdir + "/build/"+ parallel + "/bin/kinebec"

	resultsdir = rootdir + "/results/relaxation/"
	if (newdic["type_collision"] == "classical"):
		resultsdir += newdic["type_collision"]+"_"+str(newdic["ndims"])+"d"
	else:
		resultsdir += newdic["type_qparticles"]+"_"+str(newdic["ndims"])+"d"
	
	resultsdir += "/"+newdic["type_initialization"]
	if (subdir != ""):
		if (subdir[0] == '/'):
			resultsdir += subdir
		else:
			resultsdir += "/"+subdir

	list_setups = []
	if (batch_mode):
		list_submit = []
	
	jh = 0
	for hbar in list_hbar:
		if (hbar < 0.):
			texthbar = ""
		else:
			newdic["hbar"] = hbar
			if (look_for_critical):
				texthbar = "hbarmax"+str(dic["percent_hbarmax"][jh])
				jh += 1
			else:
				texthbar = "hbar"+str(hbar)
		
		for time_method in list_time_method:
			newdic["time_method"] = time_method
			texttimemethod = time_method
			
			for N in list_N:
				newdic["Nvx"] = N
				newdic["Nvy"] = N
				if (newdic["ndims"] == 3):
					newdic["Nvz"] = N
				textN = "N"+str(N)
			
				for residual in list_residual:
					newdic["use_residual"] = residual
					
					if (residual):
						textresidual = "residual"
					else:
						textresidual = "noresidual"

				
					for L in list_L:
						newdic["L"] = L
						textL = "L"+str(L)
						
						for rescaling in list_rescaling:
							newdic["use_rescaling"] = rescaling
					
							if (rescaling):
								textrescaling = "rescale"
							else:
								textrescaling = "norescale"
							
							for j in range(len(list_convolution)):
								newdic["force_convolution"] = list_convolution[j]
								if (newdic["force_convolution"]):
									[newdic["P1_int"], newdic["P2_int"], newdic["P3_int"]] = list_int_degrees[j]
									textconvol = "convol"+str(newdic["P1_int"])+"-"+str(newdic["P2_int"])+"-"+str(newdic["P3_int"])
								else:
									if (newdic["ndims"] == 2):
										newdic["M_int"] = list_int_degrees[j]
										textconvol = "noconvol"+str(newdic["M_int"])
									else:
										[newdic["M1_int"], newdic["M2_int"]] = list_int_degrees[j]
										textconvol = "noconvol"+str(newdic["M1_int"])+"-"+str(newdic["M2_int"])
							
								if (hbar < 0.):
									namejob = ""
								else:
									namejob = texthbar+"_"
								namejob += textL+"_"+textrescaling+"_"+textresidual+"_"+texttimemethod+"_"+textconvol+"_"+textN
								nametest = resultsdir+"/" + namejob
								
								namejob = newdic["type_qparticles"]+"_"+str(newdic["ndims"])+"d_"+newdic["type_initialization"]+"_"+namejob
								#print(namejob)

								newdic["outputdir"] = nametest
								if (not os.path.isdir(newdic["outputdir"])):
									cmd = "mkdir -p "+newdic["outputdir"]
									subprocess.call(cmd, shell=True)
								namesetupfile = nametest+".dat"
								write_setupfile(newdic, namesetupfile)
								list_setups.append([namesetupfile, N])
								if (batch_mode):
									namesubmitfile = nametest+"_submit.slurm"
									if (parallel == "cuda"):
										write_script_zeus_slurm(namesetupfile, nameexec_kinebec, nCPUs=(N**dic["ndims"]), parallel=parallel, force_archlimits=force_archlimits, namejob=namejob, memory=memory, mailuser=mailuser, namesubmitfile=namesubmitfile, resume=resume, walltime=walltime, memprofile_freq=memprofile_freq)
									else:
										write_script_zeus_slurm(namesetupfile, nameexec_kinebec, nCPUs=nCPUs, parallel=parallel, force_archlimits=force_archlimits, namejob=namejob, memory=memory, mailuser=mailuser, namesubmitfile=namesubmitfile, resume=resume, walltime=walltime, memprofile_freq=memprofile_freq)
									list_submit.append(namesubmitfile)

	if (run):
		if (batch_mode):
			for submit in list_submit:
				cmd = "sbatch "+submit
				subprocess.call(cmd, shell=True)
		else:
			i = 1
			hostname = subprocess.check_output(['uname', '-n'])
			hostname = hostname.decode("utf-8").split('\n')[0]
			title = "["+hostname+"] Relaxation ("
			if (newdic["type_collision"] == "classical"):
				title += newdic["type_collision"]+" "+str(newdic["ndims"])+"d)"
			else:
				title += newdic["type_qparticles"]+" "+str(newdic["ndims"])+"d)"
			for setupfile in list_setups:
				if ((memprofile_freq > 0.) and (mprof_exec != "")):
					memprofile_logfile = setupfile[0].replace(".dat", "_memprofile.dat")
					mprof_prefix = mprof_exec + " -T "+str(memprofile_freq)+" -o "+memprofile_logfile+" "
				else:
					mprof_prefix = ""
				cmd = mprof_prefix + cmdrun(setupfile[1], nCPUs=nCPUs, parallel=parallel, force_archlimits=force_archlimits) + nameexec_kinebec + " " + setupfile[0]
				if (parallel == "cuda"):
					k = int(np.ceil(np.sqrt(setupfile[1]**newdic["ndims"])))
					cmd += " -nb "+str(k)+" -nt "+str(k)
				
				if (resume):
					cmd += " --resume"
				subprocess.call(cmd, shell=True)
				message = oldmessage+"Job finished on file "+setupfile[0]+" ("+str(i)+" / "+str(len(list_setups))+")"
				oldmessage = send_pushbullet_notification(title, message)
				i += 1
			if (oldmessage != ""):
				send_pushbullet_notification(title, oldmessage, force=True)
	else:
		for setupfile in list_setups:
			if ((memprofile_freq > 0.) and (mprof_exec != "")):
				memprofile_logfile = setupfile[0].replace(".dat", "_memprofile.dat")
				mprof_prefix = mprof_exec + " -T "+str(memprofile_freq)+" -o "+memprofile_logfile+" "
			else:
				mprof_prefix = ""
			cmd = mprof_prefix + cmdrun(setupfile[1], nCPUs=nCPUs, parallel=parallel, force_archlimits=force_archlimits) + nameexec_kinebec + " " + setupfile[0]
			if (parallel == "cuda"):
				k = int(np.ceil(np.sqrt(setupfile[1]**newdic["ndims"])))
				cmd += " -nb "+str(k)+" -nt "+str(k)
			if (resume):
				cmd += " --resume"
			print(cmd)






dic = {}

# Dimensionality (2 or 3)
dic["ndims"] = 2

# Type of particles ("classical", "fermions" or "bosons")
dic["type_particles"] = "fermions"

if ((dic["type_particles"] == "fermions") and (dic["ndims"] == 2)):
	Rmax = 2.
elif ((dic["type_particles"] == "fermions") and (dic["ndims"] == 3)):
	Rmax = 5./3. * np.sqrt(10./np.pi)
elif ((dic["type_particles"] == "bosons") and (dic["ndims"] == 3)):
	Rmax = ((specfuncs.zeta(1.5))**(2.5)) / ((specfuncs.zeta(2.5))**(1.5))
else:
	Rmax = 0.

# List of numbers of points in each direction
dic["N"] = [64]

# Final time
dic["Tfinal"] = 1.

# Number of time iterations
dic["Nt"] = 10

# List of time semi-discretizations ("Euler" and/or "RK2")
dic["time_method"] = ["RK2"]

# List of residual method activations
dic["use_residual"] = [True]

# List of velocity rescaling method activations
dic["rescaling"] = [False]

# Definition of the classical maxwellian
dic["type_initialization"] = "ball_indicator"
dic["rho0"] = 1.
dic["u0"] = [0., 0.]
dic["e0"] = 1.

# List of sizes of the bounding box in each direction
dic["L"] = [6.]

# Activate the discrete convolution method
dic["force_convolution"] = [True]
dic["int_degrees"] = [[16, 2, 2]]

# List of hbar values
if ((dic["type_particles"] == "fermions") or ((dic["type_particles"] == "bosons") and (dic["ndims"] == 3))):
	if (dic["type_initialization"] == "classical_maxwellian"):
		eint = 0.5*dic["ndims"]*dic["sigma0"]
	elif (dic["type_initialization"] == "ball_indicator"):
		eint = dic["e0"]

	hbarmax = (Rmax/dic["rho0"])**(1./dic["ndims"]) * np.sqrt(4.*np.pi*eint/dic["ndims"])

	#dic["percent_hbarmax"] = [0.1, 0.2, 0.5, 0.8, 0.9, 0.95, 0.99, 1., 1.01, 1.05, 1.1, 1.2, 1.5]
	dic["percent_hbarmax"] = [0.1]
	dic["hbarmax"] = hbarmax




test_relaxation(dic, parallel="mpi", nCPUs=2, force_archlimits=False, run=True, resume=False, walltime=[288,0,0], memprofile_freq=0.1)

