#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tests_kinebec import *




def test_BKW(dic, parallel="serial", nCPUs="max", memprofile_freq=-1., force_archlimits=False, run=False):
	newdic = {}
	oldmessage = ""
	
	# Dimension in velocity
	# Allowed values : 2 or 3
	newdic["ndims"] = 2

	# Type of collisions
	# Allowed values : "classical" or "quantum"
	newdic["type_collision"] = "classical"

	# Type of initialization
	# Allowed values : "from_file", "classical_maxwellian", "classical_bimaxwellian"
	#                  "quantum_maxwellian" (only for quantum collisions) or "BKW" (only for 2D classical collisions)
	newdic["type_initialization"] = "BKW"

	# Generate default setup based on the choices above
	newdic = kinebec_default_setup(newdic)

	# Keep only norms
	newdic["save_results"] = dic["save_results"]
	newdic["Nt"] = dic["Nt"]
	newdic["Tfinal"] = dic["Tfinal"]
	newdic["type_cutoff"] = "kS"

	list_N = dic["N"]
	list_time_methods = dic["time_method"]
	list_L = dic["L"]
	list_cutoff = dic["cutoff"]
	list_rescaling = dic["rescaling"]
	list_residual = dic["residual"]

	batch_mode = activate_batch_mode()
	if (batch_mode):
		mailuser = ""
		memory = ""
	
	# Environment variables
	hostname = subprocess.check_output(['uname', '-n'])
	hostname = hostname.decode("utf-8").split('\n')[0]
	if (hostname == 'mouton-Latitude-5480'):
		rootdir = "/home/mouton/KINEBEC/code"
	elif (hostname == "mouton-OptiPlex-7050"):
		rootdir = "/home/mouton/KINEBEC/code"
	elif ("mathcalc" in hostname):
		rootdir = "/home/mouton/KINEBEC"
	elif ("zeus" in hostname):
		rootdir = "/home/amouton/KINEBEC"

	nameexec_kinebec = rootdir + "/build/" + parallel + "/bin/kinebec"
	resultsdir = rootdir + "/results/BKW"

	list_setups = []
	if (batch_mode):
		list_submit = []

	for N in list_N:
		newdic["Nvx"] = N
		newdic["Nvy"] = N
		textN = "N"+str(N)
		
		for cutoff in list_cutoff:
			newdic["cutoff_radius"] = cutoff
			textcutoff = "cutoff"+str(cutoff)+"S"

			for L in list_L:
				newdic["L"] = L
				textL = "L"+str(L)
				
				for rescaling in list_rescaling:
					newdic["use_rescaling"] = rescaling
							
					if (rescaling):
						textrescaling = "rescale"
					else:
						textrescaling = "norescale"
					
					for residual in list_residual:
						dic["use_residual"] = residual
						
						if (residual):
							textresidual = "residual"
						else:
							textresidual = "noresidual"

						for time_method in list_time_methods:
							newdic["time_method"] = time_method
							texttimemethod = time_method
							
							namejob = texttimemethod+"_"+textcutoff+"_"+textL+"_"+textrescaling+"_"+textresidual+"_"+textN
							nametest = resultsdir+"/" + namejob
							
							namejob = "KINEBEC_"+namejob

							newdic["outputdir"] = nametest
							if (not os.path.isdir(newdic["outputdir"])):
								cmd = "mkdir -p "+newdic["outputdir"]
								subprocess.call(cmd, shell=True)
							namesetupfile = nametest+".dat"
							write_setupfile(newdic, namesetupfile)
							list_setups.append([namesetupfile, N])

							if (batch_mode):
								namesubmitfile = nametest+"_submit.slurm"
								if (parallel == "cuda"):
									write_script_zeus_slurm(namesetupfile, nameexec_kinebec, nCPUs=(N**dic["ndims"]), parallel=parallel, force_archlimits=force_archlimits, namejob=namejob, memory=memory, mailuser=mailuser, namesubmitfile=namesubmitfile, resume=resume, walltime=walltime, memprofile_freq=memprofile_freq)
								else:
									write_script_zeus_slurm(namesetupfile, nameexec_kinebec, nCPUs=nCPUs, parallel=parallel, force_archlimits=force_archlimits, namejob=namejob, memory=memory, mailuser=mailuser, namesubmitfile=namesubmitfile, resume=resume, walltime=walltime, memprofile_freq=memprofile_freq)
								list_submit.append(namesubmitfile)

	if (run):
		if (batch_mode):
			for submit in list_submit:
				cmd = "sbatch "+submit
				subprocess.call(cmd, shell=True)
		else:
			i = 1
			hostname = subprocess.check_output(['uname', '-n'])
			hostname = hostname.decode("utf-8").split('\n')[0]
			title = "["+hostname+"] BKW test (classical 2d)"
			for setupfile in list_setups:
				if ((memprofile_freq > 0.) and (mprof_exec != "")):
					memprofile_logfile = setupfile[0].replace(".dat", "_memprofile.dat")
					mprof_prefix = mprof_exec + " -T "+str(memprofile_freq)+" -o "+memprofile_logfile+" "
				else:
					mprof_prefix = ""
				cmd = mprof_prefix + cmdrun(setupfile[1], nCPUs=nCPUs, parallel=parallel, force_archlimits=force_archlimits) + nameexec_kinebec + " " + setupfile[0]
				subprocess.call(cmd, shell=True)
				message = oldmessage+"Job finished on file "+setupfile[0]+" ("+str(i)+" / "+str(len(list_setups))+")"
				oldmessage = send_pushbullet_notification(title, message)
				i += 1
			if (oldmessage != ""):
				send_pushbullet_notification(title, oldmessage, force=True)
				
	else:
		for setupfile in list_setups:
			if ((memprofile_freq > 0.) and (mprof_exec != "")):
				memprofile_logfile = setupfile[0].replace(".dat", "_memprofile.dat")
				mprof_prefix = mprof_exec + " -T "+str(memprofile_freq)+" -o "+memprofile_logfile+" "
			else:
				mprof_prefix = ""
			cmd = mprof_prefix + cmdrun(setupfile[1], nCPUs=nCPUs, parallel=parallel, force_archlimits=force_archlimits) + nameexec_kinebec + " " + setupfile[0]
			if (parallel == "cuda"):
				k = int(np.ceil(np.sqrt(setupfile[1]**newdic["ndims"])))
				cmd += " -nb "+str(k)+" -nt "+str(k)
			if (resume):
				cmd += " --resume"
			print(cmd)



dic = {}

# Save all results or norms only ("only_norms" or "all")
dic["save_results"] = "all"

# Number of time steps
dic["Nt"] = 4000

# Final time
dic["Tfinal"] = 40.

# List of time methods ("Euler" or "RK2")
dic["time_method"] = ["RK2"]

# List of mesh grid sizes
#dic["N"] = [16, 32, 64, 128]
dic["N"] = [64]

# List of bounding box sizes
#dic["L"] = [3., 6.]
dic["L"] = [6.]

# List of cutoff constants
#dic["cutoff"] = [1.,2.]
dic["cutoff"] = [1.]

# Activate the velocity rescaling method
dic["rescaling"] = [True, False]

# Activate the residual method
dic["residual"] = [True, False]

test_BKW(dic, parallel="omp", run=True)
