#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tests_kinebec import *



def test_spectral_accuracy(dic, parallel="serial", nCPUs="max", walltime=[1,0,0], memprofile_freq=-1., force_archlimits=False, run=False):
	newdic = {}
	newdic["ndims"] = dic["ndims"]

	newdic["type_qparticles"] = dic["type_particles"]
	if (dic["type_particles"] == "classical"):
		newdic["type_collision"] = "classical"
	else:
		newdic["type_collision"] = "quantum"
	newdic["type_initialization"] = newdic["type_collision"]+"_maxwellian"

	newdic = kinebec_default_setup(newdic)

	# Restore the type of particles
	newdic["type_qparticles"] = dic["type_particles"]

	# Keep only norms
	newdic["save_results"] = "all"		# "all" or "only_norms"
	newdic["Nt"] = 2
	newdic["Tfinal"] = 1.
	newdic["use_residual"] = False
	newdic["time_method"] = "Euler"
	newdic["rho0"] = dic["rho0"]
	newdic["type_cutoff"] = "kS"

	list_convolution = dic["force_convolution"]
	list_int_degrees = dic["int_degrees"]

	list_N = dic["N"]
	list_L = dic["L"]
	list_cutoff = dic["cutoff"]
	list_sigma0 = dic["sigma0"]
	list_u0 = dic["u0"]
	list_rescaling = dic["rescaling"]
	if (newdic["type_collision"] == "classical"):
		list_hbar = [-1.]
	else:
		list_hbar = dic["hbar"]

	oldmessage = ""
	batch_mode = activate_batch_mode()
	if (batch_mode):
		mailuser = ""
		memory = ""

	# Environment variables
	hostname = subprocess.check_output(['uname', '-n'])
	hostname = hostname.decode("utf-8").split('\n')[0]
	if (hostname == 'mouton-Latitude-5480'):
		rootdir = "/home/mouton/KINEBEC/code"
	elif (hostname == "mouton-OptiPlex-7050"):
		rootdir = "/home/mouton/KINEBEC/code"
	elif ("mathcalc" in hostname):
		rootdir = "/home/mouton/KINEBEC"
	elif ("zeus" in hostname):
		rootdir = "/home/amouton/KINEBEC"
	
	nameexec_kinebec = rootdir + "/build/" + parallel + "/bin/spectral_accuracy"

	nameexec_fugacity = rootdir + "/build/" + parallel + "/bin/compute_fugacity"
	resultsdir = rootdir + "/results/spectral_accuracy/"
	if (newdic["type_collision"] == "classical"):
		resultsdir += newdic["type_collision"]+"_"+str(newdic["ndims"])+"d"
	else:
		resultsdir += newdic["type_qparticles"]+"_"+str(newdic["ndims"])+"d"

	list_setups = []
	if (batch_mode):
		list_submit = []
	
	for hbar in list_hbar:
		if (hbar < 0.):
			texthbar = ""
		else:
			newdic["hbar"] = hbar
			texthbar = "hbar"+str(hbar)
		
		for sigma0 in list_sigma0:
			newdic["sigma0"] = sigma0
			textsigma0 = "sigma"+str(sigma0)
			
			# We compute the internal energy
			if (hbar >= 0.):
				outputfile = rootdir+"/charac_qmaxwellian_"+str(newdic["type_qparticles"])+"_"+str(newdic["ndims"])+"d_"+"rho"+str(newdic["rho0"])+"_sigma"+str(newdic["sigma0"])+"_hbar"+str(newdic["hbar"])
			else:
				outputfile = ""
			
			[newdic["m0"],newdic["e0"],zfug,T] = compute_internal_energy(nameexec_fugacity, newdic["rho0"], newdic["sigma0"], hbar, newdic["type_qparticles"], newdic["ndims"], outputfile)
			"""
			print("m0 = "+str(newdic["m0"]))
			print("e0 = "+str(newdic["e0"]))
			print("z  = "+str(zfug))
			print("T  = "+str(T))
			exit()
			"""
			for N in list_N:
				newdic["Nvx"] = N
				newdic["Nvy"] = N
				if (newdic["ndims"] == 3):
					newdic["Nvz"] = N
				textN = "N"+str(N)
			
				for cutoff in list_cutoff:
					newdic["cutoff_radius"] = cutoff
					textcutoff = "cutoff"+str(cutoff)+"S"
				
					for L in list_L:
						newdic["L"] = L
						textL = "L"+str(L)
						
						if (newdic["m0"] < 0.):
							for u0 in list_u0:
								newdic["u0"] = u0
								if (abs(max(u0)) < 1e-12):
									textu0 = "noshift"
								else:
									textu0 = "shift"
						
								for rescaling in list_rescaling:
									newdic["use_rescaling"] = rescaling
							
									if (rescaling):
										textrescaling = "rescale"
									else:
										textrescaling = "norescale"
									
									for j in range(len(list_convolution)):
										newdic["force_convolution"] = list_convolution[j]
										if (newdic["force_convolution"]):
											[newdic["P1_int"], newdic["P2_int"], newdic["P3_int"]] = list_int_degrees[j]
											textconvol = "convol"+str(newdic["P1_int"])+"-"+str(newdic["P2_int"])+"-"+str(newdic["P3_int"])
										else:
											if (newdic["ndims"] == 2):
												newdic["M_int"] = list_int_degrees[j]
												textconvol = "noconvol"+str(newdic["M_int"])
											else:
												[newdic["M1_int"], newdic["M2_int"]] = list_int_degrees[j]
												textconvol = "noconvol"+str(newdic["M1_int"])+"-"+str(newdic["M2_int"])
									
										if (hbar < 0.):
											namejob = ""
										else:
											namejob = texthbar+"_"
										namejob += textcutoff+"_"+textL+"_"+textrescaling+"_"+textsigma0+"_"+textu0+"_"+textconvol+"_"+textN
										nametest = resultsdir+"/" + namejob
										
										namejob = newdic["type_qparticles"]+"_"+str(newdic["ndims"])+"d_"+namejob

										newdic["outputdir"] = nametest
										if (not os.path.isdir(newdic["outputdir"])):
											cmd = "mkdir -p "+newdic["outputdir"]
											subprocess.call(cmd, shell=True)
										namesetupfile = nametest+".dat"
										write_setupfile(newdic, namesetupfile)
										list_setups.append([namesetupfile, N])
										if (batch_mode):
											namesubmitfile = nametest+"_submit.slurm"
											if (parallel == "cuda"):
												write_script_zeus_slurm(namesetupfile, nameexec_kinebec, nCPUs=(N**dic["ndims"]), parallel=parallel, force_archlimits=force_archlimits, namejob=namejob, memory=memory, mailuser=mailuser, namesubmitfile=namesubmitfile, resume=resume, walltime=walltime, memprofile_freq=memprofile_freq)
											else:
												write_script_zeus_slurm(namesetupfile, nameexec_kinebec, nCPUs=nCPUs, parallel=parallel, force_archlimits=force_archlimits, namejob=namejob, memory=memory, mailuser=mailuser, namesubmitfile=namesubmitfile, resume=resume, walltime=walltime, memprofile_freq=memprofile_freq)
											list_submit.append(namesubmitfile)

	if (run):
		if (batch_mode):
			for submit in list_submit:
				cmd = "sbatch "+submit
				subprocess.call(cmd, shell=True)
		else:
			i = 1
			hostname = subprocess.check_output(['uname', '-n'])
			hostname = hostname.decode("utf-8").split('\n')[0]
			title = "["+hostname+"] Spectral accuracy ("
			if (newdic["type_collision"] == "classical"):
				title += newdic["type_collision"]+" "+str(newdic["ndims"])+"d)"
			else:
				title += newdic["type_qparticles"]+" "+str(newdic["ndims"])+"d)"
			for setupfile in list_setups:
				if ((memprofile_freq > 0.) and (mprof_exec != "")):
					memprofile_logfile = setupfile[0].replace(".dat", "_memprofile.dat")
					mprof_prefix = mprof_exec + " -T "+str(memprofile_freq)+" -o "+memprofile_logfile+" "
				else:
					mprof_prefix = ""
				cmd = mprof_prefix + cmdrun(setupfile[1], nCPUs=nCPUs, parallel=parallel, force_archlimits=force_archlimits) + nameexec_kinebec + " " + setupfile[0]
				subprocess.call(cmd, shell=True)
				message = oldmessage+"Job finished on file "+setupfile[0]+" ("+str(i)+" / "+str(len(list_setups))+")"
				oldmessage = send_pushbullet_notification(title, message)
				i += 1
			if (oldmessage != ""):
				send_pushbullet_notification(title, oldmessage, force=True)
				
	else:
		for setupfile in list_setups:
			if ((memprofile_freq > 0.) and (mprof_exec != "")):
				memprofile_logfile = setupfile[0].replace(".dat", "_memprofile.dat")
				mprof_prefix = mprof_exec + " -T "+str(memprofile_freq)+" -o "+memprofile_logfile+" "
			else:
				mprof_prefix = ""
			cmd = mprof_prefix + cmdrun(setupfile[1], nCPUs=nCPUs, parallel=parallel, force_archlimits=force_archlimits) + nameexec_kinebec + " " + setupfile[0]
			if (parallel == "cuda"):
				k = int(np.ceil(np.sqrt(setupfile[1]**newdic["ndims"])))
				cmd += " -nb "+str(k)+" -nt "+str(k)
			if (resume):
				cmd += " --resume"
			print(cmd)


dic = {}

# Dimensionality (2 or 3)
dic["ndims"] = 2

# Type of particles ("classical", "fermions" or "bosons")
dic["type_particles"] = "fermions"

dic["rho0"] = 1.

# List of hbar values (only for non-classical particles)
dic["hbar"] = [0.1, 3.]

# List of sizes of the bounding box in each direction
dic["L"] = [2., 4., 6., 8., 10.]

# List of values of k in the cutoff definition R = k*S
dic["cutoff"] = [1.]

# List of thermal energy values
dic["sigma0"] = [0.5, 1., 2.]

# List of velocity rescaling method activations
dic["rescaling"] = [True, False]

# List of averaged velocity values
# Truncation radius of initial rescaled distribution
if (dic["ndims"] == 3):
	S = 2.*np.pi/(3.+np.sqrt(2.))
	Sv = 4.*S/np.pi
	ub = [Sv/2., Sv/2., Sv/np.sqrt(2.)]
	u0 = [0., 0., 0.]
else:
	S = 2.*np.pi/(3.+np.sqrt(2.))
	Sv = 4.*S/np.pi
	ub = [Sv/np.sqrt(2.), Sv/np.sqrt(2.)]
	u0 = [0.,0.]


dic["u0"] = [u0, ub]


dic["force_convolution"] = [True]
dic["int_degrees"] = [[16, 2, 2]]

# List of numbers of points in each direction
dic["N"] = [16]

test_spectral_accuracy(dic, parallel="mpi", nCPUs=8, force_archlimits=False, walltime=[1,0,0], run=True)

# List of numbers of points in each direction
dic["N"] = [32, 64, 128, 256]

test_spectral_accuracy(dic, parallel="mpi", nCPUs=16, force_archlimits=False, walltime=[1,0,0], run=True)

dic["type_particles"] = "bosons"

# List of numbers of points in each direction
dic["N"] = [16]

test_spectral_accuracy(dic, parallel="mpi", nCPUs=8, force_archlimits=False, walltime=[1,0,0], run=True)

# List of numbers of points in each direction
dic["N"] = [32, 64, 128, 256]

test_spectral_accuracy(dic, parallel="mpi", nCPUs=16, force_archlimits=False, walltime=[1,0,0], run=True)
