#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tests_kinebec import *


def test_speedup(dic, parallel="serial", memprofile_freq=-1., walltime=[1,0,0], force_archlimits=False, run=False):
	newdic = {}
	oldmessage = ""
	# Dimension in velocity
	# Allowed values : 2 or 3
	newdic["ndims"] = dic["ndims"]

	# Type of collisions
	# Allowed values : "classical" or "quantum"
	newdic["type_collision"] = "quantum"

	# Type of initialization
	# Allowed values : "from_file", "classical_maxwellian", "classical_bimaxwellian"
	#                  "quantum_maxwellian" (only for quantum collisions) or "BKW" (only for 2D classical collisions)
	newdic["type_initialization"] = "quantum_maxwellian"


	# Generate default setup based on the choices above
	newdic = kinebec_default_setup(newdic)
	
	# Keep only norms
	newdic["save_results"] = "only_norms"		# "all" or "only_norms"
	newdic["Nt"] = 2
	newdic["Tfinal"] = 1.
	newdic["use_residual"] = False
	newdic["time_method"] = "Euler"
	newdic["type_qparticles"] = "fermions"
	newdic["rho0"] = 1.

	list_N = dic["N"]
	list_nCPUs = dic["nCPUs"]
	newdic["L"] = 6.
	newdic["type_cutoff"] = "kS"
	newdic["cutoff_radius"] = 1.
	newdic["sigma0"] = 1.
	if (newdic["ndims"] == 2):
		newdic["u0"] = [0.,0.]
	elif (newdic["ndims"] == 3):
		newdic["u0"] = [0.,0.,0.]
	newdic["hbar"] = 0.1
	newdic["use_rescaling"] = False

	list_convolution = dic["force_convolution"]
	list_int_degrees = dic["int_degrees"]
	
	list_setups = []
	list_mpiruns = []

	batch_mode = activate_batch_mode()
	if (batch_mode):
		mailuser = ""
		memory = ""

	if (batch_mode):
		list_submit = []

	# Environment variables
	hostname = subprocess.check_output(['uname', '-n'])
	hostname = hostname.decode("utf-8").split('\n')[0]
	if (hostname == 'mouton-Latitude-5480'):
		rootdir = "/home/mouton/KINEBEC/code"
	elif (hostname == "mouton-OptiPlex-7050"):
		rootdir = "/home/mouton/KINEBEC/code"
	elif ("mathcalc" in hostname):
		rootdir = "/home/mouton/KINEBEC"
	elif ("zeus" in hostname):
		rootdir = "/home/amouton/KINEBEC"
	
	nameexec_kinebec = rootdir + "/build/"+ parallel + "/bin/spectral_accuracy"

	if (not("zeus" in hostname)):
		# Get the max number of logical processors
		nprocs_max = psutil.cpu_count(logical=False)

		# Get the max number of threads
		nthreads_max = psutil.cpu_count()
	else:
		nprocs_max = 64
		nthreads_max = 64

	if (parallel == "mpi"):
		nprocesses_max = nprocs_max
	elif (parallel == "omp"):
		nprocesses_max = nthreads_max
	else:
		nprocesses_max = 1

	nameexec_kinebec = rootdir + "/build/" + parallel + "/bin/spectral_accuracy"

	nameexec_fugacity = rootdir + "/build/" + parallel + "/bin/compute_fugacity"
	resultsdir = rootdir + "/results/speedup/fermions_"+str(newdic["ndims"])+"d"
	
	outputfile = rootdir+"/charac_qmaxwellian_"+str(newdic["type_qparticles"])+"_"+str(newdic["ndims"])+"d"
	[newdic["m0"],newdic["e0"],zfug,T] = compute_internal_energy(nameexec_fugacity, newdic["rho0"], newdic["sigma0"], newdic["hbar"], newdic["type_qparticles"], newdic["ndims"], outputfile)
	

	for j in range(len(list_convolution)):
		newdic["force_convolution"] = list_convolution[j]
		if (newdic["force_convolution"]):
			[newdic["P1_int"], newdic["P2_int"], newdic["P3_int"]] = list_int_degrees[j]
			textconvol = "convol"+str(newdic["P1_int"])+"-"+str(newdic["P2_int"])+"-"+str(newdic["P3_int"])
		else:
			if (newdic["ndims"] == 2):
				newdic["M_int"] = list_int_degrees[j]
				textconvol = "noconvol"+str(newdic["M_int"])
			else:
				[newdic["M1_int"], newdic["M2_int"]] = list_int_degrees[j]
				textconvol = "noconvol"+str(newdic["M1_int"])+"-"+str(newdic["M2_int"])
		
		for N in list_N:
			newdic["Nvx"] = N
			newdic["Nvy"] = N
			if (newdic["ndims"] == 3):
				newdic["Nvz"] = N
			textN = "N"+str(N)

			nprocesses_opt = max([N//2, 1])
			
			for nCPUs in list_nCPUs:
				if (((force_archlimits) and (nCPUs <= nprocesses_opt)) or (not(force_archlimits) and (nCPUs <= nprocesses_max))):
					textnCPUs = "nCPUs"+str(nCPUs)

					namejob = parallel + "_" + textconvol + "_" + textN + "_" + textnCPUs
					nametest = resultsdir+"/" + namejob
					
					namejob = newdic["type_qparticles"]+"_"+str(newdic["ndims"])+"d_"+namejob

					newdic["outputdir"] = nametest
					if (not os.path.isdir(newdic["outputdir"])):
						cmd = "mkdir -p "+newdic["outputdir"]
						subprocess.call(cmd, shell=True)
					namesetupfile = nametest+".dat"
					write_setupfile(newdic, namesetupfile)
					list_setups.append([namesetupfile, N])
					list_mpiruns.append(nCPUs)

					if (batch_mode):
						namesubmitfile = nametest+"_submit.slurm"
						if (parallel == "cuda"):
							write_script_zeus_slurm(namesetupfile, nameexec_kinebec, nCPUs=(N**dic["ndims"]), parallel=parallel, force_archlimits=force_archlimits, namejob=namejob, memory=memory, mailuser=mailuser, namesubmitfile=namesubmitfile, walltime=walltime, memprofile_freq=memprofile_freq)
						else:
							write_script_zeus_slurm(namesetupfile, nameexec_kinebec, nCPUs=nCPUs, parallel=parallel, force_archlimits=force_archlimits, namejob=namejob, memory=memory, mailuser=mailuser, namesubmitfile=namesubmitfile, walltime=walltime, memprofile_freq=memprofile_freq)
						list_submit.append(namesubmitfile)

	if (run):
		if (batch_mode):
			for submit in list_submit:
				cmd = "sbatch "+submit
				subprocess.call(cmd, shell=True)
		else:
			i = 1
			hostname = subprocess.check_output(['uname', '-n'])
			hostname = hostname.decode("utf-8").split('\n')[0]
			title = "["+hostname+"] Speedup (fermions "+str(newdic["ndims"])+"d)"
			for setupfile in list_setups:
				if ((memprofile_freq > 0.) and (mprof_exec != "")):
					memprofile_logfile = setupfile[0].replace(".dat", "_memprofile.dat")
					mprof_prefix = mprof_exec + " -T "+str(memprofile_freq)+" -o "+memprofile_logfile+" "
				else:
					mprof_prefix = ""

				cmdtorun = cmdrun(setupfile[1], nCPUs=list_mpiruns[i-1], parallel=parallel, force_archlimits=force_archlimits)
				if (parallel == "omp"):
					cmd = cmdtorun + mprof_prefix
				else:
					cmd = mprof_prefix + cmdtorun
				cmd += nameexec_kinebec + " " + setupfile[0]
				
				subprocess.call(cmd, shell=True)
				message = oldmessage+"Job finished on file "+setupfile[0]+" ("+str(i)+" / "+str(len(list_setups))+")"
				oldmessage = send_pushbullet_notification(title, message)
				i += 1
			if (oldmessage != ""):
				send_pushbullet_notification(title, oldmessage, force=True)

	else:
		i = 1
		buf = ""
		for setupfile in list_setups:
			if ((memprofile_freq > 0.) and (mprof_exec != "")):
				memprofile_logfile = setupfile[0].replace(".dat", "_memprofile.dat")
				mprof_prefix = mprof_exec + " -T "+str(memprofile_freq)+" -o "+memprofile_logfile+" "
			else:
				mprof_prefix = ""
			cmd = mprof_prefix + cmdrun(setupfile[1], nCPUs=list_mpiruns[i-1], parallel=parallel, force_archlimits=force_archlimits) + nameexec_kinebec + " " + setupfile[0]
			if (parallel == "cuda"):
				k = int(np.ceil(np.sqrt(setupfile[1]**newdic["ndims"])))
				cmd += " -nb "+str(k)+" -nt "+str(k)
			buf += cmd+"\n"
			i += 1
		print(buf)




dic = {}

dic["ndims"] = 3


if (hostname == 'mouton-Latitude-5480'):
	rootdir = "/home/mouton/KINEBEC/code"
	dic["nCPUs"] = [1, 2, 4]
elif (hostname == "mouton-OptiPlex-7050"):
	rootdir = "/home/mouton/KINEBEC/code"
	dic["nCPUs"] = [1]
elif ("mathcalc" in hostname):
	rootdir = "/home/mouton/KINEBEC"
	dic["nCPUs"] = [1, 2, 4, 8, 16]
elif ("zeus" in hostname):
	rootdir = "/home/amouton/KINEBEC"
	dic["nCPUs"] = [1, 2, 4, 8, 16, 32, 64]


parallel = "mpi"

if (dic["ndims"] == 2):
	dic["N"] = [8, 16, 32, 64, 128, 256]
	dic["force_convolution"] = [True, False]
	dic["int_degrees"] = [[16,2,2], 4]
else:
	dic["N"] = [8, 16, 32, 64, 128, 256]
	dic["force_convolution"] = [True, False]
	dic["int_degrees"] = [[16,2,2], [4,4]]

test_speedup(dic, parallel="mpi", force_archlimits=False, run=False, memprofile_freq=0.1)
