#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tests_kinebec import *
from visualization import *


def visu_relaxation_comparison(resultsdirs, id_diag, legends, title, ylabel, outputpng, xlim=None, yscale="linear", linecolors=None, linestyles=None):
	list_linecolors = ["blue", "green", "red", "magenta", "cyan", "yellow", "black"]
	list_linestyles = ["solid", "dashed", "dashdot", "dashdotdotted"]

	mintab = []
	maxtab = []
	lmax = 0

	xlabel = "Time"
	
	mintab = []
	maxtab = []

	for isave in range(2):
		x_list = []
		y_list = []
		thelinestyles = []
		thelinecolors = []
		labels = []
	
		i = 0
		j = 0
		for resultsdir in resultsdirs:
			listfiles = glob.glob(resultsdir+"/*_"+id_diag+".txt")
			if (len(listfiles) > 1):
				if (id_diag != "entropy"):
					print("Error: "+resultsdir+" contains several files with name *_"+id_diag+".txt")
					for namefile in listfiles:
						print("  "+namefile)
					exit()
				else:
					for namefile in listfiles:
						if ("rel_entropy.txt" in namefile):
							listfiles.remove(namefile)
			elif (len(listfiles) < 1):
				print("Error: could not find any file "+resultsdir+"/*_"+id_diag+".txt")
				exit()

			f = open(listfiles[0], "r")
			lines = f.readlines()
			x = []
			y = []
			for line in lines:
				data = line.split(" ")
				time = float(data[0].replace(",", ""))
				val = float(data[1].replace(",", ""))
				if ((xlim != None) and (len(xlim) == 2)):
					if ((time >= min(xlim)) and (time <= max(xlim))):
						x.append(time)
						y.append(val)
			f.close()
			x_list.append(x)
			y_list.append(y)

			mintab.append(min(y))
			maxtab.append(max(y))

			if ((linecolors != None) and (linestyles != None)) and (((len(linecolors) == len(resultsdirs)) and (len(linestyles) == len(resultsdirs)))):
				thelinecolor = linecolors[i]
				thelinestyle = linestyles[i]
				i += 1
			else:
				thelinecolor = list_linecolors[i]
				thelinestyle = list_linestyles[j]
				i += 1
				if (i == len(list_linecolors)):
					i = 0
					j += 1
				if (j == len(list_linestyles)):
					j = 0
					i = 0
			thelinecolors.append(thelinecolor)
			thelinestyles.append(thelinestyle)

		rab = 0.#0.1*(max(maxtab)-min(mintab))
		ylim = [min(mintab)-rab, max(maxtab)+rab]

		plot_curves(x_list, y_list, xlabel, ylabel, title, outputpng, labels="", linecolors=thelinecolors, linestyles=thelinestyles, yscale=yscale, ylim=ylim)



def visu_relaxation_curves(resultsdir, id_diag, label_diag, yscale="linear"):

	list_linecolors = ["blue", "green", "red", "magenta", "cyan", "yellow", "black"]
	list_linestyles = ["solid", "dashed", "dashdot", "dashdotdotted"]
	
	list_rescale = [True, False]
	list_residual = [True, False]
	
	mintab = []
	maxtab = []
	lmax = 0
	ylim = ""
	
	xlabel = "Time"
	ylabel = label_diag
	
	for isave in range(2):
		title = label_diag
		
		x_list = []
		y_list = []
		linestyles = []
		linecolors = []
		labels = []
		
		outputpng = resultsdir+"/"+id_diag+".png"

		i_0 = 0
		for rescale in list_rescale:
			if (rescale):
				text_rescale = "rescale"
				label_rescale = "with rescaling"
			else:
				text_rescale = "norescale"
				label_rescale = "without rescaling"
			
			thelinecolor = list_linecolors[i_0]
			i_0 += 1
			
			i_1 = 0
			for residual in list_residual:
				if (residual):
					text_residual = "residual"
					label_residual = "with residual"
				else:
					text_residual = "noresidual"
					label_residual = "without residual"
				
				thelinestyle = list_linestyles[i_1]
				i_1 += 1
				
				namejob = text_rescale+"_"+text_residual
				namedir = resultsdir+"/"+namejob
				filetoopen = namedir+"/"+id_diag+".txt"
				
				thelabel = label_rescale+", "+label_residual
				lmax = max([lmax, len(thelabel)])
				
				f = open(filetoopen, "r")
				lines = f.readlines()
				x = []
				y = []
				for line in lines:
					data = line.split(" ")
					time = float(data[0])
					if (("entropy" in id_diag) and (not("rel_entropy" in id_diag))):
						val = -float(data[-1])
					else:
						val = float(data[-1])
					x.append(time)
					y.append(val)
				f.close()
				
				x_list.append(x)
				y_list.append(y)
				labels.append(thelabel)
				
				linecolors.append(thelinecolor)
				linestyles.append(thelinestyle)
		
		# At this point, we have loaded the data corresponding to the full set of curves to be plotted
		[theminval, themaxval] = plot_curves(x_list, y_list, xlabel, ylabel, title, outputpng, labels=labels, linecolors=linecolors, linestyles=linestyles, yscale=yscale)



def plot_time_diags(type_qparticles, ndim):
	batch_mode = activate_batch_mode()
	if (batch_mode):
		mailuser = ""
		memory = ""
		rootdir = "/home/amouton/KINEBEC"
	else:
		rootdir = "/home/mouton/KINEBEC/code"

	resultsdir = rootdir+"/results/relaxation/"+type_qparticles+"_"+str(ndim)+"d"

	dic_diags = {}
	dic_diags["rel_entropy"] = ["Relative entropy", "log"]
	dic_diags["mass"] = ["Mass", "linear"]
	dic_diags["ux"] = ["Average x-velocity", "linear"]
	dic_diags["uy"] = ["Average y-velocity", "linear"]
	if (ndim == 3):
		dic_diags["uz"] = ["Average z-velocity", "linear"]
	dic_diags["ekin"] = ["Kinetic energy", "linear"]
	dic_diags["Linfnorm"] = ["Linf norm in space", "log"]
	dic_diags["entropy"] = ["Entropy", "log"]
	dic_diags["L1normrelaxation"] = ["Relaxation error in L1-norm", "log"]
	dic_diags["temperature"] = ["Temperature", "linear"]

	for key in dic_diags.keys():
		id_diag = "diags"+"_"+key
		visu_relaxation_curves(resultsdir, id_diag, dic_diags[key][0], yscale=dic_diags[key][1])


def plot_dynamics_2d(resultsdir, prefixfile, indices, ndigits, method):

	minfile = resultsdir+"/"+prefixfile+"_minf.txt"
	f = open(minfile, "r")
	lines = f.readlines()
	minf = float(lines[0].split(" ")[-1])
	for line in lines:
		minf = min([minf, float(line.split(" ")[-1])])
	f.close()

	maxfile = resultsdir+"/"+prefixfile+"_maxf.txt"
	f = open(maxfile, "r")
	lines = f.readlines()
	maxf = float(lines[0].split(" ")[-1])
	for line in lines:
		maxf = max([maxf, float(line.split(" ")[-1])])
	f.close()

	listpngs = []
	
	if (method in ["surface", "surface_clim", "color", "color_clim"]):
		outputmovie = resultsdir+"/"+prefixfile+"_"+method+".mp4"
		outputdir = resultsdir+"/"+method
		if (not(os.path.isdir(outputdir))):
			os.mkdir(outputdir)
	else:
		projection = method[2]
		if (projection == "x"):
			outputmovie = resultsdir+"/"+prefixfile+"_slicex.mp4"
			outputdir = resultsdir+"/slicex"
			if (not(os.path.isdir(outputdir))):
				os.mkdir(outputdir)
		elif (projection == "y"):
			outputmovie = resultsdir+"/"+prefixfile+"_slicey.mp4"
			outputdir = resultsdir+"/slicey"
			if (not(os.path.isdir(outputdir))):
				os.mkdir(outputdir)
	
	tmppath = resultsdir+"/tmp"
	
	for i in indices:
		inputh5 = resultsdir+"/"+prefixfile+"_"+str(i).rjust(ndigits, '0')+".h5"
		time = extract_from_h5(inputh5, "/distribution/Time")[0]
		distrib = extract_from_h5(inputh5, "/distribution/f")
		title = "Distribution at time t = {:.5e}".format(time)
	
		L = extract_from_h5(inputh5, "/distribution/L")[0]
		type_omega = extract_from_h5(inputh5, "/distribution/type_omega")[0]
		if (type_omega != 1):
			omega = extract_from_h5(inputh5, "/distribution/omega")[0]
			ux = extract_from_h5(inputh5, "/distribution/Average_ux")[0]
			uy = extract_from_h5(inputh5, "/distribution/Average_uy")[0]
		else:
			omega = 1.
			ux = 0.
			uy = 0.
		xlim = [ux-L/omega, ux+L/omega]
		ylim = [uy-L/omega, uy+L/omega]

		if (method in ["surface", "surface_clim", "color", "color_clim"]):
			outputpng = outputdir+"/"+prefixfile+"_"+str(i).rjust(ndigits, '0')+"_"+method+".png"
		else:
			projection = method[2]
			if (projection == "x"):
				outputpng = outputdir+"/"+prefixfile+"_"+str(i).rjust(ndigits, '0')+"_slicex.png"
			elif (projection == "y"):
				outputpng = outputdir+"/"+prefixfile+"_"+str(i).rjust(ndigits, '0')+"_slicey.png"
		
		if (method == "surface_clim"):
			plot_surface_2d(distrib, title, xlim=xlim, ylim=ylim, minmax=[minf,maxf], outputpng=outputpng)
			listpngs.append(outputpng)
		elif (method == "color_clim"):
			plot_pseudocolor_2d(distrib, title, xlim=xlim, ylim=ylim, minmax=[minf,maxf], outputpng=outputpng, pngheightpx=600, pngwidthpx=800)
			listpngs.append(outputpng)
		elif (method == "surface"):
			plot_surface_2d(distrib, title, xlim=xlim, ylim=ylim, outputpng=outputpng)
			listpngs.append(outputpng)
		elif (method == "color"):
			plot_pseudocolor_2d(distrib, title, xlim=xlim, ylim=ylim, outputpng=outputpng, pngheightpx=600, pngwidthpx=800)
			listpngs.append(outputpng)
		else:
			p1 = method[0]
			p2 = method[1]
			projection = method[2]
			title = method[3]+" at time t =  {:.5e}".format(time)
			ylabel ="distribution"
			if ((projection == "x") or (projection == "both")):
				outputpng = outputdir+"/"+prefixfile+"_"+str(i).rjust(ndigits, '0')+"_"+"slice_x.png"
				plot_slice_curve(inputh5, p1, p2, projection, ylabel=ylabel, title=title, outputpng=outputpng, pngheightpx=600, pngwidthpx=800)
				listpngs.append(outputpng)
			if ((projection == "y") or (projection == "both")):
				outputpng = outputdir+"/"+prefixfile+"_"+str(i).rjust(ndigits, '0')+"_"+"slice_y.png"
				plot_slice_curve(inputh5, p1, p2, projection, ylabel=ylabel, title=title, outputpng=outputpng, pngheightpx=600, pngwidthpx=800)
				listpngs.append(outputpng)
				
		print("File "+inputh5+" done")
	
	encode_video(listpngs, tmppath, outputmovie, slowinglevel=1., forcesave=True)




nstart = 0
nend = 500
list_iso = []
list_vol = []
for i in range(nstart, nend+1):
	inputh5 = "/home/mouton/KINEBEC/code/results/diags_"+str(i).rjust(4,'0')+".h5"
	distrib = extract_from_h5(inputh5, "/distribution/f")
	xlim = [-6.,6.]
	ylim = xlim
	zlim = xlim
	isovalues = [np.max(distrib)*0.05, np.max(distrib)*0.6]
	#isovalues = [np.max(distrib)*0.05]
	xlabel = "vx"
	ylabel = "vy"
	zlabel = "vz"
	time = extract_from_h5(inputh5, "/distribution/Time")[0]
	title = "distribution at t = "+str(time)
	png_iso = "/home/mouton/KINEBEC/code/results/diags_iso_"+str(i).rjust(4,'0')+".png"
	list_iso.append(png_iso)
	#plot_isosurfaces_3d(distrib, isovalues, title=title, xlim=xlim, ylim=xlim, zlim=zlim, xlabel=xlabel, ylabel=ylabel, zlabel=zlabel, zoom_factor=0.9, outputpng=png_iso, pngheight=600, pngwidth=800)
	png_vol = "/home/mouton/KINEBEC/code/results/diags_vol_"+str(i).rjust(4,'0')+".png"
	list_vol.append(png_vol)
	#plot_volume_3d(distrib, title=title, xlim=xlim, ylim=xlim, zlim=zlim, xlabel=xlabel, ylabel=ylabel, zlabel=zlabel, zoom_factor=0.9, outputpng=png_vol, pngheight=600, pngwidth=800)

tmppath = "/home/mouton/KINEBEC/code/results/tmp"
outputmovie = "/home/mouton/KINEBEC/code/results/diags_iso.mp4"
encode_video(list_iso, tmppath, outputmovie, slowinglevel=1., forcesave=True)
outputmovie = "/home/mouton/KINEBEC/code/results/diags_vol.mp4"
encode_video(list_vol, tmppath, outputmovie, slowinglevel=1., forcesave=True)
exit()



	
"""
hbartab = [0.95, 0.99, 1., 1.01, 1.05, 1.1, 1.2, 1.5]#[0.1, 0.2, 0.5, 0.8, 0.9, 0.95, 0.99, 1., 1.01, 1.05, 1.1, 1.2, 1.5]
testcasetab = ["ball_indicator", "classical_maxwellian"]
tmax_relax = [30.]*2
tmax_ball = tmax_relax + [23.9325, 16.3525, 7.19, 4., 2.0525, 0.735]
tmax_maxwell = tmax_relax + [30., 30., 13.685, 7.94, 3.5525, 1.0625]
dt = 30./12000.
nt = {}
nt["ball_indicator"] = [round(t/dt) for t in tmax_ball]
nt["classical_maxwellian"] = [round(t/dt) for t in tmax_maxwell]


for j in range(len(hbartab)):
	hbar = hbartab[j]
	for testcase in testcasetab:
		resultsdir = '/home/mouton/KINEBEC/code/results/relaxation/fermions_2d/'+testcase+'/hbarmax'+str(hbar)+'_L8.0_norescale_residual_RK2_N128'
		prefixfile = 'diags'
		ndiags = nt[testcase][j]
		ndigits = 5

		namefile = resultsdir+'/'+prefixfile+'_entropy.txt'
		f = open(namefile, 'r')
		time = []
		data = []
		for line in f.readlines():
			if (len(time) > ndiags):
				break
			[t, d] = line.split(' ')
			t = float(t)
			d = float(d.replace('\n', ''))
			time.append(t)
			data.append(d)
		f.close()

		outputdir = resultsdir+"/entropy"
		#list_output = []
		if (not(os.path.isdir(outputdir))):
			os.mkdir(outputdir)
		for i in range(ndiags):
			outputpng = outputdir+"/"+prefixfile+"_"+str(i).rjust(ndigits, '0')+"_entropy.png"
			plot_curves([time[0:(i+1)]], [data[0:(i+1)]], "time", "Entropy", "Evolution of entropy", outputpng, linewidth=2, yscale="linear", xlim=[min(time), max(time)], pngheightpx=600, pngwidthpx=800)
			print(outputpng+" done")
			#list_output.append(outputpng)
	#x = input("finished "+testcase+" with hbar = "+str(hbar)+"*hbarmax")
		
"""




id_diag = "maxf"
ylabel = 'Max(f)'
title = 'Max(f)'
hbartab = [0.1, 0.2, 0.5, 0.8, 0.9, 0.95, 0.99, 1.,  1.01, 1.05, 1.1]

#tmax =    [30., 30., 30., 30., 30., 30.,  30.,  30., 30.,  13.5, 7.8]
#tmax =    [30., 30., 30., 30., 30., 30.,  30.,  23.9, 16.1,  7.1,   3.9]
#tslice =  [30., 30., 30., 30., 30., 30.,  30.,  30,  30.,  9.,  4.5]
#inddiag = [round(t/0.0025) for t in tslice]

tmax =    [10., 10., 10., 10., 10., 10.,  2.57,  10., 10.,  10.,  10.]

#inddiag = [round(t/0.01) for t in tslice]
for k in range(len(hbartab)):
	hbar = hbartab[k]
	xlim = [0., tmax[k]]
	resultsdirs = []
	legends = []
	testcase = "classical_maxwellian"
	hbarmax_percents = [hbar]
	for hbarmax_percent in hbarmax_percents:
		rootdir = "./results/relaxation/fermions_3d/"+testcase+"/hbarmax"
		resultdir = rootdir + str(hbarmax_percent)+"_L6.0_norescale_residual_RK2_N32"
		resultsdirs.append(resultdir)
		legend = r'$\hbar = '+str(hbarmax_percent)+r'\, \hbar^{*}$'
		legends.append(legend)
	outputfile = "./"+testcase+"_hbarmax"+str(hbar)+"_"+id_diag+".png"
	#visu_relaxation_comparison(resultsdirs, id_diag, legends, title, ylabel, outputfile, xlim=xlim, yscale="linear", linecolors=["black"], linestyles=["solid"])
	"""
	inputh5 = resultdir+"/diags_"+str(inddiag[k]).rjust(5, '0')+".h5"
	p1 = [-8.,0.]
	p2 = [8., 0.]
	outputpng = outputfile = "./"+testcase+"_hbarmax"+str(hbar)+"_slicex_t"+str(tslice[k])+".png"
	title_slice = "f(vx,0) at time t = "+str(tslice[k])
	plot_slice_curve(inputh5, p1, p2, "vx", linewidth=2, ylabel="", title=title_slice, outputpng=outputpng)
	"""
"""
inputh5 = "/home/mouton/KINEBEC/code/results/spectral_accuracy/bosons_2d/hbar3.0_cutoff1.0S_L6.0_norescale_sigma0.5_noshift_N128/diags.h5"
p1 = [-6., 0.]
p2 = [6., 0.]
title_slice = "f(vx,0)"
outputpng = "bosons_2d_L6_sigma05.png"
plot_slice_curve(inputh5, p1, p2, "vx", linewidth=2, ylabel="", title=title_slice, outputpng=outputpng)

inputh5 = "/home/mouton/KINEBEC/code/results/spectral_accuracy/bosons_2d/hbar3.0_cutoff1.0S_L6.0_norescale_sigma1.0_noshift_N128/diags.h5"
p1 = [-6., 0.]
p2 = [6., 0.]
title_slice = "f(vx,0)"
outputpng = "bosons_2d_L6_sigma1.png"
plot_slice_curve(inputh5, p1, p2, "vx", linewidth=2, ylabel="", title=title_slice, outputpng=outputpng)
"""

"""
ndiags = 2000
ndigits = 4




hbartab = [0.99, 1.01]

for hbar in hbartab:
	resultsdir = "./results/relaxation/fermions_2d/hbarmax"+str(hbar)+"_L8.0_norescale_residual_RK2_N128"
	prefixfile = "diags"
	ndiags = 2000
	ndigits = 4
	indices = list(np.arange(0,ndiags+1))

	method = [[-8.,0.], [8.,0.], "x", "f(x,0)"]
	plot_dynamics_2d(resultsdir, prefixfile, indices, ndigits, method)
	method = "color"
	plot_dynamics_2d(resultsdir, prefixfile, indices, ndigits, method)

	namefile = resultsdir+"/"+prefixfile+"_entropy.txt"
	f = open(namefile, 'r')
	time = []
	data = []
	for line in f.readlines():
		if (len(time) > ndiags):
			break
		[t, d] = line.split(" ")
		t = float(t)
		d = float(d.replace('\n', ''))
		time.append(t)
		data.append(d)
	f.close()

	outputdir = resultsdir+"/entropy"
	list_output = []
	if (not(os.path.isdir(outputdir))):
		os.mkdir(outputdir)
	for i in range(ndiags+1):
		outputpng = outputdir+"/"+prefixfile+"_"+str(i).rjust(ndigits, '0')+"_entropy.png"
		plot_curves([time[0:(i+1)]], [data[0:(i+1)]], "time", "Entropy", "Evolution of entropy", outputpng, linewidth=2, yscale="linear", xlim=[min(time), max(time)], pngheightpx=600, pngwidthpx=800)
		list_output.append(outputpng)

	tmppath = resultsdir+"/tmp"
	outputmovie = resultsdir+"/diags_entropy.mp4"
	encode_video(list_output, tmppath, outputmovie, forcesave=True)


"""




"""
list_output = []
outputdir = "./results/relaxation/fermions_2d/compare_0.99_1.01"
if (not(os.path.isdir(outputdir))):
	os.mkdir(outputdir)
for i in range(ndiags+1):
	hbar = 0.99
	resultsdir = "./results/relaxation/fermions_2d/hbarmax"+str(hbar)+"_L8.0_norescale_residual_RK2_N128"
	filecolor_hbar099 = resultsdir+"/color/diags_"+str(i).rjust(ndigits, '0')+"_color.png"
	fileslicex_hbar099 = resultsdir+"/slicex/diags_"+str(i).rjust(ndigits, '0')+"_slice_x.png"
	fileentropy_hbar099 = resultsdir+"/entropy/diags_"+str(i).rjust(ndigits, '0')+"_entropy.png"

	hbar = 1.01
	resultsdir = "./results/relaxation/fermions_2d/hbarmax"+str(hbar)+"_L8.0_norescale_residual_RK2_N128"
	filecolor_hbar101 = resultsdir+"/color/diags_"+str(i).rjust(ndigits, '0')+"_color.png"
	fileslicex_hbar101 = resultsdir+"/slicex/diags_"+str(i).rjust(ndigits, '0')+"_slice_x.png"
	fileentropy_hbar101 = resultsdir+"/entropy/diags_"+str(i).rjust(ndigits, '0')+"_entropy.png"

	fileoutput = outputdir+"/diags_"+str(i).rjust(ndigits, '0')+".png"
	cmd = "montage -mode Concatenate -tile 3x2 -geometry +0+0+0 -background none "
	cmd += filecolor_hbar099+" "+fileslicex_hbar099+" "+fileentropy_hbar099+" "
	cmd += filecolor_hbar101+" "+fileslicex_hbar101+" "+fileentropy_hbar101
	cmd += " "+fileoutput

	os.system(cmd)
	#cmd = "mogrify -trim "+fileoutput
	#cmd = "optipng "+fileoutput
	#os.system(cmd)
	
	list_output.append(fileoutput)
	#print(fileoutput)

tmppath = outputdir+"/tmp"
outputmovie = outputdir+"/compare_hbarmax_0.99_1.01_color_slicex_entropy.mp4"

encode_video(list_output, tmppath, outputmovie, forcesave=True)
"""


inputh5 = "./results/spectral_accuracy/bosons_3d/hbar3.0_cutoff1.0S_L2.0_norescale_sigma0.5_noshift_N64/diags.h5"
data = extract_from_h5(inputh5, "/distribution/f")
plot_volume_3d(data, title, xlim=[-2.,2.], ylim=[-2.,2.], zlim=[-2.,2.], minmax=None, outputpng="test.png")