#!/usr/bin/env python
# -*- coding: utf-8 -*-

from visualization import *



def get_text_label(key, entry):
	if (key == "L"):
		thetext = "L"+str(entry)
		thelabel = "L = "+str(entry)
	elif (key == "cutoff"):
		thetext = "cutoff"+str(entry)+"S"
		thelabel = "R = "+str(entry)+"S"
	elif (key == "sigma"):
		thetext = "sigma"+str(entry)
		thelabel = "sigma = "+str(entry)
	elif (key == "u0"):
		if (max([abs(max(entry)), abs(min(entry))]) < 1e-12):
			thetext = "noshift"
		else:
			thetext = "shift"
		thelabel = "u0 = ("
		for i_d in range(len(entry)-1):
			thelabel += str(entry[i_d])
			thelabel += ","
		thelabel += str(entry[-1])
		thelabel += ")"
	elif (key == "rescale"):
		if (entry):
			thetext = "rescale"
			thelabel = "with rescaling"
		else:
			thetext = "norescale"
			thelabel = "without rescaling"
	elif (key == "hbar"):
		thetext = "hbar"+str(entry)
		thelabel = "hbar = "+str(entry)

	return [thetext, thelabel]



def get_namesubdir(dic_out, dic_in, struct_namedir):
	subdir = ""
	i = 0
	for key in struct_namedir:
		for key_in in dic_in.keys():
			if (key_in == key):
				# Found the key in dic_in
				[thetext, thelabel] = get_text_label(key, dic_in[key])
				subdir += thetext
		for key_out in dic_out.keys():
			if (key_out == key):
				# Found the key in dic_out
				[thetext, thelabel] = get_text_label(key, dic_out[key])
				subdir += thetext
		i += 1
		if (i < len(struct_namedir)-1):
			subdir += "_"

	return subdir


def visu_spectral_accuracy(rootdir, struct_namedir, list_dics_outplot, dic_lists_inplot, list_N):

	list_linecolors = ["blue", "green", "red", "magenta", "cyan", "yellow", "black"]
	list_linestyles = ["solid", "dashed", "dashdot", "dashdotdotted"]
	list_markers = ["P", "X", "*", "o", "v", "s"]

	xlabel = "N"
	ylabel = "Linf norm of residual"

	mintab = []
	maxtab = []
	lmax = 0
	ylim = ""

	for isave in range(2):
		for i_out in range(len(list_dics_outplot)):
			dic_out = list_dics_outplot[i_out]

			# We are able to get the title of the output plot, the name of the produced png
			title = "Residual ("
			outputpng = rootdir+"/"
			i_key = 0
			for key in dic_out.keys():
				entry = dic_out[key]
				[thetext, thelabel] = get_text_label(key, entry)
				title += thelabel
				outputpng += thetext
				if (i_key < len(dic_out.keys())-1):
					title += ", "
					outputpng += "_"
				else:
					title += ")"
					outputpng += ".png"
				i_key += 1

			keys_in = list(dic_lists_inplot.keys())
			key_in_0 = keys_in[0]
			i_0 = 0
			dic_in = {}
			x_list = []
			y_list = []
			linestyles = []
			linecolors = []
			markers = []
			labels = []
			for val_in_0 in dic_lists_inplot[key_in_0]:
				dic_in[key_in_0] = val_in_0
				thelinecolor = list_linecolors[i_0]

				[text_0, label_0] = get_text_label(key_in_0, val_in_0)

				if (len(keys_in) > 1):
					key_in_1 = keys_in[1]
					i_1 = 0
					for val_in_1 in dic_lists_inplot[key_in_1]:
						dic_in[key_in_1] = val_in_1
						themarker = list_markers[i_1]

						[text_1, label_1] = get_text_label(key_in_1, val_in_1)

						if (len(keys_in) > 2):
							key_in_2 = key_in[2]
							i_2 = 0
							for val_in_2 in dic_lists_inplot[key_in_2]:
								dic_in[key_in_2] = val_in_2
								thelinestyle = list_linestyles[i_2]

								[text_2, label_2] = get_text_label(key_in_2, val_in_2)

								rootsubdir = get_namesubdir(dic_out, dic_in, struct_namedir)
								rootsubdir = rootdir+"/"+rootsubdir

								if (len(keys_in) > 3):
									print("Error: cannot handle more than 3 categories of curves on a same plot")
									exit()

								thelabel = label_0+", "+label_1+", "+label_2
								lmax = max([lmax, len(thelabel)])


								LinfL2_list = []
								N_list = []
								for N in list_N:
									textN = "N"+str(N)
							
									subdir = rootdir+"/"+rootsubdir+"_"+textN
									namefile = subdir+"/diags_LinfnormQ.dat"
									try:
										f = open(namefile, "r")
										lines = f.readlines()
										LinfL2 = float(lines[-1].split(" ")[-1])
										LinfL2_list.append(LinfL2)
										N_list.append(N)
										f.close()
									except FileNotFoundError:
										pass

								x_list.append(N_list)
								y_list.append(LinfL2_list)
								labels.append(thelabel)

								markers.append(themarker)
								linestyles.append(thelinestyle)
								linecolors.append(thelinecolor)

								i_2 += 1
						else:
							rootsubdir = get_namesubdir(dic_out, dic_in, struct_namedir)

							thelabel = label_0+", "+label_1
							lmax = max([lmax, len(thelabel)])

							LinfL2_list = []
							N_list = []
							for N in list_N:
								textN = "N"+str(N)
							
								subdir = rootdir+"/"+rootsubdir+"_"+textN
								namefile = subdir+"/diags_LinfnormQ.dat"
								try:
									f = open(namefile, "r")
									lines = f.readlines()
									LinfL2 = float(lines[-1].split(" ")[-1])
									LinfL2_list.append(LinfL2)
									N_list.append(N)
									f.close()
								except FileNotFoundError:
									pass

							x_list.append(N_list)
							y_list.append(LinfL2_list)
							labels.append(thelabel)
				
							markers.append(themarker)
							linecolors.append(thelinecolor)

						i_1 += 1
				else:
					rootsubdir = get_namesubdir(dic_out, dic_in, struct_namedir)

					thelabel = label_0
					lmax = max([lmax, len(thelabel)])


					LinfL2_list = []
					N_list = []
					for N in list_N:
						textN = "N"+str(N)
							
						subdir = rootdir+"/"+rootsubdir+"_"+textN
						namefile = subdir+"/diags_LinfnormQ.dat"
						try:
							f = open(namefile, "r")
							lines = f.readlines()
							LinfL2 = float(lines[-1].split(" ")[-1])
							LinfL2_list.append(LinfL2)
							N_list.append(N)
							f.close()
						except FileNotFoundError:
							pass

					x_list.append(N_list)
					y_list.append(LinfL2_list)
					labels.append(thelabel)

					linecolors.append(thelinecolor)

				i_0 += 1

			# At this point, we have loaded the data corresponding to the full set of curves to be plotted
			if (len(keys_in) <= 2):
				linestyles = ""
			if (len(keys_in) == 1):
				markers = ""
			[theminval, themaxval] = plot_curves(x_list, y_list, xlabel, ylabel, title, outputpng, labels=labels, markers=markers, linecolors=linecolors, linestyles=linestyles, yscale="log", ylim=ylim)
			if (theminval > 0.):
				mintab.append(theminval)
			maxtab.append(themaxval)
		ylim = [min(mintab), max(maxtab)]
		print("{0:e}".format(ylim[0]))



def extract_data_spectral_accuracy(rho, list_u0, list_sigma, list_L, list_hbar, list_rescale, list_cutoff, list_convolution, list_int_degrees, list_N):
	listdic = []
	textrho = "rho"+str(rho)
	for u0 in list_u0:
		if (max([abs(max(u0)), abs(min(u0))]) < 1e-12):
			text_u0 = "noshift"
		else:
			text_u0 = "shift"
		for sigma in list_sigma:
			text_sigma = "sigma"+str(sigma)
			for L in list_L:
				text_L = "L"+str(L)
				for rescale in list_rescale:
					if (rescale):
						text_rescale = "rescale"
					else:
						text_rescale = "norescale"
					for cutoff in list_cutoff:
						text_cutoff = "cutoff"+str(cutoff)+"S"
						for j in range(len(list_convolution)):
							if (list_convolution[j]):
								text_convol = "convol"+str(list_int_degrees[j][0])+"-"+str(list_int_degrees[j][1])+"-"+str(list_int_degrees[j][2])
							else:
								if (len(u0) == 2):
									text_convol = "noconvol"+str(list_int_degrees[j])
								elif (len(u0) == 3):
									text_convol = "noconvol"+str(list_int_degrees[j][0])+"-"+str(list_int_degrees[j][1])
							for N in list_N:
								text_N = "N"+str(N)
								if (len(list_hbar) > 0):
									for hbar in list_hbar:
										text_hbar = "hbar"+str(hbar)

										dic = {}
										dic["u0"] = u0
										dic["sigma"] = sigma
										dic["L"] = L
										dic["rescale"] = rescale
										dic["cutoff"] = cutoff
										dic["N"] = N
										dic["hbar"] = hbar
										dic["force_convolution"] = list_convolution[j]
										dic["int_degrees"] = list_int_degrees[j]

										namesubdir = text_hbar+"_"+text_cutoff+"_"+text_L+"_"+text_rescale+"_"+text_sigma+"_"+text_u0+"_"+text_convol+"_"+text_N
										namefile = rootdir+"/"+namesubdir+"/diags_LinfnormQ.dat"
										
										try:
											f = open(namefile, "r")
											dic["LinfQ"] = float(f.readlines()[0].split(" ")[-1])
											f.close()
										except:
											dic["LinfQ"] = -1.
										listdic.append(dic)
								else:
									dic = {}
									dic["u0"] = u0
									dic["sigma"] = sigma
									dic["L"] = L
									dic["rescale"] = rescale
									dic["cutoff"] = cutoff
									dic["N"] = N
									dic["force_convolution"] = list_convolution[j]
									dic["int_degrees"] = list_int_degrees[j]

									namesubdir = text_cutoff+"_"+text_L+"_"+text_rescale+"_"+text_sigma+"_"+text_u0+"_"+text_convol+"_"+text_N
									namefile = rootdir+"/"+namesubdir+"/diags_LinfnormQ.dat"
									
									try:
										f = open(namefile, "r")
										dic["LinfQ"] = float(f.readlines()[0].split(" ")[-1])
										f.close()
									except:
										dic["LinfQ"] = -1.
									listdic.append(dic)
	return listdic






resultsdir = "./results/spectral_accuracy"


type_qparticles = "fermions"          # "classical" or "fermions" or "bosons"
ndim = 2                              # 2 or 3

rootdir = resultsdir+"/"+type_qparticles+"_"+str(ndim)+"d"

S = 2.*np.pi/(3.+np.sqrt(2.))
Sv = 4.*S/(np.pi)

if (ndim == 2):
	ub = [Sv/np.sqrt(2.), Sv/np.sqrt(2.)]
	u0 = [0., 0.]
	list_N = [16, 32, 64, 128, 256]
else:
	ub = [Sv/2., Sv/2., Sv/np.sqrt(2.)]
	u0 = [0., 0., 0.]
	list_N = [16, 32, 64]

list_u0 = [ub, u0]
list_L = [2., 4., 6., 8., 10.]
list_hbar = [0.1, 3.]
list_cutoff = [1.]
list_sigma = [0.5, 1., 2.]
list_rescale = [True, False]
rho = 1.

list_convol = [True]
list_int_degrees = [[16,2,2]]


listdic = extract_data_spectral_accuracy(rho, list_u0, list_sigma, list_L, list_hbar, list_rescale, list_cutoff, list_convol, list_int_degrees, list_N)

buf = "\\begin{table}\n"
buf += "\\begin{tabular}{|c|c|c|c|c|c|}\n"
buf += "\\hline\n"
buf += "\\multirow{2}{*}{$L$} & \\multirow{2}{*}{Grid} & \\multicolumn{2}{|c|}{$\\mathbf{u}^{0} = \\mathbf{0}$} & \\multicolumn{2}{|c|}{$\\mathbf{u}^{0} = \\mathbf{u}_{b}$} \\\\\n"
buf += "\\cline{3-6}\n"
buf += "& & with rescaling & without rescaling & with rescaling & without rescaling \\\\\n"
buf += "\\hline\n"
for sigma in list_sigma:
	for L in list_L:
		buf += "\\multirow{"+str(len(list_N))+"}{*}{\\begin{tabular}{c}$L = "+str(L)+"$ \\\\ $\sigma = "+str(sigma)+"$\end{tabular}} "
		for N in list_N:
			for dic in listdic:
				if ((dic["N"] == N) and (abs(dic["L"] - L) < 1e-12) and (abs(dic["sigma"]-sigma) < 1e-12)):
					if (dic["rescale"]):
						if (max([abs(max(dic["u0"])), abs(min(dic["u0"]))]) < 1e-12):
							if (dic["LinfQ"] > 0.):
								noshift_rescale = "{0:.5e}".format(dic["LinfQ"])
							else:
								noshift_rescale = "-"
						else:
							if (dic["LinfQ"] > 0.):
								shift_rescale = "{0:.5e}".format(dic["LinfQ"])
							else:
								shift_rescale = "-"
					else:
						if (max([abs(max(dic["u0"])), abs(min(dic["u0"]))]) < 1e-12):
							if (dic["LinfQ"] > 0.):
								noshift_norescale = "{0:.5e}".format(dic["LinfQ"])
							else:
								noshift_norescale = "-"
						else:
							if (dic["LinfQ"] > 0.):
								shift_norescale = "{0:.5e}".format(dic["LinfQ"])
							else:
								shift_norescale = "-"
			buf += "& $"+str(N)+"^{"+str(ndim)+"}$ & "+noshift_rescale+" & "+noshift_norescale+" & "+shift_rescale+" & "+shift_norescale+"\\\\\n"
		buf += "\\hline\n"
buf += "\\end{tabular}\n"
buf += "\\end{table}\n"
print(buf)
exit()


if (type_qparticles == "classical"):
	struct_namedir = ["cutoff","L","rescale","sigma","u0","N"]
	list_hbar = []
else:
	struct_namedir = ["hbar","cutoff","L","rescale","sigma","u0","N"]


# List of the parameters of which we want to compare the impact on a same graph
dic_lists_inplot = {}
dic_lists_inplot["L"] = list_L
dic_lists_inplot["rescale"] = list_rescale

# List of the other parameters
list_dics_outplot = []
if (type_qparticles != "classical"):
	for hbar in list_hbar:
		for cutoff in list_cutoff:
			for sigma in list_sigma:
				for u0 in list_u0:
					dic = {}
					dic["hbar"] = hbar
					dic["cutoff"] = cutoff
					dic["sigma"] = sigma
					dic["u0"] = u0
					list_dics_outplot.append(dic)
else:
	for cutoff in list_cutoff:
		for sigma in list_sigma:
			for u0 in list_u0:
				dic = {}
				dic["cutoff"] = cutoff
				dic["sigma"] = sigma
				dic["u0"] = u0
				list_dics_outplot.append(dic)

visu_spectral_accuracy(rootdir, struct_namedir, list_dics_outplot, dic_lists_inplot, list_N)



# List of the parameters of which we want to compare the impact on a same graph
dic_lists_inplot = {}
dic_lists_inplot["cutoff"] = list_cutoff
dic_lists_inplot["rescale"] = list_rescale

# List of the other parameters
list_dics_outplot = []
if (type_qparticles != "classical"):
	for hbar in list_hbar:
		for L in list_L:
			for sigma in list_sigma:
				for u0 in list_u0:
					dic = {}
					dic["hbar"] = hbar
					dic["L"] = L
					dic["sigma"] = sigma
					dic["u0"] = u0
					list_dics_outplot.append(dic)
else:
	for L in list_L:
		for sigma in list_sigma:
			for u0 in list_u0:
				dic = {}
				dic["L"] = L
				dic["sigma"] = sigma
				dic["u0"] = u0
				list_dics_outplot.append(dic)

visu_spectral_accuracy(rootdir, struct_namedir, list_dics_outplot, dic_lists_inplot, list_N)





