#!/usr/bin/env python
# -*- coding: utf-8 -*-

from visualization import *
from tests_kinebec import *


hostname = subprocess.check_output(['uname', '-n'])
hostname = hostname.decode("utf-8").split('\n')[0]

def extract_ridge_resident_memory(namefile):
	[resident_memory_global, time_global, resident_memory_children, time_children] = read_memprofile(namefile)
	ridge_memory_global = max(resident_memory_global)
	ridge_memory_children = []
	if (resident_memory_children != []):
		for R in range(len(resident_memory_children)):
			ridge_memory_children.append(max(resident_memory_children[R]))
	
	return [ridge_memory_global, ridge_memory_children]


def extract_timecost(namefile):
	f = open(namefile, "r")
	lines = f.readlines()

	if ("classical" in lines[1]):
		qparticles = "classical"
	else:
		qparticles = "quantum"

	shape = list(map(int, lines[3].split(" ")))
	ndims = len(shape)

	i0 = 4

	if ("Number of integration points" in lines[4]):
		force_convolution = True
		ntimes = 1
		i0 += 2
	else:
		force_convolution = False
		if (qparticles == "classical"):
			ntimes = 2
		else:
			ntimes = 6

	if ("Number of threads" in lines[i0]):
		parallel = "omp"
		nthreads = int(lines[i0+1])
		nCPUs = 1
		i0 += 3
	elif ("Number of processes" in lines[i0]):
		nCPUs = int(lines[i0+1])
		if (nCPUs > 1):
			parallel = "mpi"
		else:
			parallel = "serial"
		i0 += 3
	elif ("Number of blocks" in lines[i0]):
		parallel = "cuda"
		nblocks = int(lines[i0+1])
		nthreadsperblock = int(lines[i0+3])
		nCPUs = 1
		i0 += 5

	timecosts = []
	hosts = []
	for i in range(nCPUs):
		data = lines[i+i0].split("|")
		timecosts.append([float(data[1+k]) for k in range(ntimes)])
		hosts.append(data[-1].replace("\n", ""))

	f.close()

	hosts = list(dict.fromkeys(hosts))

	return [timecosts, hosts]

def theoretical_Qc(N, ndims):
	return (N**ndims)*np.log(N)

def theoretical_Q1q(N, ndims):
	return (N**(2*ndims))*np.log(N)



ndims = 3

if (hostname == 'mouton-Latitude-5480'):
	rootdir = "/home/mouton/KINEBEC/code"
	list_nCPUs = [1, 2, 4]
elif (hostname == "mouton-OptiPlex-7050"):
	rootdir = "/home/mouton/KINEBEC/code"
	#list_nCPUs = [1, 2, 4]
	list_nCPUs = [1]
elif ("mathcalc" in hostname):
	rootdir = "/home/mouton/KINEBEC"
	list_nCPUs = [1, 2, 4, 8, 16]
elif ("zeus" in hostname):
	rootdir = "/home/amouton/KINEBEC"
	list_nCPUs = [1, 2, 4, 8, 16, 32, 64]


parallel = "mpi"

if (ndims == 2):
	list_N = [8, 16, 32, 64, 128, 256]
	list_convolution = [True, False]
	list_int_nbpts = [[16,2,2], [4]]
else:
	#list_N = [8, 16, 32, 64, 128]
	list_N = [8, 16, 32, 64]
	#list_convolution = [True, False]
	#list_int_nbpts = [[16,2,2], [4,4]]
	list_convolution = [True]
	list_int_nbpts = [[16,2,2]]


listdics = []
for i in range(len(list_convolution)):
	force_convolution = list_convolution[i]
	int_nbpts = list_int_nbpts[i]
	for N in list_N:
		if (parallel in ["serial", "mpi", "omp"]):
			for nCPUs in list_nCPUs:
				if (nCPUs <= N/2):
					dic = {}
					dic["parallel"] = parallel
					dic["nCPUs"] = nCPUs
					dic["force_convolution"] = force_convolution
					dic["int_nbpts"] = int_nbpts
					dic["N"] = N
					listdics.append(dic)

rootdir += "/results/speedup/fermions_"+str(ndims)+"d"


x = []
memory = []
timecost = []

for dic in listdics:
	namefile = rootdir + "/" + dic["parallel"]
	if dic["force_convolution"]:
		namefile += "_convol"+str(dic["int_nbpts"][0])+"-"+str(dic["int_nbpts"][1])+"-"+str(dic["int_nbpts"][2])
	else:
		namefile += "_noconvol"+str(dic["int_nbpts"][0])
		if (ndims == 3):
			namefile += "-"+str(dic["int_nbpts"][1])
		
	namefile += "_N"+str(dic["N"])
	if ((dic["parallel"] == "mpi") or (dic["parallel"] == "omp") or (dic["parallel"] == "serial")):
		namefile += "_nCPUs"+str(dic["nCPUs"])
	elif (dic["parallel"] == "cuda"):
		namefile += "_nblocks"+str(dic["nblocks"])+"_nthreads"+str(dic["nthreads_per_blocks"])

	memprofile = namefile + "_memprofile.dat"
	[ridge_memory_global, ridge_memory_children] = extract_ridge_resident_memory(memprofile)

	if ((dic["parallel"] == "mpi") or (dic["parallel"] == "omp") or (dic["parallel"] == "serial")):
		timefile = namefile + "/diags_timecost_collision_"+str(dic["nCPUs"])+".txt"
	elif (dic["parallel"] == "cuda"):
		timefile = namefile + "/diags_timecost_collision_"+str(dic["nblocks"])+"_blocks_"+str(dic["nthreads_per_blocks"])+"_threads.txt"
	[timecosts, hosts] = extract_timecost(timefile)

	x.append(dic["N"])
	memory.append(ridge_memory_children[0])
	timecost.append(timecosts[0])


plot_curves([x], [memory], "N", "Ridge memory", "Ridge memory for a serial run", "./test_memory", yscale="log")
plot_curves([x], [timecost], "N", "Time cost", "Time cost for a serial run", "./test_timecost", yscale="log")



exit()









ndims = 3
"""
if (ndims == 2):
	# For 2d case
	list_nCPUs = [1, 2, 4, 8, 16, 32]
	list_N = [16, 32, 64, 128]
	rootdir = "./results/speedup/fermions_2d"
elif (ndims == 3):
	# For 3d case
	list_nCPUs = [1, 2, 4, 8, 16, 32]
	list_N = [8, 16, 32, 64]
	rootdir = "./results/speedup/fermions_3d"
"""
ndims = 2
list_nCPUs = [1,2,4,8]
list_N = [128]
rootdir = "./results"
prefix = "diags"

list_dics = []
for N in list_N:
	for nCPUs in list_nCPUs:
		#namefile = rootdir+"/N"+str(N)+"_nCPUs"+str(nCPUs)+"/"+prefix+"_timecost_collision_"+str(nCPUs)+".txt"
		namefile = rootdir+"/"+prefix+"_timecost_collision_"+str(nCPUs)+".txt"
		
		#with open(namefile, "r") as f:
		try:
			f = open(namefile, "r")
			dic = {}
			dic["nCPUs"] = nCPUs
			dic["N"] = N
			alllines = f.readlines()
			line = alllines[4]
			use_omp = ("threads" in line)
			lines = alllines[7:]

			keys = ["time_Q1c", "time_Q2c", "time_Q1q", "time_Q2q", "time_Q3q", "time_Q4q"]
			for key in keys:
				dic[key] = [1e12,0.,0.]
			dic["hosts"] = []
			for line in lines:
				if '|' in line:
					data = line.split("\n")[0].split("|")[1:]
					for i in range(len(keys)):
						dic[keys[i]] = [min(dic[keys[i]][0], float(data[i])), dic[keys[i]][1]+float(data[i]), max(dic[keys[i]][2], float(data[i]))]
					dic["hosts"].append(data[-1])
				else:
					data = list(map(float, line.split("\n")[0].split(" ")[1:]))
					for i in range(len(keys)):
						dic[keys[i]] = [min(dic[keys[i]][0], float(data[i])), dic[keys[i]][1]+float(data[i]), max(dic[keys[i]][2], float(data[i]))]

				if use_omp:
					break	
			if (not(use_omp)):
				for i in range(len(keys)):
					dic[keys[i]][1] /= nCPUs
			list_dics.append(dic)

			f.close()
			
		except FileNotFoundError:
			pass



list_sizemesh = []
list_timecost_Q1c = []
list_timecost_Q2c = []
list_timecost_Q1q = []
list_timecost_Q2q = []
list_timecost_Q3q = []
list_timecost_Q4q = []
list_theoretical_Q1q = []
list_theoretical_Qc = []
for dic in list_dics:
	if (dic["nCPUs"] == 1):
		list_sizemesh.append(dic["N"])
		list_timecost_Q1c.append(dic["time_Q1c"][1])
		list_timecost_Q2c.append(dic["time_Q2c"][1])
		list_timecost_Q1q.append(dic["time_Q1q"][1])
		list_timecost_Q2q.append(dic["time_Q2q"][1])
		list_timecost_Q3q.append(dic["time_Q3q"][1])
		list_timecost_Q4q.append(dic["time_Q4q"][1])
		list_theoretical_Q1q.append(theoretical_Q1q(dic["N"], ndims))
		list_theoretical_Qc.append(theoretical_Qc(dic["N"], ndims))
		


x_list = [list_sizemesh, list_sizemesh, list_sizemesh, list_sizemesh, list_sizemesh, list_sizemesh, list_sizemesh, list_sizemesh]
y_list = [list_timecost_Q1c, list_timecost_Q2c, list_timecost_Q1q, list_timecost_Q2q, list_timecost_Q3q, list_timecost_Q4q, list_theoretical_Q1q, list_theoretical_Qc]
labels = ["Q1c", "Q2c", "Q1q", "Q2q", "Q3q", "Q4q", "N^"+str(2*ndims)+"*log(N)", "N^"+str(ndims)+"*log(N)"]
linestyles = ["solid"]*6 + ["dashed"]*2
xlabel = "N"
ylabel = "time cost"
title = "Time cost - "+str(ndims)+"D case (serial)"
ylim = [min(min(y_list)), max(max(y_list))]
outputpng = rootdir+"/timecost_serial.png"
yscale = "log"
plot_curves(x_list, y_list, xlabel, ylabel, title, outputpng, labels=labels, yscale=yscale, ylim=ylim, linestyles=linestyles)

x_list = [list_sizemesh, list_sizemesh]
y_list = [list_timecost_Q1c, [list_theoretical_Qc[i]/list_theoretical_Qc[0]*list_timecost_Q1c[0] for i in range(len(list_theoretical_Qc))]]
labels = ["Q1c", "N^"+str(ndims)+"*log(N)"]
linestyles = ["solid"] + ["dashed"]
xlabel = "N"
ylabel = "time cost"
title = "Time cost of Q1c - "+str(ndims)+"D case (serial)"
ylim = [min(min(y_list)), max(max(y_list))]
outputpng = rootdir+"/timecost_serial_Q1c.png"
yscale = "log"
plot_curves(x_list, y_list, xlabel, ylabel, title, outputpng, labels=labels, yscale=yscale, ylim=ylim, linestyles=linestyles)

x_list = [list_sizemesh, list_sizemesh]
y_list = [list_timecost_Q2c, [list_theoretical_Qc[i]/list_theoretical_Qc[0]*list_timecost_Q2c[0] for i in range(len(list_theoretical_Qc))]]
labels = ["Q2c", "N^"+str(ndims)+"*log(N)"]
linestyles = ["solid"] + ["dashed"]
xlabel = "N"
ylabel = "time cost"
title = "Time cost of Q2c - "+str(ndims)+"D case (serial)"
ylim = [min(min(y_list)), max(max(y_list))]
outputpng = rootdir+"/timecost_serial_Q2c.png"
yscale = "log"
plot_curves(x_list, y_list, xlabel, ylabel, title, outputpng, labels=labels, yscale=yscale, ylim=ylim, linestyles=linestyles)

x_list = [list_sizemesh, list_sizemesh]
y_list = [list_timecost_Q1q, [list_theoretical_Q1q[i]/list_theoretical_Q1q[0]*list_timecost_Q1q[0] for i in range(len(list_theoretical_Q1q))]]
labels = ["Q1q", "N^"+str(2*ndims)+"*log(N)"]
linestyles = ["solid"] + ["dashed"]
xlabel = "N"
ylabel = "time cost"
title = "Time cost of Q1q - "+str(ndims)+"D case (serial)"
ylim = [min(min(y_list)), max(max(y_list))]
outputpng = rootdir+"/timecost_serial_Q1q.png"
yscale = "log"
plot_curves(x_list, y_list, xlabel, ylabel, title, outputpng, labels=labels, yscale=yscale, ylim=ylim, linestyles=linestyles)

x_list = [list_sizemesh, list_sizemesh]
y_list = [list_timecost_Q2q, [list_theoretical_Qc[i]/list_theoretical_Qc[0]*list_timecost_Q2q[0] for i in range(len(list_theoretical_Qc))]]
labels = ["Q2q", "N^"+str(ndims)+"*log(N)"]
linestyles = ["solid"] + ["dashed"]
xlabel = "N"
ylabel = "time cost"
title = "Time cost of Q2q - "+str(ndims)+"D case (serial)"
ylim = [min(min(y_list)), max(max(y_list))]
outputpng = rootdir+"/timecost_serial_Q2q.png"
yscale = "log"
plot_curves(x_list, y_list, xlabel, ylabel, title, outputpng, labels=labels, yscale=yscale, ylim=ylim, linestyles=linestyles)

x_list = [list_sizemesh, list_sizemesh]
y_list = [list_timecost_Q3q, [list_theoretical_Qc[i]/list_theoretical_Qc[0]*list_timecost_Q3q[0] for i in range(len(list_theoretical_Qc))]]
labels = ["Q3q", "N^"+str(ndims)+"*log(N)"]
linestyles = ["solid"] + ["dashed"]
xlabel = "N"
ylabel = "time cost"
title = "Time cost of Q3q - "+str(ndims)+"D case (serial)"
ylim = [min(min(y_list)), max(max(y_list))]
outputpng = rootdir+"/timecost_serial_Q3q.png"
yscale = "log"
plot_curves(x_list, y_list, xlabel, ylabel, title, outputpng, labels=labels, yscale=yscale, ylim=ylim, linestyles=linestyles)

x_list = [list_sizemesh, list_sizemesh]
y_list = [list_timecost_Q4q, [list_theoretical_Qc[i]/list_theoretical_Qc[0]*list_timecost_Q4q[0] for i in range(len(list_theoretical_Qc))]]
labels = ["Q4q", "N^"+str(ndims)+"*log(N)"]
linestyles = ["solid"] + ["dashed"]
xlabel = "N"
ylabel = "time cost"
title = "Time cost of Q4q - "+str(ndims)+"D case (serial)"
ylim = [min(min(y_list)), max(max(y_list))]
outputpng = rootdir+"/timecost_serial_Q4q.png"
yscale = "log"
plot_curves(x_list, y_list, xlabel, ylabel, title, outputpng, labels=labels, yscale=yscale, ylim=ylim, linestyles=linestyles)





for N in list_N:
	list_timecost_Q1c = []
	list_timecost_Q2c = []
	list_timecost_Q1q = []
	list_timecost_Q2q = []
	list_timecost_Q3q = []
	list_timecost_Q4q = []
	list_speedup = []
	timecost_Q1c_serial = 0.
	timecost_Q2c_serial = 0.
	timecost_Q1q_serial = 0.
	timecost_Q2q_serial = 0.
	timecost_Q3q_serial = 0.
	timecost_Q4q_serial = 0.
	for dic in list_dics:
		if (dic["N"] == N):
			list_timecost_Q1c.append(dic["time_Q1c"][1])
			list_timecost_Q2c.append(dic["time_Q2c"][1])
			list_timecost_Q1q.append(dic["time_Q1q"][1])
			list_timecost_Q2q.append(dic["time_Q2q"][1])
			list_timecost_Q3q.append(dic["time_Q3q"][1])
			list_timecost_Q4q.append(dic["time_Q4q"][1])
			list_speedup.append(float(dic["nCPUs"]))
			if (dic["nCPUs"] == 1):
				timecost_Q1c_serial = dic["time_Q1c"][1]
				timecost_Q2c_serial = dic["time_Q2c"][1]
				timecost_Q1q_serial = dic["time_Q1q"][1]
				timecost_Q2q_serial = dic["time_Q2q"][1]
				timecost_Q3q_serial = dic["time_Q3q"][1]
				timecost_Q4q_serial = dic["time_Q4q"][1]
	
	list_speedup_Q1c = []
	for i in range(len(list_timecost_Q1c)):
		list_speedup_Q1c.append(timecost_Q1c_serial / list_timecost_Q1c[i])
	list_speedup_Q2c = []
	for i in range(len(list_timecost_Q2c)):
		list_speedup_Q2c.append(timecost_Q2c_serial / list_timecost_Q2c[i])
	list_speedup_Q1q = []
	for i in range(len(list_timecost_Q1q)):
		list_speedup_Q1q.append(timecost_Q1q_serial / list_timecost_Q1q[i])
	list_speedup_Q2q = []
	for i in range(len(list_timecost_Q2q)):
		list_speedup_Q2q.append(timecost_Q2q_serial / list_timecost_Q2q[i])
	list_speedup_Q3q = []
	for i in range(len(list_timecost_Q3q)):
		list_speedup_Q3q.append(timecost_Q3q_serial / list_timecost_Q3q[i])
	list_speedup_Q4q = []
	for i in range(len(list_timecost_Q4q)):
		list_speedup_Q4q.append(timecost_Q4q_serial / list_timecost_Q4q[i])
			
	x_list = [list_speedup, list_speedup, list_speedup, list_speedup, list_speedup, list_speedup, list_speedup]
	y_list = [list_speedup_Q1c, list_speedup_Q2c, list_speedup_Q1q, list_speedup_Q2q, list_speedup_Q3q, list_speedup_Q4q, list_speedup]
	
	labels = ["Q1c", "Q2c", "Q1q", "Q2q", "Q3q", "Q4q", "optimal"]
	linestyles = ["solid"]*6+["dashed"]
	xlabel = "nCPUs"
	ylabel = "Speedup"
	if (use_omp):
		title = "OpenMP speedup - "+str(N)+"^"+str(ndims)+" grid"
	else:
		title = "MPI speedup - "+str(N)+"^"+str(ndims)+" grid"
	outputpng = rootdir+"/speedup_N"+str(N)+".png"
	yscale = "linear"
	xlim = [min(list_speedup), max(list_speedup)]
	plot_curves(x_list, y_list, xlabel, ylabel, title, outputpng, labels=labels, yscale=yscale, xlim=xlim, ylim="", linestyles=linestyles)
