#include "./src/IO/adios_interface.h"


void adios_write_0d_double(const char* namefile, const char* namegroup, const char* namefield, double data){
	int sizefloc_Mb;
	int64_t id_group;
	int64_t id_file;
	int istatus;
	int64_t i64status;
	uint64_t groupsize, totalsize;

	sizefloc_Mb = 1;
#ifdef key_mpi
	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif

	// Initialize Adios environment
	adios_init_noxml(MPI_COMM_WORLD);

	// Allocate transport buffer
	adios_allocate_buffer(ADIOS_BUFFER_ALLOC_NOW, sizefloc_Mb);

	// Declare Adios group
	istatus = adios_declare_group(&id_group, namegroup, "", adios_flag_yes);

	// Specify transport method
	istatus = adios_select_method(id_group, "MPI", "", "");

	// Define the variables to be saved
	i64status = adios_define_var(id_group, namefield, "", adios_double, 0, 0, 0);

	// Open the BP file for writing by the communicator
	istatus = adios_open(&id_file, namegroup, namefile, "u", MPI_COMM_WORLD);

	// Compute the size of the group
	groupsize = sizeof(double);
	istatus = adios_group_size(id_file, groupsize, &totalsize);

	// Write data
	adios_write(id_file, namefield, (void*)&data);

	// Close Adios environment
	adios_close (id_file);

#ifdef key_mpi
	adios_finalize(rank);
	// Impose a MPI Barrier
	MPI_Barrier(MPI_COMM_WORLD);
#endif
}


void adios_write_0d_int(const char* namefile, const char* namegroup, const char* namefield, int data){
	int sizefloc_Mb;
	int64_t id_group;
	int64_t id_file;
	int istatus;
	int64_t i64status;
	uint64_t groupsize, totalsize;

	sizefloc_Mb = 1;
#ifdef key_mpi
	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif

	// Initialize Adios environment
	adios_init_noxml(MPI_COMM_WORLD);

	// Allocate transport buffer
	adios_allocate_buffer(ADIOS_BUFFER_ALLOC_NOW, sizefloc_Mb);

	// Declare Adios group
	istatus = adios_declare_group(&id_group, namegroup, "", adios_flag_yes);

	// Specify transport method
	istatus = adios_select_method(id_group, "MPI", "", "");

	// Define the variables to be saved
	i64status = adios_define_var(id_group, namefield, "", adios_integer, 0, 0, 0);

	// Open the BP file for writing by the communicator
	istatus = adios_open(&id_file, namegroup, namefile, "u", MPI_COMM_WORLD);

	// Compute the size of the group
	groupsize = sizeof(double);
	istatus = adios_group_size(id_file, groupsize, &totalsize);

	// Write data
	adios_write(id_file, namefield, (void*)&data);

	// Close Adios environment
	adios_close (id_file);

#ifdef key_mpi
	adios_finalize(rank);
	// Impose a MPI Barrier
	MPI_Barrier(MPI_COMM_WORLD);
#endif
}


void adios_write_2d(const char* namefile, const char* namegroup, const char* namefield, double** floc, PARAMETERS parameters){
	int sizefloc_Mb;
	int64_t id_group;
	int64_t id_file;
	int istatus;
	int64_t i64status;
	uint64_t groupsize, totalsize;

	double* mesh_vx;
	double* mesh_vy;

#ifdef key_mpi
	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif
	sizefloc_Mb = (int)parameters.nvxloc*(int)parameters.nvyloc*sizeof(double)/(1024*1024) + 1;

	// Initialize Adios environment
	adios_init_noxml(MPI_COMM_WORLD);

	// Allocate transport buffer
	adios_allocate_buffer(ADIOS_BUFFER_ALLOC_NOW, sizefloc_Mb);

	// Declare Adios group
	istatus = adios_declare_group(&id_group, namegroup, "", adios_flag_yes);

	// Specify transport method
	istatus = adios_select_method(id_group, "MPI", "", "");

	// Define the variables to be saved
	i64status = adios_define_var(id_group, "nvx", "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "nvy", "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "nvxloc", "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "ivx0loc", "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "nvyloc", "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "ivy0loc", "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "L", "", adios_double, 0, 0, 0);
	i64status = adios_define_var(id_group, "mesh_vx", "", adios_double, "nvxloc", "nvx", "ivx0loc");
	i64status = adios_define_var(id_group, "mesh_vy", "", adios_double, "nvyloc", "nvy", "ivy0loc");

	i64status = adios_define_var(id_group, namefield, "", adios_double, "nvxloc,nvyloc", "nvx,nvy", "ivx0loc,ivy0loc");

	// Define the mesh
	adios_define_schema_version(id_group, "1.1");
	adios_define_mesh_rectilinear("nvx,nvy", "mesh_vx,mesh_vy", "2", id_group, "mesh");
	adios_define_mesh_timevarying("no", id_group, "mesh");
	adios_define_var_mesh(id_group, namefield, "mesh");
	adios_define_var_centering(id_group, namefield, "point");

	// Create the BP file for writing by the communicator
	istatus = adios_open(&id_file, namegroup, namefile, "u", MPI_COMM_WORLD);

	// Compute the size of the group
	groupsize = sizeof(int)*2 + sizeof(double) + sizeof(ptrdiff_t)*4 + sizeof(double)*((int)parameters.nvxloc+(int)parameters.nvyloc) + sizeof(double)*parameters.nvx*parameters.nvy;
	istatus = adios_group_size(id_file, groupsize, &totalsize);

	// Write data
	adios_write(id_file, "nvx", (void*)&parameters.nvx);
	adios_write(id_file, "nvy", (void*)&parameters.nvy);
	adios_write(id_file, "L", (void*)&parameters.L);
	adios_write(id_file, "nvxloc", (void*)&parameters.nvxloc);
	adios_write(id_file, "nvyloc", (void*)&parameters.nvyloc);
	adios_write(id_file, "ivx0loc", (void*)&parameters.ivx0loc);
	adios_write(id_file, "ivy0loc", (void*)&parameters.ivy0loc);
	mesh_vx = allocation_1d(parameters.nvxloc);
	for (int i = 0; i < (int)parameters.nvxloc; i++)
		mesh_vx[i] = -parameters.L + (i+(int)parameters.ivx0loc)*(2.*parameters.L)/parameters.nvx;
	adios_write(id_file, "mesh_vx", mesh_vx);
	free_1d(mesh_vx);
	mesh_vy = allocation_1d(parameters.nvyloc);
	for (int j = 0; j < (int)parameters.nvyloc; j++)
		mesh_vy[j] = -parameters.L + (j+(int)parameters.ivy0loc)*(2.*parameters.L)/parameters.nvy;
	adios_write(id_file, "mesh_vy", mesh_vy);
	free_1d(mesh_vy);

	adios_write(id_file, namefield, floc[0]);

	// Close Adios environment
	adios_close (id_file);

#ifdef key_mpi
	adios_finalize(rank);
	// Impose a MPI Barrier
	MPI_Barrier(MPI_COMM_WORLD);
#endif

}




void adios_write_3d(const char* namefile, const char* namegroup, const char* namefield, double*** floc, PARAMETERS parameters){
	int sizefloc_Mb;
	int64_t id_group;
	int64_t id_file;
	int istatus;
	int64_t i64status;
	uint64_t groupsize, totalsize;

	double* mesh_vx;
	double* mesh_vy;
	double* mesh_vz;

#ifdef key_mpi
	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif
	sizefloc_Mb = (int)parameters.nvxloc*(int)parameters.nvyloc*(int)parameters.nvzloc*sizeof(double)/(1024*1024) + 1;

	// Initialize Adios environment
	adios_init_noxml(MPI_COMM_WORLD);

	// Allocate transport buffer
	adios_allocate_buffer(ADIOS_BUFFER_ALLOC_NOW, sizefloc_Mb);

	// Declare Adios group
	istatus = adios_declare_group(&id_group, namegroup, "", adios_flag_yes);

	// Specify transport method
	istatus = adios_select_method(id_group, "MPI", "", "");

	// Define the variables to be saved
	i64status = adios_define_var(id_group, "nvx" , "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "nvy" , "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "nvz" , "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "L" , "", adios_double, 0, 0, 0);
	i64status = adios_define_var(id_group, "nvxloc" , "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "ivx0loc" , "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "nvyloc" , "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "ivy0loc" , "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "nvzloc" , "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "ivz0loc" , "", adios_integer, 0, 0, 0);
	i64status = adios_define_var(id_group, "mesh_vx", "", adios_double, "nvxloc", "nvx", "ivx0loc");
	i64status = adios_define_var(id_group, "mesh_vy", "", adios_double, "nvyloc", "nvy", "ivy0loc");
	i64status = adios_define_var(id_group, "mesh_vz", "", adios_double, "nvzloc", "nvz", "ivz0loc");
	i64status = adios_define_var(id_group, namefield, "", adios_double, "nvxloc,nvyloc,nvzloc", "nvx,nvy,nvz", "ivx0loc,ivy0loc,ivz0loc");

	// Define the mesh
	adios_define_schema_version(id_group, "1.1");
	adios_define_mesh_rectilinear("nvx,nvy,nvz", "mesh_vx,mesh_vy,mesh_vz", "3", id_group, "mesh");
	adios_define_mesh_timevarying("no", id_group, "mesh");
	adios_define_var_mesh(id_group, namefield, "mesh");
	adios_define_var_centering(id_group, namefield, "point");

	// Create the BP file for writing by the communicator
	istatus = adios_open(&id_file, namegroup, namefile, "u", MPI_COMM_WORLD);

	// Compute the size of the global group
	groupsize = sizeof(int)*3 + sizeof(double) + sizeof(ptrdiff_t)*6 + sizeof(double)*(parameters.nvx+parameters.nvy+parameters.nvz) + sizeof(double)*(int)parameters.nvxloc*(int)parameters.nvyloc*(int)parameters.nvzloc;
	istatus = adios_group_size(id_file, groupsize, &totalsize);

	// Write data
	adios_write(id_file, "nvx", (void*)&parameters.nvx);
	adios_write(id_file, "nvy", (void*)&parameters.nvy);
	adios_write(id_file, "nvz", (void*)&parameters.nvz);
	adios_write(id_file, "L", (void*)&parameters.L);
	adios_write(id_file, "nvxloc", (void*)&parameters.nvxloc);
	adios_write(id_file, "ivx0loc", (void*)&parameters.ivx0loc);
	adios_write(id_file, "nvyloc", (void*)&parameters.nvyloc);
	adios_write(id_file, "ivy0loc", (void*)&parameters.ivy0loc);
	adios_write(id_file, "nvzloc", (void*)&parameters.nvzloc);
	adios_write(id_file, "ivz0loc", (void*)&parameters.ivz0loc);
	mesh_vx = allocation_1d(parameters.nvxloc);
	for (int i = 0; i < (int)parameters.nvxloc; i++)
		mesh_vx[i] = -parameters.L + (i+(int)parameters.ivx0loc)*(2.*parameters.L)/parameters.nvx;
	adios_write(id_file, "mesh_vx", mesh_vx);
	free_1d(mesh_vx);
	mesh_vy = allocation_1d(parameters.nvyloc);
	for (int j = 0; j < (int)parameters.nvyloc; j++)
		mesh_vy[j] = -parameters.L + (j+(int)parameters.ivy0loc)*(2.*parameters.L)/parameters.nvy;
	adios_write(id_file, "mesh_vy", mesh_vy);
	free_1d(mesh_vy);
	mesh_vz = allocation_1d(parameters.nvzloc);
	for (int k = 0; k < (int)parameters.nvzloc; k++)
		mesh_vz[k] = -parameters.L + (k+(int)parameters.ivz0loc)*(2.*parameters.L)/parameters.nvz;
	adios_write(id_file, "mesh_vz", mesh_vz);
	free_1d(mesh_vz);
	adios_write(id_file, namefield, floc[0][0]);

	// Close Adios environment
	adios_close (id_file);

#ifdef key_mpi
	adios_finalize(rank);
	// Impose a MPI Barrier
	MPI_Barrier(MPI_COMM_WORLD);
#endif
}


void adios_read_0d_double(const char* namefile, const char* namegroup, const char* namefield, double *data){
	int istatus;
	ADIOS_SELECTION *selection = NULL;

#ifdef key_mpi
	// Initialize reading method
	istatus = adios_read_init_method(ADIOS_READ_METHOD_BP, MPI_COMM_WORLD, "verbose=3");

	// Open the file
	ADIOS_FILE *f = adios_read_open(namefile, ADIOS_READ_METHOD_BP, MPI_COMM_WORLD, ADIOS_LOCKMODE_NONE, 0);
	if (f == NULL)
		error(adios_errmsg());

	// Read the data subset
	istatus = adios_schedule_read(f, selection, namefield, 0, 1, data);
	istatus = adios_perform_reads(f, 1);

	// Close file
	istatus = adios_read_close(f);

	// Delete metadata about reading method
	istatus = adios_read_finalize_method(ADIOS_READ_METHOD_BP);
#endif
}


void adios_read_0d_int(const char* namefile, const char* namegroup, const char* namefield, int *data){
	int istatus;
	ADIOS_SELECTION *selection = NULL;

#ifdef key_mpi
	// Initialize reading method
	istatus = adios_read_init_method(ADIOS_READ_METHOD_BP, MPI_COMM_WORLD, "verbose=3");

	// Open the file
	ADIOS_FILE *f = adios_read_open(namefile, ADIOS_READ_METHOD_BP, MPI_COMM_WORLD, ADIOS_LOCKMODE_NONE, 0);
	if (f == NULL)
		error(adios_errmsg());

	// Read the data subset
	istatus = adios_schedule_read(f, selection, namefield, 0, 1, data);
	istatus = adios_perform_reads(f, 1);

	// Close file
	istatus = adios_read_close(f);

	// Delete metadata about reading method
	istatus = adios_read_finalize_method(ADIOS_READ_METHOD_BP);
#endif
}


void adios_read_2d(const char* namefile, const char* namegroup, const char* namefield, double ***data, PARAMETERS *parameters){
	int istatus;
	int itemp, itemp2;
	uint64_t start[2];
	uint64_t count[2];
	char message[MAXSTRINGLENGTH];

#ifdef key_mpi
	int rank, commsize;
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	// Initialize reading method
	istatus = adios_read_init_method(ADIOS_READ_METHOD_BP, MPI_COMM_WORLD, "verbose=3");

	// Open the file
	ADIOS_FILE *f = adios_read_open(namefile, ADIOS_READ_METHOD_BP, MPI_COMM_WORLD, ADIOS_LOCKMODE_NONE, 0);
	if (f == NULL)
		error(adios_errmsg());

	// Read the value of L
	istatus = adios_schedule_read(f, NULL, "L", 0, 1, &parameters->L);
	istatus = adios_perform_reads(f, 1);

	// Inquire the field
	ADIOS_VARINFO *v = adios_inq_var(f, namefield);

	// Extract metadata about the inquired field
	if (v->ndim != 2){
		sprintf(message, "Error: file %s does not contain a 2D field named %s", namefile, namefield);
		error(message);
	}
	parameters->nvx = v->dims[0];
	parameters->nvy = v->dims[1];

	// Build the partition table associated to the current communicator MPI_COMM_WORLD
	// Warning: it can be different from the partition table used for writting the file
#ifdef key_fftw_mpi
	// Build the partition table
	// The global array of size nx*ny will be splitted in parts of size nxloc*ny
	parameters->alloc_local = fftw_mpi_local_size_2d(parameters->nvx, parameters->nvy, MPI_COMM_WORLD, &parameters->nvxloc, &parameters->ivx0loc);
	parameters->ivy0loc = 0;
	parameters->nvyloc = parameters->nvy;

	start[0] = (uint64_t)parameters->ivx0loc;
	start[1] = (uint64_t)parameters->ivy0loc;
	count[0] = (uint64_t)parameters->nvxloc;
	count[1] = (uint64_t)parameters->nvyloc;

	parameters->ivx0tab = allocation_1d_int(commsize);
	itemp = (int)parameters->ivx0loc;
	istatus = MPI_Allgather(&itemp, 1, MPI_INT, parameters->ivx0tab, 1, MPI_INT, MPI_COMM_WORLD);
	parameters->nvxtab = allocation_1d_int(commsize);
	itemp2 = (int)parameters->nvxloc;
	istatus = MPI_Allgather(&itemp2, 1, MPI_INT, parameters->nvxtab, 1, MPI_INT, MPI_COMM_WORLD);

	parameters->ivy0tab = allocation_1d_int(commsize);
	itemp = (int)parameters->ivy0loc;
	istatus = MPI_Allgather(&itemp, 1, MPI_INT, parameters->ivy0tab, 1, MPI_INT, MPI_COMM_WORLD);
	parameters->nvytab = allocation_1d_int(commsize);
	itemp2 = (int)parameters->nvyloc;
	istatus = MPI_Allgather(&itemp2, 1, MPI_INT, parameters->nvytab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	start[0] = 0;
	start[1] = 0;
	count[0] = parameters->nvx;
	count[1] = parameters->nvy;
#endif

	// Prepare the selection of data to be read
	ADIOS_SELECTION *selection = adios_selection_boundingbox(v->ndim, start, count);

	// Allocate data
	*data = allocation_2d((int)count[0], (int)count[1]);

	// Read the data subset
	istatus = adios_schedule_read(f, selection, namefield, 0, 1, (void *)&(*data)[0][0]);
	istatus = adios_perform_reads(f, 1);

	// Delete the metadata about the selection
	adios_selection_delete(selection);

	// Delete metadata about the inquired field
	adios_free_varinfo(v);

	// Close file
	istatus = adios_read_close(f);

	// Delete metadata about reading method
	istatus = adios_read_finalize_method(ADIOS_READ_METHOD_BP);
#endif
}


void adios_read_3d(const char* namefile, const char* namegroup, const char* namefield, double**** data, PARAMETERS *parameters){
	int istatus;
	int itemp, itemp2;
	char message[MAXSTRINGLENGTH];
	uint64_t start[3];
	uint64_t count[3];

#ifdef key_mpi
	int rank, commsize;
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	// Initialize reading method
	istatus = adios_read_init_method(ADIOS_READ_METHOD_BP, MPI_COMM_WORLD, "verbose=3");

	// Open the file
	ADIOS_FILE *f = adios_read_open(namefile, ADIOS_READ_METHOD_BP, MPI_COMM_WORLD, ADIOS_LOCKMODE_NONE, 0);
	if (f == NULL)
		error(adios_errmsg());

	// Read the value of L
	istatus = adios_schedule_read(f, NULL, "L", 0, 1, &parameters->L);
	istatus = adios_perform_reads(f, 1);

	// Inquire the field
	ADIOS_VARINFO *v = adios_inq_var(f, namefield);

	// Extract metadata about the inquired field
	if (v->ndim != 3){
		sprintf(message, "Error: file %s does not contain a 3D field named %s", namefile, namefield);
		error(message);
	}
	parameters->nvx = v->dims[0];
	parameters->nvy = v->dims[1];
	parameters->nvz = v->dims[2];

	// Build the partition table associated to the current communicator MPI_COMM_WORLD
	// Warning: it can be different from the partition table used for writting the file
#ifdef key_fftw_mpi
	// Build the partition table
	// The global array of size nx*ny will be splitted in parts of size nxloc*ny
	parameters->alloc_local = fftw_mpi_local_size_3d(parameters->nvx, parameters->nvy, parameters->nvz, MPI_COMM_WORLD, &parameters->nvxloc, &parameters->ivx0loc);
	parameters->ivy0loc = 0;
	parameters->nvyloc = parameters->nvy;
	parameters->ivz0loc = 0;
	parameters->nvzloc = parameters->nvz;

	start[0] = (uint64_t)parameters->ivx0loc;
	start[1] = (uint64_t)parameters->ivy0loc;
	start[2] = (uint64_t)parameters->ivz0loc;
	count[0] = (uint64_t)parameters->nvxloc;
	count[1] = (uint64_t)parameters->nvyloc;
	count[2] = (uint64_t)parameters->nvzloc;

	parameters->ivx0tab = allocation_1d_int(commsize);
	itemp = (int)parameters->ivx0loc;
	istatus = MPI_Allgather(&itemp, 1, MPI_INT, parameters->ivx0tab, 1, MPI_INT, MPI_COMM_WORLD);
	parameters->nvxtab = allocation_1d_int(commsize);
	itemp2 = (int)parameters->nvxloc;
	istatus = MPI_Allgather(&itemp2, 1, MPI_INT, parameters->nvxtab, 1, MPI_INT, MPI_COMM_WORLD);

	parameters->ivy0tab = allocation_1d_int(commsize);
	itemp = (int)parameters->ivy0loc;
	istatus = MPI_Allgather(&itemp, 1, MPI_INT, parameters->ivy0tab, 1, MPI_INT, MPI_COMM_WORLD);
	parameters->nvytab = allocation_1d_int(commsize);
	itemp2 = (int)parameters->nvyloc;
	istatus = MPI_Allgather(&itemp2, 1, MPI_INT, parameters->nvytab, 1, MPI_INT, MPI_COMM_WORLD);

	parameters->ivz0tab = allocation_1d_int(commsize);
	itemp = (int)parameters->ivz0loc;
	istatus = MPI_Allgather(&itemp, 1, MPI_INT, parameters->ivz0tab, 1, MPI_INT, MPI_COMM_WORLD);
	parameters->nvztab = allocation_1d_int(commsize);
	itemp2 = (int)parameters->nvzloc;
	istatus = MPI_Allgather(&itemp2, 1, MPI_INT, parameters->nvztab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	start[0] = 0;
	start[1] = 0;
	start[2] = 0;
	count[0] = parameters->nvx;
	count[1] = parameters->nvy;
	count[2] = parameters->nvz;
#endif

	// Prepare the selection of data to be read
	ADIOS_SELECTION *selection = adios_selection_boundingbox(v->ndim, start, count);

	// Allocate data
	*data = allocation_3d((int)count[0], (int)count[1], (int)count[2]);

	// Read the data subset
	istatus = adios_schedule_read(f, selection, namefield, 0, 1, (void *)&(*data)[0][0][0]);
	istatus = adios_perform_reads(f, 1);

	// Delete the metadata about the selection
	adios_selection_delete(selection);

	// Delete metadata about the inquired field
	adios_free_varinfo(v);

	// Close file
	istatus = adios_read_close(f);

	// Delete metadata about reading method
	istatus = adios_read_finalize_method(ADIOS_READ_METHOD_BP);
#endif
}
