#include "./src/IO/read_data.h"


void read_logfile_lastline(double *data, const char *namefile){
	FILE *file;
	double time;
	char buf[3*MAXSTRINGLENGTH];
	char *sstatus;
	int istatus;
	char backupfile[MAXSTRINGLENGTH];
	
	int rank;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
#endif

#ifdef key_mpi
	int init;
	MPI_Initialized(&init);
	if (init){
		MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif
		if (rank == 0){
			sprintf(backupfile, "%s_backup", namefile);
			// Extract the last line of the log file
			sprintf(buf, "tail -1 %s > %s", namefile, backupfile);
			istatus = system(buf);
			// Read the line and extract the data
			file = fopen(backupfile, "r");
			sstatus = fgets(buf, sizeof(buf), file);
			istatus = sscanf(buf, "%lf %lf", &time, data);
			fclose(file);
			// Clean the temporary file
			sprintf(buf, "rm %s", backupfile);
			istatus = system(buf);
		}
#ifdef key_mpi
		MPI_Bcast(data, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	}
#endif
}



void read_0d_double(const char* namefile, const char* namegroup, const char* namefield, double *data){
	const char* extension = get_filename_ext(namefile);
	if (strcmp(extension, "bp") == 0){
#if defined(key_adios) || defined(key_adios_nompi)
		adios_read_0d_double(namefile, namegroup, namefield, data);
#else
		error("Kinebec has not been compiled with Adios dependency. Please re-compile it or choose another output file format");
#endif
	}
	else if (strcmp(extension, "h5") == 0){
#if defined(key_phdf5) || defined(key_hdf5)
		hdf5_read_0d_double(namefile, namegroup, namefield, data);
#else
		error("Kinebec has not been compiled with Hdf5 dependency. Please re-compile it or choose another output file format");
#endif
	}
	else
		error("There is no alternative to adios_interface or hdf5_interface for reading data at present time");
}


void read_0d_int(const char* namefile, const char* namegroup, const char* namefield, int *data){
	const char* extension = get_filename_ext(namefile);
	if (strcmp(extension, "bp") == 0){
#if defined(key_adios) || defined(key_adios_nompi)
		adios_read_0d_int(namefile, namegroup, namefield, data);
#else
		error("Kinebec has not been compiled with Adios dependency. Please re-compile it or choose another output file format");
#endif
	}
	else if (strcmp(extension, "h5") == 0){
#if defined(key_phdf5) || defined(key_hdf5)
		hdf5_read_0d_int(namefile, namegroup, namefield, data);
#else
		error("Kinebec has not been compiled with Hdf5 dependency. Please re-compile it or choose another output file format");
#endif
	}
	else
		error("There is no alternative to adios_interface or hdf5_interface for reading data at present time");
}


void read_2d(const char* namefile, const char* namegroup, const char* namefield, double ***data, PARAMETERS *parameters){
	const char* extension = get_filename_ext(namefile);
	if (strcmp(extension, "bp") == 0){
#if defined(key_adios) || defined(key_adios_nompi)
		adios_read_2d(namefile, namegroup, namefield, data, parameters);
#else
		error("Kinebec has not been compiled with Adios dependency. Please re-compile it or choose another output file format");
#endif
	}
	else if (strcmp(extension, "h5") == 0){
#if defined(key_phdf5) || defined(key_hdf5)
		hdf5_read_2d(namefile, namegroup, namefield, data, parameters);
#else
		error("Kinebec has not been compiled with Hdf5 dependency. Please re-compile it or choose another output file format");
#endif
	}
	else
		error("There is no alternative to adios_interface or hdf5_interface for reading data at present time");
}


void read_3d(const char* namefile, const char* namegroup, const char* namefield, double ****data, PARAMETERS *parameters){
	const char* extension = get_filename_ext(namefile);
	if (strcmp(extension, "bp") == 0){
#if defined(key_adios) || defined(key_adios_nompi)
		adios_read_3d(namefile, namegroup, namefield, data, parameters);
#else
		error("Kinebec has not been compiled with Adios dependency. Please re-compile it or choose another output file format");
#endif
	}
	else if (strcmp(extension, "h5") == 0){
#if defined(key_phdf5) || defined(key_hdf5)
		hdf5_read_3d(namefile, namegroup, namefield, data, parameters);
#else
		error("Kinebec has not been compiled with Hdf5 dependency. Please re-compile it or choose another output file format");
#endif
	}
	else
		error("There is no alternative to adios_interface or hdf5_interface for reading data at present time");
}




int search_in_file(const char* namefile, const char* dataname){
	const char* extension = get_filename_ext(namefile);
	if (strcmp(extension, "bp") == 0){
#if defined(key_adios) || defined(key_adios_nompi)
                error("The function adios_search_in_file is not implemented yet.");
#else
		error("Kinebec has not been compiled with Adios dependency. Please re-compile it or choose another output file format");
#endif
	}
	else if (strcmp(extension, "h5") == 0){
#if defined(key_phdf5) || defined(key_hdf5)
		return hdf5_search_in_file(namefile, dataname);
#else
		error("Kinebec has not been compiled with Hdf5 dependency. Please re-compile it or choose another output file format");
#endif
	}
	else
		error("There is no alternative to adios_interface or hdf5_interface for reading data at present time");
}

