#include "./src/IO/read_parameters.h"



int count_commas(char* buf){
	int res = 0;
	char *c;

	c = buf;
	while (((int)(*c)) != ((int)'\0')){
		if (((int)(*c)) == ((int)(',')))
			res++;
		c++;
	}

	return res;
}



PARAMETERS read_setup_file(const char* namefile){
	PARAMETERS parameters;

	FILE* f;
	char buf[MAXSTRINGLENGTH];
	char message[MAXSTRINGLENGTH];
	int istatus, ibuf;
	char* sstatus;
	double constant_cutoff;
	double pi = 4.*atan(1.);
	int ncommas;

	double vsquare;

	int rank;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
#endif

	if (rank == 0){
		f = fopen(namefile, "r");
		if (f != NULL){

			// Read header of setup file
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			if (strcmp(buf, "# Setup file for a spectral simulation of 2D Classical Boltzmann equation #\n") == 0){
				parameters.ndim = DIM2;
				parameters.type_collision = CLASSICAL_COLLISIONS;
			}
			else if (strcmp(buf, "# Setup file for a spectral simulation of 3D Classical Boltzmann equation #\n") == 0){
				parameters.ndim = DIM3;
				parameters.type_collision = CLASSICAL_COLLISIONS;
			}
			else if (strcmp(buf, "# Setup file for a spectral simulation of 2D Quantum Boltzmann equation #\n") == 0){
				parameters.ndim = DIM2;
				parameters.type_collision = QUANTUM_COLLISIONS;
			}
			else if (strcmp(buf, "# Setup file for a spectral simulation of 3D Quantum Boltzmann equation #\n") == 0){
				parameters.ndim = DIM3;
				parameters.type_collision = QUANTUM_COLLISIONS;
			}
			else
				error("Problem with the global header of setup file");

			sstatus = fgets(buf, sizeof(buf), f);

			// Read parameters for initial distribution
			sstatus = fgets(buf, sizeof(buf), f);
			if (strcmp(buf, "# Parameters for defining the initial distribution\n") == 0){
				parameters.type_f0 = FROM_FILE;
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%s", parameters.filef0);

				parameters.rho0 = -1.;
				parameters.T0 = -1.;
				parameters.sigma0 = -1.;
			}
			else if (strcmp(buf, "# Parameters for defining the initial distribution (Maxwellian)\n") == 0){
				parameters.type_f0 = CLASSICAL_MAXWELLIAN;
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf", &parameters.rho0);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				
				parameters.u0 = allocation_1d(parameters.ndim+1);
				ncommas = count_commas(buf);
				switch (parameters.ndim){
					case DIM2 :
						if (ncommas == 1)
							istatus = sscanf(buf, "%lf,%lf", &parameters.u0[0], &parameters.u0[1]);
						else
							error("Please provide the two components of the averaged velocity by separating them with a comma.");
						break;
					case DIM3 :
						if (ncommas == 2)
							istatus = sscanf(buf, "%lf,%lf,%lf", &parameters.u0[0], &parameters.u0[1], &parameters.u0[2]);
						else
							error("Please provide the three components of the averaged velocity by separating them with commas.");
						break;
				}
				
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				//istatus = sscanf(buf, "%lf", &parameters.T0);
				//parameters.sigma0 = sqrt(parameters.T0);
				//parameters.e0 = -1.;

				istatus = sscanf(buf, "%lf", &(parameters.sigma0));
				parameters.e0 = 0.5*(parameters.ndim+1.)*parameters.sigma0;
				parameters.T0 = -1.;

				parameters.zfug = -1.;
				parameters.m0 = -1.;
				strcpy(parameters.filef0, "");
			}

			else if (strcmp(buf, "# Parameters for defining the initial distribution (ball indicator)\n") == 0){
				parameters.type_f0 = BALL_INDICATOR;
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf", &parameters.rho0);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				
				parameters.u0 = allocation_1d(parameters.ndim+1);
				ncommas = count_commas(buf);
				switch (parameters.ndim){
					case DIM2 :
						if (ncommas == 1)
							istatus = sscanf(buf, "%lf,%lf", &parameters.u0[0], &parameters.u0[1]);
						else
							error("Please provide the two components of the averaged velocity by separating them with a comma.");
						break;
					case DIM3 :
						if (ncommas == 2)
							istatus = sscanf(buf, "%lf,%lf,%lf", &parameters.u0[0], &parameters.u0[1], &parameters.u0[2]);
						else
							error("Please provide the three components of the averaged velocity by separating them with commas.");
						break;
				}
				
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				//istatus = sscanf(buf, "%lf", &parameters.T0);
				//parameters.sigma0 = sqrt(parameters.T0);
				//parameters.e0 = -1.;

				istatus = sscanf(buf, "%lf", &(parameters.e0));
				parameters.T0 = -1.;
				parameters.sigma0 = -1.;

				parameters.zfug = -1.;
				parameters.m0 = -1.;
				strcpy(parameters.filef0, "");
			}


			else if (strcmp(buf, "# Parameters for defining the initial distribution (bi-Maxwellian)\n") == 0){
				parameters.type_f0 = CLASSICAL_BIMAXWELLIAN;
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf", &parameters.rho0);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				parameters.u1 = allocation_1d(parameters.ndim+1);
				ncommas = count_commas(buf);
				switch (parameters.ndim){
					case DIM2 :
						if (ncommas == 1)
							istatus = sscanf(buf, "%lf,%lf", &parameters.u1[0], &parameters.u1[1]);
						else
							error("Please provide the two components of velocity u1 by separating them with a comma.");
						break;
					case DIM3 :
						if (ncommas == 2)
							istatus = sscanf(buf, "%lf,%lf,%lf", &parameters.u1[0], &parameters.u1[1], &parameters.u1[2]);
						else
							error("Please provide the three components of velocity u1 by separating them with commas.");
						break;
				}
				
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				parameters.u2 = allocation_1d(parameters.ndim+1);
				ncommas = count_commas(buf);
				switch (parameters.ndim){
					case DIM2 :
						if (ncommas == 1)
							istatus = sscanf(buf, "%lf,%lf", &parameters.u2[0], &parameters.u2[1]);
						else
							error("Please provide the two components of velocity u2 by separating them with a comma.");
						break;
					case DIM3 :
						if (ncommas == 2)
							istatus = sscanf(buf, "%lf,%lf,%lf", &parameters.u2[0], &parameters.u2[1], &parameters.u2[2]);
						else
							error("Please provide the three components of velocity u2 by separating them with commas.");
						break;
				}

				parameters.u0 = allocation_1d(parameters.ndim+1);
				for (int d = 0; d <= parameters.ndim; d++)
					parameters.u0[d] = 0.5*(parameters.u1[d]+parameters.u2[d]);
				
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				// istatus = sscanf(buf, "%lf", &parameters.T0);
				// parameters.sigma0 = sqrt(parameters.T0);

				// vsquare = 0.;
				// for (int d = 0; d <= parameters.ndim; d++)
				// 	vsquare += pow(parameters.u1[d]-parameters.u2[d], 2);
				// parameters.e0 = (vsquare/8. + 0.5*(parameters.ndim+1.)*parameters.T0);
				istatus = sscanf(buf, "%lf", &parameters.sigma0);
				vsquare = 0.;
				for (int d = 0; d <= parameters.ndim; d++)
					vsquare += pow(parameters.u1[d]-parameters.u2[d], 2);
				parameters.e0 = (vsquare/8. + 0.5*(parameters.ndim+1.)*parameters.sigma0);
				parameters.T0 = -1.;

				parameters.zfug = -1.;
				parameters.m0 = -1.;
				strcpy(parameters.filef0, "");
			}
			else if (strcmp(buf, "# Parameters for defining the initial distribution (multi-Maxwellian)\n") == 0){
				parameters.type_f0 = CLASSICAL_MULTIMAXWELLIAN;
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf", &parameters.rho0);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%d", &parameters.nmaxwellians);
				parameters.u = allocation_2d(parameters.nmaxwellians, parameters.ndim+1);
				sstatus = fgets(buf, sizeof(buf), f);
				switch (parameters.ndim){
					case DIM2 :
						for (int k = 0; k < parameters.nmaxwellians; k++){
							sstatus = fgets(buf, sizeof(buf), f);
							ncommas = count_commas(buf);
							if (ncommas == 1)
								istatus = sscanf(buf, "%lf,%lf", &parameters.u[k][0], &parameters.u[k][1]);
							else{
								sprintf(buf, "Please provide the two components of velocity u[%d] by separating them with a comma.", k);
								error(buf);
							}
						}
						break;
					case DIM3 :
						for (int k = 0; k < parameters.nmaxwellians; k++){
							sstatus = fgets(buf, sizeof(buf), f);
							ncommas = count_commas(buf);
							if (ncommas == 2)
								istatus = sscanf(buf, "%lf,%lf,%lf", &parameters.u[k][0], &parameters.u[k][1], &parameters.u[k][2]);
							else{
								sprintf(buf, "Please provide the three components of velocity u[%d] by separating them with commas.", k);
								error(buf);
							}
						}
						break;
				}

				parameters.u0 = allocation_1d(parameters.ndim+1);
				for (int d = 0; d <= parameters.ndim; d++){
					parameters.u0[d] = 0.;
					for (int k = 0; k < parameters.nmaxwellians; k++)
						parameters.u0[d] += parameters.u[k][d];
					parameters.u0[d] /= parameters.nmaxwellians;
				}
				
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				// istatus = sscanf(buf, "%lf", &parameters.T0);
				// parameters.sigma0 = sqrt(parameters.T0);

				// vsquare = 0.;
				// for (int d = 0; d <= parameters.ndim; d++)
				// 	vsquare += pow(parameters.u1[d]-parameters.u2[d], 2);
				// parameters.e0 = (vsquare/8. + 0.5*(parameters.ndim+1.)*parameters.T0);
				istatus = sscanf(buf, "%lf", &parameters.sigma0);
				vsquare = 0.;
				for (int k = 0; k < parameters.nmaxwellians; k++){
					for (int l = 0; l < parameters.nmaxwellians; l++){
						if (k < l){
							for (int d = 0; d <= parameters.ndim; d++)
								vsquare += pow(parameters.u[k][d]-parameters.u[l][d], 2);
						}
					}
				}

				parameters.e0 = (vsquare/(2.*pow(parameters.nmaxwellians, 2)) + 0.5*(parameters.ndim+1.)*parameters.sigma0);
				parameters.T0 = -1.;

				parameters.zfug = -1.;
				parameters.m0 = -1.;
				strcpy(parameters.filef0, "");
			}
			else if (strcmp(buf, "# Parameters for defining the initial distribution (quantum Maxwellian)\n") == 0){
				parameters.type_f0 = QUANTUM_MAXWELLIAN;
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf", &parameters.rho0);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				parameters.u0 = allocation_1d(parameters.ndim+1);
				ncommas = count_commas(buf);
				switch (parameters.ndim){
					case DIM2 :
						if (ncommas == 1)
							istatus = sscanf(buf, "%lf,%lf", &parameters.u0[0], &parameters.u0[1]);
						else
							error("Please provide the two components of the averaged velocity by separating them with a comma.");
						break;
					case DIM3 :
						if (ncommas == 2)
							istatus = sscanf(buf, "%lf,%lf,%lf", &parameters.u0[0], &parameters.u0[1], &parameters.u0[2]);
						else
							error("Please provide the three components of the averaged velocity by separating them with commas.");
						break;
				}
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf", &parameters.e0);
				parameters.T0 = -1.;
				parameters.sigma0 = -1.;

				parameters.zfug = -1.;
				parameters.m0 = -1.;
				// The temperature, the fugacity and the critical mass will be computed once hbar is extracted (see below)
				strcpy(parameters.filef0, "");
			}
			else if (strcmp(buf, "# Parameters for defining the Boltzmann-Krook-Wu (BKW) solution\n") == 0){
				parameters.type_f0 = CLASSICAL_BKW;
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf", &parameters.sigma0);
				//parameters.T0 = pow(parameters.sigma0, 2);
				parameters.rho0 = 1.;
				parameters.u0 = allocation_1d(2);
				parameters.u0[0] = 0.;
				parameters.u0[1] = 0.;

				parameters.e0 = pow(parameters.sigma0, 2) + 0.5*pow(parameters.u0[0]-parameters.u0[1], 2);

				parameters.T0 = -1.;
				parameters.zfug = -1.;
				parameters.m0 = -1.;
				strcpy(parameters.filef0, "");
			}
			else
				error("Problem with the header of initial distribution parameters in setup file");

			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);

			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			istatus = sscanf(buf, "%lf", &parameters.Ccross);
			switch(parameters.ndim){
				case DIM2 :
					if (parameters.type_f0 == CLASSICAL_BKW){
						//parameters.Ccross = 1./(2.*pi);
						parameters.Ccross = pow(parameters.sigma0,2) * 1./(2.*pi);
					}

					parameters.gamma = 0.;
					break;

				case DIM3 :
					parameters.gamma = 1.;
					break;

				default :
					error("Cannot identify the dimensionality of the simulation");
			}

			if (parameters.type_f0 != CLASSICAL_BKW){
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf", &parameters.knudsen);
			}
			else
				parameters.knudsen = 1.;

			if (parameters.type_collision == QUANTUM_COLLISIONS){
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%d", &parameters.type_qparticles);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf", &parameters.hbar);
			}
			else{
				parameters.type_qparticles = -1;
				parameters.hbar = -1.;
			}

			if (parameters.type_f0 != FROM_FILE){
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				if ((parameters.type_collision == CLASSICAL_COLLISIONS) || (parameters.type_collision == QUANTUM_COLLISIONS)){
					istatus = sscanf(buf, "%d", &ibuf);
					switch (ibuf){
						case 1:
							parameters.type_cutoff = R_IS_KS;
							break;
						case 2:
							parameters.type_cutoff = R_IS_GIVEN;
							break;
						default:
							error("Unknown definition of cut-off type.");
					}

					sstatus = fgets(buf, sizeof(buf), f);
					sstatus = fgets(buf, sizeof(buf), f);
					istatus = sscanf(buf, "%lf", &constant_cutoff);
				}
				else
					error("Problem with the collision kernel definition\n");
			}
			else{
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf", &parameters.cutoff);
				if (parameters.cutoff <= 0.)
					parameters.type_cutoff = FROM_FILE;
				else
					parameters.type_cutoff = R_IS_GIVEN;
			}

			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			istatus = sscanf(buf, "%d", &ibuf);
			parameters.force_convolution = (ibuf == 1);

			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			ncommas = count_commas(buf);
			switch(parameters.ndim){
				case DIM2:
					if (parameters.force_convolution){
						if (ncommas == 2)
							istatus = sscanf(buf, "%d,%d,%d", &(parameters.P1_int), &(parameters.P2_int), &(parameters.P3_int));
						else
							error("Please provide the three numbers of integration points by separating them with commas.");
						parameters.mint1 = 0;
						parameters.mint2 = 0;
						parameters.mint = 0;
					}
					else{
						if (ncommas == 0)
							istatus = sscanf(buf, "%d", &(parameters.mint));
						else
							error("Please provide the unique number of integration points without any comma.");
						parameters.mint1 = 0;
						parameters.mint2 = 0;
						parameters.P1_int = 0;
						parameters.P2_int = 0;
						parameters.P3_int = 0;
					}
					/*
					sstatus = fgets(buf, sizeof(buf), f);
					istatus = sscanf(buf, "%d", &parameters.mint);
					parameters.mint1 = 0;
					parameters.mint2 = 0;
					*/
					break;
				case DIM3:
					if (parameters.force_convolution){
						if (ncommas == 2)
							istatus = sscanf(buf, "%d,%d,%d", &(parameters.P1_int), &(parameters.P2_int), &(parameters.P3_int));
						else
							error("Please provide the three numbers of integration points by separating them with commas.");
						parameters.mint1 = 0;
						parameters.mint2 = 0;
						parameters.mint = 0;
					}
					else{
						if (ncommas == 1)
							istatus = sscanf(buf, "%d,%d", &(parameters.mint1), &(parameters.mint2));
						else
							error("Please provide the two numbers of integration points by separating them with a comma.");
						parameters.mint = parameters.mint1 * parameters.mint2;
						parameters.P1_int = 0;
						parameters.P2_int = 0;
						parameters.P3_int = 0;
					}
					/*
					sstatus = fgets(buf, sizeof(buf), f);
					istatus = sscanf(buf, "%d", &parameters.mint1);
					sstatus = fgets(buf, sizeof(buf), f);
					istatus = sscanf(buf, "%d", &parameters.mint2);
					parameters.mint = parameters.mint1*parameters.mint2;
					*/
					break;
				default :
					error("Problem with dimensionality of the setup file");
			}

			if (parameters.type_f0 != FROM_FILE){
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);

				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf", &parameters.L);

				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%d", &parameters.type_support);

				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%d", &parameters.nvx);
				if (parameters.nvx % 2 != 0)
					error("The number of points in vx direction must be a multiple of 2");

				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%d", &parameters.nvy);
				if (parameters.nvy % 2 != 0)
					error("The number of points in vy direction must be a multiple of 2");

				if (parameters.ndim == DIM3){
					sstatus = fgets(buf, sizeof(buf), f);
					sstatus = fgets(buf, sizeof(buf), f);
					istatus = sscanf(buf, "%d", &parameters.nvz);
					if (parameters.nvz % 2 != 0)
						error("The number of points in vz direction must be a multiple of 2");
				}
				else
					parameters.nvz = 0;
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%d", &ibuf);
				switch (parameters.type_collision){
					case CLASSICAL_COLLISIONS:
						switch (ibuf){
							case 1:
								// Velocity rescaling is not used
								parameters.type_omega = OMEGA_IS_NOTUSED;
								break;
							case 2:
								// Velocity rescaling is used with omega = sqrt(2*ekin/(d*rho))
								parameters.type_omega = OMEGA_IS_SQRTEKINORHO;
								break;
							default:
								error("Something is wrong with the reading of type_omega");
						}
						break;
					case QUANTUM_COLLISIONS:
						switch (ibuf){
							case 1:
								// Velocity rescaling is not used
								parameters.type_omega = OMEGA_IS_NOTUSED;
								break;
							case 2:
								// Velocity rescaling is used with omega = sqrt(temperature)
								parameters.type_omega = OMEGA_IS_SQRTTEMPERATURE;
								break;
							default:
								error("Something is wrong with the reading of type_omega");
						}
						break;
					default:
						error("Problem with the collision kernel definition\n");
				}

				switch(parameters.type_support){
					case NO_ALIASING :
						switch (parameters.type_cutoff){
							case R_IS_KS:
								parameters.supportradius = 2.*pi/(2.*constant_cutoff + 1. + sqrt(2.));
								parameters.cutoff = constant_cutoff*parameters.supportradius;
								break;
							case R_IS_GIVEN:
								parameters.supportradius = (2.*pi-2.*constant_cutoff)/(1.+sqrt(2.));
								parameters.cutoff = constant_cutoff;
								break;
							default:
								error("Problem with the type of support and/or the type of cut-off.");
						}
						
						break;
					default :
						error("Problem with the definition of the initial distribution support\n");
				}
			}
			else{
				parameters.L = -1.;
				parameters.supportradius = -1.;
				parameters.nvx = 0;
				parameters.nvy = 0;
				parameters.nvz = 0;
				parameters.type_support = -1;
				parameters.type_omega = -1;
			}

			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);

			if (parameters.type_f0 == FROM_FILE){
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%d", &parameters.type_Tinit);
				parameters.type_Tinit = 3-parameters.type_Tinit;
			}
			else
				parameters.type_Tinit = SETTOZERO;

			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			istatus = sscanf(buf, "%lf", &parameters.Tfinal);

			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			istatus = sscanf(buf, "%d", &parameters.nt);

			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			istatus = sscanf(buf, "%d", &parameters.time_method);
			
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			istatus = sscanf(buf, "%d", &parameters.use_residual);
			parameters.use_residual = (parameters.use_residual == 1);

			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			istatus = sscanf(buf, "%s", parameters.outputdir);

			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			istatus = sscanf(buf, "%s", parameters.prefix);

			sstatus = fgets(buf, sizeof(buf), f);
			if (strcmp(buf, "# Output file format\n") != 0){
				// Cannot read output file format
#if defined(key_adios) || defined(key_adios_nompi)
				parameters.output_fmt = FMT_BP;
#elif defined(key_phdf5) || defined(key_hdf5)
				parameters.output_fmt = FMT_H5;
#else
				error("There is no alternative to adios_interface or hdf5_interface for saving data at present time");
#endif
			}
			else{
				// Read output file format
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%d", &parameters.output_fmt);
			}
			
			// Read the boolean that indicates if only norms and errors are saved
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			sstatus = fgets(buf, sizeof(buf), f);
			istatus = sscanf(buf, "%d", &ibuf);
			parameters.save_norms_only = (ibuf == 2);
			if (parameters.save_norms_only)
				printf("Save only norms and errors\n");
			fclose(f);
		}
		else{
			// File does not exist
			sprintf(message, "Setup file %s cannot be found", namefile);
			error(message);
		}
	}

#ifdef key_mpi
	MPI_Bcast(&parameters.ndim, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.L, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.type_f0, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.filef0, MAXSTRINGLENGTH, MPI_CHAR, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.nvx, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.nvy, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.nvz, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.type_omega, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.rho0, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.T0, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.e0, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.sigma0, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	switch (parameters.type_f0){
		case CLASSICAL_MAXWELLIAN :
		case QUANTUM_MAXWELLIAN :
		case BALL_INDICATOR:
		case CLASSICAL_BKW :
			if (rank != 0)
				parameters.u0 = allocation_1d(parameters.ndim+1);
			MPI_Bcast(parameters.u0, parameters.ndim+1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
			break;
		case CLASSICAL_BIMAXWELLIAN :
			if (rank != 0){
				parameters.u0 = allocation_1d(parameters.ndim+1);
				parameters.u1 = allocation_1d(parameters.ndim+1);
				parameters.u2 = allocation_1d(parameters.ndim+1);
			}
			MPI_Bcast(parameters.u0, parameters.ndim+1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
			MPI_Bcast(parameters.u1, parameters.ndim+1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
			MPI_Bcast(parameters.u2, parameters.ndim+1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
			break;
		case CLASSICAL_MULTIMAXWELLIAN:
			MPI_Bcast(&(parameters.nmaxwellians), 1, MPI_INT, 0, MPI_COMM_WORLD);
			if (rank != 0){
				parameters.u0 = allocation_1d(parameters.ndim+1);
				parameters.u = allocation_2d(parameters.nmaxwellians, parameters.ndim+1);
			}
			MPI_Bcast(parameters.u0, parameters.ndim+1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
			MPI_Bcast(parameters.u[0], parameters.nmaxwellians*(parameters.ndim+1), MPI_DOUBLE, 0, MPI_COMM_WORLD);
			break;
	}

	MPI_Bcast(&parameters.supportradius, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.type_qparticles, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.hbar, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.zfug, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.m0, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.cutoff, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.type_support, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.type_collision, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.knudsen, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.Ccross, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.gamma, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.type_cutoff, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.mint, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.mint1, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.mint2, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.force_convolution, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.P1_int, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.P2_int, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.P3_int, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.type_Tinit, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.Tfinal, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.time_method, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.use_residual, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.nt, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.outputdir, MAXSTRINGLENGTH, MPI_CHAR, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.prefix, MAXSTRINGLENGTH, MPI_CHAR, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.output_fmt, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parameters.save_norms_only, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif

	parameters.exact_solution = (parameters.type_f0 == CLASSICAL_BKW);

	return parameters;
}
