#include "./src/IO/write_data.h"



void write_logfile(double time, double data, const char* namefile, int newfile){
	FILE *file;
#ifdef key_mpi
	int rank, init;
	MPI_Initialized(&init);
	if (init){
		MPI_Comm_rank(MPI_COMM_WORLD, &rank);
		if (rank == 0){
			if (newfile == 1)
				file = fopen(namefile, "w");
			else
				file = fopen(namefile, "ab+");
			fprintf(file, "%e %.15e\n", time, data);
			fclose(file);
		}
	}
	else{
		if (newfile == 1)
			file = fopen(namefile, "w");
		else
			file = fopen(namefile, "ab+");
		fprintf(file, "%e %.15e\n", time, data);
		fclose(file);
	}
#else
	if (newfile == 1)
		file = fopen(namefile, "w");
	else
		file = fopen(namefile, "ab+");
	fprintf(file, "%e %.15e\n", time, data);
	fclose(file);
#endif
}



void cut_logfile(const char *namefile, int nlines){
	char cmd[3*MAXSTRINGLENGTH];
	int rank;
	char backupfile[MAXSTRINGLENGTH];
	int istatus;
	
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
#endif
	if (rank == 0){
		sprintf(backupfile, "%s_backup", namefile);
		sprintf(cmd, "head -n +%d %s > %s", nlines, namefile, backupfile);
		istatus = system(cmd);
		sprintf(cmd, "mv %s %s", backupfile, namefile);
		istatus = system(cmd);
	}
}



void write_0d_double(const char* namefile, const char* namegroup, const char* namefield, double data){
	const char* ext = get_filename_ext(namefile);
	if (strcmp(ext, "bp") == 0){
#if defined(key_adios) || defined(key_adios_nompi)
		adios_write_0d_double(namefile, namegroup, namefield, data);
#else
		error("Kinebec has not been compiled with Adios dependency. Please re-compile it or choose another output file format");
#endif
	}
	else if (strcmp(ext, "h5") == 0){
#if defined(key_phdf5) || defined(key_hdf5)
		hdf5_write_0d_double(namefile, namegroup, namefield, data);
#else
		error("Kinebec has not been compiled with Hdf5 dependency. Please re-compile it or choose another output file format");
#endif
	}
	else
		error("There is no alternative to adios_interface or hdf5_interface for saving data at present time");
}

void write_0d_int(const char* namefile, const char* namegroup, const char* namefield, int data){
	const char* ext = get_filename_ext(namefile);
	if (strcmp(ext, "bp") == 0){
#if defined(key_adios) || defined(key_adios_nompi)
		adios_write_0d_int(namefile, namegroup, namefield, data);
#else
		error("Kinebec has not been compiled with Adios dependency. Please re-compile it or choose another output file format");
#endif
	}
	else if (strcmp(ext, "h5") == 0){
#if defined(key_phdf5) || defined(key_hdf5)
		hdf5_write_0d_int(namefile, namegroup, namefield, data);
#else
		error("Kinebec has not been compiled with Hdf5 dependency. Please re-compile it or choose another output file format");
#endif
	}
	else
		error("There is no alternative to adios_interface or hdf5_interface for saving data at present time");
}



void write_2d(const char* namefile, const char* namegroup, const char* namefield, double** floc, PARAMETERS parameters){
	const char* ext = get_filename_ext(namefile);
	if (strcmp(ext, "bp") == 0){
#if defined(key_adios) || defined(key_adios_nompi)
		adios_write_2d(namefile, namegroup, namefield, floc, parameters);
#else
		error("Kinebec has not been compiled with Adios dependency. Please re-compile it or choose another output file format");
#endif
	}
	else if (strcmp(ext, "h5") == 0){
#if defined(key_phdf5) || defined(key_hdf5)
		hdf5_write_2d(namefile, namegroup, namefield, floc, parameters);
#else
		error("Kinebec has not been compiled with Hdf5 dependency. Please re-compile it or choose another output file format");
#endif
	}
	else
		error("There is no alternative to adios_interface or hdf5_interface for saving data at present time");
}


void write_3d(const char* namefile, const char* namegroup, const char* namefield, double*** floc, PARAMETERS parameters){
	const char* ext = get_filename_ext(namefile);
	if (strcmp(ext, "bp") == 0){
#if defined(key_adios) || defined(key_adios_nompi)
		adios_write_3d(namefile, namegroup, namefield, floc, parameters);
#else
		error("Kinebec has not been compiled with Adios dependency. Please re-compile it or choose another output file format");
#endif
	}
	else if (strcmp(ext, "h5") == 0){
#if defined(key_phdf5) || defined(key_hdf5)
		hdf5_write_3d(namefile, namegroup, namefield, floc, parameters);
#else
		error("Kinebec has not been compiled with Hdf5 dependency. Please re-compile it or choose another output file format");
#endif
	}
	else
		error("There is no alternative to adios_interface or hdf5_interface for saving data at present time");
}



void write_enriched_parameters(PARAMETERS parameters){
	int rank;
	char namefile[2*MAXSTRINGLENGTH+17];
	FILE *f;

#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
#endif

	if (rank == 0){
		sprintf(namefile, "%s/%s_parameters.txt", parameters.outputdir, parameters.prefix);
		f = fopen(namefile, "w");

		fprintf(f, "# Dimensionality in velocity\n");
		fprintf(f, "%d\n", parameters.ndim+1);

		fprintf(f, "# Type of initial distribution\n");
		switch (parameters.type_f0){
			case FROM_FILE:
				fprintf(f, "from file\n");
				break;
			case CLASSICAL_MAXWELLIAN:
				fprintf(f, "classical maxwellian\n");
				break;
			case CLASSICAL_BIMAXWELLIAN:
				fprintf(f, "classical bimaxwellian\n");
				break;
			case CLASSICAL_MULTIMAXWELLIAN:
				fprintf(f, "classical multi-maxwellian\n");
				break;
			case QUANTUM_MAXWELLIAN:
				fprintf(f, "quantum maxwellian\n");
				break;
			case CLASSICAL_BKW:
				fprintf(f, "classical BKW\n");
				break;
			case BALL_INDICATOR:
				fprintf(f, "ball indicator\n");
				break;
			default:
				error("Unknown type of initial distribution");
		}
		if (parameters.type_f0 != FROM_FILE){
			fprintf(f, "# Density of initial distribution\n");
			fprintf(f, "%.15e\n", parameters.rho0);

			if (parameters.type_f0 == CLASSICAL_BIMAXWELLIAN){
				fprintf(f, "# Velocities of initial distribution\n");
				fprintf(f, "%.15e,%.15e", parameters.u1[0], parameters.u1[1]);
				if (parameters.ndim == DIM3)
					fprintf(f, ",%.15e", parameters.u1[2]);
				fprintf(f, "\n");
				fprintf(f, "%.15e,%.15e", parameters.u2[0], parameters.u2[1]);
				if (parameters.ndim == DIM3)
					fprintf(f, ",%.15e", parameters.u2[2]);
				fprintf(f, "\n");
			}
			if (parameters.type_f0 == CLASSICAL_MULTIMAXWELLIAN){
				fprintf(f, "# Number of velocities in initial distribution\n");
				fprintf(f, "%d\n", parameters.nmaxwellians);
				fprintf(f, "# Velocity vectors of initial distribution\n");
				for (int k = 0; k < parameters.nmaxwellians; k++){
					fprintf(f, "%.15e,%.15e", parameters.u[k][0], parameters.u[k][1]);
				if (parameters.ndim == DIM3)
					fprintf(f, ",%.15e", parameters.u[k][2]);
				fprintf(f, "\n");
				}
			}

			fprintf(f, "# Average velocity of initial distribution\n");
			fprintf(f, "%.15e,%.15e", parameters.u0[0], parameters.u0[1]);
			if (parameters.ndim == DIM3)
				fprintf(f, ",%.15e", parameters.u0[2]);
			fprintf(f, "\n");

			fprintf(f, "# Thermal energy of initial distribution\n");
			fprintf(f, "%.15e\n", parameters.sigma0);

			fprintf(f, "# Internal energy of initial distribution\n");
			fprintf(f, "%.15e\n", parameters.e0);

			fprintf(f, "# Temperature of initial distribution\n");
			fprintf(f, "%.15e\n", parameters.T0);

			if (parameters.type_f0 == QUANTUM_MAXWELLIAN){
				fprintf(f, "# Fugacity of initial distribution\n");
				fprintf(f, "%.15e\n", parameters.zfug);
				fprintf(f, "# Critical mass of initial distribution\n");
				fprintf(f, "%.15e\n", parameters.m0);
			}
		}



		fprintf(f, "# Type of collisions\n");
		switch(parameters.type_collision){
			case CLASSICAL_COLLISIONS:
				fprintf(f, "classical\n");
				break;
			case QUANTUM_COLLISIONS:
				fprintf(f, "quantum\n");
				break;
			default:
				error("Unknown type of collisions");
		}
		if (parameters.type_collision == QUANTUM_COLLISIONS){
			fprintf(f, "# Type of quantum particles\n");
			switch (parameters.type_qparticles){
				case FERMIONS:
					fprintf(f, "fermions\n");
					break;
				case BOSONS:
					fprintf(f, "bosons\n");
					break;
				default:
					error("Unknown type of particles");
			}
			fprintf(f, "# Value of hbar\n");
			fprintf(f, "%.15e\n", parameters.hbar);

			fprintf(f, "# Quantum degeneracy\n");
			if (parameters.condensation)
				fprintf(f, "True\n");
			else
				fprintf(f, "False\n");

		}

		fprintf(f, "# Existence of exact solution\n");
		if (parameters.exact_solution)
			fprintf(f, "True\n");
		else
			fprintf(f, "False\n");


		fprintf(f, "# Knudsen number\n");
		fprintf(f, "%.15e\n", parameters.knudsen);
		fprintf(f, "# Amplitude of cross-section\n");
		fprintf(f, "%.15e\n", parameters.Ccross);
		fprintf(f, "# Exponent of cross-section radial part\n");
		fprintf(f, "%.15e\n", parameters.gamma);

		fprintf(f, "# Value of L\n");
		fprintf(f, "%.15e\n", parameters.L);

		fprintf(f, "# Definition of cutoff radius R in the collision integral\n");
		switch (parameters.type_cutoff){
			case R_IS_KS:
				fprintf(f, "R=k*S\n");
				break;
			case R_IS_GIVEN:
				fprintf(f, "given\n");
				break;
			default:
				error("Unknown definition of cutoff");
		}
		fprintf(f, "# Cutoff radius R in the collision integral\n");
		fprintf(f, "%.15e\n", parameters.cutoff);

		fprintf(f, "# Link between cutoff radius R and support radius S\n");
		switch (parameters.type_support){
			case NO_ALIASING:
				fprintf(f, "Anti-aliasing\n");
				break;
			default:
				error("Unknown type of initial distribution support");
		}
		fprintf(f, "# Support radius of initial distribution\n");
		fprintf(f, "%.15e\n", parameters.supportradius);

		fprintf(f, "# Linf-norm of initial distribution that has been cut\n");
		fprintf(f, "%.15e\n", parameters.maxcut);

		fprintf(f, "# Velocity grid size\n");
		fprintf(f, "%d,%d", parameters.nvx, parameters.nvy);
		if (parameters.ndim == DIM3)
			fprintf(f, ",%d", parameters.nvz);
		fprintf(f, "\n");
		fprintf(f, "# Number of cross-section integration points\n");
		switch (parameters.ndim){
			case DIM2:
				fprintf(f, "%d\n", parameters.mint);
				break;
			case DIM3:
				fprintf(f, "%d,%d\n", parameters.mint1, parameters.mint2);
				break;
			default:
				error("Unknown dimensionality");
		}

		fprintf(f, "# Type of time initialization\n");
		switch(parameters.type_Tinit){
			case FROM_FILE:
				fprintf(f, "from file\n");
				break;
			case SETTOZERO:
				fprintf(f, "set to 0\n");
				break;
			default:
				error("Unknown type of time initialization");
		}
		fprintf(f, "# Initial time\n");
		fprintf(f, "%.15e\n", parameters.Tinit);
		fprintf(f, "# Final time\n");
		fprintf(f, "%.15e\n", parameters.Tfinal);
		fprintf(f, "# Time semi-discretization\n");
		switch (parameters.time_method){
			case EULER:
				fprintf(f, "Euler\n");
				break;
			case RK2:
				fprintf(f, "RK2\n");
				break;
			default:
				error("Unknown time method");
		}

		fprintf(f, "# Use of residual method\n");
		if (parameters.use_residual)
			fprintf(f, "True\n");
		else
			fprintf(f, "False\n");
		
		fprintf(f, "# Use of velocity rescaling method\n");
		if (parameters.type_omega == OMEGA_IS_NOTUSED)
			fprintf(f, "False\n");
		else
			fprintf(f, "True\n");
		
		fprintf(f, "# Number of time steps\n");
		fprintf(f, "%d\n", parameters.nt);

		fprintf(f, "# Output directory\n");
		fprintf(f, "%s\n", parameters.outputdir);

		fprintf(f, "# Prefix of output files\n");
		fprintf(f, "%s\n", parameters.prefix);

		fprintf(f, "# Resumed simulation\n");
		if (parameters.resume)
			fprintf(f, "True\n");
		else
			fprintf(f, "False\n");
		
		fprintf(f, "# Initial time index\n");
		fprintf(f, "%d\n", parameters.n_init);

		fprintf(f, "# Output file format\n");
		switch (parameters.output_fmt){
			case FMT_BP:
				fprintf(f, "Adios");
				break;
			case FMT_H5:
				fprintf(f, "Hdf5\n");
				break;
			default:
				error("Unknown output file format");
		}

		fprintf(f, "# Save all results\n");
		if (parameters.save_norms_only)
			fprintf(f, "False\n");
		else
			fprintf(f, "True\n");

		fclose(f);
	}
}