#include "./src/IO/write_parameters.h"


void write_setup_file(const char* namefile, int ndim, int collision_kind, int f0_kind){
	FILE* f = NULL;

	f = fopen(namefile, "w+");

	switch(collision_kind){
		case CLASSICAL_COLLISIONS :
			switch(ndim){
				case DIM2 :
					fprintf(f, "###########################################################################\n");
					fprintf(f, "# Setup file for a spectral simulation of 2D Classical Boltzmann equation #\n");
					fprintf(f, "###########################################################################\n");
					break;
				case DIM3 :
					fprintf(f, "###########################################################################\n");
					fprintf(f, "# Setup file for a spectral simulation of 3D Classical Boltzmann equation #\n");
					fprintf(f, "###########################################################################\n");
					break;
				default :
					error("Unknown value for 2nd argument in write_setup_file");
			}
			break;
		case QUANTUM_COLLISIONS :
			switch(ndim){
				case DIM2 :
					fprintf(f, "#########################################################################\n");
					fprintf(f, "# Setup file for a spectral simulation of 2D Quantum Boltzmann equation #\n");
					fprintf(f, "#########################################################################\n");
					break;
				case DIM3 :
					fprintf(f, "#########################################################################\n");
					fprintf(f, "# Setup file for a spectral simulation of 3D Quantum Boltzmann equation #\n");
					fprintf(f, "#########################################################################\n");
					break;
				default :
					error("Unknown value for 2nd argument in write_setup_file");
			}
			break;
		default :
			error("Unknown value for 3rd argument in write_setup_file");
	}

	switch(f0_kind){
		case FROM_FILE :
			fprintf(f, "# Parameters for defining the initial distribution\n");
			fprintf(f, "##################################################\n");
			fprintf(f, "# File that contains the initial distribution (only .bp or .h5 files are handled)\n");
			fprintf(f, "./\n");
			break;

		case CLASSICAL_MAXWELLIAN :
			fprintf(f, "# Parameters for defining the initial distribution (Maxwellian)\n");
			fprintf(f, "###############################################################\n");
			fprintf(f, "# Initial density rho (strictly positive constant)\n");
			fprintf(f, "1.\n");
			switch (ndim){
				case DIM2 :
					fprintf(f, "# Initial averaged velocity (ux,uy) (uniform 2d vector) (Warning: do not forget the separating comma)\n");
					fprintf(f, "0.,0.\n");
					break;
				case DIM3 :
					fprintf(f, "# Initial averaged velocity (ux,uy,uz) (uniform 3d vector) (Warning: do not forget the separating commas)\n");
					fprintf(f, "0.,0.,0.\n");
					
			}
			fprintf(f, "# Initial thermal energy sigma (strictly positive constant)\n");
			fprintf(f, "1.\n");
			break;
		
		case BALL_INDICATOR :
			fprintf(f, "# Parameters for defining the initial distribution (ball indicator)\n");
			fprintf(f, "###################################################################\n");
			fprintf(f, "# Initial density rho (strictly positive constant)\n");
			fprintf(f, "1.\n");
			switch (ndim){
				case DIM2 :
					fprintf(f, "# Initial averaged velocity (ux,uy) (uniform 2d vector) (Warning: do not forget the separating comma)\n");
					fprintf(f, "0.,0.\n");
					break;
				case DIM3 :
					fprintf(f, "# Initial averaged velocity (ux,uy,uz) (uniform 3d vector) (Warning: do not forget the separating commas)\n");
					fprintf(f, "0.,0.,0.\n");
					
			}
			fprintf(f, "# Initial internal energy e (strictly positive constant)\n");
			fprintf(f, "1.\n");
			break;

		case CLASSICAL_BIMAXWELLIAN :
			fprintf(f, "# Parameters for defining the initial distribution (bi-Maxwellian)\n");
			fprintf(f, "##################################################################\n");
			fprintf(f, "# Initial density rho (strictly positive constant)\n");
			fprintf(f, "1.\n");
			switch (ndim){
				case DIM2 :
					fprintf(f, "# Initial velocity (ux_1,uy_1) (uniform 2d vector) (Warning: do not forget the separating comma)\n");
					fprintf(f, "0.,0.\n");
					fprintf(f, "# Initial velocity (ux_2,uy_2) (uniform 2d vector) (Warning: do not forget the separating comma)\n");
					fprintf(f, "0.,0.\n");
					break;
				case DIM3 :
					fprintf(f, "# Initial velocity (ux_1,uy_1,uz_1) (uniform 3d vector) (Warning: do not forget the separating commas)\n");
					fprintf(f, "0.,0.,0.\n");
					fprintf(f, "# Initial velocity (ux_2,uy_2,uz_2) (uniform 3d vector) (Warning: do not forget the separating commas)\n");
					fprintf(f, "0.,0.,0.\n");
					break;
			}
			fprintf(f, "# Initial thermal energy sigma (strictly positive constant)\n");
			fprintf(f, "1.\n");
			break;
		
		case CLASSICAL_MULTIMAXWELLIAN :
			fprintf(f, "# Parameters for defining the initial distribution (multi-Maxwellian)\n");
			fprintf(f, "#####################################################################\n");
			fprintf(f, "# Initial density rho, common to all maxwellians (strictly positive constant)\n");
			fprintf(f, "1.\n");
			fprintf(f, "# Number of maxwellians to be considered (at least one)\n");
			fprintf(f, "1\n");
			switch (ndim){
				case DIM2 :
					fprintf(f, "# Initial velocities (ux_k,uy_k) (uniform 2d vectors) (Warning: one line per vector) (Warning: do not forget the separating comma)\n");
					fprintf(f, "0.,0.\n");
					break;
				case DIM3 :
					fprintf(f, "# Initial velocity (ux_k,uy_k,uz_k) (uniform 3d vector) (Warning: one line per vector) (Warning: do not forget the separating commas)\n");
					fprintf(f, "0.,0.,0.\n");
					break;
			}
			fprintf(f, "# Initial thermal energy sigma, common to all maxwellians (strictly positive constant)\n");
			fprintf(f, "1.\n");
			break;

		case QUANTUM_MAXWELLIAN :
			fprintf(f, "# Parameters for defining the initial distribution (quantum Maxwellian)\n");
			fprintf(f, "#######################################################################\n");
			fprintf(f, "# Initial density rho (strictly positive constant)\n");
			fprintf(f, "1.\n");
			switch (ndim){
				case DIM2 :
					fprintf(f, "# Initial averaged velocity (ux,uy) (uniform 2d vector) (Warning: do not forget the separating comma)\n");
					fprintf(f, "0.,0.\n");
					break;
				case DIM3 :
					fprintf(f, "# Initial averaged velocity (ux,uy,uz) (uniform 3d vector) (Warning: do not forget the separating commas)\n");
					fprintf(f, "0.,0.,0.\n");
					
			}
			fprintf(f, "# Initial internal energy e (strictly positive constant)\n");
			fprintf(f, "1.\n");
			break;
			
		case CLASSICAL_BKW :
			fprintf(f, "# Parameters for defining the Boltzmann-Krook-Wu (BKW) solution\n");
			fprintf(f, "###############################################################\n");
			fprintf(f, "# Initial temperature (positive constant)\n");
			fprintf(f, "1.\n");
			break;
			
		default :
			error("Unknown value for 4th argument in write_setup_file");
	}

	fprintf(f, "# Parameters for defining the collision operator\n");
	fprintf(f, "################################################\n");
	switch(ndim){
		case DIM2 :
			fprintf(f, "# Value of C in cross-section definition for Maxwellian molecules B(|u|,theta) = C\n");
			fprintf(f, "1.\n");
			break;

		case DIM3 :
			fprintf(f, "# Value of C in cross-section definition for variable hard sphere model B(|u|,theta) = C*|u|\n");
			fprintf(f, "1.\n");
			break;

		default :
			error("Unknown value for 2nd argument in write_setup_file");
	}
	switch(collision_kind){
		case CLASSICAL_COLLISIONS :
			if (f0_kind != CLASSICAL_BKW){
				fprintf(f, "# Knudsen number\n");
				fprintf(f, "1.\n");
			}
			break;
		case QUANTUM_COLLISIONS :
			fprintf(f, "# Knudsen number\n");
			fprintf(f, "1.\n");
			fprintf(f, "# Kind of quantum particles\n");
			fprintf(f, "# 1 - Bosons\n");
			fprintf(f, "# 2 - Fermions\n");
			fprintf(f, "1\n");
			fprintf(f, "# Rescaled Planck constant\n");
			fprintf(f, "1.\n");
			break;
		default :
			error("Unknown value for 2nd argument in write_setup_file");
	}


	if (f0_kind != FROM_FILE){
		fprintf(f, "# Cut-off support B(0,R) for the integrals within the collision kernel\n");
		fprintf(f, "# 1 - R = k*S        where B(0,S) is the support of initial distribution and k >= 1\n");
		fprintf(f, "# 2 - R = k          where k in [0,pi] is given by the user\n");
		fprintf(f, "1\n");
		
		fprintf(f, "# Provide a value for the constant k\n");
		fprintf(f, "1.\n");
	}
	else{
		fprintf(f, "# Parameter R for cut-off support B(0,R) for the integrals within the collision kernel\n");
		fprintf(f, "# Provide a non-positive value for using the value of R provided in the initial distribution file\n");
		fprintf(f, "-1.\n");
	}

	fprintf(f, "# Use quadrature rules for computing cross-section modes and enforce discrete convolutions\n");
	fprintf(f, "# -1- Yes\n");
	fprintf(f, "# -2- No\n");
	fprintf(f, "2\n");

	fprintf(f, "# Number of integration points\n");
	fprintf(f, "# If you use quadrature rules, please provide 3 integers separated by commas for defining the accuracy orders.\n");
	switch (ndim){
		case DIM2 :
			fprintf(f, "# Else, provide the number of integration points on half-circle (one integer >= 4)\n");
			fprintf(f, "4\n");
			break;
		
		case DIM3 :
			fprintf(f, "# Else, provide the numbers of integration points on half-circle (two integers >= 4 separated by a comma)\n");
			fprintf(f, "4,4\n");
			break;
		
		default :
			error("Unknown value for 2nd argument in write_setup_file");
	}

	/*switch(ndim){
		case DIM2 :
			fprintf(f, "# Number of integration points on half-circle (one integer >= 4)\n");
			fprintf(f, "4\n");
			break;
		case DIM3 :
			fprintf(f, "# Number of integration points on half-sphere (two integers >= 4)\n");
			fprintf(f, "4\n");
			fprintf(f, "4\n");
			break;
		default :
			error("Unknown value for 2nd argument in write_setup_file");
	}*/

	if (f0_kind != FROM_FILE){
		switch(ndim){
			case DIM2 :
				fprintf(f, "# Parameters for defining the velocity domain [-L,L]^2 and the velocity grid\n");
				fprintf(f, "############################################################################\n");
				break;
			case DIM3 :
				fprintf(f, "# Parameters for defining the velocity domain [-L,L]^3 and the velocity grid\n");
				fprintf(f, "############################################################################\n");
				break;
			default:
				error("Unknown value for 2nd argument in write_setup_file");
		}
		fprintf(f, "# Value of L (positive constant)\n");
		fprintf(f, "8.\n");
		fprintf(f, "# Relation between L and the support B(0,S) of initial distribution\n");
		fprintf(f, "# 1 - S maximal such that 2*pi >= 2*R + (1+sqrt(2))*S\n");
		fprintf(f, "1\n");

		fprintf(f, "# Number of points in vx direction (must be a multiple of 2)\n");
		fprintf(f, "64\n");
		fprintf(f, "# Number of points in vy-direction (must be a multiple of 2)\n");
		fprintf(f, "64\n");
		if (ndim == DIM3){
			fprintf(f, "# Number of points in vz-direction (must be a multiple of 2)\n");
			fprintf(f, "64\n");
		}
		
		fprintf(f, "# Use of velocity rescaling method\n");
		fprintf(f, "# -1- Do not use it\n");
		switch (collision_kind){
			case CLASSICAL_COLLISIONS:
				switch (ndim){
					case DIM2:
						fprintf(f, "# -2- Use it with omega = sqrt(kinetic energy / mass)\n");
						break;
					case DIM3:
						fprintf(f, "# -2- Use it with omega = sqrt(2/3 * kinetic energy / mass)\n");
						break;
					default :
						error("Unknown value for 2nd argument in write_setup_file");
				}
				break;
			case QUANTUM_COLLISIONS:
				fprintf(f, "# -2- Use it with omega = sqrt(temperature)\n");
				break;
			default:
				error("Unknown value for 3rd argument in write_setup_file");
		}
		fprintf(f, "1\n");
	}

	fprintf(f, "# Parameters for defining the time domain\n");
	fprintf(f, "#########################################\n");
	if (f0_kind == FROM_FILE){
		fprintf(f, "# Initial time\n");
		fprintf(f, "# 1 - Set to 0\n");
		fprintf(f, "# 2 - Associated to the external initial distribution\n");
		fprintf(f, "1\n");
	}
	fprintf(f, "# Final time\n");
	fprintf(f, "1.\n");

	fprintf(f, "# Parameters of the numerical method in time\n");
	fprintf(f, "############################################\n");
	fprintf(f, "# Number of time steps\n");
	fprintf(f, "100\n");
	fprintf(f, "# Numerical solver in time\n");
	fprintf(f, "# 1 - Euler method\n");
	fprintf(f, "# 2 - Runge-Kutta 2 method\n");
	fprintf(f, "1\n");
	fprintf(f, "# Use of residual method\n");
	fprintf(f, "# 1 - Yes\n");
	fprintf(f, "# 2 - No\n");
	fprintf(f, "1\n");


	fprintf(f, "# Parameters for output results\n");
	fprintf(f, "###############################\n");
	fprintf(f, "# Output directory (no slash symbol at the end)\n");
	fprintf(f, "./results\n");
	fprintf(f, "# Prefix name of output files\n");
	fprintf(f, "diags\n");
#if (defined(key_adios) || defined(key_adios_nompi)) && (defined(key_phdf5) || defined(key_hdf5))
	fprintf(f, "# Output file format\n");
	fprintf(f, "# 1 - Adios (.bp format)\n");
	fprintf(f, "# 2 - Hdf5 (.h5 format)\n");
	fprintf(f, "2\n");
#elif (defined(key_adios) || defined(key_adios_nompi) || defined(key_phdf5) || defined(key_hdf5))
#else
	error("There is no alternative to adios_interface or hdf5_interface for saving data at present time");
#endif
	fprintf(f, "# Save data\n");
	fprintf(f, "# 1 - All results (2D/3D, norms, errors)\n");
	fprintf(f, "# 2 - Only norms and errors\n");
	fprintf(f, "1\n");

	fclose(f);
}

