#include "./src/UI/generate_setup_file.h"


int nb_alloc_1d, nb_alloc_2d, nb_alloc_3d;
int nb_alloc_1d_char;
int nb_alloc_1d_int;


void print_setup_generator_help(void)
{
	printf("The way to use this program is the following\n");
	printf("    generate_setup_file [setup_file]\n");
	printf("  For printing this manual\n");
	printf("    generate_setup_file -help\n");
}



void setup_generator_ui(int *ndim, int *collision_kind, int *f0_kind)
{
	int istatus;
	int ibuf;

	*ndim = 0;
	*collision_kind = 0;
	*f0_kind = 0;
	ibuf = 0;

	printf("\n");
	printf("****************************\n");
	printf("KINEBEC setup file generator\n");
	printf("****************************\n");

	printf("\n");
	printf("Choose velocity dimensionality\n");
	printf("1 - 2D (vx,vy)\n");
	printf("2 - 3D (vx,vy,vz)\n");

	while ((*ndim != DIM2) && (*ndim != DIM3))
	{
		printf("\n");
		printf("Your choice : ");
		istatus = scanf("%d", ndim);
	}

	printf("\n");
	printf("Choose collision kernel kind\n");
	printf("1 - Classical collisions\n");
	printf("2 - Quantum collisions\n");
	while ((*collision_kind != QUANTUM_COLLISIONS) && (*collision_kind != CLASSICAL_COLLISIONS))
	{
		printf("\n");
		printf("Your choice : ");
		istatus = scanf("%d", &*collision_kind);
	}

	printf("\n");
	printf("Choose initial distribution\n");
	switch(*collision_kind)
	{
		case CLASSICAL_COLLISIONS :
			// Initial distribution for classical collisions
			printf("1 - From file\n");
			printf("2 - Classical Maxwellian distribution\n");
			printf("3 - Classical Bi-Maxwellian distribution\n");
			printf("4 - Classical multi-Maxwellian distribution\n");
			printf("5 - Ball indicator\n");
			if (*ndim == DIM2){
				printf("6 - Classical Boltzmann-Krook-Wu exact solution\n");
				while ((ibuf != 1) && (ibuf != 2) && (ibuf != 3) && (ibuf != 4) && (ibuf != 5) && (ibuf != 6))
				{
					printf("\n");
					printf("Your choice : ");
					istatus = scanf("%d", &ibuf);
				}
			}
			else{
				while ((ibuf != 1) && (ibuf != 2) && (ibuf != 3) && (ibuf != 4) && (ibuf != 5))
				{
					printf("\n");
					printf("Your choice : ");
					istatus = scanf("%d", &ibuf);
				}
			}
			switch (ibuf){
				case 1:
					*f0_kind = FROM_FILE;
					break;
				case 2:
					*f0_kind = CLASSICAL_MAXWELLIAN;
					break;
				case 3:
					*f0_kind = CLASSICAL_BIMAXWELLIAN;
					break;
				case 4:
					*f0_kind = CLASSICAL_MULTIMAXWELLIAN;
					break;
				case 5:
					*f0_kind = BALL_INDICATOR;
					break;
				case 6:
					*f0_kind = CLASSICAL_BKW;
					break;
				default:
					error("Problem in generate_setup_file.c");
			}
			
			break;
		case QUANTUM_COLLISIONS :
			// Initial distribution for classical collisions
			printf("1 - From file\n");
			printf("2 - Classical Maxwellian distribution\n");
			printf("3 - Classical Bi-Maxwellian distribution\n");
			printf("4 - Classical multi-Maxwellian distribution\n");
			printf("5 - Quantum Maxwellian distribution\n");
			printf("6 - Ball indicator\n");
			while ((ibuf != 1) && (ibuf != 2) && (ibuf != 3) && (ibuf != 4) && (ibuf != 5) && (ibuf != 6))
			{
				printf("\n");
				printf("Your choice : ");
				istatus = scanf("%d", &ibuf);
			}
			
			switch (ibuf){
				case 1:
					*f0_kind = FROM_FILE;
					break;
				case 2:
					*f0_kind = CLASSICAL_MAXWELLIAN;
					break;
				case 3:
					*f0_kind = CLASSICAL_BIMAXWELLIAN;
					break;
				case 4:
					*f0_kind = CLASSICAL_MULTIMAXWELLIAN;
					break;
				case 5:
					*f0_kind = QUANTUM_MAXWELLIAN;
					break;
				case 6:
					*f0_kind = BALL_INDICATOR;
					break;
				default:
					error("Problem in generate_setup_file.c");
			}
			
			break;
		default :
			error("Unknown collision kernel kind");
	}
}



void main(int argc, char *argv[])
{
	int ndim, collision_kind, f0_kind;

	if (argc > 1)
	{
		if (strcmp(argv[1], "-help") == 0)
		{
			print_setup_generator_help();
		}
		else
		{
			setup_generator_ui(&ndim, &collision_kind, &f0_kind);
			write_setup_file(argv[1], ndim, collision_kind, f0_kind);
		}
	}
	else
	{
		setup_generator_ui(&ndim, &collision_kind, &f0_kind);
		write_setup_file("./parameters", ndim, collision_kind, f0_kind);
	}
}
