#!/usr/bin/env python
# -*- coding: utf-8 -*-

import subprocess
import sys
import os
import numpy as np

try:
	mprof_exec = subprocess.check_output(['which', 'mprof']).decode("utf-8").split('\n')[0]+" run --multiprocess"
	mprof_exec = mprof_exec.replace("~", "/home/amouton")
except subprocess.CalledProcessError:
	mprof_exec = ""

def write_script_zeus_slurm(namesetupfile, nameexec, nCPUs=1, force_archlimits=False, namejob="KINEBEC", memprofile_freq=-1., memory="", mailuser="", namesubmitfile="", parallel="serial", walltime=[1,0,0], resume=False):
	buf = "#!/bin/sh\n"
	
	"""
	if (nCPUs > 16):
		nnodes = nCPUs//16
	else:
		nnodes = 1
	ntaskspernode = nCPUs//nnodes
	
	buf += "#SBATCH --nodes="+str(nnodes)+"\n"
	buf += "#SBATCH --ntasks-per-node="+str(ntaskspernode)+"\n"
	"""

	if (parallel == "serial"):
		buf += "#SBATCH --ntasks=1\n"
	elif (parallel == "omp"):
		"""
		if (nCPUs > 16):
			nnodes = nCPUs//16
		else:
			ntasks = 1
		ncorespertask = nCPUs//ntasks
		buf += "#SBATCH --ntasks="+str(ntasks)+"\n"
		buf += "#SBATCH --cpus-per-task="+str(ncorespertask)+"\n"
		"""
		buf += "#SBATCH --nodes=1\n"
		buf += "#SBATCH --ntasks-per-node=1\n"
		buf += "#SBATCH --cpus-per-task="+str(nCPUs)+"\n"
	elif (parallel == "mpi"):
		#buf += "#SBATCH --ntasks="+str(nCPUs)+"\n"
		if (nCPUs > 16):
			nnodes = nCPUs//16
		else:
			nnodes = 1
		ntaskspernode = nCPUs//nnodes
	
		buf += "#SBATCH --nodes="+str(nnodes)+"\n"
		buf += "#SBATCH --ntasks-per-node="+str(ntaskspernode)+"\n"
	elif (parallel == "cuda"):
		buf += "#SBATCH --nodes=1\n"
		buf += "#SBATCH --ntasks-per-node=1\n"
		buf += "#SBATCH --gres=gpu:1\n"


	buf += "#SBATCH --job-name="+namejob+"\n"
	if (walltime != "max"):
		buf += "#SBATCH --time="+str(walltime[0])+":"+str(walltime[1])+":"+str(walltime[2])+"\n"
	if (memory != ""):
		buf += "#SBATCH --mem="+memory+"\n"
	
	buf += "#SBATCH --output="+namejob+".out\n"
	buf += "#SBATCH --error="+namejob+".err\n"
	
	#buf += "#SBATCH --exclude=nyx-09-07,nyx-09-08,nyx-09-09,nyx-09-10,nyx-09-16,nyx-09-19,nyx-09-20,nyx-09-21,nyx-09-22,nyx-09-23,nyx-09-24,nyx-09-30,nyx-09-31,nyx-09-32\n"
	#buf += "#SBATCH --exclude=nyx-09-[01-40],orion-14-[01-24]\n"
	
	if (mailuser != ""):
		buf += "#SBATCH --mail-type=ALL\n"
		buf += "#SBATCH --mail-user="+mailuser+"\n"
	
	buf += "module load fftw/3.3.4/gcc/4.8.5\n"
	if (parallel == "mpi"):
		buf += "module load gcc/4.8.5/openmpi/1.8.4 hdf5/1.8.18/gcc/4.8.5/mpi\n"
		buf += "module load hdf5/1.8.18/gcc/4.8.5/mpi\n"
		if (nCPUs > 1):
			buf += "/share/compilers/gcc/4.8.5/openmpi/1.8.4/bin/mpirun -mca pml ob1 -np "+str(nCPUs)+" "
	elif (parallel == "omp"):
		buf += "module load hdf5/1.8.18/gcc/4.8.5/seq\n"
		buf += "export OMP_NUM_THREADS="+str(nCPUs)+"\n"
	elif (parallel == "cuda"):
		buf += "module load nvidia/cuda/10.0/compilers\n"
		buf += "module load gcc/6.1.0/compilers\n"
	
	if ((memprofile_freq > 0.) and (mprof_exec != "")):
		memprofile_logfile = namesetupfile.replace(".dat", "_memprofile.dat")
		mprof_prefix = mprof_exec + " -T "+str(memprofile_freq)+" -o "+memprofile_logfile+" "
	else:
		mprof_prefix = ""
	buf += mprof_prefix + nameexec+" "+namesetupfile
	if (parallel == "cuda"):
		# WARNING: nCPUs stands for the total number of grid points
		k = int(np.ceil(np.sqrt(nCPUs)))
		buf += " -nb "+str(k)+" -nt "+str(k)
	if (resume):
		buf += " --resume"
	buf += "\n"
	
	if (namesubmitfile == ""):
		namesubmitfile = namesetupfile+"_submit.slurm"
	
	f = open(namesubmitfile, 'w')
	f.write(buf)
	f.close()
	


def write_setupfile(dic, namesetupfile):
	buf = ""
	if (dic["type_collision"] == "classical"):
		if (dic["ndims"] == 2):
			buf += "###########################################################################\n"
			buf += "# Setup file for a spectral simulation of 2D Classical Boltzmann equation #\n"
			buf += "###########################################################################\n"
		elif (dic["ndims"] == 3):
			buf += "###########################################################################\n"
			buf += "# Setup file for a spectral simulation of 3D Classical Boltzmann equation #\n"
			buf += "###########################################################################\n"
	elif (dic["type_collision"] == "quantum"):
		if (dic["ndims"] == 2):
			buf += "#########################################################################\n"
			buf += "# Setup file for a spectral simulation of 2D Quantum Boltzmann equation #\n"
			buf += "#########################################################################\n"
		elif (dic["ndims"] == 3):
			buf += "#########################################################################\n"
			buf += "# Setup file for a spectral simulation of 3D Quantum Boltzmann equation #\n"
			buf += "#########################################################################\n"
	
	if (dic["type_initialization"] == "from_file"):
		buf += "# Parameters for defining the initial distribution\n"
		buf += "##################################################\n"
		buf += "# File that contains the initial distribution (only .bp or .h5 files are handled)\n"
		buf += dic["init_file"]+"\n"
		
	elif (dic["type_initialization"] == "classical_maxwellian"):
		buf += "# Parameters for defining the initial distribution (Maxwellian)\n"
		buf += "###############################################################\n"
		buf += "# Initial density rho (strictly positive constant)\n"
		buf += str(dic["rho0"])+"\n"
		if (dic["ndims"] == 2):
			buf += "# Initial averaged velocity (ux,uy) (uniform 2d vector) (Warning: do not forget the separating comma)\n"
			buf += str(dic["u0"][0])+","+str(dic["u0"][1])+"\n"
		if (dic["ndims"] == 3):
			buf += "# Initial averaged velocity (ux,uy,uz) (uniform 3d vector) (Warning: do not forget the separating commas)\n"
			buf += str(dic["u0"][0])+","+str(dic["u0"][1])+","+str(dic["u0"][2])+"\n"
		buf += "# Initial thermal energy sigma (strictly positive constant)\n"
		buf += str(dic["sigma0"])+"\n"
	
	elif (dic["type_initialization"] == "ball_indicator"):
		buf += "# Parameters for defining the initial distribution (ball indicator)\n"
		buf += "###################################################################\n"
		buf += "# Initial density rho (strictly positive constant)\n"
		buf += str(dic["rho0"])+"\n"
		if (dic["ndims"] == 2):
			buf += "# Initial averaged velocity (ux,uy) (uniform 2d vector) (Warning: do not forget the separating comma)\n"
			buf += str(dic["u0"][0])+","+str(dic["u0"][1])+"\n"
		if (dic["ndims"] == 3):
			buf += "# Initial averaged velocity (ux,uy,uz) (uniform 3d vector) (Warning: do not forget the separating commas)\n"
			buf += str(dic["u0"][0])+","+str(dic["u0"][1])+","+str(dic["u0"][2])+"\n"
		buf += "# Initial internal energy e (strictly positive constant)\n"
		buf += str(dic["e0"])+"\n"
	
	elif (dic["type_initialization"] == "classical_bimaxwellian"):
		buf += "# Parameters for defining the initial distribution (bi-Maxwellian)\n"
		buf += "##################################################################\n"
		buf += "# Initial density rho (strictly positive constant)\n"
		buf += str(dic["rho0"])+"\n"
		if (dic["ndims"] == 2):
			buf += "# Initial velocity (ux_1,uy_1) (uniform 2d vector) (Warning: do not forget the separating comma)\n"
			buf += str(dic["u1"][0])+","+str(dic["u1"][1])+"\n"
			buf += "# Initial velocity (ux_2,uy_2) (uniform 2d vector) (Warning: do not forget the separating comma)\n"
			buf += str(dic["u2"][0])+","+str(dic["u2"][1])+"\n"
		if (dic["ndims"] == 3):
			buf += "# Initial velocity (ux_1,uy_1,uz_1) (uniform 3d vector) (Warning: do not forget the separating commas)\n"
			buf += str(dic["u1"][0])+","+str(dic["u1"][1])+","+str(dic["u1"][2])+"\n"
			buf += "# Initial velocity (ux_2,uy_2,uz_2) (uniform 3d vector) (Warning: do not forget the separating commas)\n"
			buf += str(dic["u2"][0])+","+str(dic["u2"][1])+","+str(dic["u2"][2])+"\n"
		buf += "# Initial thermal energy sigma (strictly positive constant)\n"
		buf += str(dic["sigma0"])+"\n"
	
	elif (dic["type_initialization"] == "classical_multimaxwellian"):
		buf += "# Parameters for defining the initial distribution (multi-Maxwellian)\n"
		buf += "#####################################################################\n"
		buf += "# Initial density rho, common to all maxwellians (strictly positive constant)\n"
		buf += str(dic["rho0"])+"\n"
		buf += "# Number of maxwellians to be considered (at least one)\n"
		buf += str(dic["nmaxwellians"])+"\n"
		buf += "# Initial velocities (ux_k,uy_k) (uniform 2d vectors) (Warning: one line per vector) (Warning: do not forget the separating comma)\n"
		for k in range(len(dic["nmaxwellians"])):
			buf += str(dic["u"][k][0])+","+str(dic["u"][k][1])
			if (dic["ndims"] == 3):
				buf += ","+str(dic["u"][k][2])
			buf += "\n"

		buf += "# Initial thermal energy sigma (strictly positive constant)\n"
		buf += str(dic["sigma0"])+"\n"
	
	elif (dic["type_initialization"] == "quantum_maxwellian"):
		buf += "# Parameters for defining the initial distribution (quantum Maxwellian)\n"
		buf += "#######################################################################\n"
		buf += "# Initial density rho (strictly positive constant)\n"
		buf += str(dic["rho0"])+"\n"
		if (dic["ndims"] == 2):
			buf += "# Initial averaged velocity (ux,uy) (uniform 2d vector) (Warning: do not forget the separating comma)\n"
			buf += str(dic["u0"][0])+","+str(dic["u0"][1])+"\n"
		if (dic["ndims"] == 3):
			buf += "# Initial averaged velocity (ux,uy,uz) (uniform 3d vector) (Warning: do not forget the separating commas)\n"
			buf += str(dic["u0"][0])+","+str(dic["u0"][1])+","+str(dic["u0"][2])+"\n"
		buf += "# Initial internal energy e (strictly positive constant)\n"
		buf += str(dic["e0"])+"\n"
	
	elif (dic["type_initialization"] == "BKW"):
		buf += "# Parameters for defining the Boltzmann-Krook-Wu (BKW) solution\n"
		buf += "###############################################################\n"
		buf += "# Initial temperature (positive constant)\n"
		buf += str(dic["sigma0"])+"\n"
		
	buf += "# Parameters for defining the collision operator\n"
	buf += "################################################\n"
	if (dic["ndims"] == 2):
		buf += "# Value of C in cross-section definition for Maxwellian molecules B(|u|,theta) = C\n"
	elif (dic["ndims"] == 3):
		buf += "# Value of C in cross-section definition for variable hard sphere model B(|u|,theta) = C*|u|\n"
	buf += str(dic["Cphi"])+"\n"

	if (dic["type_initialization"] != "BKW"):
		buf += "# Knudsen number\n"
		buf += str(dic["knudsen"])+"\n"
	
	if (dic["type_collision"] == "quantum"):
		buf += "# Kind of quantum particles\n"
		buf += "# 1 - Bosons\n"
		buf += "# 2 - Fermions\n"
		buf += str(int(dic["type_qparticles"] == "bosons")+2*int(dic["type_qparticles"] == "fermions"))+"\n"
		buf += "# Rescaled Planck constant\n"
		buf += str(dic["hbar"])+"\n"
	
	if (dic["type_initialization"] != "from_file"):
		buf += "# Cut-off support B(0,R) for the integrals within the collision kernel\n"
		buf += "# 1 - R = k*S        where B(0,S) is the support of initial distribution and k >= 1\n"
		buf += "# 2 - R = k          where k in [0,pi] is given by the user\n"
		buf += str(int(dic["type_cutoff"] == "kS") + 2*int(dic["type_cutoff"] == "k")) + "\n"
		
		buf += "# Provide a value for the constant k\n"
		buf += str(dic["cutoff_radius"])+"\n"
	else:
		buf += "# Parameter R for cut-off support B(0,R) for the integrals within the collision kernel\n"
		buf += "# Provide a non-positive value for using the value of R provided in the initial distribution file\n"
		buf += str(dic["cutoff_radius"])+"\n"


	buf += "# Use quadrature rules for computing cross-section modes and enforce discrete convolutions\n"
	buf += "# -1- Yes\n"
	buf += "# -2- No\n"
	buf += str(int(dic["force_convolution"])+2*int(not(dic["force_convolution"])))+"\n"

	buf += "# Number of integration points\n"
	buf += "# If you use quadrature rules, please provide 3 integers separated by commas for defining the accuracy orders.\n"
	
	if (dic["ndims"] == 2):
		#buf += "# Number of integration points on half-circle (one integer >= 4)\n"
		buf += "# Else, provide the number of integration points on half-circle (one integer >= 4)\n"
		if (dic["force_convolution"]):
			buf += str(dic["P1_int"])+","+str(dic["P2_int"])+","+str(dic["P3_int"])+"\n"
		else:
			buf += str(dic["M_int"])+"\n"
	elif (dic["ndims"] == 3):
		#buf += "# Number of integration points on half-sphere (two integers >= 4)\n"
		buf += "# Else, provide the numbers of integration points on half-circle (two integers >= 4 separated by a comma)\n";
		if (dic["force_convolution"]):
			buf += str(dic["P1_int"])+","+str(dic["P2_int"])+","+str(dic["P3_int"])+"\n"
		else:
			buf += str(dic["M1_int"])+","+str(dic["M2_int"])+"\n"

	buf += "# Parameters for defining the velocity domain [-L,L]^2 and the velocity grid\n"
	buf += "############################################################################\n"
	if (dic["type_initialization"] != "from_file"):
		buf += "# Value of L (positive constant)\n"
		buf += str(dic["L"])+"\n"
		buf += "# Relation between L and the support B(0,S) of initial distribution\n"
		buf += "# 1 - S maximal such that pi >= (R + (1+sqrt(2))*S) / 2\n"
		buf += str(int(dic["type_support"] == "anti-aliasing"))+"\n"
	
		buf += "# Number of points in vx direction (must be a multiple of 2)\n"
		buf += str(dic["Nvx"])+"\n"
		buf += "# Number of points in vy direction (must be a multiple of 2)\n"
		buf += str(dic["Nvy"])+"\n"
		if (dic["ndims"] == 3):
			buf += "# Number of points in vz direction (must be a multiple of 2)\n"
			buf += str(dic["Nvz"])+"\n"
	
		buf += "# Use of velocity rescaling method\n"
		buf += "# -1- Do not use it\n"
		if (dic["type_collision"] == "classical"):
			if (dic["ndims"] == 2):
				buf += "# -2- Use it with omega = sqrt(kinetic energy / mass)\n"
			elif (dic["ndims"] == 3):
				buf += "# -2- Use it with omega = sqrt(2/3 * kinetic energy / mass)\n"
		elif (dic["type_collision"] == "quantum"):
			buf += "# -2- Use it with omega = sqrt(temperature)\n"
		buf += str(int(dic["use_rescaling"])+1)+"\n"
	
	buf += "# Parameters for defining the time domain\n"
	buf += "#########################################\n"
	if (dic["type_initialization"] == "from_file"):
		buf += "# Initial time\n"
		buf += "# 1 - Set to 0\n"
		buf += "# 2 - Associated to the external initial distribution\n"
		buf += str(int(dic["type_Tinit"] == "reset")+2*int(dic["type_Tinit"] == "extract"))+"\n"
	buf += "# Final time\n"
	buf += str(dic["Tfinal"])+"\n"
	
	buf += "# Parameters of the numerical method in time\n"
	buf += "############################################\n"
	buf += "# Number of time steps\n"
	buf += str(dic["Nt"])+"\n"
	buf += "# Numerical solver in time\n"
	buf += "# 1 - Euler method\n"
	buf += "# 2 - Runge-Kutta 2 method\n"
	buf += str(int(dic["time_method"] == "Euler") + 2*int(dic["time_method"] == "RK2"))+"\n"
	buf += "# Use of residual method\n"
	buf += "# 1 - Yes\n"
	buf += "# 2 - No\n"
	buf += str(int(not(dic["use_residual"]))+1)+"\n"
	
	buf += "# Parameters for output results\n"
	buf += "###############################\n"
	buf += "# Output directory (no slash symbol at the end)\n"
	buf += dic["outputdir"]+"\n"
	buf += "# Prefix name of output files\n"
	buf += dic["prefix"]+"\n"
	if ("format_results" in dic.keys()):
		buf += "# Output file format\n"
		buf += "# 1 - Adios (.bp format)\n"
		buf += "# 2 - Hdf5 (.h5 format)\n"
		buf += str(int(dic["format_results"] == "ADIOS") + 2*int(dic["format_results"] == "HDF5"))+"\n"
	
	buf += "# Save data\n"
	buf += "# 1 - All results (2D/3D, norms, errors)\n"
	buf += "# 2 - Only norms and errors\n"
	buf += str(int(dic["save_results"] == "all") + 2*int(dic["save_results"] == "only_norms"))
	
	f = open(namesetupfile, 'w')
	f.write(buf)
	f.close()



def read_setupfile(namesetupfile):
	dic = {}
	f = open(namesetupfile, "r")
	lines = f.readlines()
	i = 1
	if (lines[i] == "# Setup file for a spectral simulation of 2D Classical Boltzmann equation #\n"):
		dic["type_collision"] = "classical"
		dic["ndims"] = 2
	elif (lines[i] == "# Setup file for a spectral simulation of 3D Classical Boltzmann equation #\n"):
		dic["type_collision"] = "classical"
		dic["ndims"] = 3
	elif (lines[i] == "# Setup file for a spectral simulation of 2D Quantum Boltzmann equation #\n"):
		dic["type_collision"] = "quantum"
		dic["ndims"] = 2
	elif (lines[i] == "# Setup file for a spectral simulation of 3D Quantum Boltzmann equation #\n"):
		dic["type_collision"] = "quantum"
		dic["ndims"] = 3
	i += 2

	if (lines[i] == "# Parameters for defining the initial distribution\n"):
		dic["type_initialization"] = "from_file"
		i += 3
		dic["init_file"] = lines[i].replace("\n", "")
	
	elif (lines[i] == "# Parameters for defining the initial distribution (Maxwellian)\n"):
		dic["type_initialization"] = "classical_maxwellian"
		i += 3
		dic["rho0"] = float(lines[i].replace("\n", ""))
		i += 2
		dic["u0"] = list(map(float, lines[i].replace("\n", "").split(",")))
		i += 2
		dic["e0"] = float(lines[i].replace("\n", ""))
		i += 1

	elif (lines[i] == "# Parameters for defining the initial distribution (ball indicator)\n"):
		dic["type_initialization"] = "ball_indicator"
		i += 3
		dic["rho0"] = float(lines[i].replace("\n", ""))
		i += 2
		dic["u0"] = list(map(float, lines[i].replace("\n", "").split(",")))
		i += 2
		dic["e0"] = float(lines[i].replace("\n", ""))
		i += 1
	
	elif (lines[i] == "# Parameters for defining the initial distribution (bi-Maxwellian)\n"):
		dic["type_initialization"] = "classical_bimaxwellian"
		i += 3
		dic["rho0"] = float(lines[i].replace("\n", ""))
		i += 2
		dic["u1"] = list(map(float, lines[i].replace("\n", "").split(",")))
		i += 2
		dic["u2"] = list(map(float, lines[i].replace("\n", "").split(",")))
		i += 2
		dic["sigma0"] = float(lines[i].replace("\n", ""))
		i += 1
	
	elif (lines[i] == "# Parameters for defining the initial distribution (multi-Maxwellian)\n"):
		dic["type_initialization"] = "classical_multimaxwellian"
		i += 3
		dic["rho0"] = float(lines[i].replace("\n", ""))
		i += 2
		dic["nmaxwellians"] = int(lines[i].replace("\n", ""))
		i += 2
		dic["u"] = []
		for j in range(dic["nmaxwellians"]):
			newu = list(map(float, lines[i].replace("\n", "").split(",")))
			dic["u"].append(newu)
			i += 1
		i += 1
		dic["sigma0"] = float(lines[i].replace("\n", ""))
		i += 1
	
	elif (lines[i] == "# Parameters for defining the initial distribution (quantum Maxwellian)\n"):
		dic["type_initialization"] = "quantum_maxwellian"
		i += 3
		dic["rho0"] = float(lines[i].replace("\n", ""))
		i += 2
		dic["u0"] = list(map(float, lines[i].replace("\n", "").split(",")))
		i += 2
		dic["e0"] = float(lines[i].replace("\n", ""))
		i += 1
	
	elif (lines[i] == "# Parameters for defining the Boltzmann-Krook-Wu (BKW) solution\n"):
		dic["type_initialization"] = "BKW"
		i += 3
		dic["sigma0"] = float(lines[i].replace("\n", ""))
		i += 1
	
	i += 3
	dic["Cphi"] = float(lines[i].replace("\n", ""))
	i += 1

	if (dic["type_initialization"] != "BKW"):
		i += 1
		dic["knudsen"] = float(lines[i].replace("\n", ""))
		i += 1
	
	if (dic["type_collision"] == "quantum"):
		i += 3
		itemp = int(lines[i].replace("\n", ""))
		if (itemp == 1):
			dic["type_qparticles"] = "bosons"
		elif (itemp == 2):
			dic["type_qparticles"] = "fermions"
		else:
			print("Error while reading the type of quantum particles")
			exit()
		
		i += 2
		dic["hbar"] = float(lines[i].replace("\n", ""))
		i += 1

	if (dic["type_initialization"] != "from_file"):
		i += 3
		itemp = int(lines[i].replace("\n", ""))
		if (itemp == 1):
			dic["type_cutoff"] = "kS"
		elif (itemp == 2):
			dic["type_cutoff"] = "k"
		else:
			print("Error while reading the type of cutoff")
			exit()
		
		i += 2
		dic["cutoff_radius"] = float(lines[i].replace("\n", ""))
		i += 1
	else:
		i += 2
		dic["cutoff_radius"] = float(lines[i].replace("\n", ""))
		i += 1

	i += 3
	
	dic["force_convolution"] = (int(lines[i].replace("\n","")) == 1)
	i += 1

	i += 4


	if (dic["ndims"] == 2):
		if (dic["force_convolution"]):
			[dic["P1_int"], dic["P2_int"], dic["P3_int"]] = map(int, lines[i].replace("\n", "").split(","))
		else:
			dic["M_int"] = int(lines[i].replace("\n", ""))
		i += 1
	elif (dic["ndims"] == 3):
		if (dic["force_convolution"]):
			[dic["P1_int"], dic["P2_int"], dic["P3_int"]] = map(int, lines[i].replace("\n", "").split(","))
		else:
			[dic["M1_int"], dic["M2_int"]] = map(int, lines[i].replace("\n", "").split(","))
			"""
			dic["M1_int"] = int(lines[i].replace("\n", ""))
			i += 1
			dic["M2_int"] = int(lines[i].replace("\n", ""))
			i += 1
			"""
		i += 1

	i += 2

	if (dic["type_initialization"] != "from_file"):
		i += 1
		dic["L"] = float(lines[i].replace("\n", ""))

		i += 3
		itemp = int(lines[i].replace("\n", ""))
		if (itemp == 1):
			dic["type_support"] = "anti-aliasing"
		else:
			print("Error while reading the cutoff of the initial distribution")
			exit()
		
		i += 2
		dic["Nvx"] = int(lines[i].replace("\n", ""))
		i += 2
		dic["Nvy"] = int(lines[i].replace("\n", ""))
		if (dic["ndims"] == 3):
			i += 2
			dic["Nvz"] = int(lines[i].replace("\n", ""))
		
		i += 4
		dic["use_rescaling"] = (int(lines[i].replace("\n", "")) == 2)
		i += 1

	i += 2
	if (dic["type_initialization"] == "from_file"):
		i += 3
		itemp = int(lines[i].replace("\n", ""))
		if (itemp == 1):
			dic["type_Tinit"] = "reset"
		elif (itemp == 2):
			dic["type_Tinit"] = "extract"
		else:
			print("Error while reading the initialization time method")
			exit()
		i += 1
	
	i += 1
	dic["Tfinal"] = float(lines[i].replace("\n", ""))
	i += 1

	i += 3
	dic["Nt"] = int(lines[i].replace("\n", ""))

	i += 4
	itemp = int(lines[i].replace("\n", ""))
	if (itemp == 1):
		dic["time_method"] = "Euler"
	elif (itemp == 2):
		dic["time_method"] = "RK2"
	else:
		print("Error while reading the time method")
		exit()
	
	i += 4
	dic["use_residual"] = (int(lines[i].replace("\n", "")) == 1)
	
	i += 4
	dic["outputdir"] = lines[i].replace("\n", "")

	i += 2
	dic["prefix"] = lines[i].replace("\n", "")

	i += 1
	if (lines[i] == "# Output file format\n"):
		i += 3
		itemp = int(lines[i].replace("\n", ""))
		if (itemp == 1):
			dic["format_results"] = "ADIOS"
		elif (itemp == 2):
			dic["format_results"] = "HDF5"
		else:
			print("Error while reading the output file format")
			exit()
		
		i += 1
	
	i += 3
	itemp = int(lines[i].replace("\n", ""))
	if (itemp == 1):
		dic["save_results"] = "all"
	elif (itemp == 2):
		dic["save_results"] = "only_norms"
	else:
		print("Error while reading the file saving method")
		exit()

	f.close()

	return dic
	






def kinebec_default_setup(dic):
	newdic = dic
	pi = 4.*np.arctan(1.)
	
	if (newdic["type_initialization"] == "from_file"):
		newdic["init_file"] = "./"
	
	elif (newdic["type_initialization"] == "classical_maxwellian"):
		newdic["rho0"] = 1.
		newdic["u0"] = newdic["ndims"]*[0.]
		newdic["sigma0"] = 1.
		newdic["L"] = 8.
		newdic["Tfinal"] = 1.
	
	elif (newdic["type_initialization"] == "ball_indicator"):
		newdic["rho0"] = 1.
		newdic["u0"] = newdic["ndims"]*[0.]
		newdic["e0"] = 1.
		newdic["L"] = 8.
		newdic["Tfinal"] = 1.
	
	elif (newdic["type_initialization"] == "classical_bimaxwellian"):
		if (newdic["type_collision"] == "classical"):
			if (newdic["ndims"] == 2):
				newdic["rho0"] = 1.25
				newdic["u1"] = [1.,0.]
				newdic["u2"] = [-1.,0.]
				newdic["sigma0"] = 0.5
				newdic["L"] = 12.
				newdic["Tfinal"] = 10.
			elif (newdic["ndims"] == 3):
				newdic["rho0"] = 1.
				newdic["u1"] = [1.,1.,0.25]
				newdic["u2"] = [-1.,-1.,-0.25]
				newdic["sigma0"] = 0.2
				newdic["L"] = 6.
				newdic["Tfinal"] = 2.
		
		elif (newdic["type_collision"] == "quantum"):
			if (newdic["type_qparticles"] == "fermions"):
				if (newdic["ndims"] == 2):
					newdic["rho0"] = 1.
					newdic["u1"] = [2.,1.]
					newdic["u2"] = [-2.,-1.]
					newdic["sigma0"] = 1.
					newdic["L"] = 12.
					newdic["Tfinal"] = 10.
				elif (newdic["ndims"] == 3):
					newdic["rho0"] = 1.
					newdic["u1"] = [1.,1.,0.5]
					newdic["u2"] = [-1.,-1.,-0.5]
					newdic["sigma0"] = 1.
					newdic["L"] = 8.
					newdic["Tfinal"] = 2.
			elif (newdic["type_qparticles"] == "bosons"):
				if (newdic["ndims"] == 2):
					newdic["rho0"] = 1.
					newdic["u1"] = [0.,0.5]
					newdic["u2"] = [0.,-0.5]
					newdic["sigma0"] = 0.25
					newdic["L"] = 8.
					newdic["Tfinal"] = 10.
				elif (newdic["ndims"] == 3):
					newdic["rho0"] = 1.
					newdic["u1"] = [1.,1.,0.5]
					newdic["u2"] = [-1.,-1.,-0.5]
					newdic["sigma0"] = 1.
					newdic["L"] = 8.
					newdic["Tfinal"] = 2.
	
	elif (newdic["type_initialization"] == "classical_multimaxwellian"):
		if (newdic["type_collision"] == "classical"):
			if (newdic["ndims"] == 2):
				newdic["rho0"] = 1.25
				newdic["nmaxwellians"] = 2
				newdic["u"] = [[1.,0.], [-1.,0.]]
				newdic["sigma0"] = 0.5
				newdic["L"] = 12.
				newdic["Tfinal"] = 10.
			elif (newdic["ndims"] == 3):
				newdic["rho0"] = 1.
				newdic["nmaxwellians"] = 2
				newdic["u"] = [[1.,1.,0.25], [-1.,-1.,-0.25]]
				newdic["sigma0"] = 0.2
				newdic["L"] = 6.
				newdic["Tfinal"] = 2.
		
		elif (newdic["type_collision"] == "quantum"):
			if (newdic["type_qparticles"] == "fermions"):
				if (newdic["ndims"] == 2):
					newdic["rho0"] = 1.
					newdic["nmaxwellians"] = 2
					newdic["u"] = [[2.,1.], [-2.,-1.]]
					newdic["sigma0"] = 1.
					newdic["L"] = 12.
					newdic["Tfinal"] = 10.
				elif (newdic["ndims"] == 3):
					newdic["rho0"] = 1.
					newdic["nmaxwellians"] = 2
					newdic["u"] = [[1.,1.,0.5], [-1.,-1.,-0.5]]
					newdic["sigma0"] = 1.
					newdic["L"] = 8.
					newdic["Tfinal"] = 2.
			elif (newdic["type_qparticles"] == "bosons"):
				if (newdic["ndims"] == 2):
					newdic["rho0"] = 1.
					newdic["nmaxwellians"] = 2
					newdic["u"] = [[0.,0.5], [0.,-0.5]]
					newdic["sigma0"] = 0.25
					newdic["L"] = 8.
					newdic["Tfinal"] = 10.
				elif (newdic["ndims"] == 3):
					newdic["rho0"] = 1.
					newdic["nmaxwellians"] = 2
					newdic["u"] = [[1.,1.,0.5], [-1.,-1.,-0.5]]
					newdic["sigma0"] = 1.
					newdic["L"] = 8.
					newdic["Tfinal"] = 2.
	
	elif (newdic["type_initialization"] == "quantum_maxwellian"):
		newdic["rho0"] = 1.
		newdic["u0"] = newdic["ndims"]*[0.]
		newdic["e0"] = 1.
		newdic["type_qparticles"] = "fermions"
		newdic["hbar"] = 0.1
		newdic["L"] = 8.
		newdic["Tfinal"] = 1.
	
	elif (dic["type_initialization"] == "BKW"):
		newdic["sigma0"] = pi/6.
		newdic["L"] = 3.
		newdic["Tfinal"] = 10.
	
	if (newdic["type_initialization"] == "BKW"):
		newdic["Cphi"] = 1./(2.*pi)
	else:
		newdic["Cphi"] = 1.
	
	newdic["knudsen"] = 1.
	
	newdic["type_cutoff"] = "kS"
	newdic["cutoff_radius"] = 1.
	
	if (newdic["type_initialization"] == "from_file"):
		newdic["cutoff_radius"] = -1.
	
	newdic["force_convolution"] = False
	if (newdic["ndims"] == 2):
		newdic["M_int"] = 4
	elif (newdic["ndims"] == 3):
		newdic["M1_int"] = 4
		newdic["M2_int"] = 4
	
	if (newdic["type_initialization"] != "from file"):
		newdic["type_support"] = "anti-aliasing"
		
		if (newdic["type_collision"] == "classical"):
			newdic["Nvx"] = 64
			newdic["Nvy"] = 64
			if (newdic["ndims"] == 3):
				newdic["Nvz"] = 64
		elif (newdic["type_collision"] == "quantum"):
			newdic["Nvx"] = 32
			newdic["Nvy"] = 32
			if (newdic["ndims"] == 3):
				newdic["Nvz"] = 32
		
		newdic["use_rescaling"] = True
	
	if (newdic["type_initialization"] == "from_file"):
		newdic["type_Tinit"] = "reset"
		newdic["Tfinal"] = 1.
	
	newdic["Nt"] = 100
	newdic["time_method"] = "RK2"
	newdic["use_residual"] = False
	
	newdic["outputdir"] = "./results"
	newdic["prefix"] = "diags"
	newdic["format_results"] = "HDF5"
	newdic["save_results"] = "all"
	
	return newdic
	
	

def generate_default_setupfile():
	dic = {}

	print("")
	print("****************************")
	print("KINEBEC setup file generator")
	print("****************************")
	print("")
	print("Choose velocity dimensionality")
	print("1 - 2D (vx,vy)")
	print("2 - 3D (vx,vy,vz)")
	ibuf_list = [1,2]
	isnotok = True
	while (isnotok):
		ibuf = input("Your choice: ")
		isnotok = not(ibuf in ibuf_list)
	dic["ndims"] = ibuf+1
	
	print("");
	print("Choose collision kernel kind")
	print("1 - Classical collisions")
	print("2 - Quantum collisions")
	ibuf_list = [1,2]
	isnotok = True
	while (isnotok):
		ibuf = input("Your choice: ")
		isnotok = not(ibuf in ibuf_list)
	if (ibuf == 1):
		dic["type_collision"] = "classical"
	elif (ibuf == 2):
		dic["type_collision"] = "quantum"
	
	print("");
	print("Choose initial distribution")
	print("1 - From file")
	print("2 - Classical Maxwellian distribution")
	print("3 - Classical Bi-Maxwellian distribution")
	if (dic["type_collision"] == "classical"):
		print("4 - Ball indicator")
		if (dic["ndims"] == 2):
			print("5 - Classical Boltzmann-Krook-Wu exact solution")
			ibuf_list = [1,2,3,4,5]
		elif (dic["ndims"] == 3):
			ibuf_list = [1,2,3,4]
	elif (dic["type_collision"] == "quantum"):
		print("4 - Quantum Maxwellian distribution")
		print("5 - Ball indicator")
		ibuf_list = [1,2,3,4,5]
	
	isnotok = True
	while (isnotok):
		ibuf = input("Your choice: ")
		isnotok = not(ibuf in ibuf_list)
	
	if (ibuf == 1):
		dic["type_initialization"] = "from_file"
	elif (ibuf == 2):
		dic["type_initialization"] = "classical_maxwellian"
	elif (ibuf == 3):
		dic["type_initialization"] = "classical_bimaxwellian"
	elif (ibuf == 4):
		if (dic["type_collision"] == "classical"):
			dic["type_initialization"] = "ball_indicator"
		elif (dic["type_collision"] == "quantum"):
			dic["type_initialization"] = "quantum_maxwellian"
	elif (ibuf == 5):
		if ((dic["type_collision"] == "classical") and (dic["ndims"] == 2)):
			dic["type_initialization"] = "BKW"
		elif (dic["type_collision"] == "quantum"):
			dic["type_initialization"] = "ball_indicator"
	
	dic = kinebec_default_setup(dic)
	
	return dic

