#include "base.h"



const char *get_filename_ext(const char *filename){
	const char *dot = strrchr(filename, '.');
	if (!dot || dot == filename)
		return "";
	return dot + 1;
}


void split_path_file(char **p, char **f, char *pf){
	extern int nb_alloc_1d_char;


	/* Find last delimiter. */
	char *z;
	for (z = pf+strlen(pf); z >= pf; z--){
		if (*z == '/' || *z == '\\')
			break;
	}
	
	if (z >= pf){
		/* There is a delimiter: construct separate
		path and filename fragments. */
		//printf("--> %i\n", z-pf);
		*p = malloc(z-pf+1);
		nb_alloc_1d_char++;
		strncpy(*p, pf, z-pf);
		(*p)[z-pf] = '\0';
		*f = malloc(strlen(z));
		nb_alloc_1d_char++;
		strcpy(*f, z+1);
	}
	else{
		/* There is no delimiter: the entire
		string must be a filename. */
		*p = NULL;
		*f = malloc(strlen(pf)+1);
		nb_alloc_1d_char++;
		strcpy(*f, pf);
	}
}


void error(const char* message){
#ifdef key_mpi
	int rank, init;
	MPI_Initialized(&init);
	if (init){
		MPI_Comm_rank(MPI_COMM_WORLD, &rank);
		if (rank == 0)
			printf(BOLDRED "Error: " RESET "%s\n", message);
		MPI_Finalize();
		exit(0);
	}
	else{
		printf(BOLDRED "Error: " RESET "%s\n", message);
		exit(0);
	}
#else
	printf(BOLDRED "Error: " RESET "%s\n", message);
	exit(0);
#endif
}


void exitsuccess(const char* message){
#ifdef key_mpi
	int rank, init;
	MPI_Initialized(&init);
	if (init){
		MPI_Comm_rank(MPI_COMM_WORLD, &rank);
		if (rank == 0)
			printf(BOLDGREEN "End of job: " RESET "%s\n", message);
		MPI_Finalize();
		exit(0);
	}
	else{
		printf(BOLDGREEN "End of job: " RESET "%s\n", message);
		exit(0);
	}
#else
	printf(BOLDGREEN "End of job: " RESET "%s\n", message);
	exit(0);
#endif
}


void message(const char* mymessage){
#ifdef key_mpi
	int rank, init;
	MPI_Initialized(&init);
	if (init){
		MPI_Comm_rank(MPI_COMM_WORLD, &rank);
		if (rank == 0)
			printf("%s", mymessage);
	}
	else
		printf("%s", mymessage);
#else
	printf("%s", mymessage);
#endif
}
