#ifndef BASE_H
#define BASE_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#undef isnan

#define RESET   "\033[0m"
#define BOLD    "\033[1m"

#define BLACK   "\033[30m"      /* Black */
#define RED     "\033[31m"      /* Red */
#define GREEN   "\033[32m"      /* Green */
#define YELLOW  "\033[33m"      /* Yellow */
#define BLUE    "\033[34m"      /* Blue */
#define MAGENTA "\033[35m"      /* Magenta */
#define CYAN    "\033[36m"      /* Cyan */
#define WHITE   "\033[37m"      /* White */
#define BOLDBLACK   "\033[1m\033[30m"      /* Bold Black */
#define BOLDRED     "\033[1m\033[31m"      /* Bold Red */
#define BOLDGREEN   "\033[1m\033[32m"      /* Bold Green */
#define BOLDYELLOW  "\033[1m\033[33m"      /* Bold Yellow */
#define BOLDBLUE    "\033[1m\033[34m"      /* Bold Blue */
#define BOLDMAGENTA "\033[1m\033[35m"      /* Bold Magenta */
#define BOLDCYAN    "\033[1m\033[36m"      /* Bold Cyan */
#define BOLDWHITE   "\033[1m\033[37m"      /* Bold White */


#ifdef key_mpi
#include <mpi.h>
#endif

#ifdef key_openmp
#include <omp.h>
#endif

#ifdef key_fftw_mpi
#include <fftw3-mpi.h>
#endif

#if defined(key_fftw) || defined(key_fftw_openmp)
#include <fftw3.h>
#endif

#ifdef key_cufft
#include <cuda_runtime.h>
#include <cufft.h>
#include <cuda.h>
#endif


#if defined(key_adios) || defined(key_adios_nompi)
#include "adios.h"
#include "adios_types.h"
#include "adios_read.h"
#endif

#if defined(key_phdf5) || defined(key_hdf5)
#include <hdf5.h>
#endif

const char *get_filename_ext(const char *);
void split_path_file(char **, char **, char *);
void error(const char*);
void exitsuccess(const char*);
void message(const char*);

#endif
