#ifndef DEFS_H
#define DEFS_H

//#include "./src/base/allocation.h"
#include "allocation.h"

#define CLASSICAL_COLLISIONS 1
#define QUANTUM_COLLISIONS 2

#define DIM2 1
#define DIM3 2


#define BOSONS 1
#define FERMIONS 2

#define FROM_FILE 1
#define CLASSICAL_MAXWELLIAN 2
#define CLASSICAL_BIMAXWELLIAN 3
#define CLASSICAL_MULTIMAXWELLIAN 4
#define QUANTUM_MAXWELLIAN 5
#define BALL_INDICATOR 6
#define CLASSICAL_BKW 7

#define R_IS_KS 2
#define R_IS_GIVEN 3

#define NO_ALIASING 1


#define EULER 1
#define RK2 2

#define OMEGA_IS_NOTUSED 1
#define OMEGA_IS_SQRTEKINORHO 2
#define OMEGA_IS_SQRTTEMPERATURE 3

#define SETTOZERO 2

#define MAXSTRINGLENGTH 1024

#define MASS 0

#define AVERAGE_UX 100
#define AVERAGE_UY 101
#define AVERAGE_UZ 102
#define AVERAGE_U 103

#define KINETIC_ENERGY 200
#define INTERNAL_ENERGY 201
#define TEMPERATURE 202
#define STRESS_ENERGY_XX 203
#define STRESS_ENERGY_XY 204
#define STRESS_ENERGY_XZ 205
#define STRESS_ENERGY_YX 206
#define STRESS_ENERGY_YY 207
#define STRESS_ENERGY_YZ 208
#define STRESS_ENERGY_ZX 209
#define STRESS_ENERGY_ZY 210
#define STRESS_ENERGY_ZZ 211
#define STRESS_ENERGY 212

#define MOMENTUM4 400
#define MOMENTUM6 600

#define ENTROPY 1000
#define RELATIVE_ENTROPY 1001

#define L1_NORM -100
#define L2_NORM -101
#define LINF_NORM -102
#define MAX_VALUE -103
#define MIN_VALUE -104

#define L1_NORM_LIMIT -200
#define L2_NORM_LIMIT -201
#define LINF_NORM_LIMIT -202

#define L1_ERROR -300
#define L2_ERROR -301
#define LINF_ERROR -302

#define L1_NORM_EX -400
#define L2_NORM_EX -401
#define LINF_NORM_EX -402

#define L1_ERROR_RELAXATION -500
#define L2_ERROR_RELAXATION -501
#define LINF_ERROR_RELAXATION -502

#define FMT_BP 1
#define FMT_H5 2

#define THRESHOLD 1e-14


typedef struct
{
	int ndim; 							// Number of dimensions in velocity (1 for 2D, 2 for 3D)
	int type_f0;						// Type of initial distribution
	double rho0;						// Initial density (constant)
	double T0;							// Initial temperature (constant)
	double sigma0;						// Initial square root of initial temperature
	double e0;							// Initial internal energy
	double m0;							// Initial critical mass (only for 3D bosons)
	double *u0, *u1, *u2, **u;			// Initial velocity (uniform vectors)
	int nmaxwellians;					// Number of maxwellian functions in the initial distribution
	char filef0[MAXSTRINGLENGTH];		// External file that contains a discrete initial distribution
	double supportradius;				// Maximal authorized support radius
	double maxcut;						// Maximal value of initial distribution outside of authorized support
	int condensation;					// Boolean that indicates if Bose-Einstein condensation is expected

	int type_qparticles; 				// Type of quantum particles
	double zfug;						// Fugacity
	double hbar;						// Rescaled Planck constant

	double knudsen; 					// Knudsen number
	int type_collision;					// Kind of collision operator (classical or quantum)
	double Ccross;						// Amplitude coefficient in cross-section definition
	double gamma;						// gamma exponent in Hard spheres, Variable Hard spheres or Coulomb cross-section
	int type_cutoff;					// Link between the cutoff support B(0,R) and the initial data support
	double cutoff;						// Parameter R for cutoff support B(0,R) in integrals

	double L;							// Size of the velocity domain [-L,L]^d
	int type_support;					// Link between the support B(0,S) of initial data and L
	int nvx, nvy, nvz;					// Total number of points in vx, vy, vz directions
	int mint;							// Total number of integration points
	int mint1, mint2;					// Numbers of integration points on half-sphere (only for 3D case)
	int force_convolution;				// Boolean for activate Gamba's method for enforcing discrete convolution in the collision kernel
	int P1_int;							// Accuracy order of Legendre quadrature in Gamba's method in the radial direction
	int P2_int;							// Accuracy order of Lebedev/Legendre quadrature in Gamba's method on the (d-1)-dimensional sphere (1st integral)
	int P3_int;							// Accuracy order of Lebedev/Legendre quadrature in Gamba's method on the (d-1)-dimensional sphere (2nd integral)

	int type_Tinit;						// Type of initial time
	double Tfinal;						// Final time of simulation
	double Tinit;						// Initial time of simulation
	int time_method;					// Numerical method in time
	int use_residual;					// Activate the use of residual method
	int type_omega;						// Definition of omega (velocity rescaling method only)
	int nt;								// Number of time steps

	char outputdir[MAXSTRINGLENGTH];	// Output directory
	char prefix[MAXSTRINGLENGTH]; 		// Output file prefix
	
	int resume;							// Boolean that activates the resume procedure
	int n_init;						// Index of initial time step (default value is 0, can be modified in the case where the --resume flag is used)

	ptrdiff_t nvxloc;					// Local number of points in vx direction
	ptrdiff_t ivx0loc;					// Local offset in vx direction
	ptrdiff_t nvyloc;					// Local number of points in vy direction
	ptrdiff_t ivy0loc;					// Local offset in vy direction
	ptrdiff_t nvzloc;					// Local number of points in vz direction
	ptrdiff_t ivz0loc;					// Local offset in vz direction

	int *ivx0tab;						// Array of all local offsets in vx_direction
	int *nvxtab;						// Array of all local numbers of points in vx direction
	int *ivy0tab;						// Array of all local offsets in vy_direction
	int *nvytab;						// Array of all local numbers of points in vy direction
	int *ivz0tab;						// Array of all local offsets in vz_direction
	int *nvztab;						// Array of all local numbers of points in vz direction
	int *kvxabstab;						// Array of marignal absolute indices in vx-direction (size = nvxloc*nvyloc*nvzloc) not centered in 0 index
	int *kvyabstab;						// Array of marignal absolute indices in vx-direction (size = nvxloc*nvyloc*nvzloc) not centered in 0 index
	int *kvzabstab;						// Array of marignal absolute indices in vx-direction (size = nvxloc*nvyloc*nvzloc) not centered in 0 index

#ifdef key_fftw_mpi
	ptrdiff_t alloc_local;				// Allocation table index
#endif
	int output_fmt;						// Output file format
	
	int exact_solution;					// Boolean that says if there exists an exact solution
	int save_norms_only;				// Boolean that says if only norms and errors are saved or not

} PARAMETERS;


typedef struct
{
	double mass;
	double mass_limit;
	double *u_av;
	double *u_av_limit;
	double ekin;
	double eint;
	double eint_limit;
	double temperature;
	double temperature_limit;
	double **stress_energy;
	double fugacity;
	double fugacity_limit;
	double critical_mass;
	double critical_mass_limit;
	double omega;
	double L1;
	double L1L1;
	double L2L1;
	double LinfL1;
	double L2;
	double L1L2;
	double L2L2;
	double LinfL2;
	double Linf;
	double L1Linf;
	double L2Linf;
	double LinfLinf;
	double maxf;
	double minf;
	double entropy;
	double mom4;
	double mom6;
	double rel_entropy;

	double L1limit;
	double L2limit;
	double Linflimit;
	double relaxerrL1;
	double relaxerrL2;
	double relaxerrLinf;
	double relaxerrL1L1;
	double relaxerrL1L2;
	double relaxerrL1Linf;
	double relaxerrL2L1;
	double relaxerrL2L2;
	double relaxerrL2Linf;
	double relaxerrLinfL1;
	double relaxerrLinfL2;
	double relaxerrLinfLinf;
	
	
	double errL1;
	double errL1L1;
	double errL2L1;
	double errLinfL1;
	double errL2;
	double errL1L2;
	double errL2L2;
	double errLinfL2;
	double errLinf;
	double errL1Linf;
	double errL2Linf;
	double errLinfLinf;
	
	double L1ex;
	double L1L1ex;
	double L2L1ex;
	double LinfL1ex;
	double L2ex;
	double L1L2ex;
	double L2L2ex;
	double LinfL2ex;
	double Linfex;
	double L1Linfex;
	double L2Linfex;
	double LinfLinfex;
	
	double **Gloc_2d;
	double ***Gloc_3d;
	double **Gloclimit_2d;
	double ***Gloclimit_3d;
	double **Glocexact_2d;
	double ***Glocexact_3d;
} DIAGS;



void free_diags(DIAGS, PARAMETERS);
void free_parameters(PARAMETERS);

#endif
