#include "./src/core/collision.h"


#if defined(key_fftw_mpi) || defined(key_fftw_openmp) || defined(key_fftw)
void collision_Q1c_fftw(fftw_complex *Ginput_spectral, fftw_complex *Q1c_spectral, double **alpha, double **alphaprime, PARAMETERS parameters){
	double pi = 4.*atan(1.);
	//int ind;
	fftw_plan plan_g_backward, plan_gprime_backward, plan_bigG_forward;
	fftw_complex *g_real;
	fftw_complex *g_spectral;
	fftw_complex *gprime_real;
	fftw_complex *gprime_spectral;
	fftw_complex *bigG_real;
	fftw_complex *bigG_spectral;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	int i;

	// Reminder : parameters.ndim = 1 for 2D, parameters.ndim = 2 for 3D
	// Reminder : for the 3D case, mint = mint1*mint2
	double C = pow(pi, parameters.ndim)/((double)parameters.mint);

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif
	
#ifdef key_fftw_mpi
	g_real = fftw_alloc_complex(parameters.alloc_local);
	g_spectral = fftw_alloc_complex(parameters.alloc_local);
	gprime_real = fftw_alloc_complex(parameters.alloc_local);
	gprime_spectral = fftw_alloc_complex(parameters.alloc_local);
	bigG_real = fftw_alloc_complex(parameters.alloc_local);
	bigG_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_g_backward = fftw_mpi_plan_dft_2d(MVX, MVY, g_spectral, g_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprime_backward = fftw_mpi_plan_dft_2d(MVX, MVY, gprime_spectral, gprime_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_mpi_plan_dft_2d(MVX, MVY, bigG_real, bigG_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_g_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, g_spectral, g_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprime_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, gprime_spectral, gprime_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, bigG_real, bigG_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	g_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	g_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	gprime_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	gprime_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigG_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigG_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_g_backward = fftw_plan_dft_2d(MVX, MVY, g_spectral, g_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprime_backward = fftw_plan_dft_2d(MVX, MVY, gprime_spectral, gprime_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_plan_dft_2d(MVX, MVY, bigG_real, bigG_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_g_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, g_spectral, g_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprime_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, gprime_spectral, gprime_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, bigG_real, bigG_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif

	// Initialize Q1c_spectral
	#pragma omp parallel for private(i) shared(NVLOC, Q1c_spectral)
	for (i = 0; i < NVLOC; i++){
		Q1c_spectral[i][0] = 0.;
		Q1c_spectral[i][1] = 0.;
	}

	for (int p = 0; p < parameters.mint; p++){
		// Compute ghat_p(k) = alpha_p(k) * fhat(k), gprimehat_p(k) = alphaprime_p(k) * fhat(k)
		// alpha_p and alphaprime_p are real numbers
		#pragma omp parallel for private(i) shared(NVLOC, g_spectral, gprime_spectral, alpha, alphaprime, Ginput_spectral, p)
		for (i = 0; i < NVLOC; i++){
			g_spectral[i][0] = alpha[p][i] * Ginput_spectral[i][0];
			g_spectral[i][1] = alpha[p][i] * Ginput_spectral[i][1];
			gprime_spectral[i][0] = alphaprime[p][i] * Ginput_spectral[i][0];
			gprime_spectral[i][1] = alphaprime[p][i] * Ginput_spectral[i][1];
		}

		// Do inverse FFTs ghat_p -> g_p, gprimehat_p -> gprime_p
		fftw_execute(plan_g_backward);
		fftw_execute(plan_gprime_backward);

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif

		// Renormalize g_p, gprime_p
		#pragma omp parallel for private(i) shared(NVLOC, g_real, gprime_real, MV)
		for (i = 0; i < NVLOC; i++){
			g_real[i][0] /= MV;
			g_real[i][1] = 0.;
			gprime_real[i][0] /= MV;
			gprime_real[i][1] = 0.;
		}
		
		// Compute G_p(k) = g_p(k) * gprime_p(k)
		// g_p and gprime_p are real numbers
		#pragma omp parallel for private(i) shared(NVLOC, g_real, gprime_real, bigG_real)
		for (i = 0; i < NVLOC; i++){
			bigG_real[i][0] = g_real[i][0] * gprime_real[i][0];
			bigG_real[i][1] = 0.;
		}

		// Do FFT G_p -> Ghat_p
		fftw_execute(plan_bigG_forward);

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif

		// Update Q1c_spectral
		#pragma omp parallel for private(i) shared(NVLOC, Q1c_spectral, C, bigG_spectral)
		for (i = 0; i < NVLOC; i++){
			Q1c_spectral[i][0] += C * bigG_spectral[i][0];
			Q1c_spectral[i][1] += C * bigG_spectral[i][1];
		}
	}

	/*for (int i = 0; i < NVLOC; i++)
		printf("Q1c_spectral[%d] = (%.15e,%.15e)\n", i, Q1c_spectral[i][0], Q1c_spectral[i][1]);
	error("stop now");*/
	// Destroy plans
	fftw_destroy_plan(plan_g_backward);
	fftw_destroy_plan(plan_gprime_backward);
	fftw_destroy_plan(plan_bigG_forward);

	// Free memory
	fftw_free(g_real);
	fftw_free(gprime_real);
	fftw_free(g_spectral);
	fftw_free(gprime_spectral);
	fftw_free(bigG_real);
	fftw_free(bigG_spectral);
}



/*
void collision_Q1c_convolution_fftw(fftw_complex *Ginput_spectral, fftw_complex *Q1c_spectral, double ***A, double ***B, double *W, PARAMETERS parameters){
	double pi = 4.*atan(1.);

	fftw_plan plan_AGhat_backward, plan_BGhat_backward, plan_H_forward;
	fftw_complex *AGhat_spectral, *AGhat_real;
	fftw_complex *BGhat_spectral, *BGhat_real;
	fftw_complex *H_real, *H_spectral;

	int i;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

#ifdef key_fftw_mpi
	AGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	AGhat_real = fftw_alloc_complex(parameters.alloc_local);
	BGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	BGhat_real = fftw_alloc_complex(parameters.alloc_local);
	H_real = fftw_alloc_complex(parameters.alloc_local);
	H_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs on A[p]*Ghat (backward), B[p]*Ghat (backward) and H[p] = IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) (forward)
	if (parameters.nvz == 0){
		plan_AGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, AGhat_spectral, AGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, BGhat_spectral, BGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_2d(MVX, MVY, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_AGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, AGhat_spectral, AGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, BGhat_spectral, BGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	AGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	AGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	BGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	BGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs on A[p]*Ghat (backward), B[p]*Ghat (backward) and H[p] = IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) (forward)
	if (parameters.nvz == 0){
		plan_AGhat_backward = fftw_plan_dft_2d(MVX, MVY, AGhat_spectral, AGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_plan_dft_2d(MVX, MVY, BGhat_spectral, BGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_2d(MVX, MVY, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_AGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, AGhat_spectral, AGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, BGhat_spectral, BGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif

	// Initialize Q1c_spectral
	#pragma omp parallel for private(i) shared(NVLOC, Q1c_spectral)
	for (i = 0; i < NVLOC; i++){
		Q1c_spectral[i][0] = 0.;
		Q1c_spectral[i][1] = 0.;
	}

	for (int p = 0; p < parameters.mint; p++){
		// Compute A[p]*Ghat and B[p]*Ghat
		// WARNING : A, B, Ghat are complex numbers
		#pragma omp parallel for private(i) shared(AGhat_spectral, BGhat_spectral, A, B, Ginput_spectral, NVLOC, p)
		for (i = 0; i < NVLOC; i++){
			AGhat_spectral[i][0] = A[p][i][0]*Ginput_spectral[i][0] - A[p][i][1]*Ginput_spectral[i][1];
			AGhat_spectral[i][1] = A[p][i][0]*Ginput_spectral[i][1] + A[p][i][1]*Ginput_spectral[i][0];

			BGhat_spectral[i][0] = B[p][i][0]*Ginput_spectral[i][0] - B[p][i][1]*Ginput_spectral[i][1];
			BGhat_spectral[i][1] = B[p][i][0]*Ginput_spectral[i][1] + B[p][i][1]*Ginput_spectral[i][0];
		}

		// Compute the backward FFT of A[p]*Ghat and B[p]*Ghat
		// WARNING : we get complex numbers
		fftw_execute(plan_AGhat_backward);
		fftw_execute(plan_BGhat_backward);
		#pragma omp parallel for private(i) shared (NVLOC, MV, AGhat_real, BGhat_real)
		for (i = 0; i < NVLOC; i++){
			AGhat_real[i][0] /= ((double)MV);
			AGhat_real[i][1] /= ((double)MV);
			BGhat_real[i][0] /= ((double)MV);
			BGhat_real[i][1] /= ((double)MV);
		}

		// Compute H[p] = IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat)
		#pragma omp parallel for private(i) shared(NVLOC, H_real, AGhat_real, BGhat_real)
		for (i = 0; i < NVLOC; i++){
			H_real[i][0] = AGhat_real[i][0]*BGhat_real[i][0] - AGhat_real[i][1]*BGhat_real[i][1];
			H_real[i][1] = AGhat_real[i][0]*BGhat_real[i][1] + AGhat_real[i][1]*BGhat_real[i][0];
		}

		// Compute the FFT of H[p]
		fftw_execute(plan_H_forward);

		// Increment Q1c_spectral with W[p] * Hhat[p]
		#pragma omp parallel for private(i) shared(NVLOC, p, H_spectral, Q1c_spectral, W)
		for (i = 0; i < NVLOC; i++){
			Q1c_spectral[i][0] += (W[p] * H_spectral[i][0]);
			Q1c_spectral[i][1] += (W[p] * H_spectral[i][1]);
		}
	}
*/
	/*for (int i = 0; i < NVLOC; i++)
		printf("Q1c_spectral[%d] = (%.15e,%.15e)\n", i, Q1c_spectral[i][0], Q1c_spectral[i][1]);
	error("stop now");*/
/*
	// Destroy plans
	fftw_destroy_plan(plan_AGhat_backward);
	fftw_destroy_plan(plan_BGhat_backward);
	fftw_destroy_plan(plan_H_forward);

	// Free memory
	fftw_free(AGhat_spectral);
	fftw_free(AGhat_real);
	fftw_free(BGhat_spectral);
	fftw_free(BGhat_real);
	fftw_free(H_real);
	fftw_free(H_spectral);
}*/


void collision_Q2c_fftw(fftw_complex *Ginput_spectral, fftw_complex *Ginput_real, fftw_complex *Q2c_spectral, double **alpha, double **alphaprime, PARAMETERS parameters){
	double pi = 4.*atan(1.);
	//int ind;
	fftw_plan plan_g_backward, plan_bigG_forward;
	fftw_complex *g_real;
	fftw_complex *g_spectral;
	fftw_complex *bigG_real;
	fftw_complex *bigG_spectral;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	int i;

	// Reminder : parameters.ndim = 1 for 2D, parameters.ndim = 2 for 3D
	// Reminder : for the 3D case, mint = mint1*mint2
	double C = pow(pi, parameters.ndim)/((double)parameters.mint);

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif
	
#ifdef key_fftw_mpi
	g_real = fftw_alloc_complex(parameters.alloc_local);
	g_spectral = fftw_alloc_complex(parameters.alloc_local);
	bigG_real = fftw_alloc_complex(parameters.alloc_local);
	bigG_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_g_backward = fftw_mpi_plan_dft_2d(MVX, MVY, g_spectral, g_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_mpi_plan_dft_2d(MVX, MVY, bigG_real, bigG_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_g_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, g_spectral, g_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, bigG_real, bigG_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	g_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	g_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigG_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigG_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_g_backward = fftw_plan_dft_2d(MVX, MVY, g_spectral, g_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_plan_dft_2d(MVX, MVY, bigG_real, bigG_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_g_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, g_spectral, g_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, bigG_real, bigG_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif

	// Initialize Q2c_spectral
	#pragma omp parallel for private(i) shared(NVLOC, Q2c_spectral)
	for (i = 0; i < NVLOC; i++){
		Q2c_spectral[i][0] = 0.;
		Q2c_spectral[i][1] = 0.;
	}

	for (int p = 0; p < parameters.mint; p++){
		// Compute ghat_p(k) = alpha_p(k) * alphaprime_p(k) * fhat(k)
		// alpha_p and alphaprime_p are real numbers
		#pragma omp parallel for private(i) shared(NVLOC, g_spectral, alpha, alphaprime, p, Ginput_spectral)
		for (i = 0; i < NVLOC; i++){
			g_spectral[i][0] = alpha[p][i]*alphaprime[p][i]*Ginput_spectral[i][0];
			g_spectral[i][1] = alpha[p][i]*alphaprime[p][i]*Ginput_spectral[i][1];
		}

		// Do inverse FFTs ghat_p -> g_p
		fftw_execute(plan_g_backward);

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif

		// Renormalize g_p
		#pragma omp parallel for private(i) shared(NVLOC, g_real, MV)
		for (i = 0; i < NVLOC; i++){
			g_real[i][0] /= MV;
			g_real[i][1] = 0.;
		}
		
		// Compute G_p(k) = g_p(k) * f(k)
		// g_p and f are real numbers
		#pragma omp parallel for private(i) shared(NVLOC, g_real, bigG_real, Ginput_real)
		for (i = 0; i < NVLOC; i++){
			bigG_real[i][0] = g_real[i][0] * Ginput_real[i][0];
			bigG_real[i][1] = 0.;
		}

		// Do FFT G_p -> Ghat_p
		fftw_execute(plan_bigG_forward);

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif

		// Update Q2c_spectral
		#pragma omp parallel for private(i) shared(NVLOC, Q2c_spectral, bigG_spectral, C)
		for (i = 0; i < NVLOC; i++){
			Q2c_spectral[i][0] += C * bigG_spectral[i][0];
			Q2c_spectral[i][1] += C * bigG_spectral[i][1];
		}
	}

	// Destroy plans
	fftw_destroy_plan(plan_g_backward);
	fftw_destroy_plan(plan_bigG_forward);

	// Free memory
	fftw_free(g_real);
	fftw_free(g_spectral);
	fftw_free(bigG_real);
	fftw_free(bigG_spectral);
}

/*
void collision_Q2c_convolution_fftw(fftw_complex *Ginput_spectral, fftw_complex *Ginput_real, fftw_complex *Q2c_spectral, double ***C, double *W, PARAMETERS parameters){
	double pi = 4.*atan(1.);

	fftw_plan plan_CGhat_backward, plan_H_forward;
	fftw_complex *CGhat_spectral, *CGhat_real;
	fftw_complex *H_real, *H_spectral;

	int i;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

#ifdef key_fftw_mpi
	CGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	CGhat_real = fftw_alloc_complex(parameters.alloc_local);
	H_real = fftw_alloc_complex(parameters.alloc_local);
	H_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs on C[p]*Ghat (backward) and H[p] = G * IFFT(C[p]*Ghat) (forward)
	if (parameters.nvz == 0){
		plan_CGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, CGhat_spectral, CGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_2d(MVX, MVY, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_CGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, CGhat_spectral, CGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	CGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	CGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs on C[p]*Ghat (backward) and H[p] = G * IFFT(C[p]*Ghat) (forward)
	if (parameters.nvz == 0){
		plan_CGhat_backward = fftw_plan_dft_2d(MVX, MVY, CGhat_spectral, CGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_2d(MVX, MVY, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_CGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, CGhat_spectral, CGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif

	// Initialize Q2c_spectral
	#pragma omp parallel for private(i) shared(NVLOC, Q2c_spectral)
	for (i = 0; i < NVLOC; i++){
		Q2c_spectral[i][0] = 0.;
		Q2c_spectral[i][1] = 0.;
	}

	for (int p = 0; p < parameters.mint; p++){
		// Compute C[p]*Ghat
		// WARNING : C, Ghat are complex numbers
		#pragma omp parallel for private(i) shared(NVLOC, p, CGhat_spectral, C, Ginput_spectral)
		for (i = 0; i < NVLOC; i++){
			CGhat_spectral[i][0] = C[p][i][0]*Ginput_spectral[i][0] - C[p][i][1]*Ginput_spectral[i][1];
			CGhat_spectral[i][1] = C[p][i][0]*Ginput_spectral[i][1] + C[p][i][1]*Ginput_spectral[i][0];
		}

		// Compute the backward FFT of C[p]*Ghat
		// WARNING : we get complex numbers
		fftw_execute(plan_CGhat_backward);
		#pragma omp parallel for private(i) shared(NVLOC, CGhat_real, MV)
		for (i = 0; i < NVLOC; i++){
			CGhat_real[i][0] /= ((double)MV);
			CGhat_real[i][1] /= ((double)MV);
		}

		// Compute H[p] = IFFT(C[p]*Ghat) * G
		// WARNING : G is real
		#pragma omp parallel for private(i) shared(NVLOC, H_real, CGhat_real, Ginput_real)
		for (i = 0; i < NVLOC; i++){
			H_real[i][0] = CGhat_real[i][0]*Ginput_real[i][0];
			H_real[i][1] = CGhat_real[i][1]*Ginput_real[i][0];
		}

		// Compute the FFT of H[p]
		fftw_execute(plan_H_forward);

		// Increment Q2c_spectral with W[p] * Hhat[p]
		#pragma omp parallel for private(i) shared(NVLOC, p, Q2c_spectral, W, H_spectral)
		for (i = 0; i < NVLOC; i++){
			Q2c_spectral[i][0] += (W[p] * H_spectral[i][0]);
			Q2c_spectral[i][1] += (W[p] * H_spectral[i][1]);
		}
	}

	// Destroy plans
	fftw_destroy_plan(plan_CGhat_backward);
	fftw_destroy_plan(plan_H_forward);

	// Free memory
	fftw_free(CGhat_spectral);
	fftw_free(CGhat_real);
	fftw_free(H_real);
	fftw_free(H_spectral);
}*/


void collision_Q1q_fftw(fftw_complex *Ginput_real, fftw_complex *Q1q_spectral, double **alpha, double **alphaprime, PARAMETERS parameters){

	double pi = 4.*atan(1.);

	fftw_complex *gj_real, *gj_spectral;
	fftw_complex *gprimej_real, *gprimej_spectral;
	fftw_complex *bigGj_real, *bigGj_spectral;
	fftw_complex *bigGprimej_real, *bigGprimej_spectral;
	fftw_plan plan_gj_backward, plan_gprimej_backward, plan_bigGj_forward, plan_bigGprimej_forward;
	double *GDIST, gdist_j;
	
	extern int crossmode_done;
	extern double ***crossmode_cos;
	extern double ***crossmode_sin;

	int itemp;
	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;
	int *M;
	int *kvxabsdist, *kvyabsdist, *kvzabsdist;

	int i, k;
	
	int kvxabs, kvyabs, kvzabs;
	int jvxabs, jvyabs, jvzabs;
	int modek[3];
	int modej[3];
	double crossmode;
	double a1, a2, a3, b1, b2, b3;

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

	int rank, commsize;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
#else
	rank = 0;
	commsize = 1;
#endif

	// Reminder : parameters.ndim = 1 for 2D, parameters.ndim = 2 for 3D
	// Reminder : for the 3D case, mint = mint1*mint2
	double C = pow(pi, parameters.ndim)/((double)parameters.mint);
	
	M = allocation_1d_int(parameters.ndim+1);
	M[0] = MVX;
	M[1] = MVY;
	if (parameters.nvz > 0)
		M[2] = MVZ;

#ifdef key_fftw_mpi
	gj_real = fftw_alloc_complex(parameters.alloc_local);
	gj_spectral = fftw_alloc_complex(parameters.alloc_local);
	gprimej_real = fftw_alloc_complex(parameters.alloc_local);
	gprimej_spectral = fftw_alloc_complex(parameters.alloc_local);
	bigGj_real = fftw_alloc_complex(parameters.alloc_local);
	bigGj_spectral = fftw_alloc_complex(parameters.alloc_local);
	bigGprimej_real = fftw_alloc_complex(parameters.alloc_local);
	bigGprimej_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_gj_backward = fftw_mpi_plan_dft_2d(MVX, MVY, gj_spectral, gj_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprimej_backward = fftw_mpi_plan_dft_2d(MVX, MVY, gprimej_spectral, gprimej_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigGj_forward = fftw_mpi_plan_dft_2d(MVX, MVY, bigGj_real, bigGj_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_bigGprimej_forward = fftw_mpi_plan_dft_2d(MVX, MVY, bigGprimej_real, bigGprimej_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_gj_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, gj_spectral, gj_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprimej_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, gprimej_spectral, gprimej_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigGj_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, bigGj_real, bigGj_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_bigGprimej_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, bigGprimej_real, bigGprimej_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	gj_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	gj_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	gprimej_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	gprimej_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigGj_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigGj_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigGprimej_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigGprimej_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_gj_backward = fftw_plan_dft_2d(MVX, MVY, gj_spectral, gj_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprimej_backward = fftw_plan_dft_2d(MVX, MVY, gprimej_spectral, gprimej_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigGj_forward = fftw_plan_dft_2d(MVX, MVY, bigGj_real, bigGj_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_bigGprimej_forward = fftw_plan_dft_2d(MVX, MVY, bigGprimej_real, bigGprimej_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_gj_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, gj_spectral, gj_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprimej_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, gprimej_spectral, gprimej_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigGj_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, bigGj_real, bigGj_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_bigGprimej_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, bigGprimej_real, bigGprimej_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif

	// Use of contiguous memory notations
	#pragma omp parallel for private(i) shared(NVLOC, Q1q_spectral)
	for (i = 0; i < NVLOC; i++){
		Q1q_spectral[i][0] = 0.;
		Q1q_spectral[i][1] = 0.;
	}

	// Loop over the communicator
#ifdef key_force_vectorize
	if (!(crossmode_done)){
		crossmode_cos = malloc(sizeof(double**)*commsize);
		crossmode_sin = malloc(sizeof(double**)*commsize);
		for (int R = 0; R < commsize; R++){
			crossmode_cos[R] = malloc(sizeof(double*)*parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]);
			crossmode_sin[R] = malloc(sizeof(double*)*parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]);
			
			kvxabsdist = allocation_1d_int(parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]);
			kvyabsdist = allocation_1d_int(parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]);
			kvzabsdist = allocation_1d_int(parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]);
			
			if (R == rank){
				#pragma omp parallel for private(k) shared(parameters, rank, kvxabsdist, kvyabsdist, kvzabsdist)
				for (k = 0; k < parameters.nvxtab[rank]*parameters.nvytab[rank]*parameters.nvztab[rank]; k++){
					kvxabsdist[k] = parameters.kvxabstab[k];
					kvyabsdist[k] = parameters.kvyabstab[k];
					kvzabsdist[k] = parameters.kvzabstab[k];
				}
			}
#ifdef key_mpi
			MPI_Bcast(kvxabsdist, parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R], MPI_INT, R, MPI_COMM_WORLD);
			MPI_Bcast(kvyabsdist, parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R], MPI_INT, R, MPI_COMM_WORLD);
			MPI_Bcast(kvzabsdist, parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R], MPI_INT, R, MPI_COMM_WORLD);
#endif
			
			for (int jdist = 0; jdist < parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]; jdist++){
				modej[0] = kvxabsdist[jdist] - MVX*(kvxabsdist[jdist] >= MVX/2);
				modej[1] = kvyabsdist[jdist] - MVY*(kvyabsdist[jdist] >= MVY/2);
				if (parameters.ndim == DIM3)
					modej[2] = kvzabsdist[jdist] - MVZ*(kvzabsdist[jdist] >= MVZ/2);
				
				crossmode_cos[R][jdist] = allocation_1d(NVXLOC*NVYLOC*NVZLOC);
				crossmode_sin[R][jdist] = allocation_1d(NVXLOC*NVYLOC*NVZLOC);
				
				#pragma omp parallel for private(k, crossmode, modek) shared(NVLOC, parameters, MVX, MVY, MVZ, pi, modej, M, R, jdist, crossmode_cos, crossmode_sin)
				for (k = 0; k < NVLOC; k++){
					modek[0] = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
					modek[1] = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
					if (parameters.ndim == DIM3)
						modek[2] = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
					
					crossmode = 0.;
					for (int d = 0; d <= parameters.ndim; d++)
						crossmode += (2.*pi*((double)modek[d])*((double)modej[d])/((double)M[d]));
					
					crossmode_cos[R][jdist][k] = cos(crossmode);
					crossmode_sin[R][jdist][k] = sin(crossmode);
				}
			}
			
			free_1d_int(kvxabsdist);
			free_1d_int(kvyabsdist);
			free_1d_int(kvzabsdist);
		}
		crossmode_done = (1==1);
	}
#endif

	for (int p = 0; p < parameters.mint; p++){
		for (int k = 0; k < NVLOC; k++){
			// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
			itemp = k/NVZLOC;
			// itemp now contains kvyloc+NVYLOC*kvxloc;
			kvzabs = k - itemp*NVZLOC;
			kvxabs = itemp/NVYLOC;
			kvyabs = itemp-kvxabs*NVYLOC;
			kvxabs += IVX0LOC;
			kvyabs += IVY0LOC;
			kvzabs += IVZ0LOC;
			if (kvxabs >= MVX/2)
				kvxabs -= MVX;
			if (kvyabs >= MVY/2)
				kvyabs -= MVY;
			if (kvzabs >= MVZ/2)
				kvzabs -= MVZ;
		}
	}


	for (int R = 0; R < commsize; R++){
		// Broadcast the part of Ginput_real that belongs to distant CPU R
		GDIST = allocation_1d(parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]);
		
#ifdef key_force_vectorize
		kvxabsdist = allocation_1d_int(parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]);
		kvyabsdist = allocation_1d_int(parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]);
		kvzabsdist = allocation_1d_int(parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]);
#endif

		if (R == rank){
			#pragma omp parallel for private(k) shared(parameters, rank, GDIST, Ginput_real, kvxabsdist, kvyabsdist, kvzabsdist)
			for (k = 0; k < parameters.nvxtab[rank]*parameters.nvytab[rank]*parameters.nvztab[rank]; k++){
				GDIST[k] = Ginput_real[k][0];
#ifdef key_force_vectorize
				kvxabsdist[k] = parameters.kvxabstab[k];
				kvyabsdist[k] = parameters.kvyabstab[k];
				kvzabsdist[k] = parameters.kvzabstab[k];
#endif
			}
		}

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
		MPI_Bcast(GDIST, parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R], MPI_DOUBLE, R, MPI_COMM_WORLD);
#ifdef key_force_vectorize
		MPI_Bcast(kvxabsdist, parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R], MPI_INT, R, MPI_COMM_WORLD);
		MPI_Bcast(kvyabsdist, parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R], MPI_INT, R, MPI_COMM_WORLD);
		MPI_Bcast(kvzabsdist, parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R], MPI_INT, R, MPI_COMM_WORLD);
#endif
#endif
		
		// Loop over the indices that belong to distant CPU R
		for (int jdist = 0; jdist < parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]; jdist++){
#ifdef key_force_vectorize
			jvxabs = kvxabsdist[jdist] - MVX*(kvxabsdist[jdist] >= MVX/2);
			jvyabs = kvyabsdist[jdist] - MVY*(kvyabsdist[jdist] >= MVY/2);
			jvzabs = kvzabsdist[jdist] - MVZ*(kvzabsdist[jdist] >= MVZ/2);
#else
			// j = jvzdist+NVZDIST*(jvydist+NVYDIST*jvxdist)
			itemp = jdist/parameters.nvztab[R];
			// itemp now contains kvyloc+NVYDIST*kvxloc;
			jvzabs = jdist - itemp*parameters.nvztab[R];
			jvxabs = itemp/parameters.nvytab[R];
			jvyabs = itemp-jvxabs*parameters.nvytab[R];
			jvxabs += parameters.ivx0tab[R];
			jvyabs += parameters.ivy0tab[R];
			jvzabs += parameters.ivz0tab[R];
			if (jvxabs >= MVX/2)
				jvxabs -= MVX;
			if (jvyabs >= MVY/2)
				jvyabs -= MVY;
			if (jvzabs >= MVZ/2)
				jvzabs -= MVZ;
#endif			
			modej[0] = jvxabs;
			modej[1] = jvyabs;
			if (parameters.ndim == DIM3)
				modej[2] = jvzabs;

			gdist_j = GDIST[jdist];

			for (int p = 0; p < parameters.mint; p++){

				// Compute g_p,j_hat = alpha[p] * exp(-2i*pi*k.j/N) and g'_p,j_hat = alpha'[p] * exp(-2i*pi*k.j/N)
				#pragma omp parallel for private(k, itemp, kvxabs, kvyabs, kvzabs, modek, crossmode) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, gj_spectral, gprimej_spectral, alpha, alphaprime, crossmode_cos, crossmode_sin, R, jdist, IVX0LOC, IVY0LOC, IVZ0LOC, MVX, MVY, MVZ, parameters, pi, modej, M, p)
				for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
					gj_spectral[k][0] = alpha[p][k] * crossmode_cos[R][jdist][k];
					gj_spectral[k][1] = alpha[p][k] * (-crossmode_sin[R][jdist][k]);
					
					gprimej_spectral[k][0] = alphaprime[p][k] * crossmode_cos[R][jdist][k];
					gprimej_spectral[k][1] = alphaprime[p][k] * (-crossmode_sin[R][jdist][k]);
#else
					// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
					itemp = k/NVZLOC;
					// itemp now contains kvyloc+NVYLOC*kvxloc;
					kvzabs = k - itemp*NVZLOC;
					kvxabs = itemp/NVYLOC;
					kvyabs = itemp-kvxabs*NVYLOC;
					kvxabs += IVX0LOC;
					kvyabs += IVY0LOC;
					kvzabs += IVZ0LOC;
					if (kvxabs >= MVX/2)
						kvxabs -= MVX;
					if (kvyabs >= MVY/2)
						kvyabs -= MVY;
					if (kvzabs >= MVZ/2)
						kvzabs -= MVZ;

					modek[0] = kvxabs;
					modek[1] = kvyabs;
					if (parameters.ndim == DIM3)
						modek[2] = kvzabs;
								
					// Compute exp(-2i*pi*k.j/N)
					crossmode = 0.;
					for (int d = 0; d <= parameters.ndim; d++)
						crossmode += (-2.*pi*((double)modek[d])*((double)modej[d])/((double)M[d]));

					gj_spectral[k][0] = alpha[p][k] * cos(crossmode);
					gj_spectral[k][1] = alpha[p][k] * sin(crossmode);


					//if ((p == 0) && (jdist == 530)){
						//printf("p = %d, j = (%d,%d), k = %d, alpha = %e, sin(crossmode) = %e\n", p, jvxabs, jvyabs, k, alpha[p][k], sin(crossmode));
						//printf("p = %d, j = %d, gj_spectral[%d] = (%e,%e), gprimej_spectral[%d] = (%e,%e)\n", p, jdist, k, gj_spectral[k][0], gj_spectral[k][1], k, gprimej_spectral[k][0], gprimej_spectral[k][1]);
					//}

					gprimej_spectral[k][0] = alphaprime[p][k] * cos(crossmode);
					gprimej_spectral[k][1] = alphaprime[p][k] * sin(crossmode);
#endif
				}
				//printf("\n");
				

				// g_p,j_hat and g'_p,j_hat are ready --> Perform inverse FFT to get g_p,j and g'_p,j
				// WARNING: g_p,j and g'_p,j are COMPLEX NUMBERS !!
				fftw_execute(plan_gj_backward);
				fftw_execute(plan_gprimej_backward);

				#pragma omp parallel for private(k) shared(NVLOC, gj_real, gprimej_real, MV)
				for (k = 0; k < NVLOC; k++){
					gj_real[k][0] /= MV;
					gj_real[k][1] /= MV;
					gprimej_real[k][0] /= MV;
					gprimej_real[k][1] /= MV;
				}
				

				// Compute G_p,j = g_p,j * F and G'_p,j = g'_p,j * F
				// Reminder : F is real but g_p,j and g'_p,j are complex
				#pragma omp parallel for private(k) shared(NVLOC, bigGj_real, bigGprimej_real, gj_real, gprimej_real, Ginput_real)
				for (k = 0; k < NVLOC; k++){
					bigGj_real[k][0] = gj_real[k][0] * Ginput_real[k][0];
					bigGj_real[k][1] = gj_real[k][1] * Ginput_real[k][0];
					bigGprimej_real[k][0] = gprimej_real[k][0] * Ginput_real[k][0];
					bigGprimej_real[k][1] = gprimej_real[k][1] * Ginput_real[k][0];
				}

				// Perform FFT to get G_p,j_hat and G'_p,j_hat
				fftw_execute(plan_bigGj_forward);
				fftw_execute(plan_bigGprimej_forward);

				//for (int i = 0; i < NVLOC; i++)
					//printf("p = %d, j = %d, bigGhat[%d] = (%e,%e), bigGprimehat[%d] = (%e,%e)\n", p, jdist, i, bigGj_spectral[i][0], bigGj_spectral[i][1], i, bigGprimej_spectral[i][0], bigGprimej_spectral[i][1]);
				//error("stop now");

				// Update Q1q_hat[k] with C/N * G_p,j_hat[k] * G'_p,j_hat[k] * exp(2i*pi*k.j/N) * Gdist[j]
				#pragma omp parallel for private(k, a1, b1, a2, b2, a3, b3, itemp, kvxabs, kvyabs, kvzabs, modek, crossmode) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, IVX0LOC, IVY0LOC, IVZ0LOC, bigGj_spectral, bigGprimej_spectral, R, jdist, MVX, MVY, MVZ, parameters, pi, modej, M, Q1q_spectral, C, gdist_j)
				for (k = 0; k < NVLOC; k++){
					a1 = bigGj_spectral[k][0];
					b1 = bigGj_spectral[k][1];
					a2 = bigGprimej_spectral[k][0];
					b2 = bigGprimej_spectral[k][1];
					//printf("p = %d, j = %d, k = %d | bigGj_spectral = (%e,%e) | bigGprimej_spectral = (%e,%e)\n", p, jdist, k, bigGj_spectral[k][0], bigGj_spectral[k][1], bigGprimej_spectral[k][0], bigGprimej_spectral[k][1]);
#ifdef key_force_vectorize
					a3 = crossmode_cos[R][jdist][k];
					b3 = crossmode_sin[R][jdist][k];
#else
					// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
					itemp = k/NVZLOC;
					// itemp now contains kvyloc+NVYLOC*kvxloc;
					kvzabs = k - itemp*NVZLOC;
					kvxabs = itemp/NVYLOC;
					kvyabs = itemp-kvxabs*NVYLOC;
					kvxabs += IVX0LOC;
					kvyabs += IVY0LOC;
					kvzabs += IVZ0LOC;
					if (kvxabs >= MVX/2)
						kvxabs -= MVX;
					if (kvyabs >= MVY/2)
						kvyabs -= MVY;
					if (kvzabs >= MVZ/2)
						kvzabs -= MVZ;
							
					modek[0] = kvxabs;
					modek[1] = kvyabs;
					if (parameters.ndim == DIM3)
						modek[2] = kvzabs;
							
					// Compute exp(2i*pi*k.j/N)
					crossmode = 0.;
					for (int d = 0; d <= parameters.ndim; d++)
						crossmode += (2.*pi*((double)modek[d])*((double)modej[d])/((double)M[d]));
					
					a3 = cos(crossmode);
					b3 = sin(crossmode);
#endif
					Q1q_spectral[k][0] += (a3*(a1*a2-b1*b2) - b3*(a2*b1+a1*b2)) * C * gdist_j;
					Q1q_spectral[k][1] += (a3*(a2*b1+a1*b2) + b3*(a1*a2-b1*b2)) * C * gdist_j;
				} // End of Q1q_hat update						
			} // End of loop over p
		} // End of loop over distant indices j
		for (k = 0; k < NVLOC; k++){
			// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
					itemp = k/NVZLOC;
					// itemp now contains kvyloc+NVYLOC*kvxloc;
					kvzabs = k - itemp*NVZLOC;
					kvxabs = itemp/NVYLOC;
					kvyabs = itemp-kvxabs*NVYLOC;
					kvxabs += IVX0LOC;
					kvyabs += IVY0LOC;
					kvzabs += IVZ0LOC;
					if (kvxabs >= MVX/2)
						kvxabs -= MVX;
					if (kvyabs >= MVY/2)
						kvyabs -= MVY;
					if (kvzabs >= MVZ/2)
						kvzabs -= MVZ;
			
			//printf("k = (%d,%d) Q1q_spectral = (%e,%e)\n", kvxabs, kvyabs, Q1q_spectral[k][0], Q1q_spectral[k][1]);
		}

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif

		free_1d(GDIST);
#ifdef key_force_vectorize
		free_1d_int(kvxabsdist);
		free_1d_int(kvyabsdist);
		free_1d_int(kvzabsdist);
#endif
	} // End of loop over the communicator: the computations are finished

	fftw_destroy_plan(plan_gj_backward);
	fftw_destroy_plan(plan_gprimej_backward);
	fftw_destroy_plan(plan_bigGj_forward);
	fftw_destroy_plan(plan_bigGprimej_forward);

	fftw_free(gj_real);
	fftw_free(gj_spectral);
	fftw_free(gprimej_real);
	fftw_free(gprimej_spectral);
	fftw_free(bigGj_real);
	fftw_free(bigGj_spectral);
	fftw_free(bigGprimej_real);
	fftw_free(bigGprimej_spectral);

	free_1d_int(M);
}

/*
void collision_Q1q_convolution_fftw(fftw_complex *Ginput_spectral, fftw_complex *Q1q_spectral, double ***A, double ***B, double ***C, double *W, PARAMETERS parameters){
	double pi = 4.*atan(1.);

	fftw_plan plan_AGhat_backward, plan_BGhat_backward, plan_CGhat_backward, plan_H_forward;
	fftw_complex *AGhat_spectral, *AGhat_real;
	fftw_complex *BGhat_spectral, *BGhat_real;
	fftw_complex *CGhat_spectral, *CGhat_real;
	fftw_complex *H_real, *H_spectral;

	int i;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

#ifdef key_fftw_mpi
	AGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	AGhat_real = fftw_alloc_complex(parameters.alloc_local);
	BGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	BGhat_real = fftw_alloc_complex(parameters.alloc_local);
	CGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	CGhat_real = fftw_alloc_complex(parameters.alloc_local);
	H_real = fftw_alloc_complex(parameters.alloc_local);
	H_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs on A[p]*Ghat (backward), B[p]*Ghat (backward), C[p]*Ghat (backward) and H[p] = IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) * IFFT(C[p]*Ghat) (forward)
	if (parameters.nvz == 0){
		plan_AGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, AGhat_spectral, AGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, BGhat_spectral, BGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, CGhat_spectral, CGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_2d(MVX, MVY, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_AGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, AGhat_spectral, AGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, BGhat_spectral, BGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, CGhat_spectral, CGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	AGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	AGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	BGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	BGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	CGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	CGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs on A[p]*Ghat (backward), B[p]*Ghat (backward), C[p]*Ghat (backward) and H[p] = IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) * IFFT(C[p]*Ghat) (forward)
	if (parameters.nvz == 0){
		plan_AGhat_backward = fftw_plan_dft_2d(MVX, MVY, AGhat_spectral, AGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_plan_dft_2d(MVX, MVY, BGhat_spectral, BGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_plan_dft_2d(MVX, MVY, CGhat_spectral, CGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_2d(MVX, MVY, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_AGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, AGhat_spectral, AGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, BGhat_spectral, BGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, CGhat_spectral, CGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif

	// Initialize Q1q_spectral
	#pragma omp parallel for private(i) shared(NVLOC, Q1q_spectral)
	for (i = 0; i < NVLOC; i++){
		Q1q_spectral[i][0] = 0.;
		Q1q_spectral[i][1] = 0.;
	}

	for (int p = 0; p < parameters.mint; p++){
		// Compute A[p]*Ghat, B[p]*Ghat, C[p]*Ghat
		// WARNING : A, B, C, Ghat are complex numbers
		#pragma omp parallel for private(i) shared(NVLOC, p, AGhat_spectral, BGhat_spectral, CGhat_spectral, A, B, C, Ginput_spectral)
		for (i = 0; i < NVLOC; i++){
			AGhat_spectral[i][0] = A[p][i][0]*Ginput_spectral[i][0] - A[p][i][1]*Ginput_spectral[i][1];
			AGhat_spectral[i][1] = A[p][i][0]*Ginput_spectral[i][1] + A[p][i][1]*Ginput_spectral[i][0];

			BGhat_spectral[i][0] = B[p][i][0]*Ginput_spectral[i][0] - B[p][i][1]*Ginput_spectral[i][1];
			BGhat_spectral[i][1] = B[p][i][0]*Ginput_spectral[i][1] + B[p][i][1]*Ginput_spectral[i][0];

			CGhat_spectral[i][0] = C[p][i][0]*Ginput_spectral[i][0] - C[p][i][1]*Ginput_spectral[i][1];
			CGhat_spectral[i][1] = C[p][i][0]*Ginput_spectral[i][1] + C[p][i][1]*Ginput_spectral[i][0];
		}

		// Compute the backward FFT of A[p]*Ghat, B[p]*Ghat, C[p]*Ghat
		// WARNING : we get complex numbers
		fftw_execute(plan_AGhat_backward);
		fftw_execute(plan_BGhat_backward);
		fftw_execute(plan_CGhat_backward);
		#pragma omp parallel for private(i) shared(NVLOC, AGhat_real, BGhat_real, CGhat_real, MV)
		for (i = 0; i < NVLOC; i++){
			AGhat_real[i][0] /= ((double)MV);
			AGhat_real[i][1] /= ((double)MV);

			BGhat_real[i][0] /= ((double)MV);
			BGhat_real[i][1] /= ((double)MV);

			CGhat_real[i][0] /= ((double)MV);
			CGhat_real[i][1] /= ((double)MV);
		}

		// Compute H[p] = IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) * IFFT(C[p]*Ghat)
		#pragma omp parallel for private(i) shared(NVLOC, H_real, AGhat_real, BGhat_real, CGhat_real)
		for (i = 0; i < NVLOC; i++){
			H_real[i][0] = AGhat_real[i][0] * (BGhat_real[i][0]*CGhat_real[i][0] - BGhat_real[i][1]*CGhat_real[i][1]) - AGhat_real[i][1] * (BGhat_real[i][1]*CGhat_real[i][0] + BGhat_real[i][0]*CGhat_real[i][1]);
			H_real[i][1] = AGhat_real[i][0] * (BGhat_real[i][1]*CGhat_real[i][0] + BGhat_real[i][0]*CGhat_real[i][1]) + AGhat_real[i][1] * (BGhat_real[i][0]*CGhat_real[i][0] - BGhat_real[i][1]*CGhat_real[i][1]);
		}

		// Compute the FFT of H[p]
		fftw_execute(plan_H_forward);

		// Increment Q1q_spectral with W[p] * Hhat[p]
		#pragma omp parallel for private(i) shared(NVLOC, p, Q1q_spectral, W, H_spectral)
		for (i = 0; i < NVLOC; i++){
			Q1q_spectral[i][0] += (W[p] * H_spectral[i][0]);
			Q1q_spectral[i][1] += (W[p] * H_spectral[i][1]);
		}
	}

	// Destroy plans
	fftw_destroy_plan(plan_AGhat_backward);
	fftw_destroy_plan(plan_BGhat_backward);
	fftw_destroy_plan(plan_CGhat_backward);
	fftw_destroy_plan(plan_H_forward);

	// Free memory
	fftw_free(AGhat_spectral);
	fftw_free(AGhat_real);
	fftw_free(BGhat_spectral);
	fftw_free(BGhat_real);
	fftw_free(CGhat_spectral);
	fftw_free(CGhat_real);
	fftw_free(H_real);
	fftw_free(H_spectral);
}*/



void collision_Q2q_fftw(fftw_complex *Ginput_spectral, fftw_complex *Ginput_real, fftw_complex *Q2q_spectral, double **alpha, double **alphaprime, PARAMETERS parameters){
	double pi = 4.*atan(1.);
	//int ind;
	fftw_plan plan_g_backward, plan_gprime_backward, plan_bigG_forward;
	fftw_complex *g_real;
	fftw_complex *g_spectral;
	fftw_complex *gprime_real;
	fftw_complex *gprime_spectral;
	fftw_complex *bigG_real;
	fftw_complex *bigG_spectral;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	int i;

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

	// Reminder : parameters.ndim = 1 for 2D, parameters.ndim = 2 for 3D
	// Reminder : for the 3D case, mint = mint1*mint2
	double C = pow(pi, parameters.ndim)/((double)parameters.mint);
	
#ifdef key_fftw_mpi
	g_real = fftw_alloc_complex(parameters.alloc_local);
	g_spectral = fftw_alloc_complex(parameters.alloc_local);
	gprime_real = fftw_alloc_complex(parameters.alloc_local);
	gprime_spectral = fftw_alloc_complex(parameters.alloc_local);
	bigG_real = fftw_alloc_complex(parameters.alloc_local);
	bigG_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_g_backward = fftw_mpi_plan_dft_2d(MVX, MVY, g_spectral, g_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprime_backward = fftw_mpi_plan_dft_2d(MVX, MVY, gprime_spectral, gprime_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_mpi_plan_dft_2d(MVX, MVY, bigG_real, bigG_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_g_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, g_spectral, g_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprime_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, gprime_spectral, gprime_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, bigG_real, bigG_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	g_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	g_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	gprime_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	gprime_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigG_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigG_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_g_backward = fftw_plan_dft_2d(MVX, MVY, g_spectral, g_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprime_backward = fftw_plan_dft_2d(MVX, MVY, gprime_spectral, gprime_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_plan_dft_2d(MVX, MVY, bigG_real, bigG_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_g_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, g_spectral, g_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_gprime_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, gprime_spectral, gprime_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, bigG_real, bigG_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif

	// Initialize Q2q_spectral
	#pragma omp parallel for private(i) shared(NVLOC, Q2q_spectral)
	for (i = 0; i < NVLOC; i++){
		Q2q_spectral[i][0] = 0.;
		Q2q_spectral[i][1] = 0.;
	}

	for (int p = 0; p < parameters.mint; p++){
		// Compute ghat_p(k) = alpha_p(k) * fhat(k), gprimehat_p(k) = alphaprime_p(k) * fhat(k)
		// alpha_p and alphaprime_p are real numbers
		#pragma omp parallel for private(i) shared(NVLOC, g_spectral, gprime_spectral, p, alpha, alphaprime, Ginput_spectral)
		for (i = 0; i < NVLOC; i++){
			g_spectral[i][0] = alpha[p][i] * Ginput_spectral[i][0];
			g_spectral[i][1] = alpha[p][i] * Ginput_spectral[i][1];
			gprime_spectral[i][0] = alphaprime[p][i] * Ginput_spectral[i][0];
			gprime_spectral[i][1] = alphaprime[p][i] * Ginput_spectral[i][1];
		}

		// Do inverse FFTs ghat_p -> g_p, gprimehat_p -> gprime_p
		fftw_execute(plan_g_backward);
		fftw_execute(plan_gprime_backward);
		// Renormalize g_p, gprime_p
		// g_p and gprime_p are real numbers
		#pragma omp parallel for private(i) shared(NVLOC, g_real, MV, gprime_real)
		for (i = 0; i < NVLOC; i++){
			g_real[i][0] /= MV;
			g_real[i][1] = 0.;
			gprime_real[i][0] /= MV;
			gprime_real[i][1] = 0.;
		}

		// Compute G_p(k) = g_p(k) * gprime_p(k) * f(k)
		// f, g_p and gprime_p are real numbers
		#pragma omp parallel for private(i) shared(NVLOC, bigG_real, g_real, gprime_real, Ginput_real)
		for (i = 0; i < NVLOC; i++){
			bigG_real[i][0] = g_real[i][0]*gprime_real[i][0]*Ginput_real[i][0];
			bigG_real[i][1] = 0.;
		}

		// Do FFT G_p -> Ghat_p
		fftw_execute(plan_bigG_forward);

		// Update Q2q_spectral
		#pragma omp parallel for private(i) shared(NVLOC, Q2q_spectral, C, bigG_spectral)
		for (i = 0; i < NVLOC; i++){
			Q2q_spectral[i][0] += C * bigG_spectral[i][0];
			Q2q_spectral[i][1] += C * bigG_spectral[i][1];
		}
	}

	// Destroy plans
	fftw_destroy_plan(plan_g_backward);
	fftw_destroy_plan(plan_gprime_backward);
	fftw_destroy_plan(plan_bigG_forward);

	// Free memory
	fftw_free(g_real);
	fftw_free(g_spectral);
	fftw_free(gprime_real);
	fftw_free(gprime_spectral);
	fftw_free(bigG_real);
	fftw_free(bigG_spectral);
}

/*
void collision_Q2q_convolution_fftw(fftw_complex *Ginput_spectral, fftw_complex *Ginput_real, fftw_complex *Q2q_spectral, double ***A, double ***B, double *W, PARAMETERS parameters){
	double pi = 4.*atan(1.);

	fftw_plan plan_AGhat_backward, plan_BGhat_backward, plan_H_forward;
	fftw_complex *AGhat_spectral, *AGhat_real;
	fftw_complex *BGhat_spectral, *BGhat_real;
	fftw_complex *H_real, *H_spectral;

	int i;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

#ifdef key_fftw_mpi
	AGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	AGhat_real = fftw_alloc_complex(parameters.alloc_local);
	BGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	BGhat_real = fftw_alloc_complex(parameters.alloc_local);
	H_real = fftw_alloc_complex(parameters.alloc_local);
	H_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs on A[p]*Ghat (backward), B[p]*Ghat (backward) and H[p] = IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) * G (forward)
	if (parameters.nvz == 0){
		plan_AGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, AGhat_spectral, AGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, BGhat_spectral, BGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_2d(MVX, MVY, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_AGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, AGhat_spectral, AGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, BGhat_spectral, BGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	AGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	AGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	BGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	BGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs on A[p]*Ghat (backward), B[p]*Ghat (backward) and H[p] = IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) * G (forward)
	if (parameters.nvz == 0){
		plan_AGhat_backward = fftw_plan_dft_2d(MVX, MVY, AGhat_spectral, AGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_plan_dft_2d(MVX, MVY, BGhat_spectral, BGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_2d(MVX, MVY, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_AGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, AGhat_spectral, AGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_BGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, BGhat_spectral, BGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif

	// Initialize Q2q_spectral
	#pragma omp parallel for private(i) shared(NVLOC, Q2q_spectral)
	for (i = 0; i < NVLOC; i++){
		Q2q_spectral[i][0] = 0.;
		Q2q_spectral[i][1] = 0.;
	}

	for (int p = 0; p < parameters.mint; p++){
		// Compute A[p]*Ghat, B[p]*Ghat
		// WARNING : A, B, Ghat are complex numbers
		#pragma omp parallel for private(i) shared(NVLOC, p, AGhat_spectral, BGhat_spectral, A, B, Ginput_spectral)
		for (i = 0; i < NVLOC; i++){
			AGhat_spectral[i][0] = A[p][i][0]*Ginput_spectral[i][0] - A[p][i][1]*Ginput_spectral[i][1];
			AGhat_spectral[i][1] = A[p][i][0]*Ginput_spectral[i][1] + A[p][i][1]*Ginput_spectral[i][0];

			BGhat_spectral[i][0] = B[p][i][0]*Ginput_spectral[i][0] - B[p][i][1]*Ginput_spectral[i][1];
			BGhat_spectral[i][1] = B[p][i][0]*Ginput_spectral[i][1] + B[p][i][1]*Ginput_spectral[i][0];
		}

		// Compute the backward FFT of A[p]*Ghat, B[p]*Ghat
		// WARNING : we get complex numbers
		fftw_execute(plan_AGhat_backward);
		fftw_execute(plan_BGhat_backward);
		#pragma omp parallel for private(i) shared(NVLOC, AGhat_real, BGhat_real, MV)
		for (i = 0; i < NVLOC; i++){
			AGhat_real[i][0] /= ((double)MV);
			AGhat_real[i][1] /= ((double)MV);

			BGhat_real[i][0] /= ((double)MV);
			BGhat_real[i][1] /= ((double)MV);
		}

		// Compute H[p] = IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) * G
		#pragma omp parallel for private(i) shared(NVLOC, H_real, AGhat_real, BGhat_real, Ginput_real)
		for (i = 0; i < NVLOC; i++){
			H_real[i][0] = AGhat_real[i][0] * (BGhat_real[i][0]*Ginput_real[i][0]) - AGhat_real[i][1] * (BGhat_real[i][1]*Ginput_real[i][0]);
			H_real[i][1] = AGhat_real[i][0] * (BGhat_real[i][1]*Ginput_real[i][0]) + AGhat_real[i][1] * (BGhat_real[i][0]*Ginput_real[i][0]);
		}

		// Compute the FFT of H[p]
		fftw_execute(plan_H_forward);

		// Increment Q2q_spectral with W[p] * Hhat[p]
		#pragma omp parallel for private(i) shared(NVLOC, p, Q2q_spectral, W, H_spectral)
		for (i = 0; i < NVLOC; i++){
			Q2q_spectral[i][0] += (W[p] * H_spectral[i][0]);
			Q2q_spectral[i][1] += (W[p] * H_spectral[i][1]);
		}
	}

	// Destroy plans
	fftw_destroy_plan(plan_AGhat_backward);
	fftw_destroy_plan(plan_BGhat_backward);
	fftw_destroy_plan(plan_H_forward);

	// Free memory
	fftw_free(AGhat_spectral);
	fftw_free(AGhat_real);
	fftw_free(BGhat_spectral);
	fftw_free(BGhat_real);
	fftw_free(H_real);
	fftw_free(H_spectral);
}*/



void collision_Q3q_fftw(fftw_complex *Ginput_spectral, fftw_complex *Ginput_real, fftw_complex *Q3q_spectral, double **alpha, double **alphaprime, PARAMETERS parameters){
	double pi = 4.*atan(1.);
	//int ind;
	fftw_plan plan_gprime_backward, plan_bigGprime_forward, plan_h_backward, plan_bigH_forward;
	fftw_complex *h_real;
	fftw_complex *h_spectral;
	fftw_complex *gprime_real;
	fftw_complex *gprime_spectral;
	fftw_complex *bigGprime_real;
	fftw_complex *bigGprime_spectral;
	fftw_complex *bigH_real;
	fftw_complex *bigH_spectral;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	int i;

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

	// Reminder : parameters.ndim = 1 for 2D, parameters.ndim = 2 for 3D
	// Reminder : for the 3D case, mint = mint1*mint2
	double C = pow(pi, parameters.ndim)/((double)parameters.mint);
	
#ifdef key_fftw_mpi
	h_real = fftw_alloc_complex(parameters.alloc_local);
	h_spectral = fftw_alloc_complex(parameters.alloc_local);
	gprime_real = fftw_alloc_complex(parameters.alloc_local);
	gprime_spectral = fftw_alloc_complex(parameters.alloc_local);
	bigGprime_real = fftw_alloc_complex(parameters.alloc_local);
	bigGprime_spectral = fftw_alloc_complex(parameters.alloc_local);
	bigH_real = fftw_alloc_complex(parameters.alloc_local);
	bigH_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_gprime_backward = fftw_mpi_plan_dft_2d(MVX, MVY, gprime_spectral, gprime_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigGprime_forward = fftw_mpi_plan_dft_2d(MVX, MVY, bigGprime_real, bigGprime_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_h_backward = fftw_mpi_plan_dft_2d(MVX, MVY, h_spectral, h_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigH_forward = fftw_mpi_plan_dft_2d(MVX, MVY, bigH_real, bigH_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_gprime_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, gprime_spectral, gprime_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigGprime_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, bigGprime_real, bigGprime_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_h_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, h_spectral, h_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigH_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, bigH_real, bigH_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	h_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	h_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	gprime_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	gprime_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigGprime_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigGprime_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigH_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigH_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_gprime_backward = fftw_plan_dft_2d(MVX, MVY, gprime_spectral, gprime_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigGprime_forward = fftw_plan_dft_2d(MVX, MVY, bigGprime_real, bigGprime_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_h_backward = fftw_plan_dft_2d(MVX, MVY, h_spectral, h_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigH_forward = fftw_plan_dft_2d(MVX, MVY, bigH_real, bigH_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_gprime_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, gprime_spectral, gprime_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigGprime_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, bigGprime_real, bigGprime_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_h_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, h_spectral, h_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigH_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, bigH_real, bigH_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif

	// Initialize Q3q_spectral
	#pragma omp parallel for private(i) shared(NVLOC, Q3q_spectral)
	for (i = 0; i < NVLOC; i++){
		Q3q_spectral[i][0] = 0.;
		Q3q_spectral[i][1] = 0.;
	}

	for (int p = 0; p < parameters.mint; p++){
		// Compute gprimehat_p(k) = alphaprime_p(k) * fhat(k)
		// alphaprime_p is a real number
		#pragma omp parallel for private(i) shared(NVLOC, gprime_spectral, p, alphaprime, Ginput_spectral)
		for (i = 0; i < NVLOC; i++){
			gprime_spectral[i][0] = alphaprime[p][i] * Ginput_spectral[i][0];
			gprime_spectral[i][1] = alphaprime[p][i] * Ginput_spectral[i][1];
		}

		// Do inverse FFTs gprimehat_p -> gprime_p
		fftw_execute(plan_gprime_backward);
		// Renormalize gprime_p
		// gprime_p is a real number
		#pragma omp parallel for private(i) shared(NVLOC, gprime_real, MV)
		for (i = 0; i < NVLOC; i++){
			gprime_real[i][0] /= MV;
			gprime_real[i][1] = 0.;
		}

		// Compute Gprime_p(k) = gprime_p(k) * f(k)
		// gprime_p and f are real numbers
		#pragma omp parallel for private(i) shared(NVLOC, bigGprime_real, gprime_real, Ginput_real)
		for (i = 0; i < NVLOC; i++){
			bigGprime_real[i][0] = gprime_real[i][0] * Ginput_real[i][0];
			bigGprime_real[i][1] = 0.;
		}

		// Do FFT Gprime_p -> Gprimehat_p
		fftw_execute(plan_bigGprime_forward);

		// Compute hhat_p(k) = alpha_p(k) * Gprimehat_p(k)
		// alpha_p is a real number
		#pragma omp parallel for private(i) shared(NVLOC, h_spectral, alpha, p, bigGprime_spectral)
		for (i = 0; i < NVLOC; i++){
			h_spectral[i][0] = alpha[p][i] * bigGprime_spectral[i][0];
			h_spectral[i][1] = alpha[p][i] * bigGprime_spectral[i][1];
		}

		// Do inverse FFTs hhat_p -> h_p
		fftw_execute(plan_h_backward);
		// Renormalize h_p
		// h_p is a real number
		#pragma omp parallel for private(i) shared(NVLOC, h_real, MV)
		for (i = 0; i < NVLOC; i++){
			h_real[i][0] /= MV;
			h_real[i][1] = 0.;
		}

		// Compute H_p(k) = h_p(k) * f(k)
		// h_p and f are real numbers
		#pragma omp parallel for private(i) shared(NVLOC, bigH_real, h_real, Ginput_real)
		for (i = 0; i < NVLOC; i++){
			bigH_real[i][0] = h_real[i][0] * Ginput_real[i][0];
			bigH_real[i][1] = 0.;
		}

		// Do FFT H_p -> Hhat_p
		fftw_execute(plan_bigH_forward);

		// Update Q3q_spectral
		#pragma omp parallel for private(i) shared(NVLOC, Q3q_spectral, C, bigH_spectral)
		for (i = 0; i < NVLOC; i++){
			Q3q_spectral[i][0] += C * bigH_spectral[i][0];
			Q3q_spectral[i][1] += C * bigH_spectral[i][1];
		}
	}

	// Destroy plans
	fftw_destroy_plan(plan_gprime_backward);
	fftw_destroy_plan(plan_h_backward);
	fftw_destroy_plan(plan_bigGprime_forward);
	fftw_destroy_plan(plan_bigH_forward);

	// Free memory
	fftw_free(h_real);
	fftw_free(h_spectral);
	fftw_free(gprime_real);
	fftw_free(gprime_spectral);
	fftw_free(bigGprime_real);
	fftw_free(bigGprime_spectral);
	fftw_free(bigH_real);
	fftw_free(bigH_spectral);
}

/*
void collision_Q3q_convolution_fftw(fftw_complex *Ginput_spectral, fftw_complex *Ginput_real, fftw_complex *Q3q_spectral, double ***A, double ***C, double *W, PARAMETERS parameters){
	double pi = 4.*atan(1.);

	fftw_plan plan_AGhat_backward, plan_CGhat_backward, plan_H_forward;
	fftw_complex *AGhat_spectral, *AGhat_real;
	fftw_complex *CGhat_spectral, *CGhat_real;
	fftw_complex *H_real, *H_spectral;

	int i;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

#ifdef key_fftw_mpi
	AGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	AGhat_real = fftw_alloc_complex(parameters.alloc_local);
	CGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	CGhat_real = fftw_alloc_complex(parameters.alloc_local);
	H_real = fftw_alloc_complex(parameters.alloc_local);
	H_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs on A[p]*Ghat (backward), C[p]*Ghat (backward) and H[p] = IFFT(A[p]*Ghat) * IFFT(C[p]*Ghat) * G (forward)
	if (parameters.nvz == 0){
		plan_AGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, AGhat_spectral, AGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, CGhat_spectral, CGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_2d(MVX, MVY, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_AGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, AGhat_spectral, AGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, CGhat_spectral, CGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	AGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	AGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	CGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	CGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs on A[p]*Ghat (backward), C[p]*Ghat (backward) and H[p] = IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) * G (forward)
	if (parameters.nvz == 0){
		plan_AGhat_backward = fftw_plan_dft_2d(MVX, MVY, AGhat_spectral, AGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_plan_dft_2d(MVX, MVY, CGhat_spectral, CGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_2d(MVX, MVY, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_AGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, AGhat_spectral, AGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, CGhat_spectral, CGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif

	// Initialize Q3q_spectral
	#pragma omp parallel for private(i) shared(NVLOC, Q3q_spectral)
	for (i = 0; i < NVLOC; i++){
		Q3q_spectral[i][0] = 0.;
		Q3q_spectral[i][1] = 0.;
	}

	for (int p = 0; p < parameters.mint; p++){
		// Compute A[p]*Ghat, C[p]*Ghat
		// WARNING : A, C, Ghat are complex numbers
		#pragma omp parallel for private(i) shared(NVLOC, p, AGhat_spectral, CGhat_spectral, A, C, Ginput_spectral)
		for (i = 0; i < NVLOC; i++){
			AGhat_spectral[i][0] = A[p][i][0]*Ginput_spectral[i][0] - A[p][i][1]*Ginput_spectral[i][1];
			AGhat_spectral[i][1] = A[p][i][0]*Ginput_spectral[i][1] + A[p][i][1]*Ginput_spectral[i][0];

			CGhat_spectral[i][0] = C[p][i][0]*Ginput_spectral[i][0] - C[p][i][1]*Ginput_spectral[i][1];
			CGhat_spectral[i][1] = C[p][i][0]*Ginput_spectral[i][1] + C[p][i][1]*Ginput_spectral[i][0];
		}

		// Compute the backward FFT of A[p]*Ghat, C[p]*Ghat
		// WARNING : we get complex numbers
		fftw_execute(plan_AGhat_backward);
		fftw_execute(plan_CGhat_backward);
		#pragma omp parallel for private(i) shared(NVLOC, AGhat_real, CGhat_real, MV)
		for (i = 0; i < NVLOC; i++){
			AGhat_real[i][0] /= ((double)MV);
			AGhat_real[i][1] /= ((double)MV);

			CGhat_real[i][0] /= ((double)MV);
			CGhat_real[i][1] /= ((double)MV);
		}

		// Compute H[p] = IFFT(A[p]*Ghat) * IFFT(C[p]*Ghat) * G
		#pragma omp parallel for private(i) shared(NVLOC, H_real, AGhat_real, CGhat_real, Ginput_real)
		for (i = 0; i < NVLOC; i++){
			H_real[i][0] = AGhat_real[i][0] * (Ginput_real[i][0]*CGhat_real[i][0]) - AGhat_real[i][1] * (Ginput_real[i][0]*CGhat_real[i][1]);
			H_real[i][1] = AGhat_real[i][0] * (Ginput_real[i][0]*CGhat_real[i][1]) + AGhat_real[i][1] * (Ginput_real[i][0]*CGhat_real[i][0]);
		}

		// Compute the FFT of H[p]
		fftw_execute(plan_H_forward);

		// Increment Q1q_spectral with W[p] * Hhat[p]
		#pragma omp parallel for private(i) shared(NVLOC, p, Q3q_spectral, W, H_spectral)
		for (i = 0; i < NVLOC; i++){
			Q3q_spectral[i][0] += (W[p] * H_spectral[i][0]);
			Q3q_spectral[i][1] += (W[p] * H_spectral[i][1]);
		}
	}

	// Destroy plans
	fftw_destroy_plan(plan_AGhat_backward);
	fftw_destroy_plan(plan_CGhat_backward);
	fftw_destroy_plan(plan_H_forward);

	// Free memory
	fftw_free(AGhat_spectral);
	fftw_free(AGhat_real);
	fftw_free(CGhat_spectral);
	fftw_free(CGhat_real);
	fftw_free(H_real);
	fftw_free(H_spectral);
}
*/



void collision_Q4q_fftw(fftw_complex *Ginput_spectral, fftw_complex *Ginput_real, fftw_complex *Q4q_spectral, double **alpha, double **alphaprime, PARAMETERS parameters){
	double pi = 4.*atan(1.);
	//int ind;
	fftw_plan plan_g_backward, plan_bigG_forward, plan_hprime_backward, plan_bigHprime_forward;
	fftw_complex *hprime_real;
	fftw_complex *hprime_spectral;
	fftw_complex *g_real;
	fftw_complex *g_spectral;
	fftw_complex *bigG_real;
	fftw_complex *bigG_spectral;
	fftw_complex *bigHprime_real;
	fftw_complex *bigHprime_spectral;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	int i;

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

	// Reminder : parameters.ndim = 1 for 2D, parameters.ndim = 2 for 3D
	// Reminder : for the 3D case, mint = mint1*mint2
	double C = pow(pi, parameters.ndim)/((double)parameters.mint);

#ifdef key_fftw_mpi
	hprime_real = fftw_alloc_complex(parameters.alloc_local);
	hprime_spectral = fftw_alloc_complex(parameters.alloc_local);
	g_real = fftw_alloc_complex(parameters.alloc_local);
	g_spectral = fftw_alloc_complex(parameters.alloc_local);
	bigG_real = fftw_alloc_complex(parameters.alloc_local);
	bigG_spectral = fftw_alloc_complex(parameters.alloc_local);
	bigHprime_real = fftw_alloc_complex(parameters.alloc_local);
	bigHprime_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_hprime_backward = fftw_mpi_plan_dft_2d(MVX, MVY, hprime_spectral, hprime_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigHprime_forward = fftw_mpi_plan_dft_2d(MVX, MVY, bigHprime_real, bigHprime_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_g_backward = fftw_mpi_plan_dft_2d(MVX, MVY, g_spectral, g_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_mpi_plan_dft_2d(MVX, MVY, bigG_real, bigG_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_hprime_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, hprime_spectral, hprime_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigHprime_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, bigHprime_real, bigHprime_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_g_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, g_spectral, g_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, bigG_real, bigG_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	hprime_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	hprime_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	g_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	g_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigG_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigG_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigHprime_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	bigHprime_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs ghat -> g, gprimehat -> gprime, G -> Ghat
	if (parameters.nvz == 0){
		plan_hprime_backward = fftw_plan_dft_2d(MVX, MVY, hprime_spectral, hprime_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigHprime_forward = fftw_plan_dft_2d(MVX, MVY, bigHprime_real, bigHprime_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_g_backward = fftw_plan_dft_2d(MVX, MVY, g_spectral, g_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_plan_dft_2d(MVX, MVY, bigG_real, bigG_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_hprime_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, hprime_spectral, hprime_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigHprime_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, bigHprime_real, bigHprime_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
		plan_g_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, g_spectral, g_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_bigG_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, bigG_real, bigG_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif


	// Initialize Q4q_spectral
	#pragma omp parallel for private(i) shared(NVLOC, Q4q_spectral)
	for (i = 0; i < NVLOC; i++){
		Q4q_spectral[i][0] = 0.;
		Q4q_spectral[i][1] = 0.;
	}

	for (int p = 0; p < parameters.mint; p++){
		// Compute ghat_p(k) = alpha_p(k) * fhat(k)
		// alpha_p is a real number
		#pragma omp parallel for private(i) shared(NVLOC, g_spectral, alpha, Ginput_spectral, p)
		for (i = 0; i < NVLOC; i++){
			g_spectral[i][0] = alpha[p][i] * Ginput_spectral[i][0];
			g_spectral[i][1] = alpha[p][i] * Ginput_spectral[i][1];
		}

		// Do inverse FFTs ghat_p -> g_p
		fftw_execute(plan_g_backward);
		// Renormalize g_p
		// g_p is a real number
		#pragma omp parallel for private(i) shared(NVLOC, g_real, MV)
		for (i = 0; i < NVLOC; i++){
			g_real[i][0] /= MV;
			g_real[i][1] = 0.;
		}

		// Compute G_p(k) = g_p(k) * f(k)
		// g_p and f are real numbers
		#pragma omp parallel for private(i) shared(NVLOC, bigG_real, g_real, Ginput_real)
		for (i = 0; i < NVLOC; i++){
			bigG_real[i][0] = g_real[i][0] * Ginput_real[i][0];
			bigG_real[i][1] = 0.;
		}

		// Do FFT G_p -> Ghat_p
		fftw_execute(plan_bigG_forward);

		// Compute hprimehat_p(k) = alphaprime_p(k) * Ghat_p(k)
		// alphaprime_p is a real number
		#pragma omp parallel for private(i) shared(NVLOC, hprime_spectral, alphaprime, p, bigG_spectral)
		for (i = 0; i < NVLOC; i++){
			hprime_spectral[i][0] = alphaprime[p][i] * bigG_spectral[i][0];
			hprime_spectral[i][1] = alphaprime[p][i] * bigG_spectral[i][1];
		}

		// Do inverse FFTs hprimehat_p -> hprime_p
		fftw_execute(plan_hprime_backward);
		// Renormalize hprime_p
		// hprime_p is a real number
		#pragma omp parallel for private(i) shared(NVLOC, hprime_real, MV)
		for (i = 0; i < NVLOC; i++){
			hprime_real[i][0] /= MV;
			hprime_real[i][1] = 0.;
		}

		// Compute Hprime_p(k) = hprime_p(k) * f(k)
		// hprime_p and f are real numbers
		#pragma omp parallel for private(i) shared(NVLOC, bigHprime_real, hprime_real, Ginput_real)
		for (i = 0; i < NVLOC; i++){
			bigHprime_real[i][0] = hprime_real[i][0] * Ginput_real[i][0];
			bigHprime_real[i][1] = 0.;
		}

		// Do FFT Hprime_p -> Hprimehat_p
		fftw_execute(plan_bigHprime_forward);

		// Update Q4q_spectral
		#pragma omp parallel for private(i) shared(NVLOC, Q4q_spectral, C, bigHprime_spectral)
		for (i = 0; i < NVLOC; i++){
			Q4q_spectral[i][0] += C * bigHprime_spectral[i][0];
			Q4q_spectral[i][1] += C * bigHprime_spectral[i][1];
		}
	}

	// Destroy plans
	fftw_destroy_plan(plan_hprime_backward);
	fftw_destroy_plan(plan_g_backward);
	fftw_destroy_plan(plan_bigHprime_forward);
	fftw_destroy_plan(plan_bigG_forward);

	// Free memory
	fftw_free(g_real);
	fftw_free(g_spectral);
	fftw_free(hprime_real);
	fftw_free(hprime_spectral);
	fftw_free(bigHprime_real);
	fftw_free(bigHprime_spectral);
	fftw_free(bigG_real);
	fftw_free(bigG_spectral);
}




/*
void collision_Q4q_convolution_fftw(fftw_complex *Ginput_spectral, fftw_complex *Ginput_real, fftw_complex *Q4q_spectral, double ***B, double ***C, double *W, PARAMETERS parameters){
	double pi = 4.*atan(1.);

	fftw_plan plan_BGhat_backward, plan_CGhat_backward, plan_H_forward;
	fftw_complex *BGhat_spectral, *BGhat_real;
	fftw_complex *CGhat_spectral, *CGhat_real;
	fftw_complex *H_real, *H_spectral;

	int i;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

#ifdef key_fftw_mpi
	BGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	BGhat_real = fftw_alloc_complex(parameters.alloc_local);
	CGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	CGhat_real = fftw_alloc_complex(parameters.alloc_local);
	H_real = fftw_alloc_complex(parameters.alloc_local);
	H_spectral = fftw_alloc_complex(parameters.alloc_local);

	// Prepare the FFTs on B[p]*Ghat (backward), C[p]*Ghat (backward) and H[p] = G * IFFT(B[p]*Ghat) * IFFT(C[p]*Ghat) (forward)
	if (parameters.nvz == 0){
		plan_BGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, BGhat_spectral, BGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, CGhat_spectral, CGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_2d(MVX, MVY, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_BGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, BGhat_spectral, BGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, CGhat_spectral, CGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#else
	BGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	BGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	CGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	CGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	H_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	// Prepare the FFTs on B[p]*Ghat (backward), C[p]*Ghat (backward) and H[p] = G * IFFT(B[p]*Ghat) * IFFT(C[p]*Ghat) (forward)
	if (parameters.nvz == 0){
		plan_BGhat_backward = fftw_plan_dft_2d(MVX, MVY, BGhat_spectral, BGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_plan_dft_2d(MVX, MVY, CGhat_spectral, CGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_2d(MVX, MVY, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
	else{
		plan_BGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, BGhat_spectral, BGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_CGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, CGhat_spectral, CGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
		plan_H_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, H_real, H_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	}
#endif

	// Initialize Q4q_spectral
	#pragma omp parallel for private(i) shared(NVLOC, Q4q_spectral)
	for (i = 0; i < NVLOC; i++){
		Q4q_spectral[i][0] = 0.;
		Q4q_spectral[i][1] = 0.;
	}

	for (int p = 0; p < parameters.mint; p++){
		// Compute B[p]*Ghat, C[p]*Ghat
		// WARNING : B, C, Ghat are complex numbers
		#pragma omp parallel for private(i) shared(NVLOC, p, BGhat_spectral, CGhat_spectral, B, C, Ginput_spectral)
		for (i = 0; i < NVLOC; i++){
			BGhat_spectral[i][0] = B[p][i][0]*Ginput_spectral[i][0] - B[p][i][1]*Ginput_spectral[i][1];
			BGhat_spectral[i][1] = B[p][i][0]*Ginput_spectral[i][1] + B[p][i][1]*Ginput_spectral[i][0];

			CGhat_spectral[i][0] = C[p][i][0]*Ginput_spectral[i][0] - C[p][i][1]*Ginput_spectral[i][1];
			CGhat_spectral[i][1] = C[p][i][0]*Ginput_spectral[i][1] + C[p][i][1]*Ginput_spectral[i][0];
		}

		// Compute the backward FFT of B[p]*Ghat, C[p]*Ghat
		// WARNING : we get complex numbers
		fftw_execute(plan_BGhat_backward);
		fftw_execute(plan_CGhat_backward);
		#pragma omp parallel for private(i) shared(NVLOC, BGhat_real, CGhat_real, MV)
		for (i = 0; i < NVLOC; i++){
			BGhat_real[i][0] /= ((double)MV);
			BGhat_real[i][1] /= ((double)MV);

			CGhat_real[i][0] /= ((double)MV);
			CGhat_real[i][1] /= ((double)MV);
		}

		// Compute H[p] = G * IFFT(B[p]*Ghat) * IFFT(C[p]*Ghat)
		#pragma omp parallel for private(i) shared(NVLOC, H_real, Ginput_real, BGhat_real, CGhat_real)
		for (i = 0; i < NVLOC; i++){
			H_real[i][0] = Ginput_real[i][0] * (BGhat_real[i][0]*CGhat_real[i][0] - BGhat_real[i][1]*CGhat_real[i][1]);
			H_real[i][1] = Ginput_real[i][0] * (BGhat_real[i][1]*CGhat_real[i][0] + BGhat_real[i][0]*CGhat_real[i][1]);
		}

		// Compute the FFT of H[p]
		fftw_execute(plan_H_forward);

		// Increment Q4q_spectral with W[p] * Hhat[p]
		#pragma omp parallel for private(i) shared(NVLOC, p, Q4q_spectral, W, H_spectral)
		for (i = 0; i < NVLOC; i++){
			Q4q_spectral[i][0] += (W[p] * H_spectral[i][0]);
			Q4q_spectral[i][1] += (W[p] * H_spectral[i][1]);
		}
	}

	// Destroy plans
	fftw_destroy_plan(plan_BGhat_backward);
	fftw_destroy_plan(plan_CGhat_backward);
	fftw_destroy_plan(plan_H_forward);

	// Free memory
	fftw_free(BGhat_spectral);
	fftw_free(BGhat_real);
	fftw_free(CGhat_spectral);
	fftw_free(CGhat_real);
	fftw_free(H_real);
	fftw_free(H_spectral);
}
*/

void collision_Qtotal_fftw(double *Gloc, double *QGloc, double **alpha, double **alphaprime, double omega, PARAMETERS parameters){	
	double pi = 4.*atan(1.);
	
	FILE *f;
	char namefile[MAXSTRINGLENGTH];
	char suffix[MAXSTRINGLENGTH];
	char hostname[MAXSTRINGLENGTH];
	char **hostlist;
	
	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	fftw_complex *QG_spectral;
	fftw_complex *QG_real;
	fftw_complex *G_spectral;
	fftw_complex *G_real;
	fftw_plan plan_G_forward, plan_QG_backward;

	int i;
	
	int rank, commsize, nthreads;
	double tend, tstart, elap;
	double *elaptab_Q1c, *elaptab_Q2c, *elaptab_Q1q, *elaptab_Q2q, *elaptab_Q3q, *elaptab_Q4q;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
#else
	rank = 0;
	commsize = 1;
#endif

#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

	double *Qtest, test;
	Qtest = allocation_1d(NVXLOC*NVYLOC*NVZLOC);

	double myalpha, mu, myomega, phi;
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		mu = pow(parameters.L/pi, parameters.ndim+1);
		myomega = 1.;
	}
	else{
		mu = 1.;
		myomega = omega;
	}
	
	if (parameters.type_collision == QUANTUM_COLLISIONS){
		if (parameters.type_qparticles == BOSONS)
			myalpha = pow(parameters.hbar,parameters.ndim+1);
		else if (parameters.type_qparticles == FERMIONS)
			myalpha = -pow(parameters.hbar,parameters.ndim+1);
	}
	else if (parameters.type_collision == CLASSICAL_COLLISIONS)
		myalpha = 0.;
	phi = myalpha * mu * pow(pi*myomega/parameters.L, parameters.ndim+1);


	elaptab_Q1c = allocation_1d(commsize);
	elaptab_Q2c = allocation_1d(commsize);
	elaptab_Q1q = allocation_1d(commsize);
	elaptab_Q2q = allocation_1d(commsize);
	elaptab_Q3q = allocation_1d(commsize);
	elaptab_Q4q = allocation_1d(commsize);

#ifdef key_fftw_mpi
	QG_spectral = fftw_alloc_complex(parameters.alloc_local);
	QG_real = fftw_alloc_complex(parameters.alloc_local);
	G_spectral = fftw_alloc_complex(parameters.alloc_local);
	G_real = fftw_alloc_complex(parameters.alloc_local);

	switch (parameters.ndim){
		case DIM2:
			plan_G_forward = fftw_mpi_plan_dft_2d(MVX, MVY, G_real, G_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_QG_backward = fftw_mpi_plan_dft_2d(MVX, MVY, QG_spectral, QG_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		case DIM3:
			plan_G_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, G_real, G_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_QG_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, QG_spectral, QG_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		default:
			error("Unknown dimensionality");
	}
#else
	QG_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	QG_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	G_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	G_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	switch (parameters.ndim){
		case DIM2:
			plan_G_forward = fftw_plan_dft_2d(MVX, MVY, G_real, G_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_QG_backward = fftw_plan_dft_2d(MVX, MVY, QG_spectral, QG_real, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		case DIM3:
			plan_G_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, G_real, G_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_QG_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, QG_spectral, QG_real, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		default:
			error("Unknown dimensionality");
	}	
#endif
/*
#ifdef key_openmp
	sprintf(namefile, "alpha_%d.txt", nthreads);
	f = fopen(namefile, "w");
	for (int kvx = 0; kvx < NVXLOC; kvx++){
		for (int kvy = 0; kvy < NVYLOC; kvy++){
			for (int kvz = 0; kvz < NVZLOC; kvz++){
				for (int p = 0; p < parameters.mint; p++)
					fprintf(f, "(%d,%d,%d) | %d | %e\n", kvx, kvy, kvz, p, alpha[p][kvz+NVZLOC*(kvy+NVYLOC*kvx)]);
			}
		}
	}
	fclose(f);
#endif


#ifdef key_openmp
	sprintf(namefile, "alphaprime_%d.txt", nthreads);
	f = fopen(namefile, "w");
	for (int kvx = 0; kvx < NVXLOC; kvx++){
		for (int kvy = 0; kvy < NVYLOC; kvy++){
			for (int kvz = 0; kvz < NVZLOC; kvz++){
				for (int p = 0; p < parameters.mint; p++)
					fprintf(f, "(%d,%d,%d) | %d | %e\n", kvx, kvy, kvz, p, alphaprime[p][kvz+NVZLOC*(kvy+NVYLOC*kvx)]);
			}
		}
	}
	fclose(f);
#endif
*/
	#pragma omp parallel for private(i) shared(NVLOC, G_real, Gloc, QGloc)
	for (i = 0; i < NVLOC; i++){
		G_real[i][0] = Gloc[i];
		G_real[i][1] = 0.;
		QGloc[i] = 0.;
		G_spectral[i][0] = 0.;
		G_spectral[i][1] = 0.;
	}

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif

/*
#ifdef key_openmp
	sprintf(namefile, "%s/%s_G_real.txt", parameters.outputdir, parameters.prefix);
	f = fopen(namefile, "w");
	for (int kvx = 0; kvx < NVXLOC; kvx++){
		for (int kvy = 0; kvy < NVYLOC; kvy++){
			for (int kvz = 0; kvz < NVZLOC; kvz++)
				fprintf(f, "(%d,%d,%d) | (%e,%e)\n", kvx, kvy, kvz, G_real[kvz+NVZLOC*(kvy+NVYLOC*kvx)][0], G_real[kvz+NVZLOC*(kvy+NVYLOC*kvx)][1]);
		}
	}
	fclose(f);
#endif
*/

	fftw_execute(plan_G_forward);
/*
#ifdef key_openmp
	sprintf(namefile, "%s/%s_G_spectral.txt", parameters.outputdir, parameters.prefix);
	f = fopen(namefile, "w");
	for (int kvx = 0; kvx < NVXLOC; kvx++){
		for (int kvy = 0; kvy < NVYLOC; kvy++){
			for (int kvz = 0; kvz < NVZLOC; kvz++)
				fprintf(f, "(%d,%d,%d) | (%e,%e)\n", kvx, kvy, kvz, G_spectral[kvz+NVZLOC*(kvy+NVYLOC*kvx)][0], G_spectral[kvz+NVZLOC*(kvy+NVYLOC*kvx)][1]);
		}
	}
	fclose(f);
#endif
*/

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
	tstart = MPI_Wtime();
#elif defined(key_openmp)
	tstart = omp_get_wtime();
#else
	tstart = (double)clock();
#endif

	collision_Q1c_fftw(G_spectral, QG_spectral, alpha, alphaprime, parameters);

#ifdef key_mpi
	tend = MPI_Wtime();
	elap = tend-tstart;
	MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q1c, 1, MPI_DOUBLE, MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
	tend = omp_get_wtime();
	elap = tend-tstart;
	elaptab_Q1c[0] = elap;
#else
	tend = (double)clock();
	elap = (tend-tstart)/CLOCKS_PER_SEC;
	elaptab_Q1c[0] = elap;
#endif

	fftw_execute(plan_QG_backward);

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif

	test = 0.;
	strcpy(namefile, "");
	strcat(namefile, parameters.outputdir);
	strcat(namefile, "/");
	strcat(namefile, parameters.prefix);
	strcat(namefile, "_Q1c_real.txt");
	f = fopen(namefile, "w");
	#pragma omp parallel for private(i) shared(NVLOC, MV, QG_real, QGloc, Qtest, test)
	for (i = 0; i < NVLOC; i++){
		QG_real[i][0] /= MV;
		QG_real[i][1] /= MV;
		fprintf(f, "Q1c_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
		QGloc[i] += QG_real[i][0];
		Qtest[i] = QG_real[i][0];
		//#pragma omp critical
		//test = fmax(fabs(Qtest[i]), test);
	}
	fclose(f);
	//printf("Linf(Q1c) = %e\n", test);
	message("Q1c computed\n");

	/*f = fopen("Q1c_serial.txt", "w");
	for (int i = 0; i < MV; i++)
		fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
	fclose(f);*/

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
	tstart = MPI_Wtime();
#elif defined(key_openmp)
	tstart = omp_get_wtime();
#else
	tstart = (double)clock();
#endif

	collision_Q2c_fftw(G_spectral, G_real, QG_spectral, alpha, alphaprime, parameters);

#ifdef key_mpi
	tend = MPI_Wtime();
	elap = tend-tstart;
	MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q2c, 1, MPI_DOUBLE, MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
	tend = omp_get_wtime();
	elap = tend-tstart;
	elaptab_Q2c[0] = elap;
#else
	tend = (double)clock();
	elap = (tend-tstart)/CLOCKS_PER_SEC;
	elaptab_Q2c[0] = elap;
#endif
	
	fftw_execute(plan_QG_backward);

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif

	test = 0.;
	strcpy(namefile, "");
	strcat(namefile, parameters.outputdir);
	strcat(namefile, "/");
	strcat(namefile, parameters.prefix);
	strcat(namefile, "_Q2c_real.txt");
	f = fopen(namefile, "w");
	#pragma omp parallel for private(i) shared(NVLOC, MV, QG_real, QGloc, Qtest, test)
	for (i = 0; i < NVLOC; i++){
		QG_real[i][0] /= MV;
		QG_real[i][1] /= MV;
		fprintf(f, "Q2c_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
		QGloc[i] -= QG_real[i][0];
		Qtest[i] = QG_real[i][0];
		//#pragma omp critical
		//test = fmax(fabs(Qtest[i]), test);
	}
	fclose(f);

	//printf("Linf(Q2c) = %e\n", test);
	message("Q2c computed\n");

	/*f = fopen("Q2c_serial.txt", "w");
	for (int i = 0; i < MV; i++)
		fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
	fclose(f);*/

	if (parameters.type_collision == QUANTUM_COLLISIONS){
#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
		tstart = MPI_Wtime();
#elif defined(key_openmp)
		tstart = omp_get_wtime();
#else
		tstart = (double)clock();
#endif

		collision_Q1q_fftw(G_real, QG_spectral, alpha, alphaprime, parameters);

#ifdef key_mpi
		tend = MPI_Wtime();
		elap = tend-tstart;
		MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q1q, 1, MPI_DOUBLE, MPI_COMM_WORLD);
		MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
		tend = omp_get_wtime();
		elap = tend-tstart;
		elaptab_Q1q[0] = elap;
#else
		tend = (double)clock();
		elap = (tend-tstart)/CLOCKS_PER_SEC;
		elaptab_Q1q[0] = elap;
#endif

		fftw_execute(plan_QG_backward);

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif
		
		test = 0.;
		strcpy(namefile, "");
		strcat(namefile, parameters.outputdir);
		strcat(namefile, "/");
		strcat(namefile, parameters.prefix);
		strcat(namefile, "_Q1q_real.txt");
		f = fopen(namefile, "w");
		#pragma omp parallel for private(i) shared(NVLOC, QG_real, MV, QGloc, phi, Qtest, test)
		for (i = 0; i < NVLOC; i++){
			QG_real[i][0] /= MV;
			QG_real[i][1] /= MV;
			fprintf(f, "Q1q_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
			QGloc[i] += phi*QG_real[i][0];
			Qtest[i] = phi*QG_real[i][0];
			//#pragma omp critical
			//test = fmax(fabs(Qtest[i]), test);
		}
		fclose(f);
		//printf("Linf(Q1q) = %e\n", test);
		message("Q1q computed\n");

		/*f = fopen("Q1q_serial.txt", "w");
		for (int i = 0; i < MV; i++)
			fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
		fclose(f);*/

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
		tstart = MPI_Wtime();
#elif defined(key_openmp)
		tstart = omp_get_wtime();
#else
		tstart = (double)clock();
#endif

		collision_Q2q_fftw(G_spectral, G_real, QG_spectral, alpha, alphaprime, parameters);

#ifdef key_mpi
		tend = MPI_Wtime();
		elap = tend-tstart;
		MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q2q, 1, MPI_DOUBLE, MPI_COMM_WORLD);
		MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
		tend = omp_get_wtime();
		elap = tend-tstart;
		elaptab_Q2q[0] = elap;
#else
		tend = (double)clock();
		elap = (tend-tstart)/CLOCKS_PER_SEC;
		elaptab_Q2q[0] = elap;
#endif

		fftw_execute(plan_QG_backward);

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif
		
		test = 0.;
		strcpy(namefile, "");
		strcat(namefile, parameters.outputdir);
		strcat(namefile, "/");
		strcat(namefile, parameters.prefix);
		strcat(namefile, "_Q2q_real.txt");
		f = fopen(namefile, "w");
		#pragma omp parallel for private(i) shared(NVLOC, QG_real, MV, QGloc, phi, Qtest, test)
		for (int i = 0; i < NVLOC; i++){
			QG_real[i][0] /= MV;
			QG_real[i][1] /= MV;
			fprintf(f, "Q2q_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
			QGloc[i] += phi*QG_real[i][0];
			Qtest[i] = phi*QG_real[i][0];
			//#pragma omp critical
			//test = fmax(fabs(Qtest[i]), test);
		}
		fclose(f);
		//printf("Linf(Q2q) = %e\n", test);
		message("Q2q computed\n");

		/*f = fopen("Q2q_serial.txt", "w");
		for (int i = 0; i < MV; i++)
			fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
		fclose(f);*/

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
		tstart = MPI_Wtime();
#elif defined(key_openmp)
		tstart = omp_get_wtime();
#else
		tstart = (double)clock();
#endif

		collision_Q3q_fftw(G_spectral, G_real, QG_spectral, alpha, alphaprime, parameters);

#ifdef key_mpi
		tend = MPI_Wtime();
		elap = tend-tstart;
		MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q3q, 1, MPI_DOUBLE, MPI_COMM_WORLD);
		MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
		tend = omp_get_wtime();
		elap = tend-tstart;
		elaptab_Q3q[0] = elap;
#else
		tend = (double)clock();
		elap = (tend-tstart)/CLOCKS_PER_SEC;
		elaptab_Q3q[0] = elap;
#endif
		
		fftw_execute(plan_QG_backward);

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif
		
		test = 0.;
		strcpy(namefile, "");
		strcat(namefile, parameters.outputdir);
		strcat(namefile, "/");
		strcat(namefile, parameters.prefix);
		strcat(namefile, "_Q3q_real.txt");
		f = fopen(namefile, "w");
		#pragma omp parallel for private(i) shared(NVLOC, QG_real, MV, QGloc, phi, Qtest, test)
		for (int i = 0; i < NVLOC; i++){
			QG_real[i][0] /= MV;
			QG_real[i][1] /= MV;
			fprintf(f, "Q3q_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
			QGloc[i] -= phi*QG_real[i][0];
			Qtest[i] = phi*QG_real[i][0];
			//#pragma omp critical
			//test = fmax(fabs(Qtest[i]), test);
		}
		fclose(f);
		//printf("Linf(Q3q) = %e\n", test);
		message("Q3q computed\n");

		/*f = fopen("Q3q_serial.txt", "w");
		for (int i = 0; i < MV; i++)
			fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
		fclose(f);*/

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
		tstart = MPI_Wtime();
#elif defined(key_openmp)
		tstart = omp_get_wtime();
#else
		tstart = (double)clock();
#endif

		collision_Q4q_fftw(G_spectral, G_real, QG_spectral, alpha, alphaprime, parameters);

#ifdef key_mpi
		tend = MPI_Wtime();
		elap = tend-tstart;
		MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q4q, 1, MPI_DOUBLE, MPI_COMM_WORLD);
		MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
		tend = omp_get_wtime();
		elap = tend-tstart;
		elaptab_Q4q[0] = elap;
#else
		tend = (double)clock();
		elap = (tend-tstart)/CLOCKS_PER_SEC;
		elaptab_Q4q[0] = elap;
#endif
		
		fftw_execute(plan_QG_backward);

#ifdef _key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif
		
		test = 0.;
		strcpy(namefile, "");
		strcat(namefile, parameters.outputdir);
		strcat(namefile, "/");
		strcat(namefile, parameters.prefix);
		strcat(namefile, "_Q4q_real.txt");
		f = fopen(namefile, "w");
		#pragma omp parallel for private(i) shared(NVLOC, QG_real, MV, QGloc, phi, Qtest, test)
		for (int i = 0; i < NVLOC; i++){
			QG_real[i][0] /= MV;
			QG_real[i][1] /= MV;
			fprintf(f, "Q4q_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
			QGloc[i] -= phi*QG_real[i][0];
			Qtest[i] = phi*QG_real[i][0];
			//#pragma omp critical
			//test = fmax(fabs(Qtest[i]), test);
		}
		fclose(f);
		//printf("Linf(Q4q) = %e\n", test);
		message("Q4q computed\n");

		/*f = fopen("Q4q_serial.txt", "w");
		for (int i = 0; i < MV; i++)
			fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
		fclose(f);*/
	}

	gethostname(hostname, MAXSTRINGLENGTH);
	hostlist = malloc(sizeof(char*)*commsize);
	for (int R = 0; R < commsize; R++){
		hostlist[R] = allocation_1d_char(MAXSTRINGLENGTH);
		if (R == rank)
			sprintf(hostlist[R], "%s", hostname);
#ifdef key_mpi
		MPI_Bcast(hostlist[R], MAXSTRINGLENGTH, MPI_CHAR, R, MPI_COMM_WORLD);
#endif
	}
	
	if (rank == 0){
		strcpy(namefile, "");
		strcat(namefile, parameters.outputdir);
		strcat(namefile, "/");
		strcat(namefile, parameters.prefix);
		strcat(namefile, "_timecost_collision_");
#ifdef key_openmp
		sprintf(suffix, "%d", nthreads);
#else
		sprintf(suffix, "%d", commsize);
#endif
		strcat(namefile, suffix);
		strcat(namefile, ".txt");
		f = fopen(namefile, "w");
		fprintf(f, "# Kind of particles\n");
		if (parameters.type_collision == CLASSICAL_COLLISIONS)
			fprintf(f, "classical\n");
		else{
			if (parameters.type_qparticles == FERMIONS)
				fprintf(f, "fermions\n");
			else
				fprintf(f, "bosons\n");
		}
		fprintf(f, "# Number of grid points\n");
		fprintf(f, "%d %d", MVX, MVY);
		if (parameters.ndim == DIM3)
			fprintf(f, " %d", MVZ);
		fprintf(f, "\n");
#ifdef key_openmp
		fprintf(f, "# Number of threads\n");
		fprintf(f, "%d\n", nthreads);
#else
		fprintf(f, "# Number of processes\n");
		fprintf(f, "%d\n", commsize);
#endif
		fprintf(f, "# Rank |    Time cost Q1c    |    Time cost Q2c    ");
		if (parameters.type_collision == QUANTUM_COLLISIONS)
			fprintf(f, "|    Time cost Q1q    |    Time cost Q2q    |    Time cost Q3q    |    Time cost Q4q    ");
		fprintf(f, "| Hostname\n");
		for (int R = 0; R < commsize; R++){
			fprintf(f, "%7d|%.15e|%.15e", R, elaptab_Q1c[R], elaptab_Q2c[R]);
			if (parameters.type_collision == QUANTUM_COLLISIONS)
				fprintf(f, "|%.15e|%.15e|%.15e|%.15e", elaptab_Q1q[R], elaptab_Q2q[R], elaptab_Q3q[R], elaptab_Q4q[R]);
			fprintf(f, "|%s", hostlist[R]);
			fprintf(f, "\n");
		}
		fclose(f);
	}

#ifdef key_mpi	
	MPI_Barrier(MPI_COMM_WORLD);
#endif

	fftw_destroy_plan(plan_G_forward);
	fftw_destroy_plan(plan_QG_backward);
	
	fftw_free(QG_spectral);
	fftw_free(QG_real);
	fftw_free(G_spectral);
	fftw_free(G_real);

	free_1d(Qtest);
	free_1d(elaptab_Q1c);
	free_1d(elaptab_Q2c);
	free_1d(elaptab_Q1q);
	free_1d(elaptab_Q2q);
	free_1d(elaptab_Q3q);
	free_1d(elaptab_Q4q);
	
	for (int R = 0; R < commsize; R++)
		free_1d_char(hostlist[R]);
	free(hostlist);
}





//void collision_Qtotal_quick_convolution_fftw(double *Gloc, double *QGloc, double ***A, double ***B, double ***C, double *W, double omega, PARAMETERS parameters){
void collision_Qtotal_quick_convolution_fftw(double *Gloc, double *QGloc, double **xitilde, double **sigma, double *W, double omega, PARAMETERS parameters){
	double pi = 4.*atan(1.);

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	fftw_complex *G_real, *G_spectral;
	fftw_complex *AGhat_spectral, *AGhat_real;
	fftw_complex *BGhat_spectral, *BGhat_real;
	fftw_complex *CGhat_spectral, *CGhat_real;
	double A_re, A_im;
	double B_re, B_im;
	double C_re, C_im;

	int kvxabs, kvyabs, kvzabs, itemp;
	double mode, normxi;
	double modek[3];

	fftw_plan plan_G_forward;
	fftw_plan plan_AGhat_backward, plan_BGhat_backward, plan_CGhat_backward;

	double contrib_Q1c, contrib_Q2c, contrib_Q1q, contrib_Q2q, contrib_Q3q, contrib_Q4q;

	int k, p;

	double tstart, tend, elap, *elaptab;
	char namefile[MAXSTRINGLENGTH];
	char suffix[MAXSTRINGLENGTH];
	char hostname[MAXSTRINGLENGTH];
	FILE *f;
	char **hostlist;

	int rank, commsize, nthreads;

#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
#else
	rank = 0;
	commsize = 1;
#endif
	elaptab = allocation_1d(commsize);

#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

	double myalpha, mu, myomega, phi;
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		mu = pow(parameters.L/pi, parameters.ndim+1);
		myomega = 1.;
	}
	else{
		mu = 1.;
		myomega = omega;
	}
	
	if (parameters.type_collision == QUANTUM_COLLISIONS){
		if (parameters.type_qparticles == BOSONS)
			myalpha = pow(parameters.hbar,parameters.ndim+1);
		else if (parameters.type_qparticles == FERMIONS)
			myalpha = -pow(parameters.hbar,parameters.ndim+1);
	}
	else if (parameters.type_collision == CLASSICAL_COLLISIONS)
		myalpha = 0.;
	phi = myalpha * mu * pow(pi*myomega/parameters.L, parameters.ndim+1);

#ifdef key_fftw_mpi
	G_real= fftw_alloc_complex(parameters.alloc_local);
	G_spectral = fftw_alloc_complex(parameters.alloc_local);
	AGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	AGhat_real = fftw_alloc_complex(parameters.alloc_local);
	BGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	BGhat_real = fftw_alloc_complex(parameters.alloc_local);
	CGhat_spectral = fftw_alloc_complex(parameters.alloc_local);
	CGhat_real = fftw_alloc_complex(parameters.alloc_local);

	switch (parameters.ndim){
		case DIM2:
			plan_G_forward = fftw_mpi_plan_dft_2d(MVX, MVY, G_real, G_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_AGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, AGhat_spectral, AGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
			plan_BGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, BGhat_spectral, BGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
			plan_CGhat_backward = fftw_mpi_plan_dft_2d(MVX, MVY, CGhat_spectral, CGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		case DIM3:
			plan_G_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, G_real, G_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_AGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, AGhat_spectral, AGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
			plan_BGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, BGhat_spectral, BGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
			plan_CGhat_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, CGhat_spectral, CGhat_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		default:
			error("Unknown dimensionality");
	}
#else
	G_real= fftw_malloc(sizeof(fftw_complex)*NVLOC);
	G_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	AGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	AGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	BGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	BGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	CGhat_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	CGhat_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	switch (parameters.ndim){
		case DIM2:
			plan_G_forward = fftw_plan_dft_2d(MVX, MVY, G_real, G_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_AGhat_backward = fftw_plan_dft_2d(MVX, MVY, AGhat_spectral, AGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
			plan_BGhat_backward = fftw_plan_dft_2d(MVX, MVY, BGhat_spectral, BGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
			plan_CGhat_backward = fftw_plan_dft_2d(MVX, MVY, CGhat_spectral, CGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		case DIM3:
			plan_G_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, G_real, G_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_AGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, AGhat_spectral, AGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
			plan_BGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, BGhat_spectral, BGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
			plan_CGhat_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, CGhat_spectral, CGhat_real, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		default:
			error("Unknown dimensionality");
	}	
#endif

	#pragma omp parallel for private(k) shared(NVLOC, QGloc, G_real, Gloc)
	for (k = 0; k < NVLOC; k++){
		QGloc[k] = 0.;
		G_real[k][0] = Gloc[k];
		G_real[k][1] = 0.;
	}

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
		tstart = MPI_Wtime();
#elif defined(key_openmp)
		tstart = omp_get_wtime();
#else
		tstart = (double)clock();
#endif

	// Compute Ghat
	fftw_execute(plan_G_forward);

	for (p = 0; p < parameters.mint; p++){
		normxi = 0.;
		for (int d = 0; d <= parameters.ndim; d++)
			normxi += pow(xitilde[p][d], 2);
		normxi = sqrt(normxi);

		// Compute A[p]*Ghat, B[p]*Ghat, C[p]*Ghat
		#pragma omp parallel for private(k, A_re, A_im, B_re, B_im, C_re, C_im, kvxabs, kvyabs, kvzabs, itemp, modek, mode) shared(p, NVXLOC, NVYLOC, NVZLOC, NVLOC, IVX0LOC, IVY0LOC, IVZ0LOC, AGhat_spectral, BGhat_spectral, CGhat_spectral, G_spectral, MVX, MVY, MVZ, parameters, xitilde, sigma, W, normxi)
		for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
			kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
			kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
			kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
			itemp = k/NVZLOC;
			kvzabs = k - itemp*NVZLOC;
			kvxabs = itemp/NVYLOC;
			kvyabs = itemp-kvxabs*NVYLOC;
			kvxabs += IVX0LOC;
			kvyabs += IVY0LOC;
			kvzabs += IVZ0LOC;
			if (kvxabs >= MVX/2)
				kvxabs -= MVX;
			if (kvyabs >= MVY/2)
				kvyabs -= MVY;
			if (kvzabs >= MVZ/2)
				kvzabs -= MVZ;
#endif
			modek[0] = kvxabs;
			modek[1] = kvyabs;
			if (parameters.ndim == DIM3)
				modek[2] = kvzabs;

			mode = 0.;
			for (int d = 0; d <= parameters.ndim; d++){
				#pragma omp atomic
				mode += (modek[d]*(0.5*xitilde[p][d] + 0.5*normxi*sigma[p][d]));
			}
			A_re = cos(mode);
			A_im = sin(mode);

			mode = 0.;
			for (int d = 0; d <= parameters.ndim; d++){
				#pragma omp atomic
				mode += (modek[d]*(0.5*xitilde[p][d] - 0.5*normxi*sigma[p][d]));
			}
			B_re = cos(mode);
			B_im = sin(mode);

			mode = 0.;
			for (int d = 0; d <= parameters.ndim; d++){
				#pragma omp atomic
				mode += (modek[d]*xitilde[p][d]);
			}
			C_re = cos(mode);
			C_im = sin(mode);

			AGhat_spectral[k][0] = A_re * G_spectral[k][0] - A_im * G_spectral[k][1];
			AGhat_spectral[k][1] = A_re * G_spectral[k][1] + A_im * G_spectral[k][0];

			BGhat_spectral[k][0] = B_re * G_spectral[k][0] - B_im * G_spectral[k][1];
			BGhat_spectral[k][1] = B_re * G_spectral[k][1] + B_im * G_spectral[k][0];
			
			CGhat_spectral[k][0] = C_re * G_spectral[k][0] - C_im * G_spectral[k][1];
			CGhat_spectral[k][1] = C_re * G_spectral[k][1] + C_im * G_spectral[k][0];
			//printf("p = %d, k = (%d,%d,%d) | AGhat = (%e,%e), BGhat = (%e,%e), CGhat = (%e,%e)\n", p, kvxabs, kvyabs, kvzabs, AGhat_spectral[k][0], AGhat_spectral[k][1], BGhat_spectral[k][0], BGhat_spectral[k][1], CGhat_spectral[k][0], CGhat_spectral[k][1]);
		}

		// Compute IFFT(A[p]*Ghat), IFFT(B[p]*Ghat), IFFT(B[p]*Ghat)
		// WARNING : they are complex numbers
		fftw_execute(plan_AGhat_backward);
		fftw_execute(plan_BGhat_backward);
		fftw_execute(plan_CGhat_backward);
		#pragma omp parallel for private(k) shared(NVLOC, AGhat_real, BGhat_real, CGhat_real, MV)
		for (k = 0; k < NVLOC; k++){
			AGhat_real[k][0] /= ((double)MV);
			AGhat_real[k][1] /= ((double)MV);

			BGhat_real[k][0] /= ((double)MV);
			BGhat_real[k][1] /= ((double)MV);

			CGhat_real[k][0] /= ((double)MV);
			CGhat_real[k][1] /= ((double)MV);
		}

		#pragma omp parallel for private(k, contrib_Q1c, contrib_Q2c, contrib_Q1q, contrib_Q2q, contrib_Q3q, contrib_Q4q) shared(p, NVLOC, W, AGhat_real, BGhat_real, CGhat_real, G_real, phi, QGloc)
		for (k = 0; k < NVLOC; k++){
			// Add W[p] * IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) to Q1c
			contrib_Q1c = (AGhat_real[k][0]*BGhat_real[k][0] - AGhat_real[k][1]*BGhat_real[k][1]);

			// Add W[p] * IFFT(C[p]*Ghat) * G to Q2c
			contrib_Q2c = (CGhat_real[k][0]*G_real[k][0]);

			// Add W[p] * IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) * IFFT(C[p]*Ghat) to Q1q
			contrib_Q1q = ((AGhat_real[k][0]*BGhat_real[k][0] - AGhat_real[k][1]*BGhat_real[k][1])*CGhat_real[k][0] - (AGhat_real[k][0]*BGhat_real[k][1] + AGhat_real[k][1]*BGhat_real[k][0])*CGhat_real[k][1]);
			
			// Add W[p] * IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) * G to Q2q
			contrib_Q2q = ((AGhat_real[k][0]*BGhat_real[k][0] - AGhat_real[k][1]*BGhat_real[k][1])*G_real[k][0]);

			// Add W[p] * IFFT(A[p]*Ghat) * IFFT(C[p]*Ghat) * G to Q3q
			contrib_Q3q = ((AGhat_real[k][0]*CGhat_real[k][0] - AGhat_real[k][1]*CGhat_real[k][1])*G_real[k][0]);

			// Add W[p] * IFFT(B[p]*Ghat) * IFFT(C[p]*Ghat) * G to Q4q
			contrib_Q4q = ((BGhat_real[k][0]*CGhat_real[k][0] - BGhat_real[k][1]*CGhat_real[k][1])*G_real[k][0]);

			QGloc[k] += (W[p] * (contrib_Q1c - contrib_Q2c + phi * (contrib_Q1q + contrib_Q2q - contrib_Q3q - contrib_Q4q)));
		}
	}

#ifdef key_mpi
	tend = MPI_Wtime();
	elap = tend-tstart;
	MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab, 1, MPI_DOUBLE, MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
	tend = omp_get_wtime();
	elap = tend-tstart;
	elaptab[0] = elap;
#else
	tend = (double)clock();
	elap = (tend-tstart)/CLOCKS_PER_SEC;
	elaptab[0] = elap;
#endif


	gethostname(hostname, MAXSTRINGLENGTH);
	hostlist = malloc(sizeof(char*)*commsize);
	for (int R = 0; R < commsize; R++){
		hostlist[R] = allocation_1d_char(MAXSTRINGLENGTH);
		if (R == rank)
			sprintf(hostlist[R], "%s", hostname);
#ifdef key_mpi
		MPI_Bcast(hostlist[R], MAXSTRINGLENGTH, MPI_CHAR, R, MPI_COMM_WORLD);
#endif
	}
	
	if (rank == 0){
		strcpy(namefile, "");
		strcat(namefile, parameters.outputdir);
		strcat(namefile, "/");
		strcat(namefile, parameters.prefix);
		strcat(namefile, "_timecost_collision_");
#ifdef key_openmp
		sprintf(suffix, "%d", nthreads);
#else
		sprintf(suffix, "%d", commsize);
#endif
		strcat(namefile, suffix);
		strcat(namefile, ".txt");
		f = fopen(namefile, "w");
		fprintf(f, "# Kind of particles\n");
		if (parameters.type_collision == CLASSICAL_COLLISIONS)
			fprintf(f, "classical\n");
		else{
			if (parameters.type_qparticles == FERMIONS)
				fprintf(f, "fermions\n");
			else
				fprintf(f, "bosons\n");
		}
		fprintf(f, "# Number of grid points\n");
		fprintf(f, "%d %d", MVX, MVY);
		if (parameters.ndim == DIM3)
			fprintf(f, " %d", MVZ);
		fprintf(f, "\n");
		if (parameters.ndim == DIM2){
			fprintf(f, "# Number of integration points (Legendre-Legendre-Legendre)\n");
			fprintf(f, "%d %d %d\n", parameters.P1_int, parameters.P2_int, parameters.P3_int);
		}
		else{
			fprintf(f, "# Number of integration points (Legendre-Lebedev-Lebedev)\n");
			fprintf(f, "%d %d %d\n", parameters.P1_int, order_table(parameters.P2_int), order_table(parameters.P3_int));
		}
#ifdef key_openmp
		fprintf(f, "# Number of threads\n");
		fprintf(f, "%d\n", nthreads);
#else
		fprintf(f, "# Number of processes\n");
		fprintf(f, "%d\n", commsize);
#endif
		fprintf(f, "# Rank |  Time cost Qtotal   |     Hostname\n");
		for (int R = 0; R < commsize; R++)
			fprintf(f, "%7d|%.15e|%s\n", R, elaptab[R], hostlist[R]);
		fclose(f);
	}

	fftw_free(G_real);
	fftw_free(G_spectral);
	fftw_free(AGhat_spectral);
	fftw_free(AGhat_real);
	fftw_free(BGhat_spectral);
	fftw_free(BGhat_real);
	fftw_free(CGhat_spectral);
	fftw_free(CGhat_real);

	fftw_destroy_plan(plan_G_forward);
	fftw_destroy_plan(plan_AGhat_backward);
	fftw_destroy_plan(plan_BGhat_backward);
	fftw_destroy_plan(plan_CGhat_backward);

	free_1d(elaptab);
	for (int R = 0; R < commsize; R++)
		free_1d_char(hostlist[R]);
	free(hostlist);
}


/*
void collision_Qtotal_force_convolution_fftw(double *Gloc, double *QGloc, double ***A, double ***B, double ***C, double *W, double omega, PARAMETERS parameters){	
	double pi = 4.*atan(1.);
	
	FILE *f;
	char namefile[MAXSTRINGLENGTH];
	char suffix[MAXSTRINGLENGTH];
	char hostname[MAXSTRINGLENGTH];
	char **hostlist;
	
	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	fftw_complex *QG_spectral;
	fftw_complex *QG_real;
	fftw_complex *G_spectral;
	fftw_complex *G_real;
	fftw_plan plan_G_forward, plan_QG_backward;

	int i;
	
	int rank, commsize, nthreads;
	double tend, tstart, elap;
	double *elaptab_Q1c, *elaptab_Q2c, *elaptab_Q1q, *elaptab_Q2q, *elaptab_Q3q, *elaptab_Q4q;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
#else
	rank = 0;
	commsize = 1;
#endif

#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

	double *Qtest, test;
	Qtest = allocation_1d(NVXLOC*NVYLOC*NVZLOC);

	double myalpha, mu, myomega, phi;
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		mu = pow(parameters.L/pi, parameters.ndim+1);
		myomega = 1.;
	}
	else{
		mu = 1.;
		myomega = omega;
	}
	
	if (parameters.type_collision == QUANTUM_COLLISIONS){
		if (parameters.type_qparticles == BOSONS)
			myalpha = pow(parameters.hbar,parameters.ndim+1);
		else if (parameters.type_qparticles == FERMIONS)
			myalpha = -pow(parameters.hbar,parameters.ndim+1);
	}
	else if (parameters.type_collision == CLASSICAL_COLLISIONS)
		myalpha = 0.;
	phi = myalpha * mu * pow(pi*myomega/parameters.L, parameters.ndim+1);


	elaptab_Q1c = allocation_1d(commsize);
	elaptab_Q2c = allocation_1d(commsize);
	elaptab_Q1q = allocation_1d(commsize);
	elaptab_Q2q = allocation_1d(commsize);
	elaptab_Q3q = allocation_1d(commsize);
	elaptab_Q4q = allocation_1d(commsize);

#ifdef key_fftw_mpi
	QG_spectral = fftw_alloc_complex(parameters.alloc_local);
	QG_real = fftw_alloc_complex(parameters.alloc_local);
	G_spectral = fftw_alloc_complex(parameters.alloc_local);
	G_real = fftw_alloc_complex(parameters.alloc_local);

	switch (parameters.ndim){
		case DIM2:
			plan_G_forward = fftw_mpi_plan_dft_2d(MVX, MVY, G_real, G_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_QG_backward = fftw_mpi_plan_dft_2d(MVX, MVY, QG_spectral, QG_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		case DIM3:
			plan_G_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, G_real, G_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_QG_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, QG_spectral, QG_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		default:
			error("Unknown dimensionality");
	}
#else
	QG_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	QG_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	G_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	G_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	switch (parameters.ndim){
		case DIM2:
			plan_G_forward = fftw_plan_dft_2d(MVX, MVY, G_real, G_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_QG_backward = fftw_plan_dft_2d(MVX, MVY, QG_spectral, QG_real, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		case DIM3:
			plan_G_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, G_real, G_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
			plan_QG_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, QG_spectral, QG_real, FFTW_BACKWARD, FFTW_ESTIMATE);
			break;
		default:
			error("Unknown dimensionality");
	}	
#endif
*/
/*
#ifdef key_openmp
	sprintf(namefile, "alpha_%d.txt", nthreads);
	f = fopen(namefile, "w");
	for (int kvx = 0; kvx < NVXLOC; kvx++){
		for (int kvy = 0; kvy < NVYLOC; kvy++){
			for (int kvz = 0; kvz < NVZLOC; kvz++){
				for (int p = 0; p < parameters.mint; p++)
					fprintf(f, "(%d,%d,%d) | %d | %e\n", kvx, kvy, kvz, p, alpha[p][kvz+NVZLOC*(kvy+NVYLOC*kvx)]);
			}
		}
	}
	fclose(f);
#endif


#ifdef key_openmp
	sprintf(namefile, "alphaprime_%d.txt", nthreads);
	f = fopen(namefile, "w");
	for (int kvx = 0; kvx < NVXLOC; kvx++){
		for (int kvy = 0; kvy < NVYLOC; kvy++){
			for (int kvz = 0; kvz < NVZLOC; kvz++){
				for (int p = 0; p < parameters.mint; p++)
					fprintf(f, "(%d,%d,%d) | %d | %e\n", kvx, kvy, kvz, p, alphaprime[p][kvz+NVZLOC*(kvy+NVYLOC*kvx)]);
			}
		}
	}
	fclose(f);
#endif
*/
/*
	#pragma omp parallel for private(i) shared(NVLOC, G_real, Gloc, QGloc)
	for (i = 0; i < NVLOC; i++){
		G_real[i][0] = Gloc[i];
		G_real[i][1] = 0.;
		QGloc[i] = 0.;
		G_spectral[i][0] = 0.;
		G_spectral[i][1] = 0.;
	}

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif
*/
/*
#ifdef key_openmp
	sprintf(namefile, "%s/%s_G_real.txt", parameters.outputdir, parameters.prefix);
	f = fopen(namefile, "w");
	for (int kvx = 0; kvx < NVXLOC; kvx++){
		for (int kvy = 0; kvy < NVYLOC; kvy++){
			for (int kvz = 0; kvz < NVZLOC; kvz++)
				fprintf(f, "(%d,%d,%d) | (%e,%e)\n", kvx, kvy, kvz, G_real[kvz+NVZLOC*(kvy+NVYLOC*kvx)][0], G_real[kvz+NVZLOC*(kvy+NVYLOC*kvx)][1]);
		}
	}
	fclose(f);
#endif
*/
/*
	fftw_execute(plan_G_forward);
*/
/*
#ifdef key_openmp
	sprintf(namefile, "%s/%s_G_spectral.txt", parameters.outputdir, parameters.prefix);
	f = fopen(namefile, "w");
	for (int kvx = 0; kvx < NVXLOC; kvx++){
		for (int kvy = 0; kvy < NVYLOC; kvy++){
			for (int kvz = 0; kvz < NVZLOC; kvz++)
				fprintf(f, "(%d,%d,%d) | (%e,%e)\n", kvx, kvy, kvz, G_spectral[kvz+NVZLOC*(kvy+NVYLOC*kvx)][0], G_spectral[kvz+NVZLOC*(kvy+NVYLOC*kvx)][1]);
		}
	}
	fclose(f);
#endif
*/
/*
#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
	tstart = MPI_Wtime();
#elif defined(key_openmp)
	tstart = omp_get_wtime();
#else
	tstart = (double)clock();
#endif

	collision_Q1c_convolution_fftw(G_spectral, QG_spectral, A, B, W, parameters);

#ifdef key_mpi
	tend = MPI_Wtime();
	elap = tend-tstart;
	MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q1c, 1, MPI_DOUBLE, MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
	tend = omp_get_wtime();
	elap = tend-tstart;
	elaptab_Q1c[0] = elap;
#else
	tend = (double)clock();
	elap = (tend-tstart)/CLOCKS_PER_SEC;
	elaptab_Q1c[0] = elap;
#endif

	fftw_execute(plan_QG_backward);

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif

	test = 0.;
	strcpy(namefile, "");
	strcat(namefile, parameters.outputdir);
	strcat(namefile, "/");
	strcat(namefile, parameters.prefix);
	strcat(namefile, "_Q1c_real.txt");
	f = fopen(namefile, "w");
	#pragma omp parallel for private(i) shared(NVLOC, MV, QG_real, QGloc, Qtest, test)
	for (i = 0; i < NVLOC; i++){
		QG_real[i][0] /= MV;
		QG_real[i][1] /= MV;
		fprintf(f, "Q1c_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
		QGloc[i] += QG_real[i][0];
		Qtest[i] = QG_real[i][0];
		//#pragma omp critical
		//test = fmax(fabs(Qtest[i]), test);
	}
	fclose(f);
	//printf("Linf(Q1c) = %e\n", test);
	message("Q1c computed\n");
*/
	/*f = fopen("Q1c_serial.txt", "w");
	for (int i = 0; i < MV; i++)
		fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
	fclose(f);*/
/*
#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
	tstart = MPI_Wtime();
#elif defined(key_openmp)
	tstart = omp_get_wtime();
#else
	tstart = (double)clock();
#endif

	collision_Q2c_convolution_fftw(G_spectral, G_real, QG_spectral, C, W, parameters);

#ifdef key_mpi
	tend = MPI_Wtime();
	elap = tend-tstart;
	MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q2c, 1, MPI_DOUBLE, MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
	tend = omp_get_wtime();
	elap = tend-tstart;
	elaptab_Q2c[0] = elap;
#else
	tend = (double)clock();
	elap = (tend-tstart)/CLOCKS_PER_SEC;
	elaptab_Q2c[0] = elap;
#endif
	
	fftw_execute(plan_QG_backward);

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif

	test = 0.;
	strcpy(namefile, "");
	strcat(namefile, parameters.outputdir);
	strcat(namefile, "/");
	strcat(namefile, parameters.prefix);
	strcat(namefile, "_Q2c_real.txt");
	f = fopen(namefile, "w");
	#pragma omp parallel for private(i) shared(NVLOC, MV, QG_real, QGloc, Qtest, test)
	for (i = 0; i < NVLOC; i++){
		QG_real[i][0] /= MV;
		QG_real[i][1] /= MV;
		fprintf(f, "Q2c_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
		QGloc[i] -= QG_real[i][0];
		Qtest[i] = QG_real[i][0];
		//#pragma omp critical
		//test = fmax(fabs(Qtest[i]), test);
	}
	fclose(f);
	//printf("Linf(Q2c) = %e\n", test);
	message("Q2c computed\n");
*/
	/*f = fopen("Q2c_serial.txt", "w");
	for (int i = 0; i < MV; i++)
		fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
	fclose(f);*/
/*
	if (parameters.type_collision == QUANTUM_COLLISIONS){
#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
		tstart = MPI_Wtime();
#elif defined(key_openmp)
		tstart = omp_get_wtime();
#else
		tstart = (double)clock();
#endif

		collision_Q1q_convolution_fftw(G_spectral, QG_spectral, A, B, C, W, parameters);

#ifdef key_mpi
		tend = MPI_Wtime();
		elap = tend-tstart;
		MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q1q, 1, MPI_DOUBLE, MPI_COMM_WORLD);
		MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
		tend = omp_get_wtime();
		elap = tend-tstart;
		elaptab_Q1q[0] = elap;
#else
		tend = (double)clock();
		elap = (tend-tstart)/CLOCKS_PER_SEC;
		elaptab_Q1q[0] = elap;
#endif

		fftw_execute(plan_QG_backward);

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif
		
		test = 0.;
		strcpy(namefile, "");
		strcat(namefile, parameters.outputdir);
		strcat(namefile, "/");
		strcat(namefile, parameters.prefix);
		strcat(namefile, "_Q1q_real.txt");
		f = fopen(namefile, "w");
		#pragma omp parallel for private(i) shared(NVLOC, QG_real, MV, QGloc, phi, Qtest, test)
		for (i = 0; i < NVLOC; i++){
			QG_real[i][0] /= MV;
			QG_real[i][1] /= MV;
			fprintf(f, "Q1q_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
			QGloc[i] += phi*QG_real[i][0];
			Qtest[i] = phi*QG_real[i][0];
			//#pragma omp critical
			//test = fmax(fabs(Qtest[i]), test);
		}
		fclose(f);
		//printf("Linf(Q1q) = %e\n", test);
		message("Q1q computed\n");
*/
		/*f = fopen("Q1q_serial.txt", "w");
		for (int i = 0; i < MV; i++)
			fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
		fclose(f);*/
/*
#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
		tstart = MPI_Wtime();
#elif defined(key_openmp)
		tstart = omp_get_wtime();
#else
		tstart = (double)clock();
#endif

		collision_Q2q_convolution_fftw(G_spectral, G_real, QG_spectral, A, B, W, parameters);

#ifdef key_mpi
		tend = MPI_Wtime();
		elap = tend-tstart;
		MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q2q, 1, MPI_DOUBLE, MPI_COMM_WORLD);
		MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
		tend = omp_get_wtime();
		elap = tend-tstart;
		elaptab_Q2q[0] = elap;
#else
		tend = (double)clock();
		elap = (tend-tstart)/CLOCKS_PER_SEC;
		elaptab_Q2q[0] = elap;
#endif

		fftw_execute(plan_QG_backward);

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif
		
		test = 0.;
		strcpy(namefile, "");
		strcat(namefile, parameters.outputdir);
		strcat(namefile, "/");
		strcat(namefile, parameters.prefix);
		strcat(namefile, "_Q2q_real.txt");
		f = fopen(namefile, "w");
		#pragma omp parallel for private(i) shared(NVLOC, QG_real, MV, QGloc, phi, Qtest, test)
		for (int i = 0; i < NVLOC; i++){
			QG_real[i][0] /= MV;
			QG_real[i][1] /= MV;
			fprintf(f, "Q2q_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
			QGloc[i] += phi*QG_real[i][0];
			Qtest[i] = phi*QG_real[i][0];
			//#pragma omp critical
			//test = fmax(fabs(Qtest[i]), test);
		}
		fclose(f);
		//printf("Linf(Q2q) = %e\n", test);
		message("Q2q computed\n");
*/
		/*f = fopen("Q2q_serial.txt", "w");
		for (int i = 0; i < MV; i++)
			fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
		fclose(f);*/
/*
#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
		tstart = MPI_Wtime();
#elif defined(key_openmp)
		tstart = omp_get_wtime();
#else
		tstart = (double)clock();
#endif
		collision_Q3q_convolution_fftw(G_spectral, G_real, QG_spectral, A, C, W, parameters);

#ifdef key_mpi
		tend = MPI_Wtime();
		elap = tend-tstart;
		MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q3q, 1, MPI_DOUBLE, MPI_COMM_WORLD);
		MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
		tend = omp_get_wtime();
		elap = tend-tstart;
		elaptab_Q3q[0] = elap;
#else
		tend = (double)clock();
		elap = (tend-tstart)/CLOCKS_PER_SEC;
		elaptab_Q3q[0] = elap;
#endif
		
		fftw_execute(plan_QG_backward);

#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif
		
		test = 0.;
		strcpy(namefile, "");
		strcat(namefile, parameters.outputdir);
		strcat(namefile, "/");
		strcat(namefile, parameters.prefix);
		strcat(namefile, "_Q3q_real.txt");
		f = fopen(namefile, "w");
		#pragma omp parallel for private(i) shared(NVLOC, QG_real, MV, QGloc, phi, Qtest, test)
		for (int i = 0; i < NVLOC; i++){
			QG_real[i][0] /= MV;
			QG_real[i][1] /= MV;
			fprintf(f, "Q3q_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
			QGloc[i] -= phi*QG_real[i][0];
			Qtest[i] = phi*QG_real[i][0];
			//#pragma omp critical
			//test = fmax(fabs(Qtest[i]), test);
		}
		fclose(f);
		//printf("Linf(Q3q) = %e\n", test);
		message("Q3q computed\n");
*/
		/*f = fopen("Q3q_serial.txt", "w");
		for (int i = 0; i < MV; i++)
			fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
		fclose(f);*/
/*
#ifdef key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
		tstart = MPI_Wtime();
#elif defined(key_openmp)
		tstart = omp_get_wtime();
#else
		tstart = (double)clock();
#endif

		collision_Q4q_convolution_fftw(G_spectral, G_real, QG_spectral, B, C, W, parameters);

#ifdef key_mpi
		tend = MPI_Wtime();
		elap = tend-tstart;
		MPI_Allgather(&elap, 1, MPI_DOUBLE, elaptab_Q4q, 1, MPI_DOUBLE, MPI_COMM_WORLD);
		MPI_Barrier(MPI_COMM_WORLD);
#elif defined (key_openmp)
		tend = omp_get_wtime();
		elap = tend-tstart;
		elaptab_Q4q[0] = elap;
#else
		tend = (double)clock();
		elap = (tend-tstart)/CLOCKS_PER_SEC;
		elaptab_Q4q[0] = elap;
#endif
		
		fftw_execute(plan_QG_backward);

#ifdef _key_mpi
		MPI_Barrier(MPI_COMM_WORLD);
#endif
		
		test = 0.;
		strcpy(namefile, "");
		strcat(namefile, parameters.outputdir);
		strcat(namefile, "/");
		strcat(namefile, parameters.prefix);
		strcat(namefile, "_Q4q_real.txt");
		f = fopen(namefile, "w");
		#pragma omp parallel for private(i) shared(NVLOC, QG_real, MV, QGloc, phi, Qtest, test)
		for (int i = 0; i < NVLOC; i++){
			QG_real[i][0] /= MV;
			QG_real[i][1] /= MV;
			fprintf(f, "Q4q_real[%d] = (%.15e,%.15e)\n", i, QG_real[i][0], QG_real[i][1]);
			QGloc[i] -= phi*QG_real[i][0];
			Qtest[i] = phi*QG_real[i][0];
			//#pragma omp critical
			//test = fmax(fabs(Qtest[i]), test);
		}
		fclose(f);
		//printf("Linf(Q4q) = %e\n", test);
		message("Q4q computed\n");
*/
		/*f = fopen("Q4q_serial.txt", "w");
		for (int i = 0; i < MV; i++)
			fprintf(f, "%5d, %.15e %.15e\n", i, QG_real[i][0], QG_real[i][1]);
		fclose(f);*/
/*	}

	gethostname(hostname, MAXSTRINGLENGTH);
	hostlist = malloc(sizeof(char*)*commsize);
	for (int R = 0; R < commsize; R++){
		hostlist[R] = allocation_1d_char(MAXSTRINGLENGTH);
		if (R == rank)
			sprintf(hostlist[R], "%s", hostname);
#ifdef key_mpi
		MPI_Bcast(hostlist[R], MAXSTRINGLENGTH, MPI_CHAR, R, MPI_COMM_WORLD);
#endif
	}
	
	if (rank == 0){
		strcpy(namefile, "");
		strcat(namefile, parameters.outputdir);
		strcat(namefile, "/");
		strcat(namefile, parameters.prefix);
		strcat(namefile, "_timecost_collision_");
#ifdef key_openmp
		sprintf(suffix, "%d", nthreads);
#else
		sprintf(suffix, "%d", commsize);
#endif
		strcat(namefile, suffix);
		strcat(namefile, ".txt");
		f = fopen(namefile, "w");
		fprintf(f, "# Kind of particles\n");
		if (parameters.type_collision == CLASSICAL_COLLISIONS)
			fprintf(f, "classical\n");
		else{
			if (parameters.type_qparticles == FERMIONS)
				fprintf(f, "fermions\n");
			else
				fprintf(f, "bosons\n");
		}
		fprintf(f, "# Number of grid points\n");
		fprintf(f, "%d %d", MVX, MVY);
		if (parameters.ndim == DIM3)
			fprintf(f, " %d", MVZ);
		fprintf(f, "\n");
#ifdef key_openmp
		fprintf(f, "# Number of threads\n");
		fprintf(f, "%d\n", nthreads);
#else
		fprintf(f, "# Number of processes\n");
		fprintf(f, "%d\n", commsize);
#endif
		fprintf(f, "# Rank |    Time cost Q1c    |    Time cost Q2c    ");
		if (parameters.type_collision == QUANTUM_COLLISIONS)
			fprintf(f, "|    Time cost Q1q    |    Time cost Q2q    |    Time cost Q3q    |    Time cost Q4q    ");
		fprintf(f, "| Hostname\n");
		for (int R = 0; R < commsize; R++){
			fprintf(f, "%7d|%.15e|%.15e", R, elaptab_Q1c[R], elaptab_Q2c[R]);
			if (parameters.type_collision == QUANTUM_COLLISIONS)
				fprintf(f, "|%.15e|%.15e|%.15e|%.15e", elaptab_Q1q[R], elaptab_Q2q[R], elaptab_Q3q[R], elaptab_Q4q[R]);
			fprintf(f, "|%s", hostlist[R]);
			fprintf(f, "\n");
		}
		fclose(f);
	}

#ifdef key_mpi	
	MPI_Barrier(MPI_COMM_WORLD);
#endif

	fftw_destroy_plan(plan_G_forward);
	fftw_destroy_plan(plan_QG_backward);
	
	fftw_free(QG_spectral);
	fftw_free(QG_real);
	fftw_free(G_spectral);
	fftw_free(G_real);

	free_1d(Qtest);
	free_1d(elaptab_Q1c);
	free_1d(elaptab_Q2c);
	free_1d(elaptab_Q1q);
	free_1d(elaptab_Q2q);
	free_1d(elaptab_Q3q);
	free_1d(elaptab_Q4q);
	
	for (int R = 0; R < commsize; R++)
		free_1d_char(hostlist[R]);
	free(hostlist);
}
*/






void free_crossmode(PARAMETERS parameters){
	extern int crossmode_done;
	extern double ***crossmode_cos;
	extern double ***crossmode_sin;

	int commsize;
#ifdef key_mpi
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
#else
	commsize = 1;
#endif

#ifdef key_force_vectorize
	if ((crossmode_done) && (!parameters.force_convolution)){
		for (int R = 0; R < commsize; R++){
			
			for (int jdist = 0; jdist < parameters.nvxtab[R]*parameters.nvytab[R]*parameters.nvztab[R]; jdist++){
				free_1d(crossmode_cos[R][jdist]);
				free_1d(crossmode_sin[R][jdist]);
			}
			free(crossmode_cos[R]);
			free(crossmode_sin[R]);
		}
		free(crossmode_cos);
		free(crossmode_sin);
	}
	crossmode_done = (1==0);
#endif
}

#endif
