#include "./src/core/diagnostics.h"


void compute_diagnostic(DIAGS *my_diags, int thediag, PARAMETERS parameters){
	char buf[MAXSTRINGLENGTH];

	double dxix, dxiy, dxiz, dxi;
	
	double mu, lambda, myomega, alpha;
	double dvx, dvy, dvz, dv;
	

	double G0, Gex0, Glimit0;
	double f0, fex0, flimit0;
	double diagloc;
	double diag = 0.;
	
	int needlimit, needexact;

	int k, d;

	double pi = 4.*atan(1.);

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz,1);
	int MV = MVX*MVY*MVZ;

	switch (thediag){
		case RELATIVE_ENTROPY:
		case L1_NORM_LIMIT:
		case L2_NORM_LIMIT:
		case LINF_NORM_LIMIT:
		case L1_ERROR_RELAXATION:
		case L2_ERROR_RELAXATION:
		case LINF_ERROR_RELAXATION:
			needlimit = (1==1);
			break;
		default:
			needlimit = (0==1);
			break;
	}

	switch (thediag){
		case L1_ERROR:
		case L2_ERROR:
		case LINF_ERROR:
		case L1_NORM_EX:
		case L2_NORM_EX:
		case LINF_NORM_EX:
			needexact = (1==1);
			break;
		default:
			needexact = (1==0);
			break;
	}
	
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		lambda = 0.;
		myomega = 1.;
		mu = pow(parameters.L/pi, parameters.ndim+1);
	}
	else{
		lambda = 1.;
		myomega = my_diags->omega;
		mu = 1.;
	}
	
	if (parameters.type_collision == QUANTUM_COLLISIONS){
		switch(parameters.type_qparticles){
			case BOSONS :
				alpha = -pow(parameters.hbar,parameters.ndim+1);
				break;
			case FERMIONS :
				alpha = pow(parameters.hbar,parameters.ndim+1);
				break;
			default :
				error("Unknown kind of quantum particles.");
		}
	}

	dxix = 2.*pi/MVX;
	dvx = 2.*(parameters.L/myomega)/MVX;
	dxiy = 2.*pi/MVY;
	dvy = 2.*(parameters.L/myomega)/MVY;
	if (parameters.nvz == 0){
		dxiz = 1.;
		dvz = 1.;
	}
	else{
		dxiz = 2.*pi/MVZ;
		dvz = 2.*(parameters.L/myomega)/MVZ;
	}
	dxi = dxix*dxiy*dxiz;
	dv = dvx*dvy*dvz;

	// Initial value of the diagnostic
	if (parameters.ndim == DIM2){
		G0 = my_diags->Gloc_2d[0][0];
		if (needlimit && (!(parameters.condensation)))
			Glimit0 = my_diags->Gloclimit_2d[0][0];
		if (needexact && (parameters.exact_solution))
			Gex0 = my_diags->Glocexact_2d[0][0];	
	}
	else if (parameters.ndim == DIM3){
		G0 = my_diags->Gloc_3d[0][0][0];
		if (needlimit && (!(parameters.condensation)))
			Glimit0 = my_diags->Gloclimit_3d[0][0][0];
		if (needexact && (parameters.exact_solution))
			Gex0 = my_diags->Glocexact_3d[0][0][0];
	}
	f0 = mu * pow(pi*myomega/parameters.L, parameters.ndim+1) * G0;
	if (!(parameters.condensation))
		flimit0 = mu * pow(pi*myomega/parameters.L, parameters.ndim+1) * Glimit0;
	if (parameters.exact_solution)
		fex0 = mu * pow(pi*myomega/parameters.L, parameters.ndim+1) * Gex0;
	else
		fex0 = 0.;
	switch (thediag){
		case MASS:
			diagloc = 0.;
			break;
		case AVERAGE_UX:
			diagloc = 0.;
			break;
		case AVERAGE_UY:
			diagloc = 0.;
			break;
		case AVERAGE_UZ:
			diagloc = 0.;
			break;
		case AVERAGE_U:
			break;
		case KINETIC_ENERGY:
			diagloc = 0.;
			break;
		case INTERNAL_ENERGY:
			diagloc = 0.;
			break;
		case TEMPERATURE:
			break;
		case STRESS_ENERGY_XX:
			diagloc = 0.;
			break;
		case STRESS_ENERGY_XY:
			diagloc = 0.;
			break;
		case STRESS_ENERGY_XZ:
			diagloc = 0.;
			break;
		case STRESS_ENERGY_YX:
			diagloc = 0.;
			break;
		case STRESS_ENERGY_YY:
			diagloc = 0.;
			break;
		case STRESS_ENERGY_YZ:
			diagloc = 0.;
			break;
		case STRESS_ENERGY_ZX:
			diagloc = 0.;
			break;
		case STRESS_ENERGY_ZY:
			diagloc = 0.;
			break;
		case STRESS_ENERGY_ZZ:
			diagloc = 0.;
			break;
		case STRESS_ENERGY:
			break;
		case MOMENTUM4:
			diagloc = 0.;
			break;
		case MOMENTUM6:
			diagloc = 0.;
			break;
		case ENTROPY:
			diagloc = 0.;
			break;
		case RELATIVE_ENTROPY:
			if (needlimit)
				diagloc = 0.;
			break;
		case L1_NORM:
			diagloc = 0.;
			break;
		case L2_NORM:
			diagloc = 0.;
			break;
		case LINF_NORM:
			diagloc = fabs(f0);
			break;
		case MAX_VALUE:
			diagloc = f0;
			break;
		case MIN_VALUE:
			diagloc = f0;
			break;
		case L1_NORM_EX:
			if (needexact)
				diagloc = 0.;
			else
				error("Cannot compute L1 norm of exact solution");
			break;
		case L2_NORM_EX:
			if (needexact)
				diagloc = 0.;
			else
				error("Cannot compute L2 norm of exact solution");
			break;
		case LINF_NORM_EX:
			if (needexact)
				diagloc = fabs(fex0);
			else
				error("Cannot compute Linf norm of exact solution");
			break;
		case L1_ERROR:
			if (needexact)
				diagloc = 0.;
			else
				error("Cannot compute L1 error");
			break;
		case L2_ERROR:
			if (needexact)
				diagloc = 0.;
			else
				error("Cannot compute L2 error");
			break;
		case LINF_ERROR:
			if (needexact)
				diagloc = fabs(f0-fex0);
			else
				error("Cannot compute Linf error");
			break;
		case L1_NORM_LIMIT:
			if (needlimit)
				diagloc = 0.;
			else
				error("Cannot compute L1 norm of limit distribution");
			break;
		case L2_NORM_LIMIT:
			if (needlimit)
				diagloc = 0.;
			else
				error("Cannot compute L2 norm of limit distribution");
			break;
		case LINF_NORM_LIMIT:
			if (needlimit)
				diagloc = fabs(flimit0);
			else
				error("Cannot compute Linf norm of limit distribution");
			break;
		case L1_ERROR_RELAXATION:
			if (needlimit)
				diagloc = 0.;
			else
				error("Cannot compute L1 relaxation error");
			break;
		case L2_ERROR_RELAXATION:
			if (needlimit)
				diagloc = 0.;
			else
				error("Cannot compute L2 relaxation error");
			break;
		case LINF_ERROR_RELAXATION:
			if (needlimit)
				diagloc = fabs(f0-flimit0);
			else
				error("Cannot compute Linf relaxation error");
			break;
		default:
			error("Unknown diagnostic.");
	}


	// Diagnostics based on discrete integrals:
	// mass, average velocity, kinetic energy, internal energy, stress energy tensor, 4th order momentum,
	// 6th order momentum, entropy, relative entropy,
	// L1 and L2 norms, errors, relaxation errors
	#pragma omp parallel for default(none) private(k, d) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, parameters, MVX, MVY, MVZ, IVX0LOC, IVY0LOC, IVZ0LOC, dxix, dxiy, dxiz, dv, pi, myomega, mu, lambda, my_diags, needexact, needlimit, thediag, diagloc, alpha)
	for (k = 0; k < NVLOC; k++){
		double G, Gex, Glimit;
		double f, fex, flimit;

		double xi[3];
		double v[3];

		double vsquare, contrib, dtemp;

		int kvxabs, kvyabs, kvzabs;
		int itemp, itest;
#ifdef key_force_vectorize
		kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
		kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
		kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
		// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
		itemp = k/NVZLOC;
		// itemp now contains kvyloc+NVYLOC*kvxloc;
		kvzabs = k - itemp*NVZLOC;
		kvxabs = itemp/NVYLOC;
		kvyabs = itemp-kvxabs*NVYLOC;
							
		kvxabs += IVX0LOC;
		kvyabs += IVY0LOC;
		kvzabs += IVZ0LOC;
							
		if (kvxabs >= MVX/2)
			kvxabs -= MVX;
		if (kvyabs >= MVY/2)
			kvyabs -= MVY;
		if (kvzabs >= MVZ/2)
			kvzabs -= MVZ;
#endif
		xi[0] = kvxabs * dxix;
		xi[1] = kvyabs * dxiy;
		v[0] = (parameters.L/(pi*myomega)*xi[0]+lambda*my_diags->u_av[0]);
		v[1] = (parameters.L/(pi*myomega)*xi[1]+lambda*my_diags->u_av[1]);
		if (parameters.ndim == DIM3){
			xi[2] = kvzabs * dxiz;
			v[2] = (parameters.L/(pi*myomega)*xi[2]+lambda*my_diags->u_av[2]);
		}
		if (parameters.ndim == DIM2){
			G = my_diags->Gloc_2d[0][k];
			if (needlimit && (!(parameters.condensation)))
				Glimit = my_diags->Gloclimit_2d[0][k];
			if (needexact && (parameters.exact_solution))
				Gex = my_diags->Glocexact_2d[0][k];	
		}
		else if (parameters.ndim == DIM3){
			G = my_diags->Gloc_3d[0][0][k];
			if (needlimit && (!(parameters.condensation)))
				Glimit = my_diags->Gloclimit_3d[0][0][k];
			if (needexact && (parameters.exact_solution))
				Gex = my_diags->Glocexact_3d[0][0][k];
		}
		f = mu * pow(pi*myomega/parameters.L, parameters.ndim+1) * G;
		if (!(parameters.condensation))
			flimit = mu * pow(pi*myomega/parameters.L, parameters.ndim+1) * Glimit;
		if (parameters.exact_solution)
			fex = mu * pow(pi*myomega/parameters.L, parameters.ndim+1) * Gex;
		else
			fex = 0.;
		
		// Default local contribution
		contrib = 0.;

		#pragma omp critical
		{
		dtemp = diagloc;

		// Compute local contribution
		if (thediag == MASS)
			contrib = f;
		else if (thediag == AVERAGE_UX)
			contrib = v[0] * f;
		else if (thediag == AVERAGE_UY)
			contrib = v[1] * f;
		else if (thediag == AVERAGE_UZ)
			contrib = v[2] * f;
		else if (thediag == AVERAGE_U)
			contrib = 0.;
			// Nothing to do now
		else if (thediag == KINETIC_ENERGY){
			vsquare = 0.;
			for (d = 0; d <= parameters.ndim; d++)
				vsquare += pow(v[d],2);
			contrib = vsquare * f;
		}
		else if (thediag == INTERNAL_ENERGY){
			vsquare = 0.;
			for (d = 0; d <= parameters.ndim; d++)
				vsquare += pow(v[d]-my_diags->u_av[d],2);
			contrib = 0.5 * vsquare * f;
		}
		else if (thediag == TEMPERATURE)
			contrib = 0.;
			// Nothing to do now
		else if (thediag == STRESS_ENERGY_XX){
			vsquare = (v[0]-my_diags->u_av[0]) * (v[0]-my_diags->u_av[0]);
			contrib = vsquare * f;
		}
		else if (thediag == STRESS_ENERGY_XY){
			vsquare = (v[0]-my_diags->u_av[0]) * (v[1]-my_diags->u_av[1]);
			contrib = vsquare * f;
		}
		else if (thediag == STRESS_ENERGY_XZ){
			vsquare = (v[0]-my_diags->u_av[0]) * (v[2]-my_diags->u_av[2]);
			contrib = vsquare * f;
		}
		else if (thediag == STRESS_ENERGY_YX){
			vsquare = (v[1]-my_diags->u_av[1]) * (v[0]-my_diags->u_av[0]);
			contrib = vsquare * f;
		}
		else if (thediag == STRESS_ENERGY_YY){
			vsquare = (v[1]-my_diags->u_av[1]) * (v[1]-my_diags->u_av[1]);
			contrib = vsquare * f;
		}
		else if (thediag == STRESS_ENERGY_YZ){
			vsquare = (v[1]-my_diags->u_av[1]) * (v[2]-my_diags->u_av[2]);
			contrib = vsquare * f;
		}
		else if (thediag == STRESS_ENERGY_ZX){
			vsquare = (v[2]-my_diags->u_av[2]) * (v[0]-my_diags->u_av[0]);
			contrib = vsquare * f;
		}
		else if (thediag == STRESS_ENERGY_ZY){
			vsquare = (v[2]-my_diags->u_av[2]) * (v[1]-my_diags->u_av[1]);
			contrib = vsquare * f;
		}
		else if (thediag == STRESS_ENERGY_ZZ){
			vsquare = (v[2]-my_diags->u_av[2]) * (v[2]-my_diags->u_av[2]);
			contrib = vsquare * f;
		}
		else if (thediag == STRESS_ENERGY)
			contrib = 0.;
			// Nothing to do now
		else if (thediag == MOMENTUM4){
			vsquare = 0.;
			for (d = 0; d <= parameters.ndim; d++)
				vsquare += pow(v[d],2);
			contrib = pow(vsquare,2) * f;
		}
		else if (thediag == MOMENTUM6){
			vsquare = 0.;
			for (d = 0; d <= parameters.ndim; d++)
				vsquare += pow(v[d],2);
			contrib = pow(vsquare,3) * f;
		}
		else if (thediag == ENTROPY){
			if(parameters.type_collision == CLASSICAL_COLLISIONS)
				contrib = fmax(THRESHOLD, f) * log(fmax(THRESHOLD, f));
			else if (parameters.type_collision == QUANTUM_COLLISIONS)
				contrib = (fmax(THRESHOLD, f)*log(fmax(THRESHOLD, f)) + fmax(THRESHOLD, (1.-alpha*f))/alpha * log(fmax(THRESHOLD, (1.-alpha*f))));
			else
				error("Unknown kind of collision operator.");
		}
		else if (thediag == RELATIVE_ENTROPY){
			if (needlimit){
				if (parameters.type_collision == CLASSICAL_COLLISIONS)
					contrib = fmax(THRESHOLD, f) * log(fmax(THRESHOLD, f/flimit));
				else if (parameters.type_collision == QUANTUM_COLLISIONS)
					contrib = (fmax(THRESHOLD, f)*log(fmax(THRESHOLD, f/flimit)) + fmax(THRESHOLD, (1.-alpha*f))/alpha * log(fmax(THRESHOLD, (1.-alpha*f)/(1.-alpha*flimit))));
				else
					error("Unknown kind of collision operator.");
			}
		}
		else if (thediag == L1_NORM)
			contrib = fabs(f);
		else if (thediag == L2_NORM)
			contrib = pow(f,2);
		else if (thediag == LINF_NORM){
			contrib = fabs(f) - diagloc;
			contrib *= (contrib > 0.);
		}
		else if (thediag == MAX_VALUE){
			contrib = f - diagloc;
			contrib *= (contrib > 0.);
		}
		else if (thediag == MIN_VALUE){
			contrib = f - diagloc;
			contrib *= (contrib < 0.);
		}
		else if (thediag == L1_NORM_EX){
			if (needexact)
				contrib = fabs(fex);
			else
				error("Cannot compute L1 norm of exact solution");
		}
		else if (thediag == L2_NORM_EX){
			if (needexact)
				contrib = pow(fex,2);
			else
				error("Cannot compute L2 norm of exact solution");
		}
		else if (thediag == LINF_NORM_EX){
			if (needexact){
				contrib = fabs(fex) - diagloc;
				contrib *= (contrib > 0.);
			}
			else
				error("Cannot compute Linf norm of exact solution");
		}
		else if (thediag == L1_ERROR){
			if (needexact)
				contrib = fabs(f-fex);
			else
				error("Cannot compute L1 error");
		}
		else if (thediag == L2_ERROR){
			if (needexact)
				contrib = pow(f-fex,2);
			else
				error("Cannot compute L2 error");
		}
		else if (thediag == LINF_ERROR){
			if (needexact){
				contrib = fabs(f-fex) - diagloc;
				contrib *= (contrib > 0.);
			}
			else
				error("Cannot compute Linf error");
		}
		else if (thediag == L1_NORM_LIMIT){
			if (needlimit)
				contrib = fabs(flimit);
			else
				error("Cannot compute L1 norm of limit distribution");
		}
		else if (thediag == L2_NORM_LIMIT){
			if (needlimit)
				contrib = pow(flimit,2);
			else
				error("Cannot compute L2 norm of limit distribution");
		}
		else if (thediag == LINF_NORM_LIMIT){
			if (needlimit){
				contrib = fabs(flimit) - diagloc;
				contrib *= (contrib > 0.);
			}
			else
				error("Cannot compute Linf norm of limit distribution");
		}
		else if (thediag == L1_ERROR_RELAXATION){
			if (needlimit)
				contrib = fabs(f-flimit);
			else
				error("Cannot compute L1 relaxation error");
		}
		else if (thediag == L2_ERROR_RELAXATION){
			if (needlimit)
				contrib = pow(f-flimit,2);
			else
				error("Cannot compute L2 relaxation error");
		}
		else if (thediag == LINF_ERROR_RELAXATION){
			if (needlimit){
				contrib = fabs(f-flimit) - diagloc;
				contrib *= (contrib > 0.);
			}
			else
				error("Cannot compute Linf relaxation error");
		}
		else
				error("Unknown diagnostic.");
		

		// Add the local contribution
		diagloc += contrib;
		}
	}

	// If there is MPI parallelization, we gather the results
	switch (thediag){
		case MASS:
		case AVERAGE_UX:
		case AVERAGE_UY:
		case AVERAGE_UZ:
		case KINETIC_ENERGY:
		case ENTROPY:
		case RELATIVE_ENTROPY:
		case MOMENTUM4:
		case MOMENTUM6:
		case INTERNAL_ENERGY:
		case STRESS_ENERGY_XX:
		case STRESS_ENERGY_XY:
		case STRESS_ENERGY_XZ:
		case STRESS_ENERGY_YX:
		case STRESS_ENERGY_YY:
		case STRESS_ENERGY_YZ:
		case STRESS_ENERGY_ZX:
		case STRESS_ENERGY_ZY:
		case STRESS_ENERGY_ZZ:
		case L1_NORM:
		case L2_NORM:
		case L1_NORM_EX:
		case L2_NORM_EX:
		case L1_ERROR:
		case L2_ERROR:
		case L1_NORM_LIMIT:
		case L2_NORM_LIMIT:
		case L1_ERROR_RELAXATION:
		case L2_ERROR_RELAXATION:
#ifdef key_mpi
			MPI_Allreduce(&diagloc, &diag, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#else
			diag = diagloc;
#endif
			diag *= dv;
			break;
		case MIN_VALUE:
#ifdef key_mpi
			MPI_Allreduce(&diagloc, &diag, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
#else
			diag = diagloc;
#endif
			break;
		case LINF_NORM:
		case MAX_VALUE:
		case LINF_NORM_EX:
		case LINF_ERROR:
		case LINF_NORM_LIMIT:
		case LINF_ERROR_RELAXATION:
#ifdef key_mpi
			MPI_Allreduce(&diagloc, &diag, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
#else
			diag = diagloc;
#endif
			break;
		
		case AVERAGE_U:
		case STRESS_ENERGY:
		case TEMPERATURE:
			break;
		default:
			error("Unknown diagnostic.");
	}
	
	int verbose = (0==1);
	// We finalize the computation
	switch (thediag){
		case MASS:
			my_diags->mass = diag;
			if (verbose){
				sprintf(buf, "mass = %e\n", my_diags->mass);
				message(buf);
			}
			break;
		case AVERAGE_UX:
			my_diags->u_av[0] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "ux = %e\n", my_diags->u_av[0]);
				message(buf);
			}
			break;
		case AVERAGE_UY:
			my_diags->u_av[1] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "uy = %e\n", my_diags->u_av[1]);
				message(buf);
			}
			break;
		case AVERAGE_UZ:
			my_diags->u_av[2] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "uz = %e\n", my_diags->u_av[2]);
				message(buf);
			}
			break;
		case KINETIC_ENERGY:
			my_diags->ekin = diag;
			if (verbose){
				sprintf(buf, "ekin = %e\n", my_diags->ekin);
				message(buf);
			}
			break;
		case ENTROPY:
			my_diags->entropy = diag;
			if (verbose){
				sprintf(buf, "entropy = %e\n", my_diags->entropy);
				message(buf);
			}
			break;
		case RELATIVE_ENTROPY:
			if (my_diags->critical_mass <= 0.){
				my_diags->rel_entropy = diag;
				if (verbose){
					sprintf(buf, "relative entropy = %e\n", my_diags->rel_entropy);
					message(buf);
				}
			}
			break;
		case MOMENTUM4:
			my_diags->mom4 = diag;
			if (verbose){
				sprintf(buf, "momentum 4th order = %e\n", my_diags->mom4);
				message(buf);
			}
			break;
		case MOMENTUM6:
			my_diags->mom6 = diag;
			if (verbose){
				sprintf(buf, "momentum 6th order = %e\n", my_diags->mom6);
				message(buf);
			}
			break;
		case INTERNAL_ENERGY:
			my_diags->eint = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "eint = %e\n", my_diags->eint);
				message(buf);
			}
			break;
		case STRESS_ENERGY_XX:
			my_diags->stress_energy[0][0] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "stress energy in x-x-directions = %e\n", my_diags->stress_energy[0][0]);
				message(buf);
			}
			break;
		case STRESS_ENERGY_XY:
			my_diags->stress_energy[0][1] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "stress energy in x-y-directions = %e\n", my_diags->stress_energy[0][1]);
				message(buf);
			}
			break;
		case STRESS_ENERGY_XZ:
			my_diags->stress_energy[0][2] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "stress energy in x-z-directions = %e\n", my_diags->stress_energy[0][2]);
				message(buf);
			}
			break;
		case STRESS_ENERGY_YX:
			my_diags->stress_energy[1][0] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "stress energy in y-x-directions = %e\n", my_diags->stress_energy[1][0]);
				message(buf);
			}
			break;
		case STRESS_ENERGY_YY:
			my_diags->stress_energy[1][1] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "stress energy in y-y-directions = %e\n", my_diags->stress_energy[1][1]);
				message(buf);
			}
			break;
		case STRESS_ENERGY_YZ:
			my_diags->stress_energy[1][2] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "stress energy in y-z-directions = %e\n", my_diags->stress_energy[1][2]);
				message(buf);
			}
			break;
		case STRESS_ENERGY_ZX:
			my_diags->stress_energy[2][0] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "stress energy in z-x-directions = %e\n", my_diags->stress_energy[2][0]);
				message(buf);
			}
			break;
		case STRESS_ENERGY_ZY:
			my_diags->stress_energy[2][1] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "stress energy in z-y-directions = %e\n", my_diags->stress_energy[2][1]);
				message(buf);
			}
			break;
		case STRESS_ENERGY_ZZ:
			my_diags->stress_energy[2][2] = diag/my_diags->mass;
			if (verbose){
				sprintf(buf, "stress energy in z-z-directions = %e\n", my_diags->stress_energy[2][2]);
				message(buf);
			}
			break;
		case MIN_VALUE:
			my_diags->minf = diag;
			if (verbose){
				sprintf(buf, "min value = %e\n", my_diags->minf);
				message(buf);
			}
			break;
		case MAX_VALUE:
			my_diags->maxf = diag;
			if (verbose){
				sprintf(buf, "max value = %e\n", my_diags->maxf);
				message(buf);
			}
			break;
		case L1_NORM:
			my_diags->L1 = diag;
			if (verbose){
				sprintf(buf, "L1-norm = %e\n", my_diags->L1);
				message(buf);
			}
			break;
		case L2_NORM:
			my_diags->L2 = sqrt(diag);
			if (verbose){
				sprintf(buf, "L2-norm = %e\n", my_diags->L2);
				message(buf);
			}
			break;
		case LINF_NORM:
			my_diags->Linf = diag;
			if (verbose){
				sprintf(buf, "Linf-norm = %e\n", my_diags->Linf);
				message(buf);
			}
			break;
		case L1_NORM_EX:
			my_diags->L1ex = diag;
			if (verbose){
				sprintf(buf, "L1-norm fex = %e\n", my_diags->L1ex);
				message(buf);
			}
			break;
		case L2_NORM_EX:
			my_diags->L2ex = sqrt(diag);
			if (verbose){
				sprintf(buf, "L2-norm fex = %e\n", my_diags->L2ex);
				message(buf);
			}
			break;
		case LINF_NORM_EX:
			my_diags->Linfex = diag;
			if (verbose){
				sprintf(buf, "Linf-norm fex = %e\n", my_diags->Linfex);
				message(buf);
			}
			break;
		case L1_NORM_LIMIT:
			if (my_diags->critical_mass <= 0.){
				my_diags->L1limit = diag;
				if (verbose){
					sprintf(buf, "L1-norm flimit = %e\n", my_diags->L1limit);
					message(buf);
				}
			}
			break;
		case L2_NORM_LIMIT:
			if (my_diags->critical_mass <= 0.){
				my_diags->L2limit = sqrt(diag);
				if (verbose){
					sprintf(buf, "L2-norm flimit = %e\n", my_diags->L2limit);
					message(buf);
				}
			}
			break;
		case LINF_NORM_LIMIT:
			if (my_diags->critical_mass <= 0.){
				my_diags->Linflimit = diag;
				if (verbose){
					sprintf(buf, "Linf-norm flimit = %e\n", my_diags->Linflimit);
					message(buf);
				}
			}
			break;
		case L1_ERROR:
			my_diags->errL1 = diag/my_diags->L1ex;
			if (verbose){
				sprintf(buf, "L1-norm error = %e\n", my_diags->errL1);
				message(buf);
			}
			break;
		case L2_ERROR:
			my_diags->errL2 = sqrt(diag)/my_diags->L2ex;
			if (verbose){
				sprintf(buf, "L2-norm error = %e\n", my_diags->errL2);
				message(buf);
			}
			break;
		case LINF_ERROR:
			my_diags->errLinf = diag/my_diags->Linfex;
			if (verbose){
				sprintf(buf, "Linf-norm error = %e\n", my_diags->errLinf);
				message(buf);
			}
			break;
		case L1_ERROR_RELAXATION:
			if (my_diags->critical_mass <= 0.){
				my_diags->relaxerrL1 = diag/my_diags->L1limit;
				if (verbose){
					sprintf(buf, "L1-norm relaxation = %e\n", my_diags->relaxerrL1);
					message(buf);
				}
			}
			break;
		case L2_ERROR_RELAXATION:
			if (my_diags->critical_mass <= 0.){
				my_diags->relaxerrL2 = sqrt(diag)/my_diags->L2limit;
				if (verbose){
					sprintf(buf, "L2-norm relaxation = %e\n", my_diags->relaxerrL2);
					message(buf);
				}
			}
			break;
		case LINF_ERROR_RELAXATION:
			if (my_diags->critical_mass <= 0.){
				my_diags->relaxerrLinf = diag/(my_diags->Linflimit);
				if (verbose){
					sprintf(buf, "Linf-norm relaxation = %e\n", my_diags->relaxerrLinf);
					message(buf);
				}
			}
			break;
		case AVERAGE_U:
		case STRESS_ENERGY:
		case TEMPERATURE:
			break;
		default:
			error("Unknown diagnostic.");
	}

	if (thediag == TEMPERATURE){
		switch (parameters.type_collision){
			case CLASSICAL_COLLISIONS:
				// The temperature is 2/d * eint
				my_diags->temperature = (2.*my_diags->eint)/(parameters.ndim+1.);
				my_diags->critical_mass = -1.;
				break;
			case QUANTUM_COLLISIONS:
				// The temperature and the fugacity are obtained by solving a non-linear system with rho and eint as source terms
				find_fugacity_temperature_criticalmass_fixedpoint(my_diags->mass, &(my_diags->eint), &(my_diags->temperature), &(my_diags->fugacity), &(my_diags->critical_mass), parameters.hbar, parameters.ndim+1, parameters.type_qparticles);
				break;
			default:
				error("Unknown type of collision.");
		}
		if (verbose){
			sprintf(buf, "temperature = %e\n", my_diags->temperature);
			message(buf);
		}
	}

	
	if (thediag == AVERAGE_U){
		compute_diagnostic(my_diags, AVERAGE_UX, parameters);
		compute_diagnostic(my_diags, AVERAGE_UY, parameters);
		if (parameters.ndim == DIM3)
			compute_diagnostic(my_diags, AVERAGE_UZ, parameters);
	}
	if (thediag == STRESS_ENERGY){
		compute_diagnostic(my_diags, STRESS_ENERGY_XX, parameters);
		compute_diagnostic(my_diags, STRESS_ENERGY_XY, parameters);
		compute_diagnostic(my_diags, STRESS_ENERGY_YX, parameters);
		compute_diagnostic(my_diags, STRESS_ENERGY_YY, parameters);
		if (parameters.ndim == DIM3){
			compute_diagnostic(my_diags, STRESS_ENERGY_XZ, parameters);
			compute_diagnostic(my_diags, STRESS_ENERGY_YZ, parameters);
			compute_diagnostic(my_diags, STRESS_ENERGY_ZX, parameters);
			compute_diagnostic(my_diags, STRESS_ENERGY_ZY, parameters);
			compute_diagnostic(my_diags, STRESS_ENERGY_ZZ, parameters);
		}
	}
#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif
}


void compute_time_diagnostics(DIAGS *my_diags, double dt, PARAMETERS parameters){
	my_diags->L1L1 = my_diags->L1L1 + dt*my_diags->L1;
	my_diags->L1L2 = my_diags->L1L2 + dt*my_diags->L2;
	my_diags->L1Linf = my_diags->L1Linf + dt*my_diags->Linf;
	my_diags->L2L1 = sqrt(pow(my_diags->L2L1, 2) + dt*pow(my_diags->L1, 2));
	my_diags->L2L2 = sqrt(pow(my_diags->L2L2, 2) + dt*pow(my_diags->L2, 2));
	my_diags->L2Linf = sqrt(pow(my_diags->L2Linf, 2) + dt*pow(my_diags->Linf, 2));
	my_diags->LinfL1 = fmax(my_diags->LinfL1, my_diags->L1);
	my_diags->LinfL2 = fmax(my_diags->LinfL2, my_diags->L2);
	my_diags->LinfLinf = fmax(my_diags->LinfLinf, my_diags->Linf);

	if (!(parameters.condensation)){
		my_diags->relaxerrL1L1 = my_diags->relaxerrL1L1 + dt*my_diags->relaxerrL1;
		my_diags->relaxerrL1L2 = my_diags->relaxerrL1L2 + dt*my_diags->relaxerrL2;
		my_diags->relaxerrL1Linf = my_diags->relaxerrL1Linf + dt*my_diags->relaxerrLinf;
		my_diags->relaxerrL2L1 = sqrt(pow(my_diags->relaxerrL2L1, 2) + dt*pow(my_diags->relaxerrL1, 2));
		my_diags->relaxerrL2L2 = sqrt(pow(my_diags->relaxerrL2L2, 2) + dt*pow(my_diags->relaxerrL2, 2));
		my_diags->relaxerrL2Linf = sqrt(pow(my_diags->relaxerrL2Linf, 2) + dt*pow(my_diags->relaxerrLinf, 2));
		my_diags->relaxerrLinfL1 = fmax(my_diags->relaxerrLinfL1, my_diags->relaxerrL1);
		my_diags->relaxerrLinfL2 = fmax(my_diags->relaxerrLinfL2, my_diags->relaxerrL2);
		my_diags->relaxerrLinfLinf = fmax(my_diags->relaxerrLinfLinf, my_diags->relaxerrLinf);
	}
	
	if (parameters.exact_solution){
		my_diags->L1L1ex = my_diags->L1L1ex + dt*my_diags->L1ex;
		my_diags->L1L2ex = my_diags->L1L2ex + dt*my_diags->L2ex;
		my_diags->L1Linfex = my_diags->L1Linfex + dt*my_diags->Linfex;
		my_diags->L2L1ex = sqrt(pow(my_diags->L2L1ex, 2) + dt*pow(my_diags->L1ex, 2));
		my_diags->L2L2ex = sqrt(pow(my_diags->L2L2ex, 2) + dt*pow(my_diags->L2ex, 2));
		my_diags->L2Linfex = sqrt(pow(my_diags->L2Linfex, 2) + dt*pow(my_diags->Linfex, 2));
		my_diags->LinfL1ex = fmax(my_diags->LinfL1ex, my_diags->L1ex);
		my_diags->LinfL2ex = fmax(my_diags->LinfL2ex, my_diags->L2ex);
		my_diags->LinfLinfex = fmax(my_diags->LinfLinfex, my_diags->Linfex);
		
		my_diags->errL1L1 = my_diags->errL1L1 + dt*my_diags->errL1;
		my_diags->errL1L2 = my_diags->errL1L2 + dt*my_diags->errL2;
		my_diags->errL1Linf = my_diags->errL1Linf + dt*my_diags->errLinf;
		my_diags->errL2L1 = sqrt(pow(my_diags->errL2L1, 2) + dt*pow(my_diags->errL1, 2));
		my_diags->errL2L2 = sqrt(pow(my_diags->errL2L2, 2) + dt*pow(my_diags->errL2, 2));
		my_diags->errL2Linf = sqrt(pow(my_diags->errL2Linf, 2) + dt*pow(my_diags->errLinf, 2));
		my_diags->errLinfL1 = fmax(my_diags->errLinfL1, my_diags->errL1);
		my_diags->errLinfL2 = fmax(my_diags->errLinfL2, my_diags->errL2);
		my_diags->errLinfLinf = fmax(my_diags->errLinfLinf, my_diags->errLinf);
	}
}


void write_xdmf_file(DIAGS diags_n, char *namexdmf, char *nameh5, char *namegroup, char *namefield_f, char *namefield_flimit, char *namefield_fexact, char *namefield_G, char *namefield_Glimit, char *namefield_Gexact, PARAMETERS parameters){
	FILE *filexdmf;

	double pi = 4.*atan(1.);
	
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	
	double lambda, myomega;
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		lambda = 0.;
		myomega = 1.;
	}
	else{
		lambda = 1.;
		myomega = diags_n.omega;
	}
	
	double dvx, dvy, dvz;
	dvx = (2.*parameters.L/myomega)/MVX;
	dvy = (2.*parameters.L/myomega)/MVY;
	if (parameters.ndim == DIM2)
		dvz = 1.;
	else
		dvz = (2.*parameters.L/myomega)/MVZ;
	
	double ovx, ovy, ovz;
	ovx = lambda*diags_n.u_av[0]-parameters.L/myomega;
	ovy = lambda*diags_n.u_av[1]-parameters.L/myomega;
	if (parameters.ndim == DIM3)
		ovz = lambda*diags_n.u_av[2]-parameters.L/myomega;
	
	char namefield[MAXSTRINGLENGTH];
	
	filexdmf = fopen(namexdmf, "w");
	fprintf(filexdmf, "<?xml version=\"1.0\" ?>\n");
	fprintf(filexdmf, "<!DOCTYPE Xdmf SYSTEM \"Xdmf.dtd\" []>\n");
	fprintf(filexdmf, "<Xdmf Version=\"2.0\">\n");
	fprintf(filexdmf, "  <Domain>\n");
	fprintf(filexdmf, "    <Grid Name=\"Mesh\" GridType=\"Uniform\">\n");
	if (parameters.ndim == DIM2)
		fprintf(filexdmf, "      <Topology TopologyType=\"2DCoRectMesh\" Dimensions=\"%d %d\"/>\n", MVX, MVY);
	else if (parameters.ndim == DIM3)
		fprintf(filexdmf, "      <Topology TopologyType=\"3DCoRectMesh\" Dimensions=\"%d %d %d\"/>\n", MVX, MVY, MVZ);
	
	if (parameters.ndim == DIM2){
		fprintf(filexdmf, "      <Geometry GeometryType=\"ORIGIN_DXDY\">\n");
		fprintf(filexdmf, "        <DataItem Format=\"XML\" Dimensions=\"2\" DataType=\"Float\" Precision=\"8\">\n");
		fprintf(filexdmf, "          %f %f\n", ovx, ovy);
	}
	else if (parameters.ndim == DIM3){
		fprintf(filexdmf, "      <Geometry GeometryType=\"ORIGIN_DXDYDZ\">\n");
		fprintf(filexdmf, "        <DataItem Format=\"XML\" Dimensions=\"3\" DataType=\"Float\" Precision=\"8\">\n");
		fprintf(filexdmf, "          %f %f %f\n", ovx, ovy, ovz);
	}
	fprintf(filexdmf, "        </DataItem>\n");
	
	if (parameters.ndim == DIM2){
		fprintf(filexdmf, "        <DataItem Format=\"XML\" Dimensions=\"2\" DataType=\"Float\" Precision=\"8\">\n");
		fprintf(filexdmf, "          %f %f\n", dvx, dvy);
	}
	else if (parameters.ndim == DIM3){
		fprintf(filexdmf, "        <DataItem Format=\"XML\" Dimensions=\"3\" DataType=\"Float\" Precision=\"8\">\n");
		fprintf(filexdmf, "          %f %f %f\n", dvx, dvy, dvz);
	}
	fprintf(filexdmf, "        </DataItem>\n");
	fprintf(filexdmf, "      </Geometry>\n");
	
	fprintf(filexdmf, "      <Attribute Name=\"%s\">\n", namefield_f);
	if (parameters.ndim == DIM2)
		fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY);
	else if (parameters.ndim == DIM3)
		fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY, MVZ);
	fprintf(filexdmf, "          ./%s", nameh5);
	fprintf(filexdmf, ":/%s/%s\n", namegroup, namefield_f);
	fprintf(filexdmf, "        </DataItem>\n");
	fprintf(filexdmf, "      </Attribute>\n");
	
	if (!(parameters.condensation)){
		fprintf(filexdmf, "      <Attribute Name=\"%s\">\n", namefield_flimit);
		if (parameters.ndim == DIM2)
			fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY);
		else if (parameters.ndim == DIM3)
			fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY, MVZ);
		fprintf(filexdmf, "          ./%s", nameh5);
		fprintf(filexdmf, ":/%s/%s\n", namegroup, namefield_flimit);
		fprintf(filexdmf, "        </DataItem>\n");
		fprintf(filexdmf, "      </Attribute>\n");
	}
		
	if (parameters.exact_solution){
		fprintf(filexdmf, "      <Attribute Name=\"%s\">\n", namefield_fexact);
		if (parameters.ndim == DIM2)
			fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY);
		else if (parameters.ndim == DIM3)
			fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY, MVZ);
		fprintf(filexdmf, "          ./%s", nameh5);
		fprintf(filexdmf, ":/%s/%s\n", namegroup, namefield_fexact);
		fprintf(filexdmf, "        </DataItem>\n");
		fprintf(filexdmf, "      </Attribute>\n");
	}
	fprintf(filexdmf, "    </Grid>\n");

	if (parameters.type_omega != OMEGA_IS_NOTUSED){
		fprintf(filexdmf, "    <Grid Name=\"Rescaled_mesh\" GridType=\"Uniform\">\n");
		if (parameters.ndim == DIM2)
			fprintf(filexdmf, "      <Topology TopologyType=\"2DCoRectMesh\" Dimensions=\"%d %d\"/>\n", MVX, MVY);
		else if (parameters.ndim == DIM3)
			fprintf(filexdmf, "      <Topology TopologyType=\"3DCoRectMesh\" Dimensions=\"%d %d %d\"/>\n", MVX, MVY, MVZ);
	
		if (parameters.ndim == DIM2){
			fprintf(filexdmf, "      <Geometry GeometryType=\"ORIGIN_DXDY\">\n");
			fprintf(filexdmf, "        <DataItem Format=\"XML\" Dimensions=\"2\" DataType=\"Float\" Precision=\"8\">\n");
			fprintf(filexdmf, "          %f %f\n", -pi, -pi);
		}
		else if (parameters.ndim == DIM3){
			fprintf(filexdmf, "      <Geometry GeometryType=\"ORIGIN_DXDYDZ\">\n");
			fprintf(filexdmf, "        <DataItem Format=\"XML\" Dimensions=\"3\" DataType=\"Float\" Precision=\"8\">\n");
			fprintf(filexdmf, "          %f %f %f\n", -pi, -pi, -pi);
		}
		fprintf(filexdmf, "        </DataItem>\n");
	
		if (parameters.ndim == DIM2){
			fprintf(filexdmf, "        <DataItem Format=\"XML\" Dimensions=\"2\" DataType=\"Float\" Precision=\"8\">\n");
			fprintf(filexdmf, "          %f %f\n", 2.*pi/MVX, 2.*pi/MVY);
		}
		else if (parameters.ndim == DIM3){
			fprintf(filexdmf, "        <DataItem Format=\"XML\" Dimensions=\"3\" DataType=\"Float\" Precision=\"8\">\n");
			fprintf(filexdmf, "          %f %f %f\n", 2.*pi/MVX, 2.*pi/MVY, 2.*pi/MVZ);
		}
		fprintf(filexdmf, "        </DataItem>\n");
		fprintf(filexdmf, "      </Geometry>\n");
	
		fprintf(filexdmf, "      <Attribute Name=\"%s\">\n", namefield_G);
		if (parameters.ndim == DIM2)
			fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY);
		else if (parameters.ndim == DIM3)
			fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY, MVZ);
		fprintf(filexdmf, "          ./%s", nameh5);
		fprintf(filexdmf, ":/%s/%s\n", namegroup, namefield_G);
		fprintf(filexdmf, "        </DataItem>\n");
		fprintf(filexdmf, "      </Attribute>\n");
		
		if (!(parameters.condensation)){
			fprintf(filexdmf, "      <Attribute Name=\"%s\">\n", namefield_Glimit);
			if (parameters.ndim == DIM2)
				fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY);
			else if (parameters.ndim == DIM3)
				fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY, MVZ);
			fprintf(filexdmf, "          ./%s", nameh5);
			fprintf(filexdmf, ":/%s/%s\n", namegroup, namefield_Glimit);
			fprintf(filexdmf, "        </DataItem>\n");
			fprintf(filexdmf, "      </Attribute>\n");
		}
		
		if (parameters.exact_solution){
			fprintf(filexdmf, "      <Attribute Name=\"%s\">\n", namefield_Gexact);
			if (parameters.ndim == DIM2)
				fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY);
			else if (parameters.ndim == DIM3)
				fprintf(filexdmf, "        <DataItem Format=\"HDF\" Dimensions=\"%d %d %d\" NumberType=\"Float\" Precision=\"8\">\n", MVX, MVY, MVZ);
			fprintf(filexdmf, "          ./%s", nameh5);
			fprintf(filexdmf, ":/%s/%s\n", namegroup, namefield_Gexact);
			fprintf(filexdmf, "        </DataItem>\n");
			fprintf(filexdmf, "      </Attribute>\n");
		}
	
		fprintf(filexdmf, "    </Grid>\n");
	}

	fprintf(filexdmf, "  </Domain>\n");
	fprintf(filexdmf, "</Xdmf>");
	fclose(filexdmf);
}



void shift_2d(double** Gin, double** Gout, PARAMETERS parameters){
	int kvxabs, kvyabs;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVLOC = NVXLOC*NVYLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MV = MVX*MVY;

	double contrib;

	int k;

#if defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
	fftw_complex* Gin_real;
	fftw_complex* G_spectral;
	fftw_complex* Gout_real;

	fftw_plan plan_backward, plan_forward;

#ifdef key_fftw_mpi
	Gin_real = fftw_alloc_complex(parameters.alloc_local);
	G_spectral = fftw_alloc_complex(parameters.alloc_local);
	Gout_real = fftw_alloc_complex(parameters.alloc_local);

	plan_forward = fftw_mpi_plan_dft_2d(MVX, MVY, Gin_real, G_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	plan_backward = fftw_mpi_plan_dft_2d(MVX, MVY, G_spectral, Gout_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
#else
	Gin_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	G_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	Gout_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);

	plan_forward = fftw_plan_dft_2d(MVX, MVY, Gin_real, G_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	plan_backward = fftw_plan_dft_2d(MVX, MVY, G_spectral, Gout_real, FFTW_BACKWARD, FFTW_ESTIMATE);
#endif

	#pragma omp parallel for private(k) shared(NVLOC, Gin_real, Gin)
	for (k = 0; k < NVLOC; k++){
		Gin_real[k][0] = Gin[0][k];
		Gin_real[k][1] = 0.;
	}

	fftw_execute(plan_forward);

	#pragma omp parallel for private(k, kvxabs, kvyabs, contrib) shared(NVLOC, NVXLOC, NVYLOC, parameters, IVX0LOC, IVY0LOC, G_spectral)
	for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
		kvxabs = parameters.kvxabstab[k];
		kvyabs = parameters.kvyabstab[k];
#else
		kvxabs = k/NVYLOC;
		kvyabs = k - kvxabs*NVYLOC;
		kvxabs += IVX0LOC;
		kvyabs += IVY0LOC;
#endif

		contrib = pow(-1., kvxabs+kvyabs);
		#pragma omp atomic		
		G_spectral[k][0] *= contrib;
		#pragma omp atomic
		G_spectral[k][1] *= contrib;
	}

	fftw_execute(plan_backward);

	#pragma omp parallel for private(k) shared(NVLOC, Gout, Gout_real, MV)
	for (k = 0; k < NVLOC; k++)
		Gout[0][k] = Gout_real[k][0]/MV;

	fftw_destroy_plan(plan_forward);
	fftw_destroy_plan(plan_backward);
	fftw_free(Gin_real);
	fftw_free(Gout_real);
	fftw_free(G_spectral);
#endif
}


void shift_3d(double*** Gin, double*** Gout, PARAMETERS parameters){
	int kvxabs, kvyabs, kvzabs;
	int itemp;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = parameters.nvz;
	int MV = MVX*MVY*MVZ;

	double contrib;

	int k;

#if defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
	fftw_complex* Gin_real;
	fftw_complex* G_spectral;
	fftw_complex* Gout_real;

	fftw_plan plan_backward, plan_forward;

#ifdef key_fftw_mpi
	Gin_real = fftw_alloc_complex(parameters.alloc_local);
	G_spectral = fftw_alloc_complex(parameters.alloc_local);
	Gout_real = fftw_alloc_complex(parameters.alloc_local);
	
	plan_forward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, Gin_real, G_spectral, MPI_COMM_WORLD, FFTW_FORWARD, FFTW_ESTIMATE);
	plan_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, G_spectral, Gout_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
#else
	Gin_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	G_spectral = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	Gout_real = fftw_malloc(sizeof(fftw_complex)*NVLOC);
	
	plan_forward = fftw_plan_dft_3d(MVX, MVY, MVZ, Gin_real, G_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	plan_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, G_spectral, Gout_real, FFTW_BACKWARD, FFTW_ESTIMATE);
#endif

	#pragma omp parallel for private(k) shared(NVLOC, Gin_real, Gin)
	for (k = 0; k < NVLOC; k++){
		Gin_real[k][0] = Gin[0][0][k];
		Gin_real[k][1] = 0.;
	}

	fftw_execute(plan_forward);

	#pragma omp parallel for private(k, kvxabs, kvyabs, kvzabs, itemp, contrib) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, parameters, IVX0LOC, IVY0LOC, IVZ0LOC, G_spectral)
	for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
		kvxabs = parameters.kvxabstab[k];
		kvyabs = parameters.kvyabstab[k];
		kvzabs = parameters.kvzabstab[k];
#else
		// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
		itemp = k/NVZLOC;
		// itemp now contains kvyloc+NVYLOC*kvxloc;
		kvzabs = k - itemp*NVZLOC;
		kvxabs = itemp/NVYLOC;
		kvyabs = itemp-kvxabs*NVYLOC;
		kvxabs += IVX0LOC;
		kvyabs += IVY0LOC;
		kvzabs += IVZ0LOC;
#endif
		contrib = pow(-1., kvxabs+kvyabs+kvzabs);
		#pragma omp atomic
		G_spectral[k][0] *= contrib;
		#pragma omp atomic
		G_spectral[k][1] *= contrib;
	}

	fftw_execute(plan_backward);

	#pragma omp parallel for private(k) shared(NVLOC, Gout, Gout_real, MV)
	for (k = 0; k < NVLOC; k++)
		Gout[0][0][k] = Gout_real[k][0]/MV;

	fftw_destroy_plan(plan_forward);
	fftw_destroy_plan(plan_backward);
	fftw_free(Gin_real);
	fftw_free(Gout_real);
	fftw_free(G_spectral);
#endif
}



void save_all_data(DIAGS mydiags, double time, double dt, const char* namefile, PARAMETERS parameters){
	double var;
	char namelog[MAXSTRINGLENGTH+30];
	char prefixfile[MAXSTRINGLENGTH];
	char namegroup[MAXSTRINGLENGTH];
	char namefield[MAXSTRINGLENGTH];
	char namefield_f[MAXSTRINGLENGTH];
	char namefield_flimit[MAXSTRINGLENGTH];
	char namefield_fexact[MAXSTRINGLENGTH];
	char namefield_G[MAXSTRINGLENGTH];
	char namefield_Glimit[MAXSTRINGLENGTH];
	char namefield_Gexact[MAXSTRINGLENGTH];
	int newfile, istatus;
	char xyz[5];
	strcpy(xyz, "xyz");
	const char* ext = get_filename_ext(namefile);
	double **Gshifted_2d;
	double ***Gshifted_3d;
	
	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	int MV = MVX*MVY*MVZ;

	double contrib;
	
	double pi = 4.*atan(1.);

	int k;

	int rank;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
#endif
	
	double lambda, myomega, mu;
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		lambda = 0.;
		myomega = 1.;
		mu = pow(parameters.L/pi, parameters.ndim+1);
	}
	else{
		lambda = 1.;
		myomega = mydiags.omega;
		mu = 1.;
	}
	
	struct stat sb;
	istatus = stat(namefile, &sb);
	if (!istatus){
		if (S_ISREG(sb.st_mode))
			istatus = remove(namefile);
	}

	if (parameters.type_Tinit == SETTOZERO){
		if  (fabs(time) < THRESHOLD)
			newfile = 1;
		else
			newfile = 0;
	}
	else if (parameters.type_Tinit == FROM_FILE){
		if (fabs(time-parameters.Tinit) < THRESHOLD)
			newfile = 1;
		else
			newfile = 0;
	}
	else if (parameters.resume)
		newfile = 0;
	else
		error("problem with parameters.type_Tinit.");
	
	strcpy(prefixfile, "");
	strcat(prefixfile, parameters.outputdir);
	strcat(prefixfile, "/");
	strcat(prefixfile, parameters.prefix);
	
	if (!parameters.save_norms_only){
		strcpy(namegroup, "distribution");

		strcpy(namefield_f, "f");
		strcpy(namefield_flimit, "f_limit");
		strcpy(namefield_fexact, "f_exact");
		if (parameters.type_omega != OMEGA_IS_NOTUSED){
			strcpy(namefield_G, "G");
			strcpy(namefield_Glimit, "G_limit");
			strcpy(namefield_Gexact, "G_exact");
		}
		else{
			strcpy(namefield_G, "");
			strcpy(namefield_Glimit, "");
			strcpy(namefield_Gexact, "");
		}
		
		if (parameters.ndim == DIM2){
			Gshifted_2d = allocation_2d(NVXLOC, NVYLOC);
			
			shift_2d(mydiags.Gloc_2d, Gshifted_2d, parameters);
			if (parameters.type_omega != OMEGA_IS_NOTUSED){
				// Gshifted_2d contains the rescaled distribution on [-pi,pi]^2
				// Save it as G
				write_2d(namefile, namegroup, namefield_G, Gshifted_2d, parameters);
			}
			// Perform the inverse rescaling to get the non-rescaled distribution on lambda*u+[-L/omega,L/omega]^2
			contrib = (mu * pow(myomega*pi/parameters.L, parameters.ndim+1));
			#pragma omp parallel for private(k) shared(NVLOC, Gshifted_2d, contrib)
			for (k = 0; k < NVLOC; k++){
				#pragma omp atomic
				Gshifted_2d[0][k] *= contrib;
			}
			
			// Gshifted_2d now contains the non-rescaled distribution
			// Save it as f
			write_2d(namefile, namegroup, namefield_f, Gshifted_2d, parameters);
			
			if (!(parameters.condensation)){
				shift_2d(mydiags.Gloclimit_2d, Gshifted_2d, parameters);
				if (parameters.type_omega != OMEGA_IS_NOTUSED){
					// Gshifted_2d contains the limit distribution rescaled with the current omega on [-pi,pi]^2
					// Save it as Glimit
					write_2d(namefile, namegroup, namefield_Glimit, Gshifted_2d, parameters);
				}
				// Perform the inverse rescaling to get the non-rescaled limit distribution on lambda*u+[-L/omega,L/omega]^2
				contrib = (mu * pow(myomega*pi/parameters.L, parameters.ndim+1));
				#pragma omp parallel for private(k) shared(NVLOC, Gshifted_2d, contrib)
				for (k = 0; k < NVLOC; k++){
					#pragma omp atomic
					Gshifted_2d[0][k] *= contrib;
				}
			
				// Gshifted_2d now contains the non-rescaled limit distribution
				// Save it as flimit
				write_2d(namefile, namegroup, namefield_flimit, Gshifted_2d, parameters);
			}

			if (parameters.exact_solution){
				shift_2d(mydiags.Glocexact_2d, Gshifted_2d, parameters);
				if (parameters.type_omega != OMEGA_IS_NOTUSED){
					// Gshifted_2d contains the exact distribution rescaled with the current omega on [-pi,pi]^2
					// Save it as Gexact
					write_2d(namefile, namegroup, namefield_Gexact, Gshifted_2d, parameters);
				}
				// Perform the inverse rescaling to get the non-rescaled exact distribution on lambda*u+[-L/omega,L/omega]^2
				contrib = (mu * pow(myomega*pi/parameters.L, parameters.ndim+1));
				#pragma omp parallel for private(k) shared(NVLOC, Gshifted_2d, contrib)
				for (k = 0; k < NVLOC; k++){
					#pragma omp atomic
					Gshifted_2d[0][k] *= contrib;
				}
				
				// Gshifted_2d now contains the non-rescaled exact distribution
				// Save it as fexact
				write_2d(namefile, namegroup, namefield_fexact, Gshifted_2d, parameters);
			}
			
			free_2d(Gshifted_2d);
		}
		else if (parameters.ndim == DIM3){
			Gshifted_3d = allocation_3d(NVXLOC, NVYLOC, NVZLOC);
			shift_3d(mydiags.Gloc_3d, Gshifted_3d, parameters);
			if (parameters.type_omega != OMEGA_IS_NOTUSED){
				// Gshifted_3d contains the rescaled distribution on [-pi,pi]^3
				// Save it as G
				write_3d(namefile, namegroup, namefield_G, Gshifted_3d, parameters);
			}
			// Perform the inverse rescaling to get the non-rescaled distribution on lambda*u+[-L/omega,L/omega]^3
			contrib = (mu * pow(myomega*pi/parameters.L, parameters.ndim+1));
			#pragma omp parallel for private(k) shared(NVLOC, Gshifted_3d, contrib)
			for (k = 0; k < NVLOC; k++){
				#pragma omp atomic
				Gshifted_3d[0][0][k] *= contrib;
			}
					
			// Gshifted_3d now contains the non-rescaled distribution
			// Save it as f
			write_3d(namefile, namegroup, namefield_f, Gshifted_3d, parameters);
			
			if (!(parameters.condensation)){
				shift_3d(mydiags.Gloclimit_3d, Gshifted_3d, parameters);
				if (parameters.type_omega != OMEGA_IS_NOTUSED){
					// Gshifted_3d contains the rescaled limit distribution on [-pi,pi]^3
					// Save it as Glimit
					write_3d(namefile, namegroup, namefield_Glimit, Gshifted_3d, parameters);
				}
				contrib = (mu * pow(myomega*pi/parameters.L, parameters.ndim+1));
				#pragma omp parallel for private(k) shared(NVLOC, Gshifted_3d, contrib)
				for (k = 0; k < NVLOC; k++){
					#pragma omp atomic
					Gshifted_3d[0][0][k] *= contrib;
				}
				
				// Gshifted_3d now contains the non-rescaled limit distribution
				// Save it as flimit
				write_3d(namefile, namegroup, namefield_flimit, Gshifted_3d, parameters);
			}
			
			if (parameters.exact_solution){
				shift_3d(mydiags.Glocexact_3d, Gshifted_3d, parameters);
				if (parameters.type_omega != OMEGA_IS_NOTUSED){
					// Gshifted_3d contains the exact distribution rescaled with the current omega on [-pi,pi]^3
					// Save it as Gexact
					write_3d(namefile, namegroup, namefield_Gexact, Gshifted_3d, parameters);
				}
				contrib = (mu * pow(myomega*pi/parameters.L, parameters.ndim+1));
				#pragma omp parallel for private(k) shared(NVLOC, Gshifted_3d, contrib)
				for (int k = 0; k < NVLOC; k++){
					#pragma omp atomic
					Gshifted_3d[0][0][k] *= contrib;
				}
				
				// Gshifted_3d now contains the non-rescaled exact distribution
				// Save it as fexact
				write_3d(namefile, namegroup, namefield_fexact, Gshifted_3d, parameters);
			}
	
			free_3d(Gshifted_3d);
		}

		if (strcmp(ext, "h5") == 0){
#if defined(key_phdf5) || defined(key_hdf5)
			FILE *filexdmf;
			size_t len = strlen(namefile);
			char namexdmf[MAXSTRINGLENGTH];
			memcpy(namexdmf, namefile, len-2);
			namexdmf[len-2] = 'x';
			namexdmf[len-1] = 'd';
			namexdmf[len] = 'm';
			namexdmf[len+1] = 'f';
			namexdmf[len+2] = '\0';

			char nameh5[MAXSTRINGLENGTH];
			char* shortnameh5 = strrchr(namefile, '/');
			len = strlen(shortnameh5);
			memcpy(nameh5, &shortnameh5[1], len-1);
			nameh5[len-1] = '\0';
#else
			error("Kinebec has not been compiled with Hdf5 dependency. Please re-compile it or choose another output file format.");
#endif

			if (rank == 0)
				write_xdmf_file(mydiags, namexdmf, nameh5, namegroup, namefield_f, namefield_flimit, namefield_fexact, namefield_G, namefield_Glimit, namefield_Gexact, parameters);
		}
		
		// Save dimensions of the velocity grid
		strcpy(namefield, "nvx");
		write_0d_int(namefile, namegroup, namefield, parameters.nvx);
		strcpy(namefield, "nvy");
		write_0d_int(namefile, namegroup, namefield, parameters.nvy);
		if (parameters.ndim == DIM3){
			strcpy(namefield, "nvz");
			write_0d_int(namefile, namegroup, namefield, parameters.nvz);
		}
		
		// Save time
		strcpy(namefield, "Time");
		write_0d_double(namefile, namegroup, namefield, time);

		// Save value of L
		strcpy(namefield, "L");
		write_0d_double(namefile, namegroup, namefield, parameters.L);

		// Save omega
		strcpy(namefield, "type_omega");
		write_0d_int(namefile, namegroup, namefield, parameters.type_omega);
		if (parameters.type_omega != OMEGA_IS_NOTUSED){
			strcpy(namefield, "omega");
			write_0d_double(namefile, namegroup, namefield, mydiags.omega);
		}

		// Save cut-off radius
		strcpy(namefield, "Cut-off");
		write_0d_double(namefile, namegroup, namefield, parameters.cutoff);
		
		// Save mass
		strcpy(namefield, "Mass");
		write_0d_double(namefile, namegroup, namefield, mydiags.mass);
		
		// Save vector velocity and temperature
		for (int d = 0; d <= parameters.ndim; d++){
			sprintf(namefield, "Average_u%c", xyz[d]);
			write_0d_double(namefile, namegroup, namefield, mydiags.u_av[d]);
		}
		for (int d = 0; d <= parameters.ndim; d++){
			for (int dd = 0; dd <= parameters.ndim; dd++){
				strcpy(namefield,  "");
				sprintf(namefield, "Stress_energy_%c%c", xyz[d], xyz[dd]);
				write_0d_double(namefile, namegroup, namefield, mydiags.stress_energy[d][dd]);
			}
		}
		
		// Save kinetic energy
		strcpy(namefield, "Kinetic_energy");
		write_0d_double(namefile, namegroup, namefield, mydiags.ekin);
		
		// Save temperature
		strcpy(namefield, "Temperature");
		write_0d_double(namefile, namegroup, namefield, mydiags.temperature);
		
		if (parameters.type_collision == QUANTUM_COLLISIONS){
			// Save fugacity
			strcpy(namefield, "Fugacity");
			write_0d_double(namefile, namegroup, namefield, mydiags.fugacity);
			
			if ((parameters.type_qparticles == BOSONS) && (parameters.ndim == DIM3)){
				// Save critical mass
				strcpy(namefield, "Critical_mass");
				write_0d_double(namefile, namegroup, namefield, fmax(mydiags.critical_mass,0.));
			}
		}
		
		// Save internal energy
		strcpy(namefield, "Internal_energy");
		write_0d_double(namefile, namegroup, namefield, mydiags.eint);

		// Save L1-norm in space
		strcpy(namefield, "L1_norm_space");
		write_0d_double(namefile, namegroup, namefield, mydiags.L1);

		// Save L1-L1 norm in time-space
		strcpy(namefield, "L1-L1_norm_time-space");
		write_0d_double(namefile, namegroup, namefield, mydiags.L1L1);

		// Save L2-L1 norm in time-space
		strcpy(namefield, "L2-L1_norm_time-space");
		write_0d_double(namefile, namegroup, namefield, mydiags.L2L1);

		// Save Linf-L1 norm in time-space
		strcpy(namefield, "Linf-L1_norm_time-space");
		write_0d_double(namefile, namegroup, namefield, mydiags.LinfL1);

		// Save L2-norm in space
		strcpy(namefield, "L2_norm_space");
		write_0d_double(namefile, namegroup, namefield, mydiags.L2);

		// Save L1-L2 norm in time-space
		strcpy(namefield, "L1-L2_norm_time-space");
		write_0d_double(namefile, namegroup, namefield, mydiags.L1L2);

		// Save L2-L2 norm in time-space
		strcpy(namefield, "L2-L2_norm_time-space");
		write_0d_double(namefile, namegroup, namefield, mydiags.L2L2);

		// Save Linf-L2 norm in time-space
		strcpy(namefield, "Linf-L2_norm_time-space");
		write_0d_double(namefile, namegroup, namefield, mydiags.LinfL2);

		// Save Linf-norm in space
		strcpy(namefield, "Linf_norm_space");
		write_0d_double(namefile, namegroup, namefield, mydiags.Linf);

		// Save L1-Linf norm in time-space
		strcpy(namefield, "L1-Linf_norm_time-space");
		write_0d_double(namefile, namegroup, namefield, mydiags.L1Linf);

		// Save L2-Linf norm in time-space
		strcpy(namefield, "L2-Linf_norm_time-space");
		write_0d_double(namefile, namegroup, namefield, mydiags.L2Linf);

		// Save Linf-Linf norm in time-space
		strcpy(namefield, "Linf-Linf_norm_time-space");
		write_0d_double(namefile, namegroup, namefield, mydiags.LinfLinf);

		// Save Max(f)
		strcpy(namefield, "Max_f");
		write_0d_double(namefile, namegroup, namefield, mydiags.maxf);

		// Save Min(f)
		strcpy(namefield, "Min_f");
		write_0d_double(namefile, namegroup, namefield, mydiags.minf);

		// Save Entropy
		strcpy(namefield, "Entropy");
		write_0d_double(namefile, namegroup, namefield, mydiags.entropy);

		if (!(parameters.condensation)){
			// Save Relative entropy
			strcpy(namefield, "Relative_entropy");
			write_0d_double(namefile, namegroup, namefield, mydiags.rel_entropy);

			// Save mass of limit state
			strcpy(namefield, "Mass_limit");
			write_0d_double(namefile, namegroup, namefield, mydiags.mass_limit);
		
			// Save vector velocity of limit state
			for (int d = 0; d <= parameters.ndim; d++){
				sprintf(namefield, "Average_u%c_limit", xyz[d]);
				write_0d_double(namefile, namegroup, namefield, mydiags.u_av_limit[d]);
			}
		
			// Save temperature of limit state
			strcpy(namefield, "Temperature_limit");
			write_0d_double(namefile, namegroup, namefield, mydiags.temperature_limit);
		
			if (parameters.type_collision == QUANTUM_COLLISIONS){
				// Save fugacity of limit state
				strcpy(namefield, "Fugacity_limit");
				write_0d_double(namefile, namegroup, namefield, mydiags.fugacity_limit);
				
				if ((parameters.type_qparticles == BOSONS) && (parameters.ndim == DIM3)){
					// Save critical mass of limit state
					strcpy(namefield, "Critical_mass_limit");
					write_0d_double(namefile, namegroup, namefield, fmax(mydiags.critical_mass_limit,0.));
				}
			}
		}

		// Save 4th order momentum
		strcpy(namefield, "4th_order_momentum");
		write_0d_double(namefile, namegroup, namefield, mydiags.mom4);

		// Save 6th order momentum
		strcpy(namefield, "6th_order_momentum");
		write_0d_double(namefile, namegroup, namefield, mydiags.mom6);
	
		if (parameters.exact_solution){
			// Save L1-norm of exact solution
			strcpy(namefield, "L1_norm_exact_space");
			write_0d_double(namefile, namegroup, namefield, mydiags.L1ex);
		
			// Save L2-norm of exact solution
			strcpy(namefield, "L2_norm_exact_space");
			write_0d_double(namefile, namegroup, namefield, mydiags.L2ex);
		
			// Save Linf-norm of exact solution
			strcpy(namefield, "Linf_norm_exact_space");
			write_0d_double(namefile, namegroup, namefield, mydiags.Linfex);
		
			// Save L1-L1-norm of exact solution
			strcpy(namefield, "L1-L1_norm_exact_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.L1L1ex);
		
			// Save L2-L1-norm of exact solution
			strcpy(namefield, "L2-L1_norm_exact_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.L2L1ex);
		
			// Save Linf-L1-norm of exact solution
			strcpy(namefield, "Linf-L1_norm_exact_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.LinfL1ex);
		
			// Save L1-L2-norm of exact solution
			strcpy(namefield, "L1-L2_norm_exact_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.L1L2ex);
		
			// Save L2-L2-norm of exact solution
			strcpy(namefield, "L2-L2_norm_exact_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.L2L2ex);
		
			// Save Linf-L2-norm of exact solution
			strcpy(namefield, "Linf-L2_norm_exact_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.LinfL2ex);
		
			// Save L1-Linf-norm of exact solution
			strcpy(namefield, "L1-Linf_norm_exact_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.L1Linfex);
		
			// Save L2-Linf-norm of exact solution
			strcpy(namefield, "L2-Linf_norm_exact_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.L2Linfex);
		
			// Save Linf-Linf-norm of exact solution
			strcpy(namefield, "Linf-Linf_norm_exact_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.LinfLinfex);
		
			// Save L1-error in space
			strcpy(namefield, "L1_error_space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errL1);
		
			// Save L1-L1 error in time-space
			strcpy(namefield, "L1-L1_error_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errL1L1);
		
			// Save L2-L1 error in time-space
			strcpy(namefield, "L2-L1_error_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errL2L1);
		
			// Save Linf-L1 error in time-space
			strcpy(namefield, "Linf-L1_error_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errLinfL1);
		
			// Save L2-error in space
			strcpy(namefield, "L2_error_space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errL2);
		
			// Save L1-L2 error in time-space
			strcpy(namefield, "L1-L2_error_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errL1L2);
		
			// Save L2-L2 error in time-space
			strcpy(namefield, "L2-L2_error_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errL2L2);
		
			// Save Linf-L2 error in time-space
			strcpy(namefield, "Linf-L2_error_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errLinfL2);
		
			// Save Linf-error in space
			strcpy(namefield, "Linf_error_space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errLinf);
		
			// Save L1-Linf error in time-space
			strcpy(namefield, "L1-Linf_error_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errL1Linf);
		
			// Save L2-Linf error in time-space
			strcpy(namefield, "L2-Linf_error_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errL2Linf);
		
			// Save Linf-Linf error in time-space
			strcpy(namefield, "Linf-Linf_error_time-space");
			write_0d_double(namefile, namegroup, namefield, mydiags.errLinfLinf);
		}

		if (!(parameters.condensation)){
			// Save L1-norm of flimit in space
			strcpy(namefield, "L1_norm_space_flimit");
			write_0d_double(namefile, namegroup, namefield, mydiags.L1limit);

			// Save L2-norm of flimit in space
			strcpy(namefield, "L2_norm_space_flimit");
			write_0d_double(namefile, namegroup, namefield, mydiags.L2limit);

			// Save Linf-norm of flimit in space
			strcpy(namefield, "Linf_norm_space_flimit");
			write_0d_double(namefile, namegroup, namefield, mydiags.Linflimit);

			// Save L1-norm of relaxation in space
			strcpy(namefield, "L1_norm_space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrL1);

			// Save L1-L1 norm of relaxation in time-space
			strcpy(namefield, "L1-L1_norm_time-space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrL1L1);

			// Save L2-L1 norm of relaxation in time-space
			strcpy(namefield, "L2-L1_norm_time-space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrL2L1);

			// Save Linf-L1 norm of relaxation in time-space
			strcpy(namefield, "Linf-L1_norm_time-space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrLinfL1);

			// Save L2-norm of relaxation in space
			strcpy(namefield, "L2_norm_space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrL2);

			// Save L1-L2 norm of relaxation in time-space
			strcpy(namefield, "L1-L2_norm_time-space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrL1L2);

			// Save L2-L2 norm of relaxation in time-space
			strcpy(namefield, "L2-L2_norm_time-space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrL2L2);

			// Save Linf-L2 norm of relaxation in time-space
			strcpy(namefield, "Linf-L2_norm_time-space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrLinfL2);

			// Save Linf-norm of relaxation in space
			strcpy(namefield, "Linf_norm_space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrLinf);

			// Save L1-Linf norm of relaxation in time-space
			strcpy(namefield, "L1-Linf_norm_time-space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrL1Linf);

			// Save L2-Linf norm of relaxation in time-space
			strcpy(namefield, "L2-Linf_norm_time-space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrL2Linf);

			// Save Linf-Linf norm of relaxation in time-space
			strcpy(namefield, "Linf-Linf_norm_time-space_relaxation");
			write_0d_double(namefile, namegroup, namefield, mydiags.relaxerrLinfLinf);
		}
	}

	// Save mass
	strcpy(namelog, prefixfile);
	strcat(namelog, "_mass.txt");
	write_logfile(time, mydiags.mass, namelog, newfile);
	
	// Save vector velocity and temperature
	for (int d = 0; d <= parameters.ndim; d++){
		sprintf(namelog, "%s_u%c.txt", prefixfile, xyz[d]);
		write_logfile(time, mydiags.u_av[d], namelog, newfile);
	}
	for (int d = 0; d <= parameters.ndim; d++){
		for (int dd = 0; dd <= parameters.ndim; dd++){
			sprintf(namelog, "%s_s%c%c.txt", prefixfile, xyz[d], xyz[dd]);
			write_logfile(time, mydiags.stress_energy[d][dd], namelog, newfile);
		}
	}

	// Save Kinetic energy
	strcpy(namelog, prefixfile);
	strcat(namelog, "_ekin.txt");
	write_logfile(time, mydiags.ekin, namelog, newfile);
	
	if (parameters.type_collision == QUANTUM_COLLISIONS){
		// Save fugacity
		strcpy(namelog, prefixfile);
		strcat(namelog, "_fugacity.txt");
		write_logfile(time, mydiags.fugacity, namelog, newfile);
		
		if ((parameters.type_qparticles == BOSONS) && (parameters.ndim == DIM3)){
			// Save critical mass
			strcpy(namelog, prefixfile);
			strcat(namelog, "_critical_mass.txt");
			write_logfile(time, fmax(mydiags.critical_mass,0.), namelog, newfile);
		}
	}
	
	// Save temperature
	strcpy(namelog, prefixfile);
	strcat(namelog, "_temperature.txt");
	write_logfile(time, mydiags.temperature, namelog, newfile);

	// Save omega
	if (parameters.type_omega != OMEGA_IS_NOTUSED){
		strcpy(namelog, prefixfile);
		strcat(namelog, "_omega.txt");
		write_logfile(time, mydiags.omega, namelog, newfile);
	}

	// Save Internal energy
	strcpy(namelog, prefixfile);
	strcat(namelog, "_eint.txt");
	write_logfile(time, mydiags.eint, namelog, newfile);

	// Save L1-norm in space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_L1norm.txt");
	write_logfile(time, mydiags.L1, namelog, newfile);

	// Save L1-L1 norm in time-space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_L1L1norm.txt");
	write_logfile(time, mydiags.L1L1, namelog, newfile);

	// Save L2-L1 norm in time-space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_L2L1norm.txt");
	write_logfile(time, mydiags.L2L1, namelog, newfile);

	// Save Linf-L1 norm in time-space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_LinfL1norm.txt");
	write_logfile(time, mydiags.LinfL1, namelog, newfile);

	// Save L2-norm in space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_L2norm.txt");
	write_logfile(time, mydiags.L2, namelog, newfile);

	// Save L1-L2 norm in time-space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_L1L2norm.txt");
	write_logfile(time, mydiags.L1L2, namelog, newfile);

	// Save L2-L2 norm in time-space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_L2L2norm.txt");
	write_logfile(time, mydiags.L2L2, namelog, newfile);

	// Save Linf-L2 norm in time-space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_LinfL2norm.txt");
	write_logfile(time, mydiags.LinfL2, namelog, newfile);

	// Save Linf-norm in space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_Linfnorm.txt");
	write_logfile(time, mydiags.Linf, namelog, newfile);

	// Save L1-Linf norm in time-space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_L1Linfnorm.txt");
	write_logfile(time, mydiags.L1Linf, namelog, newfile);

	// Save L2-Linf norm in time-space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_L2Linfnorm.txt");
	write_logfile(time, mydiags.L2Linf, namelog, newfile);

	// Save Linf-Linf norm in time-space
	strcpy(namelog, prefixfile);
	strcat(namelog, "_LinfLinfnorm.txt");
	write_logfile(time, mydiags.LinfLinf, namelog, newfile);

	// Save Max(f)
	strcpy(namelog, prefixfile);
	strcat(namelog, "_maxf.txt");
	write_logfile(time, mydiags.maxf, namelog, newfile);

	// Save Min(f)
	strcpy(namelog, prefixfile);
	strcat(namelog, "_minf.txt");
	write_logfile(time, mydiags.minf, namelog, newfile);

	// Save Entropy
	strcpy(namelog, prefixfile);
	strcat(namelog, "_entropy.txt");
	write_logfile(time, mydiags.entropy, namelog, newfile);

	if (!(parameters.condensation)){
		// Save Relative entropy
		strcpy(namelog, prefixfile);
		strcat(namelog, "_rel_entropy.txt");
		write_logfile(time, mydiags.rel_entropy, namelog, newfile);
	}

	// Save 4th order momentum
	strcpy(namelog, prefixfile);
	strcat(namelog, "_momentum4.txt");
	write_logfile(time, mydiags.mom4, namelog, newfile);

	// Save 6th order momentum
	strcpy(namelog, prefixfile);
	strcat(namelog, "_momentum6.txt");
	write_logfile(time, mydiags.mom6, namelog, newfile);
	
	if (parameters.exact_solution){
		// Save L1-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1normexact.txt");
		write_logfile(time, mydiags.L1ex, namelog, newfile);
	
		// Save L2-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2normexact.txt");
		write_logfile(time, mydiags.L2ex, namelog, newfile);
		
		// Save Linf-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_Linfnormexact.txt");
		write_logfile(time, mydiags.Linfex, namelog, newfile);
		
		// Save L1-L1-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1L1normexact.txt");
		write_logfile(time, mydiags.L1L1ex, namelog, newfile);
	
		// Save L2-L1-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2L1normexact.txt");
		write_logfile(time, mydiags.L2L1ex, namelog, newfile);
		
		// Save Linf-L1-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_LinfL1normexact.txt");
		write_logfile(time, mydiags.LinfL1ex, namelog, newfile);
		
		// Save L1-L2-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1L2normexact.txt");
		write_logfile(time, mydiags.L1L2ex, namelog, newfile);
		
		// Save L2-L2-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2L2normexact.txt");
		write_logfile(time, mydiags.L2L2ex, namelog, newfile);
		
		// Save Linf-L2-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_LinfL2normexact.txt");
		write_logfile(time, mydiags.LinfL2ex, namelog, newfile);
		
		// Save L1-Linf-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1Linfnormexact.txt");
		write_logfile(time, mydiags.L1Linfex, namelog, newfile);
		
		// Save L2-Linf-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2Linfnormexact.txt");
		write_logfile(time, mydiags.L2Linfex, namelog, newfile);
		
		// Save Linf-Linf-norm of exact solution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_LinfLinfnormexact.txt");
		write_logfile(time, mydiags.LinfLinfex, namelog, newfile);
		
		// Save L1-error in space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1error.txt");
		write_logfile(time, mydiags.errL1, namelog, newfile);

		// Save L1-L1 error in time-space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1L1error.txt");
		write_logfile(time, mydiags.errL1L1, namelog, newfile);

		// Save L2-L1 error in time-space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_errL2L1norm.txt");
		write_logfile(time, mydiags.errL2L1, namelog, newfile);

		// Save Linf-L1 error in time-space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_LinfL1error.txt");
		write_logfile(time, mydiags.errLinfL1, namelog, newfile);

		// Save L2-error in space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2error.txt");
		write_logfile(time, mydiags.errL2, namelog, newfile);

		// Save L1-L2 error in time-space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1L2error.txt");
		write_logfile(time, mydiags.errL1L2, namelog, newfile);

		// Save L2-L2 error in time-space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2L2error.txt");
		write_logfile(time, mydiags.errL2L2, namelog, newfile);

		// Save Linf-L2 error in time-space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_LinfL2error.txt");
		write_logfile(time, mydiags.errLinfL2, namelog, newfile);

		// Save Linf-error in space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_Linferror.txt");
		write_logfile(time, mydiags.errLinf, namelog, newfile);

		// Save L1-Linf error in time-space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1Linferror.txt");
		write_logfile(time, mydiags.errL1Linf, namelog, newfile);

		// Save L2-Linf error in time-space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2Linferror.txt");
		write_logfile(time, mydiags.errL2Linf, namelog, newfile);

		// Save Linf-Linf error in time-space
		strcpy(namelog, prefixfile);
		strcat(namelog, "_LinfLinferror.txt");
		write_logfile(time, mydiags.errLinfLinf, namelog, newfile);
	}

	if (!(parameters.condensation)){
		// Save L1-norm of limit distribution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1normlimit.txt");
		write_logfile(time, mydiags.L1limit, namelog, newfile);
	
		// Save L2-norm of limit distribution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2normlimit.txt");
		write_logfile(time, mydiags.L2limit, namelog, newfile);
		
		// Save Linf-norm of limit distribution
		strcpy(namelog, prefixfile);
		strcat(namelog, "_Linfnormlimit.txt");
		write_logfile(time, mydiags.Linflimit, namelog, newfile);

		// Save L1-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1normrelaxation.txt");
		write_logfile(time, mydiags.relaxerrL1, namelog, newfile);
	
		// Save L2-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2normrelaxation.txt");
		write_logfile(time, mydiags.relaxerrL2, namelog, newfile);
		
		// Save Linf-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_Linfnormrelaxation.txt");
		write_logfile(time, mydiags.relaxerrLinf, namelog, newfile);
	
		// Save L1-L1-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1L1normrelaxation.txt");
		write_logfile(time, mydiags.relaxerrL1L1, namelog, newfile);
	
		// Save L1-L2-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1L2normrelaxation.txt");
		write_logfile(time, mydiags.relaxerrL1L2, namelog, newfile);

		// Save L1-Linf-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L1Linfnormrelaxation.txt");
		write_logfile(time, mydiags.relaxerrL1Linf, namelog, newfile);

		// Save L2-L1-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2L1normrelaxation.txt");
		write_logfile(time, mydiags.relaxerrL2L1, namelog, newfile);
	
		// Save L2-L2-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2L2normrelaxation.txt");
		write_logfile(time, mydiags.relaxerrL2L2, namelog, newfile);

		// Save L2-Linf-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_L2Linfnormrelaxation.txt");
		write_logfile(time, mydiags.relaxerrL2Linf, namelog, newfile);

		// Save Linf-L1-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_LinfL1normrelaxation.txt");
		write_logfile(time, mydiags.relaxerrLinfL1, namelog, newfile);
	
		// Save Linf-L2-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_LinfL2normrelaxation.txt");
		write_logfile(time, mydiags.relaxerrLinfL2, namelog, newfile);

		// Save Linf-Linf-norm of relaxation
		strcpy(namelog, prefixfile);
		strcat(namelog, "_LinfLinfnormrelaxation.txt");
		write_logfile(time, mydiags.relaxerrLinfLinf, namelog, newfile);
	}
}


int integrity_test(const char *namefile, int ndim, int type_collision, int type_qparticles, int exact_solution){
	int type_omega, exists_type_omega, condensation;
	double m0, zfug;
	int exists = (1==1);
	
	
	exists = (exists && search_in_file(namefile, "distribution/L"));
	exists = (exists && search_in_file(namefile, "distribution/Time"));
	exists = (exists && search_in_file(namefile, "distribution/Cut-off"));
	
	exists = (exists && search_in_file(namefile, "distribution/f"));
	exists = (exists && search_in_file(namefile, "distribution/nvx"));
	exists = (exists && search_in_file(namefile, "distribution/nvy"));
	if (ndim == DIM3)
		exists = (exists && search_in_file(namefile, "distribution/nvz"));
	
	if (type_collision == QUANTUM_COLLISIONS){
		exists = (exists && search_in_file(namefile, "distribution/Fugacity"));

		if (type_qparticles == BOSONS){
			if (ndim == DIM3){
				exists = (exists && search_in_file(namefile, "distribution/Critical_mass"));
				if (exists)
					read_0d_double(namefile, "distribution", "Critical_mass", &m0);
				else{
					// The critical mass namefield should be saved for any 3D bosons case. So the integrity test failed.
					return exists;
				}
				condensation = (m0 > 0.);
			}
			else{
				// No critical mass to read for the 2D bosons case
				condensation = (1==0);
			}
		}
		else{
			// Fermi-Dirac quantum degeneracy is characterized by artificial setting z < 0 at present time
			if (exists)
				read_0d_double(namefile, "distribution", "Fugacity", &zfug);
			else{
				// The fugacity namefield should be saved for any quantum case. So the integrity test failed.
				return exists;
			}
			condensation = (zfug < 0.);
			m0 = -1.;
		}
	}
	else{
		condensation = (1==0);
		m0 = -1.;
	}
		
	if (!condensation)
		exists = (exists && search_in_file(namefile, "distribution/f_limit"));
	
	if (exact_solution)
		exists = (exists && search_in_file(namefile, "distribution/f_exact"));
	
	exists = (exists && search_in_file(namefile, "distribution/type_omega"));
	if (exists){
		read_0d_int(namefile, "distribution", "type_omega", &type_omega);
		if (type_omega != OMEGA_IS_NOTUSED){
			exists = (exists && search_in_file(namefile, "distribution/G"));
			exists = (exists && search_in_file(namefile, "distribution/omega"));
			
			if (exact_solution)
				exists = (exists && search_in_file(namefile, "distribution/G_exact"));
			
			// Test G_limit for any case without quantum degeneracy
			if (!condensation)
				exists = (exists && search_in_file(namefile, "distribution/G_limit"));
		}
	}
	else{
		// The type_omega namefield should be saved for any simulation. So the integrity test failed.
		return exists;
	}
	
	exists = (exists && search_in_file(namefile, "distribution/L"));
	exists = (exists && search_in_file(namefile, "distribution/Time"));
	exists = (exists && search_in_file(namefile, "distribution/Cut-off"));
	
	exists = (exists && search_in_file(namefile, "distribution/Mass"));

	exists = (exists && search_in_file(namefile, "distribution/Average_ux"));
	exists = (exists && search_in_file(namefile, "distribution/Average_uy"));
	if (ndim == DIM3)
		exists = (exists && search_in_file(namefile, "distribution/Average_uz"));

	exists = (exists && search_in_file(namefile, "distribution/Stress_energy_xx"));
	exists = (exists && search_in_file(namefile, "distribution/Stress_energy_xy"));
	exists = (exists && search_in_file(namefile, "distribution/Stress_energy_yx"));
	exists = (exists && search_in_file(namefile, "distribution/Stress_energy_yy"));
	if (ndim == DIM3){
		exists = (exists && search_in_file(namefile, "distribution/Stress_energy_xz"));
		exists = (exists && search_in_file(namefile, "distribution/Stress_energy_yz"));
		exists = (exists && search_in_file(namefile, "distribution/Stress_energy_zx"));
		exists = (exists && search_in_file(namefile, "distribution/Stress_energy_zy"));
		exists = (exists && search_in_file(namefile, "distribution/Stress_energy_zz"));
	}
	
	exists = (exists && search_in_file(namefile, "distribution/Kinetic_energy"));
	exists = (exists && search_in_file(namefile, "distribution/Internal_energy"));
	exists = (exists && search_in_file(namefile, "distribution/Temperature"));
	
	exists = (exists && search_in_file(namefile, "distribution/Entropy"));
	if (!condensation)
		exists = (exists && search_in_file(namefile, "distribution/Relative_entropy"));
		
	exists = (exists && search_in_file(namefile, "distribution/4th_order_momentum"));
	exists = (exists && search_in_file(namefile, "distribution/6th_order_momentum"));
	
	exists = (exists && search_in_file(namefile, "distribution/Min_f"));
	exists = (exists && search_in_file(namefile, "distribution/Max_f"));
	
	exists = (exists && search_in_file(namefile, "distribution/L1_norm_space"));
	exists = (exists && search_in_file(namefile, "distribution/L2_norm_space"));
	exists = (exists && search_in_file(namefile, "distribution/Linf_norm_space"));
	
	exists = (exists && search_in_file(namefile, "distribution/L1-L1_norm_time-space"));
	exists = (exists && search_in_file(namefile, "distribution/L1-L2_norm_time-space"));
	exists = (exists && search_in_file(namefile, "distribution/L1-Linf_norm_time-space"));
	exists = (exists && search_in_file(namefile, "distribution/L2-L1_norm_time-space"));
	exists = (exists && search_in_file(namefile, "distribution/L2-L2_norm_time-space"));
	exists = (exists && search_in_file(namefile, "distribution/L2-Linf_norm_time-space"));
	exists = (exists && search_in_file(namefile, "distribution/Linf-L1_norm_time-space"));
	exists = (exists && search_in_file(namefile, "distribution/Linf-L2_norm_time-space"));
	exists = (exists && search_in_file(namefile, "distribution/Linf-Linf_norm_time-space"));
	
	if (!condensation){
		exists = (exists && search_in_file(namefile, "distribution/L1_norm_space_flimit"));
		exists = (exists && search_in_file(namefile, "distribution/L2_norm_space_flimit"));
		exists = (exists && search_in_file(namefile, "distribution/Linf_norm_space_flimit"));
		
		exists = (exists && search_in_file(namefile, "distribution/L1_norm_space_relaxation"));
		exists = (exists && search_in_file(namefile, "distribution/L2_norm_space_relaxation"));
		exists = (exists && search_in_file(namefile, "distribution/Linf_norm_space_relaxation"));
	
		exists = (exists && search_in_file(namefile, "distribution/L1-L1_norm_time-space_relaxation"));
		exists = (exists && search_in_file(namefile, "distribution/L1-L2_norm_time-space_relaxation"));
		exists = (exists && search_in_file(namefile, "distribution/L1-Linf_norm_time-space_relaxation"));
		exists = (exists && search_in_file(namefile, "distribution/L2-L1_norm_time-space_relaxation"));
		exists = (exists && search_in_file(namefile, "distribution/L2-L2_norm_time-space_relaxation"));
		exists = (exists && search_in_file(namefile, "distribution/L2-Linf_norm_time-space_relaxation"));
		exists = (exists && search_in_file(namefile, "distribution/Linf-L1_norm_time-space_relaxation"));
		exists = (exists && search_in_file(namefile, "distribution/Linf-L2_norm_time-space_relaxation"));
		exists = (exists && search_in_file(namefile, "distribution/Linf-Linf_norm_time-space_relaxation"));
		
		exists = (exists && search_in_file(namefile, "distribution/Mass_limit"));

		exists = (exists && search_in_file(namefile, "distribution/Average_ux_limit"));
		exists = (exists && search_in_file(namefile, "distribution/Average_uy_limit"));
		if (ndim == DIM3)
			exists = (exists && search_in_file(namefile, "distribution/Average_uz_limit"));
		
		exists = (exists && search_in_file(namefile, "distribution/Temperature_limit"));
		
		if (type_collision == QUANTUM_COLLISIONS){
			exists = (exists && search_in_file(namefile, "distribution/Fugacity_limit"));
			if ((type_qparticles == BOSONS) && (ndim == DIM3))
				exists = (exists && search_in_file(namefile, "distribution/Critical_mass_limit"));
		}
	}
	
	if (exact_solution){
		exists = (exists && search_in_file(namefile, "distribution/L1_norm_exact_space"));
		exists = (exists && search_in_file(namefile, "distribution/L2_norm_exact_space"));
		exists = (exists && search_in_file(namefile, "distribution/Linf_norm_exact_space"));
		
		exists = (exists && search_in_file(namefile, "distribution/L1-L1_norm_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/L1-L2_norm_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/L1-Linf_norm_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/L2-L1_norm_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/L2-L2_norm_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/L2-Linf_norm_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/Linf-L1_norm_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/Linf-L2_norm_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/Linf-Linf_norm_exact_time-space"));
		
		exists = (exists && search_in_file(namefile, "distribution/L1_error_exact_space"));
		exists = (exists && search_in_file(namefile, "distribution/L2_error_exact_space"));
		exists = (exists && search_in_file(namefile, "distribution/Linf_error_exact_space"));
		
		exists = (exists && search_in_file(namefile, "distribution/L1-L1_error_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/L1-L2_error_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/L1-Linf_error_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/L2-L1_error_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/L2-L2_error_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/L2-Linf_error_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/Linf-L1_error_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/Linf-L2_error_exact_time-space"));
		exists = (exists && search_in_file(namefile, "distribution/Linf-Linf_error_exact_time-space"));
	}
	
	return exists;
}



void cut_logfiles(const char *prefixfile, int nlines){
	struct stat mystat;
	char **namefiles;
	int K = 76;
	int k = 0;
	char xyz[5];
	char **norms;
	strcpy(xyz, "xyz");
	norms = malloc(sizeof(char*)*3);
	norms[0] = allocation_1d_char(4);
	norms[1] = allocation_1d_char(4);
	norms[2] = allocation_1d_char(6);
	sprintf(norms[0], "L1");
	sprintf(norms[1], "L2");
	sprintf(norms[2], "Linf");
	
	namefiles = malloc(sizeof(char*)*K);
	for (int i = 0; i < K; i++)
		namefiles[i] = allocation_1d_char(MAXSTRINGLENGTH);

	sprintf(namefiles[k], "%s_mass.txt", prefixfile);
	k++;
	
	for (int d = 0; d <= 2; d++){
		sprintf(namefiles[k], "%s_u%c.txt", prefixfile, xyz[d]);
		k++;
	}
	
	for (int d = 0; d <= 2; d++){
		for (int dd = 0; dd <= 2; dd++){
			sprintf(namefiles[k], "%s_s%c%c.txt", prefixfile, xyz[d], xyz[dd]);
			k++;
		}
	}
	sprintf(namefiles[k], "%s_ekin.txt", prefixfile);
	k++;
	
	sprintf(namefiles[k], "%s_fugacity.txt", prefixfile);
	k++;
	
	sprintf(namefiles[k], "%s_critical_mass.txt", prefixfile);
	k++;
	
	sprintf(namefiles[k], "%s_temperature.txt", prefixfile);
	k++;
	
	sprintf(namefiles[k], "%s_omega.txt", prefixfile);
	k++;

	sprintf(namefiles[k], "%s_eint.txt", prefixfile);
	k++;
	
	sprintf(namefiles[k], "%s_maxf.txt", prefixfile);
	k++;
	
	sprintf(namefiles[k], "%s_minf.txt", prefixfile);
	k++;
	
	sprintf(namefiles[k], "%s_entropy.txt", prefixfile);
	k++;
	
	sprintf(namefiles[k], "%s_rel_entropy.txt", prefixfile);
	k++;
	
	sprintf(namefiles[k], "%s_momentum4.txt", prefixfile);
	k++;
	
	sprintf(namefiles[k], "%s_momentum6.txt", prefixfile);
	k++;
	
	for (int n = 0; n < 3; n++){
		sprintf(namefiles[k], "%s_%snorm.txt", prefixfile, norms[n]);
		k++;
		for (int nn = 0; nn < 3; nn++){
			sprintf(namefiles[k], "%s_%s%snorm.txt", prefixfile, norms[n], norms[nn]);
			k++;
		}
		sprintf(namefiles[k], "%s_%snormexact.txt", prefixfile, norms[n]);
		k++;
		for (int nn = 0; nn < 3; nn++){
			sprintf(namefiles[k], "%s_%s%snormexact.txt", prefixfile, norms[n], norms[nn]);
			k++;
		}
		sprintf(namefiles[k], "%s_%serror.txt", prefixfile, norms[n]);
		k++;
		for (int nn = 0; nn < 3; nn++){
			sprintf(namefiles[k], "%s_%s%serror.txt", prefixfile, norms[n], norms[nn]);
			k++;
		}
		
		sprintf(namefiles[k], "%s_%snormlimit.txt", prefixfile, norms[n]);
		k++;
		
		sprintf(namefiles[k], "%s_%snormrelaxation.txt", prefixfile, norms[n]);
		k++;
		for (int nn = 0; nn < 3; nn++){
			sprintf(namefiles[k], "%s_%s%snormrelaxation.txt", prefixfile, norms[n], norms[nn]);
			k++;
		}
	}
	
	for (int i = 0; i < K; i++){
		if (stat(namefiles[i], &mystat) == 0)
			cut_logfile(namefiles[i], nlines);
		//else
			//printf("File %s not found\n", namefiles[i]);
	}
	
	for (int i = 0; i < K; i++)
		free_1d_char(namefiles[i]);
	free(namefiles);
	free_1d_char(norms[0]);
	free_1d_char(norms[1]);
	free_1d_char(norms[2]);
	free(norms);

}
