#include "./src/core/exact_solutions.h"


double classical_BKW_2d(double vx, double vy, double sigma, double time){
	double vsquare = (vx*vx+vy*vy);
	double S = 1. - 0.5*exp(-time*sigma*sigma/8.);
	double pi = 4.*atan(1.);
	
	return (2.*S - 1. + (1.-S)/(2.*S) * vsquare/(sigma*sigma)) * exp(-vsquare/(2.*S*sigma*sigma)) / (2.*pi*S*S*sigma*sigma);
}


double dt_classical_BKW_2d(double vx, double vy, double sigma, double time){
	double vsquare = (vx*vx+vy*vy);
	double S = 1. - 0.5*exp(-time*sigma*sigma/8.);
	double dS = exp(-time*sigma*sigma/8.)*sigma*sigma/16.;
	double pi = 4.*atan(1.);
	
	return -exp(-vsquare/(2.*S*sigma*sigma)) * dS * (S-1.) * (8.*S*S*pow(sigma,4) - 8.*S*sigma*sigma*vsquare + vsquare*vsquare) / (8.*pi*pow(S,5)*pow(sigma,6));
}


void exact_solution(double *Gexact, double time, double omega, double *u, PARAMETERS parameters){
	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz,1);
	
	double v[3];
	double xi[3];

	int k;
	
	int kvxabs, kvyabs, kvzabs;
	
	double lambda, mu, myomega;
	
	double fexact;
	
	double pi = 4.*atan(1.);
	double dxix, dxiy, dxiz;
	dxix = 2.*pi/MVX;
	dxiy = 2.*pi/MVY;
	if (parameters.ndim == DIM3)
		dxiz = 2.*pi/MVZ;
	else
		dxiz = 1.;
	
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		lambda = 0.;
		myomega = 1.;
		mu = pow(parameters.L/pi, parameters.ndim+1);
	}
	else{
		lambda = 1.;
		myomega = omega;
		mu = 1.;
	}
	
	switch (parameters.type_f0){
		case CLASSICAL_BKW :
			if (parameters.ndim != DIM2)
				error("Classical BKW test case is only defined for 2D classical collisions");
			#pragma omp parallel for private(k, kvxabs, kvyabs, xi, v, fexact) shared(NVXLOC, NVYLOC, parameters, MVX, MVY, IVX0LOC, IVY0LOC, dxix, dxiy, pi, myomega, lambda, mu, u, time, Gexact)
			for (k = 0; k < NVXLOC*NVYLOC; k++){
#ifdef key_force_vectorize
				kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
				kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
#else
				// k = kvyloc+NVYLOC*kvxloc
				kvxabs = k/NVYLOC;
				kvyabs = k-kvxabs*NVYLOC;
				kvxabs += IVX0LOC;
				kvyabs += IVY0LOC;
				if (kvxabs >= MVX/2)
					kvxabs -= MVX;
				if (kvyabs >= MVY/2)
					kvyabs -= MVY;
#endif
				xi[0] = kvxabs * dxix;
				xi[1] = kvyabs * dxiy;
				
				v[0] = parameters.L*xi[0]/(pi*myomega) + lambda*u[0];
				v[1] = parameters.L*xi[1]/(pi*myomega) + lambda*u[1];
					
				fexact = classical_BKW_2d(v[0], v[1], parameters.sigma0, time);
				Gexact[k] = fexact / (mu * pow(pi*myomega/parameters.L, parameters.ndim+1));
			}
			break;
			
		default:
			error("Unknown exact solution");
	}
}
