#include "time_loop.h"


#ifdef key_cufft
extern void collision_Qtotal_cufft(double *, double *, double **, double **, double, PARAMETERS);
extern void collision_Qtotal_quick_convolution_cufft(double *, double *, double **, double **, double *, double, PARAMETERS);
#endif

double Linf_QG(double *QGloc, PARAMETERS parameters){
	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(1,parameters.nvz);
	int MV = MVX*MVY*MVZ;

	double pi = 4.*atan(1.);

	double LinfQGloc = fabs(QGloc[0]);
	double LinfQG = 0.;
	double contrib;
	int k;
	
	// Compute local contribution to Linf norm of Q(G)
	#pragma omp parallel for private(contrib, k) shared(NVLOC, QGloc, LinfQGloc)
	for (k = 0; k < NVLOC; k++){
		#pragma omp critical
		{
			contrib = fabs(QGloc[k]) - LinfQGloc;
			contrib *= (contrib > 0.);
			LinfQGloc += contrib;
		}
	}
#ifdef key_mpi
	MPI_Allreduce(&LinfQGloc, &LinfQG, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
#else
	LinfQG = LinfQGloc;
#endif

	return LinfQG;
}


void moms012_QG(double *mom0QG, double *mom1QG, double *mom2QG, double *QGloc, double *u, double omega, PARAMETERS parameters){
	double dxi, dxix, dxiy, dxiz;
	double xi[3];
	
	double vsquare;
	double v[3];
	
	double mom0QGloc, *mom1QGloc, mom2QGloc;
	
	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(1,parameters.nvz);
	int MV = MVX*MVY*MVZ;
	
	int kvxabs, kvyabs, kvzabs, itemp;

	double contrib_mom0;
	double contrib_mom1[3];
	double contrib_mom2;

	int k, d;

	double pi = 4.*atan(1.);
	
	double lambda, myomega, mu;
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		lambda = 0.;
		myomega = 1.;
		mu = pow(parameters.L/pi, parameters.ndim+1);
	}
	else{
		lambda = 1.;
		myomega = omega;
		mu = 1.;
	}
	
	dxix = 2.*pi/MVX;
	dxiy = 2.*pi/MVY;
	if (parameters.ndim == DIM3)
		dxiz = 2.*pi/MVZ;
	else
		dxiz = 1.;
	dxi = dxix*dxiy*dxiz;
	
	mom1QGloc = allocation_1d(parameters.ndim+1);

	// We compute the momenta of QGn
	mom0QGloc = 0.;
	for (int d = 0; d <= parameters.ndim; d++)
		mom1QGloc[d] = 0.;
	mom2QGloc = 0.;
	
	#pragma omp parallel for private(k, d, kvxabs, kvyabs, kvzabs, itemp, xi, v, vsquare, contrib_mom0, contrib_mom1, contrib_mom2)	shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, parameters, MVX, MVY, MVZ, IVX0LOC, IVY0LOC, IVZ0LOC, dxix, dxiy, dxiz, dxi, pi, myomega, lambda, u, mom0QGloc, mom1QGloc, mom2QGloc, QGloc)
	for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
		kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
		kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
		kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
		// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
		itemp = k/NVZLOC;
		// itemp now contains kvyloc+NVYLOC*kvxloc;
		kvzabs = k - itemp*NVZLOC;
		kvxabs = itemp/NVYLOC;
		kvyabs = itemp-kvxabs*NVYLOC;
											
		kvxabs += IVX0LOC;
		kvyabs += IVY0LOC;
		kvzabs += IVZ0LOC;
											
		if (kvxabs >= MVX/2)
			kvxabs -= MVX;
		if (kvyabs >= MVY/2)
			kvyabs -= MVY;
		if (kvzabs >= MVZ/2)
			kvzabs -= MVZ;
#endif
		xi[0] = kvxabs * dxix;
		xi[1] = kvyabs * dxiy;
		if (parameters.ndim == DIM3)
			xi[2] = kvzabs * dxiz;
		v[0] = parameters.L*xi[0]/(pi*myomega) + lambda*u[0];
		v[1] = parameters.L*xi[1]/(pi*myomega) + lambda*u[1];
		if (parameters.ndim == DIM3)
			v[2] = parameters.L*xi[2]/(pi*myomega) + lambda*u[2];
		vsquare = 0.;
		for (d = 0; d <= parameters.ndim; d++)
			vsquare += pow(v[d]-u[d], 2);
		
		contrib_mom0 = dxi * QGloc[k];
		for (d = 0; d <= parameters.ndim; d++)
			contrib_mom1[d] = dxi * QGloc[k] * v[d];
		contrib_mom2 = 0.5 * dxi * QGloc[k] * vsquare;

		#pragma omp atomic
		mom0QGloc += contrib_mom0;
		for (d = 0; d <= parameters.ndim; d++){
			#pragma omp atomic
			mom1QGloc[d] += contrib_mom1[d];
		}
		#pragma omp atomic
		mom2QGloc += contrib_mom2;
	}

#ifdef key_mpi
	MPI_Allreduce(&mom0QGloc, mom0QG, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	MPI_Allreduce(mom1QGloc, mom1QG, parameters.ndim+1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	MPI_Allreduce(&mom2QGloc, mom2QG, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#else
	*mom0QG = mom0QGloc;
	for (int d = 0; d <= parameters.ndim; d++)
		mom1QG[d] = mom1QGloc[d];
	*mom2QG = mom2QGloc;
#endif

	free_1d(mom1QGloc);
}


void time_loop(DIAGS *diags_n, PARAMETERS parameters){
	char buf[MAXSTRINGLENGTH];
	FILE *f;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int NVLOC = NVXLOC*NVYLOC*NVZLOC;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(1,parameters.nvz);
	int MV = MVX*MVY*MVZ;
	
	int ndigits;
	char *namefile_data;
	char prefixfile[MAXSTRINGLENGTH];
	
	double time, dt;
	int a = 0;
	
	double lambda, mu, C;
	double pi = 4.*atan(1.);

	double **alpha, **alphaprime;
	//double ***Amode, ***Bmode, ***Cmode, *Wmode;
	double **xitilde, **sigma, *W;
	int P;
	
	double *Gloclimitn;
	double *QGloclimitn;

	double *Gloclimitstar;
	double *QGloclimitstar;

	double *Gloclimit;
	double *QGloclimit;

	double *Glocn;
	double *QGlocn;

	double *Glocstar;
	double *QGlocstar;

	double *Glocexn;

	double **Gv_loc;

	double rhon, rhostar, rhonp1;
	double *uavn, *uavstar, *uavnp1;
	double eintn, eintstar, eintnp1;
	double Tn, Tstar, Tnp1;
	double zn, zstar, znp1;
	double m0n, m0star, m0np1;
	double omegan, omegastar, omeganp1;

	double dxix, dxiy, dxiz;
	double xi[3];

	int kvxabs, kvyabs, kvzabs;
	int itemp;

	int k, d;

	double usquare;

	double mom0QG, *mom1QG, mom2QG;
	double mom0QGlimit, *mom1QGlimit, mom2QGlimit;
	double LinfQG;

	double contrib;

	double *div_G_loc;
	
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		lambda = 0.;
		mu = pow(parameters.L/pi, parameters.ndim+1);
	}
	else{
		lambda = 1.;
		mu = 1.;
	}

	// Set G^n
	switch (parameters.ndim){
		case DIM2:
			Glocn = diags_n->Gloc_2d[0];
			break;
		case DIM3:
			Glocn = diags_n->Gloc_3d[0][0];
			break;
		default:
			error("Unknown dimensionality");
	}

	// Set Gexact
	if (parameters.exact_solution){
		switch (parameters.ndim){
			case DIM2:
				Glocexn = diags_n->Glocexact_2d[0];
				break;
			case DIM3:
				Glocexn = diags_n->Glocexact_3d[0][0];
				break;
			default:
				error("Unknown dimensionality");
		}
	}

	// Allocate Q(G^n)
	QGlocn = allocation_1d(NVLOC);

	// Allocate Q(Glimit^n) (set to 0 if Glimit cannot be discretized or if residual method is not considered)
	QGloclimitn = allocation_1d(NVLOC);

	// Associate Q(Glimit) to Q(Glimit^n) for non-rescaling method
	QGloclimit = QGloclimitn;

	// Allocate G* and Q(G*) (only for RK2 method)
	if (parameters.time_method == RK2){
		Glocstar = allocation_1d(NVLOC);
		QGlocstar = allocation_1d(NVLOC);
	}

	// Allocate Q(Glimit^*) (only for RK2 method)
	// (set to 0 if Glimit cannot be discretized or if residual method is not considered)
	if ((parameters.time_method == RK2) && (!(parameters.condensation)))
		Gloclimitstar = allocation_1d(NVLOC);
	if ((parameters.time_method == RK2) && (parameters.type_omega != OMEGA_IS_NOTUSED))
		QGloclimitstar = allocation_1d(NVLOC);

	if (parameters.type_omega != OMEGA_IS_NOTUSED){
		uavn = allocation_1d(parameters.ndim+1);
		uavnp1 = allocation_1d(parameters.ndim+1);
		if (parameters.time_method == RK2)
			uavstar = allocation_1d(parameters.ndim+1);
		Gv_loc = allocation_2d(parameters.ndim+1, NVLOC);
		div_G_loc = allocation_1d(NVLOC);
		mom1QG = allocation_1d(parameters.ndim+1);
		mom1QGlimit = allocation_1d(parameters.ndim+1);
	}

	// Fill alphaloc and alphaprimeloc
	message("Compute collision modes... ");
	if (parameters.force_convolution){
		/*Amode = allocation_3d(parameters.mint, NVLOC, 2);
		Bmode = allocation_3d(parameters.mint, NVLOC, 2);
		Cmode = allocation_3d(parameters.mint, NVLOC, 2);
		Wmode = allocation_1d(parameters.mint);
		compute_ABCW_force_convolution(Wmode, Amode, Bmode, Cmode, parameters);*/
      	compute_points_weights_force_convolution(parameters.cutoff, parameters.P1_int, parameters.P2_int, parameters.P3_int, parameters.ndim+1, &xitilde, &sigma, &W, &P);

	}
	else{
		alpha = allocation_2d(parameters.mint, NVLOC);
		alphaprime = allocation_2d(parameters.mint, NVLOC);
		expanded_collision_modes(alpha, alphaprime, parameters);
	}
	message("done\n");
	
	
	ndigits = fmin((int)(log10(ceil(parameters.nt)))+1, INT_MAX);
	namefile_data = allocation_1d_char(MAXSTRINGLENGTH+5+ndigits);
	strcpy(prefixfile, "");
	strcat(prefixfile, parameters.outputdir);
	strcat(prefixfile, "/");
	strcat(prefixfile, parameters.prefix);
	
	time = parameters.Tinit;
	if (parameters.resume)
		dt = parameters.Tfinal/parameters.nt;
	else
		dt = (parameters.Tfinal-time)/parameters.nt;
	
	// Verify if the code has to be prematurely stopped
	if (parameters.condensation){
		if (parameters.type_qparticles == BOSONS){
			// We expect a blow-up of the distribution as t -> infinity
			if (diags_n->Linf > 1./THRESHOLD)
				a++;
			else
				a = 0;
			if (a > 5)
				exitsuccess("Blow-up of the numerical solution");
		}
		else{
			// We expect a quantum degeneracy of the distribution
			// WARNING: at present time, we cannot identify the limit distribution
			LinfQG = Linf_QG(QGlocn, parameters);
			if (LinfQG < THRESHOLD)
				a++;
			else
				a = 0;
			if (a > 5)
				exitsuccess("Steady state reached");
		}
	}
	else{
		if (diags_n->relaxerrL1 < THRESHOLD)
			a++;
		else
			a = 0;
		if (a > 5)
			exitsuccess("Steady state reached");
	}
	
	if (!(parameters.resume)){
		// Save initial state
		switch(parameters.output_fmt){
			case FMT_BP:
#if defined(key_adios) || defined(key_adios_nompi)
				sprintf(namefile_data, "%s_%0*d.bp", prefixfile, ndigits, parameters.n_init);
				break;
#else
				error("KINEBEC has not been compiled with Adios dependency");
#endif
			case FMT_H5:
#if defined(key_phdf5) || defined(key_hdf5)
				sprintf(namefile_data, "%s_%0*d.h5", prefixfile, ndigits, parameters.n_init);
				break;
#else
				error("KINEBEC has not been compiled with Hdf5 dependency");
#endif
			default:
				error("Unknown output format (only .bp or .h5 are allowed)");
		}
		save_all_data(*diags_n, time, dt, namefile_data, parameters);
	}

	if (!(parameters.condensation)){
		switch (parameters.ndim){
			case DIM2:
				Gloclimitn = diags_n->Gloclimit_2d[0];
				break;
			case DIM3:
				Gloclimitn = diags_n->Gloclimit_3d[0][0];
				break;
			default:
				error("Unknown dimensionality");
		}
		Gloclimit = Gloclimitn;

		if (parameters.use_residual){
#ifdef key_cufft
			if (parameters.force_convolution)
				collision_Qtotal_quick_convolution_cufft(Gloclimitn, QGloclimitn, xitilde, sigma, W, diags_n->omega, parameters);
			else
				collision_Qtotal_cufft(Gloclimitn, QGloclimitn, alpha, alphaprime, diags_n->omega, parameters);
#elif defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
			if (parameters.force_convolution)
				collision_Qtotal_quick_convolution_fftw(Gloclimitn, QGloclimitn, xitilde, sigma, W, diags_n->omega, parameters);
				//collision_Qtotal_force_convolution_fftw(Gloclimitn, QGloclimitn, Amode, Bmode, Cmode, Wmode, diags_n->omega, parameters);
			else
				collision_Qtotal_fftw(Gloclimitn, QGloclimitn, alpha, alphaprime, diags_n->omega, parameters);
#endif
		}
		else{
			// If residual method is not considered, set Q(Glimit^n) to 0
			#pragma omp parallel for private(k) shared(NVLOC, QGloclimitn)
			for (k = 0; k < NVLOC; k++)
				QGloclimitn[k] = 0.;
		}
	}
	else{
		// Define dummy Glimit^n and Q(Glimit^n) filled with 0
		Gloclimitn = allocation_1d(NVLOC);
		#pragma omp parallel for private(k) shared(NVLOC, Gloclimitn, QGloclimitn)
		for (k = 0; k < NVLOC; k++){
			Gloclimitn[k] = 0.;
			QGloclimitn[k] = 0.;
		}
		Gloclimit = Gloclimitn;
	}
	
	// For RK2 method, initialize Q(Glimit^*) to Q(Glimit^n)
	// It will be 0 if the residual method is not used or if there is a condensate
	if ((parameters.time_method == RK2) && (parameters.type_omega != OMEGA_IS_NOTUSED)){
		if ((parameters.use_residual) && (!(parameters.condensation))){
			#pragma omp parallel for private(k) shared(NVLOC, QGloclimitstar, QGloclimitn)
			for (k = 0; k < NVLOC; k++)
				QGloclimitstar[k] = QGloclimitn[k];
		}
		else{
			#pragma omp parallel for private(k) shared(NVLOC, QGloclimitstar)
			for (k = 0; k < NVLOC; k++)
				QGloclimitstar[k] = 0.;
		}
		
	}

	if (parameters.resume){
		if (parameters.n_init >= parameters.nt)
			message("No more time iterations to do.\n");
		else{
			sprintf(buf, "Resuming time loop at time %e...\n", time);
			message(buf);
		}
	}
	else
		message("Starting time loop...\n");

	for (int n = parameters.n_init; n < parameters.nt; n++){
		// Compute the collision term Q(G^n)
#ifdef key_cufft
		if (parameters.force_convolution)
			collision_Qtotal_quick_convolution_cufft(Glocn, QGlocn, xitilde, sigma, W, diags_n->omega, parameters);
		else
			collision_Qtotal_cufft(Glocn, QGlocn, alpha, alphaprime, diags_n->omega, parameters);
#elif defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
		//printf("Compute QGlocn\n");
		if (parameters.force_convolution)
			collision_Qtotal_quick_convolution_fftw(Glocn, QGlocn, xitilde, sigma, W, diags_n->omega, parameters);
			//collision_Qtotal_force_convolution_fftw(Glocn, QGlocn, Amode, Bmode, Cmode, Wmode, diags_n->omega, parameters);
		else
			collision_Qtotal_fftw(Glocn, QGlocn, alpha, alphaprime, diags_n->omega, parameters);
#endif
		if (parameters.type_omega == OMEGA_IS_NOTUSED){
			// We have omega = 1, lambda = 0 and mu = (L/pi)^dv
			C = (1./parameters.knudsen) * mu * pow(pi*diags_n->omega/parameters.L, -parameters.gamma);

			switch(parameters.time_method){
				case EULER :
					// Apply Euler explicit method and get G^(n+1) = G^n + dt * C * (Q(G^n)-Q(Glimit))
					// Q(Glimit) = 0 if the residual method is not used or if there is a condensate
					#pragma omp parallel for private(k, contrib) shared(NVLOC, QGlocn, QGloclimit, Glocn, dt, C)
					for (k = 0; k < NVLOC; k++){
						/*contrib = dt * C*(QGlocn[k]-QGloclimit[k]);
						#pragma omp atomic
						Glocn[k] += contrib;*/
						Glocn[k] += dt * C*(QGlocn[k]-QGloclimit[k]);
					}
					
					break;

				case RK2:
					// Apply first fractional step of Runge-Kutta 2 method and get G* = G^n + dt/2 * C * (Q(G^n)-Q(Glimit))
					// Q(Glimit) = 0 if the residual method is not used or if there is a condensate
					#pragma omp parallel for private(k, contrib) shared(NVLOC, QGlocn, QGloclimit, Glocstar, Glocn, dt, C)
					for (k = 0; k < NVLOC; k++){
						/*contrib = Glocn[k] + 0.5*dt * C*(QGlocn[k]-QGloclimit[k]);
						Glocstar[k] = contrib;*/
						Glocstar[k] = Glocn[k] + 0.5*dt * C*(QGlocn[k]-QGloclimit[k]);
					}
					
					// Compute Q(G*) (reminder: omega = 1 here)
#ifdef key_cufft
					if (parameters.force_convolution)
						collision_Qtotal_quick_convolution_cufft(Glocstar, QGlocstar, xitilde, sigma, W, diags_n->omega, parameters);
					else
						collision_Qtotal_cufft(Glocstar, QGlocstar, alpha, alphaprime, diags_n->omega, parameters);
#elif defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
					//printf("Compute QGlocstar\n");
					if (parameters.force_convolution)
						collision_Qtotal_quick_convolution_fftw(Glocstar, QGlocstar, xitilde, sigma, W, diags_n->omega, parameters);
						//collision_Qtotal_force_convolution_fftw(Glocstar, QGlocstar, Amode, Bmode, Cmode, Wmode, diags_n->omega, parameters);
					else
						collision_Qtotal_fftw(Glocstar, QGlocstar, alpha, alphaprime, diags_n->omega, parameters);
#endif
					// Apply last step of Runge-Kutta 2 method and get G^(n+1) = G^n + dt * C * (Q(G*)-Q(Glimit))
					// Q(Glimit) = 0 if the residual method is not used or if there is a condensate
					#pragma omp parallel for private(k, contrib) shared(NVLOC, Glocn, dt, C, QGlocstar, QGloclimit)
					for (k = 0; k < NVLOC; k++){
						/*contrib = dt * C*(QGlocstar[k]-QGloclimit[k]);
						#pragma omp atomic
						Glocn[k] += contrib;*/
						Glocn[k] += dt * C*(QGlocstar[k]-QGloclimit[k]);
					}
					
					break;
				
				default:
					error("Unknown time method");
			}

			// Compute rho^(n+1), u^(n+1), e^(n+1), T^(n+1), z^(n+1), m0^(n+1)
			compute_diagnostic(diags_n, MASS, parameters);
			compute_diagnostic(diags_n, AVERAGE_U, parameters);
			compute_diagnostic(diags_n, INTERNAL_ENERGY, parameters);
			compute_diagnostic(diags_n, TEMPERATURE, parameters);
		}
		else{
			// We consider the time loop adapted for velocity rescaling method
			switch(parameters.time_method){
				case EULER :
					// Extract the momenta at time step tn
					rhon = diags_n->mass;
					for (d = 0; d <= parameters.ndim; d++)
						uavn[d] = diags_n->u_av[d];
					eintn = diags_n->eint;
					omegan = diags_n->omega;
					Tn = diags_n->temperature;
					if (parameters.type_collision == QUANTUM_COLLISIONS){
						zn = diags_n->fugacity;
						if ((parameters.ndim == DIM3) && (parameters.type_qparticles == BOSONS))
							m0n = diags_n->critical_mass;
					}

					// Compute the momenta of Q(G^n)
					moms012_QG(&mom0QG, mom1QG, &mom2QG, QGlocn, uavn, omegan, parameters);
					
					// Compute the momenta of Q(Glimit^n)
					// (only for residual methods without condensation)
					if ((parameters.use_residual) && (!(parameters.condensation)))
						moms012_QG(&mom0QGlimit, mom1QGlimit, &mom2QGlimit, QGloclimitn, uavn, omegan, parameters);
					else{
						// Set these momenta to 0
						mom0QGlimit = 0.;
						for (d = 0; d <= parameters.ndim; d++)
							mom1QGlimit[d] = 0.;
						mom2QGlimit = 0.;
					}

					// Compute rho^(n+1), u^(n+1), eint^(n+1)
					C = (1./parameters.knudsen) * pow(mu,2) * pow(pi*omegan/parameters.L, -parameters.gamma);
					rhonp1 = rhon + dt * C * (mom0QG-mom0QGlimit);
					for (d = 0; d <= parameters.ndim; d++)
						uavnp1[d] = (rhon*uavn[d] + dt * C * (mom1QG[d]-mom1QGlimit[d]))/rhonp1;
					eintnp1 = (rhon*eintn + dt * C * (mom2QG-mom2QGlimit))/rhonp1;

					// Compute T^(n+1), z^(n+1), m0^(n+1)
					switch (parameters.type_collision){
						case CLASSICAL_COLLISIONS:
							// The temperature is 2/d * eint
							Tnp1 = (2.*eintnp1)/(parameters.ndim+1.);
							break;
						case QUANTUM_COLLISIONS:
							// The temperature, fugacity and critical mass are obtained by solving a non-linear system with rho and eint as source terms
							Tnp1 = -1.;
							m0np1 = -1;
							find_fugacity_temperature_criticalmass_fixedpoint(rhonp1, &eintnp1, &Tnp1, &znp1, &m0np1, parameters.hbar, parameters.ndim+1, parameters.type_qparticles);
							break;
						default:
							error("Unknown type of collision");
					}

					// Compute an estimate of omega^(n+1)
					switch (parameters.type_omega){
						case OMEGA_IS_SQRTEKINORHO:
							usquare = 0.;
							for (d = 0; d <= parameters.ndim; d++)
								usquare += pow(uavnp1[d],2);
							omeganp1 = sqrt(Tnp1+usquare/(parameters.ndim+1.));
							break;
						case OMEGA_IS_SQRTTEMPERATURE:
							omeganp1 = sqrt(Tnp1);
							break;
						default:
							error("Something is wrong with time_loop");
					}

					// Compute the vector function in the divergence
					#pragma omp parallel for private(k, d, kvxabs, kvyabs, kvzabs, itemp, xi) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, parameters, MVX, MVY, MVZ, IVX0LOC, IVY0LOC, IVZ0LOC, dxix, dxiy, dxiz, Gv_loc, Glocn, omeganp1, omegan, dt, pi, uavnp1, uavn)
					for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
						kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
						kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
						kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
						// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
						itemp = k/NVZLOC;
						// itemp now contains kvyloc+NVYLOC*kvxloc;
						kvzabs = k - itemp*NVZLOC;
						kvxabs = itemp/NVYLOC;
						kvyabs = itemp-kvxabs*NVYLOC;
															
						kvxabs += IVX0LOC;
						kvyabs += IVY0LOC;
						kvzabs += IVZ0LOC;
															
						if (kvxabs >= MVX/2)
							kvxabs -= MVX;
						if (kvyabs >= MVY/2)
							kvyabs -= MVY;
						if (kvzabs >= MVZ/2)
							kvzabs -= MVZ;
#endif
						xi[0] = kvxabs * dxix;
						xi[1] = kvyabs * dxiy;
						if (parameters.ndim == DIM3)
							xi[2] = kvzabs * dxiz;
						
						for (d = 0; d <= parameters.ndim; d++)
							Gv_loc[d][k] = Glocn[k]*((omeganp1-omegan)/(dt*omegan) * xi[d] - (pi*omegan/parameters.L) * (uavnp1[d]-uavn[d])/dt);
					}
					
					// Compute the divergence term
#if defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
					compute_divergence_fftw(Gv_loc, div_G_loc, parameters);
#endif

					// Compute G^(n+1) = G^n + dt * C * (Q(G^n)-Q(Glimit^n)) - dt*div(Gvect^n)
					C = (1./parameters.knudsen) * mu * pow(pi*omegan/parameters.L, -parameters.gamma);
					#pragma omp parallel for private(k, contrib) shared(NVLOC, Glocn, dt, C, QGloclimitn, div_G_loc)
					for (k = 0; k < NVLOC; k++){
						/*contrib = dt * (C*(QGlocn[k]-QGloclimitn[k])-div_G_loc[k]);
						#pragma omp atomic
						Glocn[k] += contrib;*/
						Glocn[k] += dt * (C*(QGlocn[k]-QGloclimitn[k])-div_G_loc[k]);
					}

					break;
				
				case RK2:
					// Extract the momenta at time step tn
					rhon = diags_n->mass;
					for (d = 0; d <= parameters.ndim; d++)
						uavn[d] = diags_n->u_av[d];
					eintn = diags_n->eint;
					omegan = diags_n->omega;
					Tn = diags_n->temperature;
					if (parameters.type_collision == QUANTUM_COLLISIONS){
						zn = diags_n->fugacity;
						if ((parameters.ndim == DIM3) && (parameters.type_qparticles == BOSONS))
							m0n = diags_n->critical_mass;
					}
					
					// Compute the momenta of Q(G^n)
					moms012_QG(&mom0QG, mom1QG, &mom2QG, QGlocn, uavn, omegan, parameters);

					// Compute the momenta of Q(Glimit^n)
					// (only for residual methods without condensation)
					if ((parameters.use_residual) && (!(parameters.condensation)))
						moms012_QG(&mom0QGlimit, mom1QGlimit, &mom2QGlimit, QGloclimitn, uavn, omegan, parameters);
					else{
						// Set these momenta to 0
						mom0QGlimit = 0.;
						for (d = 0; d <= parameters.ndim; d++)
							mom1QGlimit[d] = 0.;
						mom2QGlimit = 0.;
					}

					// Compute rho^*, u^*, eint^*
					C = (1./parameters.knudsen) * pow(mu,2) * pow(pi*omegan/parameters.L, -parameters.gamma);
					rhostar = rhon + 0.5*dt * C * (mom0QG-mom0QGlimit);
					for (d = 0; d <= parameters.ndim; d++)
						uavstar[d] = (rhon*uavn[d] + 0.5*dt * C * (mom1QG[d]-mom1QGlimit[d]))/rhostar;
					eintstar = (rhon*eintn + 0.5*dt * C * (mom2QG-mom2QGlimit))/rhostar;

					// Compute T^*, z^*, m0^*
					switch (parameters.type_collision){
						case CLASSICAL_COLLISIONS:
							// The temperature is 2/d * eint
							Tstar = (2.*eintstar)/(parameters.ndim+1.);
							break;
						case QUANTUM_COLLISIONS:
							// The temperature, fugacity and critical mass are obtained by solving a non-linear system with rho and eint as source terms
							Tstar = -1.;
							m0star = -1;
							find_fugacity_temperature_criticalmass_fixedpoint(rhostar, &eintstar, &Tstar, &zstar, &m0star, parameters.hbar, parameters.ndim+1, parameters.type_qparticles);
							break;
						default:
							error("Unknown type of collision");
					}

					// Compute an estimate of omega^*
					switch (parameters.type_omega){
						case OMEGA_IS_SQRTEKINORHO:
							usquare = 0.;
							for (d = 0; d <= parameters.ndim; d++)
								usquare += pow(uavstar[d],2);
							omegastar = sqrt(Tstar+usquare/(parameters.ndim+1.));
							break;
						case OMEGA_IS_SQRTTEMPERATURE:
							omegastar = sqrt(Tstar);
							break;
						default:
							error("Something is wrong with time_loop");
					}

					// Compute the vector function in the divergence
					#pragma omp parallel for private(k, d, kvxabs, kvyabs, kvzabs, itemp, xi) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, MVX, MVY, MVZ, IVX0LOC, IVY0LOC, IVZ0LOC, parameters, dxix, dxiy, dxiz, Gv_loc, Glocn, omegastar, omegan, dt, pi, uavstar, uavn)
					for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
						kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
						kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
						kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
						// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
						itemp = k/NVZLOC;
						// itemp now contains kvyloc+NVYLOC*kvxloc;
						kvzabs = k - itemp*NVZLOC;
						kvxabs = itemp/NVYLOC;
						kvyabs = itemp-kvxabs*NVYLOC;
															
						kvxabs += IVX0LOC;
						kvyabs += IVY0LOC;
						kvzabs += IVZ0LOC;
															
						if (kvxabs >= MVX/2)
							kvxabs -= MVX;
						if (kvyabs >= MVY/2)
							kvyabs -= MVY;
						if (kvzabs >= MVZ/2)
							kvzabs -= MVZ;
#endif
						xi[0] = kvxabs * dxix;
						xi[1] = kvyabs * dxiy;
						if (parameters.ndim == DIM3)
							xi[2] = kvzabs * dxiz;
						
						for (d = 0; d <= parameters.ndim; d++)
							Gv_loc[d][k] = Glocn[k]*((omegastar-omegan)/(0.5*dt*omegan) * xi[d] - (pi*omegan/parameters.L) * (uavstar[d]-uavn[d])/(0.5*dt));
					}
					
					// Compute the divergence term
#if defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
					compute_divergence_fftw(Gv_loc, div_G_loc, parameters);
#endif
					// Compute G^* = G^n + 0.5*dt * C * (Q(G^n)-Q(Glimit^n)) - 0.5*dt*div(Gvect^n) 
					C = (1./parameters.knudsen) * mu * pow(pi*omegan/parameters.L, -parameters.gamma);
					#pragma omp parallel for private(k) shared(NVLOC, Glocstar, Glocn, dt, C, QGlocn, QGloclimitn, div_G_loc)
					for (k = 0; k < NVLOC; k++)
						Glocstar[k] = Glocn[k] + 0.5*dt * (C*(QGlocn[k]-QGloclimitn[k])-div_G_loc[k]);
					
					// Compute Q(G^*)
#ifdef key_cufft
					if (parameters.force_convolution)
						collision_Qtotal_quick_convolution_cufft(Glocstar, QGlocstar, xitilde, sigma, W, omegastar, parameters);
					else
						collision_Qtotal_cufft(Glocstar, QGlocstar, alpha, alphaprime, omegastar, parameters);
#elif defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
					if (parameters.force_convolution)
						collision_Qtotal_quick_convolution_fftw(Glocstar, QGlocstar, xitilde, sigma, W, omegastar, parameters);
						//collision_Qtotal_force_convolution_fftw(Glocstar, QGlocstar, Amode, Bmode, Cmode, Wmode, omegastar, parameters);
					else
						collision_Qtotal_fftw(Glocstar, QGlocstar, alpha, alphaprime, omegastar, parameters);
#endif

					// Compute the momenta of Q(G^*)
					moms012_QG(&mom0QG, mom1QG, &mom2QG, QGlocstar, uavstar, omegastar, parameters);

					// Compute Glimit^* for the following part of iteration
					// (only without condensation)
					if (!(parameters.condensation)){
						switch (parameters.type_collision){
							case CLASSICAL_COLLISIONS:
								classical_Maxwellian_stationary(Gloclimitstar, diags_n->mass_limit, diags_n->u_av_limit, diags_n->temperature_limit, omegastar, uavstar, parameters);
								break;
				
							case QUANTUM_COLLISIONS:
								quantum_Maxwellian_stationary(Gloclimitstar, diags_n->fugacity_limit, diags_n->u_av_limit, diags_n->temperature_limit, diags_n->critical_mass_limit, omegastar, uavstar, parameters);
								break;
				
							default:
								error("Unknown type of collisions");
						}
					}

					// Compute Q(Glimit^*) for the following part of the iteration
					// (only for residual method without condensation)
					if ((parameters.use_residual) && (!(parameters.condensation))){
#ifdef key_cufft
						if (parameters.force_convolution)
							collision_Qtotal_quick_convolution_cufft(Gloclimitstar, QGloclimitstar, xitilde, sigma, W, omegastar, parameters);
						else
							collision_Qtotal_cufft(Gloclimitstar, QGloclimitstar, alpha, alphaprime, omegastar, parameters);
#elif defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
						if (parameters.force_convolution)
							collision_Qtotal_quick_convolution_fftw(Gloclimitstar, QGloclimitstar, xitilde, sigma, W, omegastar, parameters);
							//collision_Qtotal_force_convolution_fftw(Gloclimitstar, QGloclimitstar, Amode, Bmode, Cmode, Wmode, omegastar, parameters);
						else
							collision_Qtotal_fftw(Gloclimitstar, QGloclimitstar, alpha, alphaprime, omegastar, parameters);
#endif
						// Compute the momenta of Q(Glimit^*)
						moms012_QG(&mom0QGlimit, mom1QGlimit, &mom2QGlimit, QGloclimitstar, uavstar, omegastar, parameters);
					}
					else{
						// Set Q(Glimit^*) to 0
						#pragma omp parallel for private(k) shared(NVLOC, QGloclimitstar)
						for (k = 0; k < NVLOC; k++)
							QGloclimitstar[k] = 0.;
						
						// Set the momenta of Q(Glimit^*) to 0
						mom0QGlimit = 0.;
						for (d = 0; d <= parameters.ndim; d++)
							mom1QGlimit[d] = 0.;
						mom2QGlimit = 0.;
					}

					// Compute rho^(n+1), u^(n+1), eint^(n+1)
					C = (1./parameters.knudsen) * pow(mu,2) * pow(pi*omegastar/parameters.L, -parameters.gamma);
					rhonp1 = rhon + dt* C*(mom0QG-mom0QGlimit);
					for (d = 0; d <= parameters.ndim; d++)
						uavnp1[d] = (rhon*uavn[d] + dt * C*(mom1QG[d]-mom1QGlimit[d]))/rhonp1;
					eintnp1 = (rhon*eintn + dt * C*(mom2QG-mom2QGlimit))/rhonp1;
					
					// Compute T^(n+1) (and z^(n+1), m0^(n+1) for the quantum case)
					switch (parameters.type_collision){
						case CLASSICAL_COLLISIONS:
							// The temperature is 2/d * eint
							Tnp1 = (2.*eintnp1)/(parameters.ndim+1.);
							break;
						case QUANTUM_COLLISIONS:
							// The temperature and the fugacity are obtained by solving a non-linear system with rho and eint as source terms
							Tnp1 = -1.;
							m0np1 = -1.;
							find_fugacity_temperature_criticalmass_fixedpoint(rhonp1, &eintnp1, &Tnp1, &znp1, &m0np1, parameters.hbar, parameters.ndim+1, parameters.type_qparticles);
							break;
						default:
							error("Unknown type of collision");
					}

					// Compute an estimate of omega^(n+1)
					switch (parameters.type_omega){
						case OMEGA_IS_SQRTEKINORHO:
							usquare = 0.;
							for (int d = 0; d <= parameters.ndim; d++)
								usquare += pow(uavnp1[d],2);
							omeganp1 = sqrt(Tnp1+usquare/(parameters.ndim+1.));
							break;
						case OMEGA_IS_SQRTTEMPERATURE:
							omeganp1 = sqrt(Tnp1);
							break;
						default:
							error("Something is wrong with time_loop");
					}

					// Compute the vector function in the divergence
					#pragma omp parallel for private(k, d, kvxabs, kvyabs, kvzabs, itemp, xi) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, parameters, MVX, MVY, MVZ, IVX0LOC, IVY0LOC, IVZ0LOC, dxix, dxiy, dxiz, Gv_loc, Glocstar, omeganp1, omegan, omegastar, dt, pi, uavnp1, uavn)
					for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
						kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
						kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
						kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
						// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
						itemp = k/NVZLOC;
						// itemp now contains kvyloc+NVYLOC*kvxloc;
						kvzabs = k - itemp*NVZLOC;
						kvxabs = itemp/NVYLOC;
						kvyabs = itemp-kvxabs*NVYLOC;
															
						kvxabs += IVX0LOC;
						kvyabs += IVY0LOC;
						kvzabs += IVZ0LOC;
															
						if (kvxabs >= MVX/2)
							kvxabs -= MVX;
						if (kvyabs >= MVY/2)
							kvyabs -= MVY;
						if (kvzabs >= MVZ/2)
							kvzabs -= MVZ;
#endif
						xi[0] = kvxabs * dxix;
						xi[1] = kvyabs * dxiy;
						xi[2] = kvzabs * dxiz;
						
						for (d = 0; d <= parameters.ndim; d++)
							Gv_loc[d][k] = Glocstar[k]*((omeganp1-omegan)/(dt*omegastar) * xi[d] - (pi*omegastar/parameters.L) * (uavnp1[d]-uavn[d])/dt);
					}
					
					// Compute the divergence term
#if defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
					compute_divergence_fftw(Gv_loc, div_G_loc, parameters);
#endif
					// Compute G^(n+1) = G^n + dt * C * (Q(G^*) - Q(Glimit^*)) - dt * div(Gvect^*)
					C = (1./parameters.knudsen) * mu * pow(pi*omegan/parameters.L, -parameters.gamma);
					#pragma omp parallel for private(k, contrib) shared(NVLOC, Glocn, dt, C, QGlocstar, QGloclimitstar, div_G_loc)
					for (k = 0; k < NVXLOC*NVYLOC*NVZLOC; k++){
						/*contrib = dt * (C*(QGlocstar[k]-QGloclimitstar[k])-div_G_loc[k]);
						#pragma omp atomic
						Glocn[k] += contrib;*/
						Glocn[k] += dt * (C*(QGlocstar[k]-QGloclimitstar[k])-div_G_loc[k]);
					}

					break;
				
				default:
					error("Unknown time method");
			}

			// Update the diagnostic list (rho^(n+1), u^(n+1), e^(n+1), T^(n+1), z^(n+1), m0^(n+1))
			diags_n->mass = rhonp1;
			for (d = 0; d <= parameters.ndim; d++)
				diags_n->u_av[d] = uavnp1[d];
			diags_n->eint = eintnp1;
			diags_n->temperature = Tnp1;
			if (parameters.type_collision == QUANTUM_COLLISIONS){
				diags_n->fugacity = znp1;
				if ((parameters.type_qparticles == BOSONS) && (parameters.ndim == DIM3))
					diags_n->critical_mass = m0np1;
			}
			diags_n->omega = omeganp1;

			// Compute Glimit^(n+1) for saving and for the following iteration
			// (only without condensation)
			if (!(parameters.condensation)){
				switch (parameters.type_collision){
					case CLASSICAL_COLLISIONS:
						classical_Maxwellian_stationary(Gloclimitn, diags_n->mass_limit, diags_n->u_av_limit, diags_n->temperature_limit, diags_n->omega, diags_n->u_av, parameters);
						break;
			
					case QUANTUM_COLLISIONS:
						quantum_Maxwellian_stationary(Gloclimitn, diags_n->fugacity_limit, diags_n->u_av_limit, diags_n->temperature_limit, diags_n->critical_mass_limit, diags_n->omega, diags_n->u_av, parameters);
						break;
				
					default:
						error("Unknown type of collisions");
				}
			}

			// Compute Q(Glimit^(n+1)) for the following iteration
			// (only for residual method without condensation)
			if ((parameters.use_residual) && (!(parameters.condensation))){
#ifdef key_cufft
				if (parameters.force_convolution)
					collision_Qtotal_quick_convolution_cufft(Gloclimitn, QGloclimitn, xitilde, sigma, W, omeganp1, parameters);
				else
					collision_Qtotal_cufft(Gloclimitn, QGloclimitn, alpha, alphaprime, omeganp1, parameters);
#elif defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
				if (parameters.force_convolution)
					collision_Qtotal_quick_convolution_fftw(Gloclimitn, QGloclimitn, xitilde, sigma, W, omeganp1, parameters);
					//collision_Qtotal_force_convolution_fftw(Gloclimitn, QGloclimitn, Amode, Bmode, Cmode, Wmode, omeganp1, parameters);
				else
					collision_Qtotal_fftw(Gloclimitn, QGloclimitn, alpha, alphaprime, omeganp1, parameters);
#endif
			}
		}

		// We already know the following diagnostics
		// rho^(n+1), u^(n+1), e^(n+1), T^(n+1), z^(n+1), m0^(n+1)
		// We now compute the remaining diagnostics

		// Kinetic energy
		compute_diagnostic(diags_n, KINETIC_ENERGY, parameters);
		// Entropy
		compute_diagnostic(diags_n, ENTROPY, parameters);
		// Relative entropy (only without condensation)
		if (!(parameters.condensation))
			compute_diagnostic(diags_n, RELATIVE_ENTROPY, parameters);
		// 4th and 6th order momenta
		compute_diagnostic(diags_n, MOMENTUM4, parameters);
		compute_diagnostic(diags_n, MOMENTUM6, parameters);
		// Stress-energy tensor
		compute_diagnostic(diags_n, STRESS_ENERGY, parameters);
		// L1, L2, Linf space norms
		compute_diagnostic(diags_n, L1_NORM, parameters);
		compute_diagnostic(diags_n, L2_NORM, parameters);
		compute_diagnostic(diags_n, LINF_NORM, parameters);
		// Min and Max bounds
		compute_diagnostic(diags_n, MAX_VALUE, parameters);
		compute_diagnostic(diags_n, MIN_VALUE, parameters);

		// If there is an exact solution, compute it at time t^(n+1)
		time += dt;
		if (parameters.exact_solution){
			exact_solution(Glocexn, time, diags_n->omega, diags_n->u_av, parameters);
			compute_diagnostic(diags_n, L1_NORM_EX, parameters);
			compute_diagnostic(diags_n, L2_NORM_EX, parameters);
			compute_diagnostic(diags_n, LINF_NORM_EX, parameters);
			compute_diagnostic(diags_n, L1_ERROR, parameters);
			compute_diagnostic(diags_n, L2_ERROR, parameters);
			compute_diagnostic(diags_n, LINF_ERROR, parameters);
		}

		// Compute relaxation diagnostics
		// (only without condensation)
		if (!(parameters.condensation)){
			compute_diagnostic(diags_n, L1_NORM_LIMIT, parameters);
			compute_diagnostic(diags_n, L2_NORM_LIMIT, parameters);
			compute_diagnostic(diags_n, LINF_NORM_LIMIT, parameters);
			compute_diagnostic(diags_n, L1_ERROR_RELAXATION, parameters);
			compute_diagnostic(diags_n, L2_ERROR_RELAXATION, parameters);
			compute_diagnostic(diags_n, LINF_ERROR_RELAXATION, parameters);
		}

		// Update time-space diagnostics
		compute_time_diagnostics(diags_n, dt, parameters);

		switch(parameters.output_fmt){
			case FMT_BP:
#if defined(key_adios) || defined(key_adios_nompi)
				sprintf(namefile_data, "%s_%0*d.bp", prefixfile, ndigits, n+1);
				break;
#else
				error("Problem: KINEBEC has not been compiled with Adios dependency");
#endif
			case FMT_H5:
#if defined(key_phdf5) || defined(key_hdf5)
				sprintf(namefile_data, "%s_%0*d.h5", prefixfile, ndigits, n+1);
				break;
#else
				error("Problem: KINEBEC has not been compiled with Hdf5 dependency");
#endif
			default:
				error("Unknown output format (only .bp or .h5 are allowed)");
		}

		// Save all data
		save_all_data(*diags_n, time, dt, namefile_data, parameters);


		sprintf(buf, "Time %f reached\n", time);
		message(buf);

		// Verify if the time loop has to be prematurely stopped
		if (parameters.condensation){
			if (parameters.type_qparticles == BOSONS){
				if (diags_n->Linf > 1./THRESHOLD)
					a++;
				else
					a = 0;
				if (a > 5)
					exitsuccess("Blow-up of the numerical solution");
			}
			else{
				// We expect a quantum degeneracy of the distribution
				// WARNING: at present time, we cannot identify the limit distribution
				LinfQG = Linf_QG(QGlocn, parameters);
				if (LinfQG < THRESHOLD)
					a++;
				else
					a = 0;
				if (a > 5)
					exitsuccess("Steady state reached");
			}
		}
		else{
			if (diags_n->relaxerrL1 < THRESHOLD)
				a++;
			else
				a = 0;
			if (a > 5)
				exitsuccess("Steady state reached");
		}
	}
	if ((parameters.n_init < parameters.nt) && (a <= 5))
		message("End of time loop\n");
		
	free_1d(QGloclimitn);

	if ((parameters.time_method == RK2) && (!(parameters.condensation)))
		free_1d(Gloclimitstar);
	if ((parameters.time_method == RK2) && (parameters.type_omega != OMEGA_IS_NOTUSED))
		free_1d(QGloclimitstar);
	
	if (parameters.force_convolution){
		free_2d(xitilde);
		free_2d(sigma);
		free_1d(W);
	}
	else{
		free_2d(alpha);
		free_2d(alphaprime);
	}
	free_1d_char(namefile_data);

	free_1d(QGlocn);

	if (parameters.time_method == RK2){
		free_1d(Glocstar);
		free_1d(QGlocstar);
	}

	if (parameters.type_omega != OMEGA_IS_NOTUSED){
		free_1d(uavn);
		free_1d(uavnp1);
		if (parameters.time_method == RK2)
			free_1d(uavstar);
		free_2d(Gv_loc);
		free_1d(div_G_loc);
		free_1d(mom1QG);
		free_1d(mom1QGlimit);
	}

}
