#include "./src/exec/compute_Qnuz.h"


int crossmode_done;
double ***crossmode_cos, ***crossmode_sin;

int nb_alloc_1d, nb_alloc_2d, nb_alloc_3d;
int nb_alloc_1d_char;
int nb_alloc_1d_int;


void print_help_Qnuz(){
	printf("Manual of compute_Qnuz function\n");
	printf("\n");
	printf("   compute_Qnuz <mandatory parameters>\n");
	printf("\n");
	printf("Mandatory parameters (no specific order is required)\n");
	printf("----------------------------------------------------\n");
	printf("\n");
	printf("  -nu <real value>                allowed values: 1., 1.5, 2. or 2.5\n");
	printf("\n");
	printf("  -qparticles <string value>      indicates the kind of quantum particles (must be ""bosons"" or ""fermions"")\n");
	printf("\n");
	printf("  -z <real value>                 give a prescribed positive value (in ]0,1[ if alpha = -1., in ]0,+inf[ if alpha = +1.\n");
	printf("\n");
    exit(0);
}



int main(int argc, char** argv){
    int n = 1;
    int istatus;
    double nu, z, Qnuz;
    int type_qparticles;
    int nuchecked = (1==0);
    int qparticleschecked = (1==0);
    int zchecked = (1==0);

    if (argc != 7)
        print_help_Qnuz();
    
    while (n < argc){
        if (strcmp(argv[n], "-nu") == 0){
            n++;
			istatus = sscanf(argv[n], "%lf", &nu);
			n++;
            nuchecked = (1==1);
        }
        else if (strcmp(argv[n], "-qparticles") == 0){
			n++;
			if (strcmp(argv[n], "bosons") == 0)
				type_qparticles = BOSONS;
			else if (strcmp(argv[n], "fermions") == 0)
				type_qparticles = FERMIONS;
			else
				print_help_Qnuz();
			n++;
            qparticleschecked = (1==1);
        }
        else if (strcmp(argv[n], "-z") == 0){
            n++;
			istatus = sscanf(argv[n], "%lf", &z);
			n++;
            zchecked = (1==1);
        }
    }

    if (!(nuchecked && qparticleschecked && zchecked))
        print_help_Qnuz();
    
    if ((type_qparticles == FERMIONS) && (z <= 0.))
        print_help_Qnuz();
    
    if ((type_qparticles == BOSONS) && ((z < 0.) || (z >= 1.)))
        print_help_Qnuz();

    Qnuz = compute_Qnuz(nu, z, type_qparticles);

    printf("%.15e\n", Qnuz);
}