#include "./src/exec/compute_fugacity.h"


int crossmode_done;
double ***crossmode_cos, ***crossmode_sin;

int nb_alloc_1d, nb_alloc_2d, nb_alloc_3d;
int nb_alloc_1d_char;
int nb_alloc_1d_int;

void print_help_fugacity(){
	printf("Manual of compute_fugacity function\n");
	printf("\n");
	printf("   compute_fugacity <mandatory parameters> [optional parameters]\n");
	printf("\n");
	printf("Mandatory parameters (no specific order is required)\n");
	printf("----------------------------------------------------\n");
	printf("\n");
	printf("  -rho <real value>               provides the prescribed mass\n");
	printf("\n");
	printf("  -d <integer value>              provides the dimensionality of the studied problem (must be 2 or 3)\n");
	printf("\n");
	printf("  -hbar <real value>              provides the value of the physical constant hbar\n");
	printf("\n");
	printf("  -qparticles <string value>      indicates the kind of studied quantum particles (must be ""bosons"" or ""fermions"")\n");
	printf("\n");
	printf("+ one of the following parameters (not both of them)\n");
	printf("----------------------------------------------------\n");
	printf("  -e <real value>                 provides the prescribed internal energy (always works)\n");
	printf("\n");
	printf("or\n");
	printf("\n");
	printf("  -T <real value>                 provides the prescribed temperature (may fail with 3D bosons)\n");
	printf("\n");
	printf("Optional parameters\n");
	printf("-------------------\n");
	printf("  -o <real value>                 output file where the fugacity, internal energy, etc. are saved\n");
	exit(0);
}


int main(int argc, char** argv){
	char outputfile[MAXSTRINGLENGTH];
	int saveresult = (1==0);
	double rho;
	double eint;
	int d;
	double hbar;
	int type_qparticles;
	double z;
	double m0;
	double T;
	int n;
	int istatus;
	int eint_checked = (1==0);
	int T_checked = (1==0);
	FILE *f;
	
	if (argc < 2)
		print_help_fugacity();
	
	n = 1;
	while (n < argc){
		if (strcmp(argv[n], "-rho") == 0){
			n++;
			istatus = sscanf(argv[n], "%lf", &rho);
			n++;
		}
		else if (strcmp(argv[n], "-d") == 0){
			n++;
			istatus = sscanf(argv[n], "%d", &d);
			n++;
		}
		else if (strcmp(argv[n], "-e") == 0){
			if (!(T_checked)){
				n++;
				istatus = sscanf(argv[n], "%lf", &eint);
				n++;
				eint_checked = (1==1);
				T = -1.;
			}
			else
				print_help_fugacity();
		}
		else if (strcmp(argv[n], "-T") == 0){
			if (!(eint_checked)){
				n++;
				istatus = sscanf(argv[n], "%lf", &T);
				n++;
				T_checked = (1==1);
				eint = -1.;
			}
			else
				print_help_fugacity();
		}
		else if (strcmp(argv[n], "-hbar") == 0){
			n++;
			istatus = sscanf(argv[n], "%lf", &hbar);
			n++;
		}
		else if (strcmp(argv[n], "-qparticles") == 0){
			n++;
			if (strcmp(argv[n], "bosons") == 0)
				type_qparticles = BOSONS;
			else if (strcmp(argv[n], "fermions") == 0)
				type_qparticles = FERMIONS;
			else
				print_help_fugacity();
			n++;
		}
		else if (strcmp(argv[n], "-o") == 0){
			n++;
			strcpy(outputfile, argv[n]);
			saveresult = (1==1);
			n++;
		}
		else
			print_help_fugacity();
	}

	//double t1, t2;
	/*printf("Dichotomy\n");
	t1 = (double)clock();
	find_fugacity_temperature_criticalmass(rho, &eint, &T, &z, &m0, hbar, d, type_qparticles);
	t2 = (double)clock();
	printf("z = %e\n", z);
	printf("Time cost = %e\n", (t2-t1)/CLOCKS_PER_SEC);*/
	/*z = -1.;
	if (T_checked)
		eint = -1.;
	if (eint_checked)
		T = -1.;

	printf("Fixed point\n");
	t1 = (double)clock();*/
	find_fugacity_temperature_criticalmass_fixedpoint(rho, &eint, &T, &z, &m0, hbar, d, type_qparticles);
	//printf("z = %.10e\n", z);
	/*t2 = (double)clock();
	printf("z = %e\n", z);
	printf("Time cost = %e\n", (t2-t1)/CLOCKS_PER_SEC);
	error("stop now");*/

	//printf("Fugacity = %e\n", z);
	
	if (saveresult){
		f = fopen(outputfile, "w");
		fprintf(f, "Mass\n");
		fprintf(f, "%.15e\n", rho);
		fprintf(f, "Internal energy\n");
		fprintf(f, "%.15e\n", eint);
		fprintf(f, "Temperature\n");
		fprintf(f, "%.15e\n", T);
		fprintf(f, "Fugacity\n");
		fprintf(f, "%.15e\n", z);
		fprintf(f, "Critical mass (stricly positive means condensation)\n");
		fprintf(f, "%.15e\n", m0);
		fprintf(f, "Type particles\n");
		switch (type_qparticles){
			case FERMIONS:
				fprintf(f, "fermions\n");
				break;
			case BOSONS:
				fprintf(f, "bosons\n");
				break;
			default:
				error("Unknown kind of particles\n");
		}
		fprintf(f, "Dimensionality\n");
		fprintf(f, "%d\n", d);
		fprintf(f, "hbar\n");
		fprintf(f, "%.15e\n", hbar);
		fclose(f);
	}
}
