#!/usr/bin/env python
# -*- coding: utf-8 -*-

import subprocess
import sys
sys.path.append("./src/UI")
from generate_setupfile import *
import multiprocessing

hostname = subprocess.check_output(['uname', '-n'])
hostname = hostname.decode("utf-8").split('\n')[0]
if (not("zeus" in hostname)):
	import psutil
import glob



"""
#########################
# Use of Pushbullet API #
########################################################################################################################################
# If you want to use the Pushbullet API and receive notifications in your smartphone, please verify that pushbullet-cli is installed.
# Then provide your authentification token in the string variable pb_token below. It should of the form "o.ABCDEFGHIJKLMNOPQRSTUVWXYZ".
#
# If you do not want the use the Pushbullet API, you have to comment the try-except block below.
try:
	from pushbullet import Pushbullet
	import time
	pb_is_ok = True
	pb_token = "o.BcZoWBGG281GgcU4YrrO8VghQs6ykTpr"
	starttime = time.time()
except ImportError as error:
        # Output expected ImportErrors.
        pb_is_ok = False
########################################################################################################################################
"""


						

def send_pushbullet_notification(title, message, force=False):
	if ('pb_is_ok' in globals()):
		global pb_is_ok
		if (pb_is_ok):
			global pb_token
			global starttime
			
			endtime = time.time()
			elaptime = endtime-starttime
			if ((elaptime > 120.) or (force)):
				pb = Pushbullet(pb_token)
				push = pb.push_note(title, message)
				starttime = endtime
				newmessage = ""
			else:
				newmessage = message+"\n"
		else:
			newmessage = ""
	else:
		newmessage = ""
		
	return newmessage


def compute_internal_energy(nameexec, rho, T, hbar, qparticles, ndim, outputfile):
	if (qparticles != "classical"):
		cmd = nameexec+" -rho "+str(rho)+" -d "+str(ndim)+" -hbar "+str(hbar)+" -qparticles "+qparticles+" -T "+str(T)+" -o "+outputfile
		subprocess.call(cmd, shell=True)
		f = open(outputfile, "r")
		lines = f.readlines()
		eint = float(lines[3])
		m0 = float(lines[9])
		z = float(lines[7])
		T = float(lines[5])
		f.close()
	else:
		eint = 0.5*ndim*T
		m0 = -1.
		z = -1.

	
	return [m0, eint, z, T]




def cmdrun(N, nCPUs="max", parallel="serial", force_archlimits=False):
	hostname = subprocess.check_output(['uname', '-n'])
	hostname = hostname.decode("utf-8").split('\n')[0]
	nprocesses_opt = max([N//2, 1])

	if (not("zeus" in hostname)):
		# Get the max number of logical processors
		nprocs_max = psutil.cpu_count(logical=False)

		# Get the max number of threads
		nthreads_max = psutil.cpu_count()

		# Reduce the max number of threads and processors for satisfying the user policy
		if ((hostname == "mathcalc1") or (hostname == "mathcalc2") or (hostname == "mathcalc3")):
			nprocs_max = max([1,nprocs_max//2])
			nthreads_max = max([1,nthreads_max//2])

	if (hostname == 'mouton-Latitude-E6430'):
		if (nCPUs != "max"):
			nprocs_max = min([4, nCPUs])
		else:
			nprocs_max = 4
		prefix_mpi = "/usr/bin/mpirun"
	elif (hostname == 'mouton-Latitude-5480'):
		if (nCPUs != "max"):
			nprocs_max = min([2, nCPUs])
		else:
			nprocs_max = 2
		prefix_mpi = "/usr/bin/mpirun"
	elif (hostname == 'mouton-OptiPlex-7050'):
		if (nCPUs != "max"):
			nprocs_max = min([2, nCPUs])
		else:
			nprocs_max = 2
		prefix_mpi = "/usr/bin/mpirun"
	elif ((hostname == "mathcalc1") or (hostname == "mathcalc2") or (hostname == "mathcalc3")):
		if (nCPUs != "max"):
			nprocs_max = min([32, nCPUs])
		else:
			nprocs_max = 32
		prefix_mpi = "/usr/bin/mpirun -mca btl ^openib"
	else:
		prefix_mpi = prefix_mpi = "/usr/bin/mpirun"
	
	if (force_archlimits):
		prefix_mpi += " --oversubscribe"
		if (nCPUs == "max"):
			nprocs = nprocesses_opt
			nthreads = nprocesses_opt
		else:
			nprocs = nCPUs
			nthreads = nCPUs
	else:
		if ((nCPUs == "max") and (parallel != "cuda")):
			nprocs = min([nprocs_max, nprocesses_opt])
			nthreads = min([nthreads_max, nprocesses_opt])
		else:
			nprocs = nCPUs
			nthreads = nCPUs

	if (parallel == "mpi"):
		cmd = prefix_mpi + " -np "+str(nprocs)+" "
	elif (parallel == "omp"):
		cmd = "export OMP_NUM_THREADS="+str(nthreads)+" && "
	elif ((parallel == "serial") or (parallel == "cuda")):
		cmd = ""
	
	return cmd



def activate_batch_mode():
	hostname = subprocess.check_output(['uname', '-n'])
	hostname = hostname.decode("utf-8").split('\n')[0]
	return ("zeus" in hostname)
